/* Generated by Cython 0.29.14 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/usr/lib/python3/dist-packages/numpy/core/include/numpy/arrayobject.h",
            "/usr/lib/python3/dist-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/usr/lib/python3/dist-packages/numpy/core/include"
        ],
        "name": "hapflk.fastphase",
        "sources": [
            "fastphase/fastphase.pyx"
        ]
    },
    "module_name": "hapflk.fastphase"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_14"
#define CYTHON_HEX_VERSION 0x001D0EF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__hapflk__fastphase
#define __PYX_HAVE_API__hapflk__fastphase
/* Early includes */
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "fastphase/fastphase.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit;
struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_ctuple_int__and_int;
typedef struct __pyx_ctuple_int__and_int __pyx_ctuple_int__and_int;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;

/* "fastphase/fastphase.pyx":601
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)
 * 
 * cdef ( int, int) idx2pair( int idx, int nK):             # <<<<<<<<<<<<<<
 *     cdef int k, l, nelem, curk
 *     nelem = 0
 */
struct __pyx_ctuple_int__and_int {
  int f0;
  int f1;
};
struct __pyx_defaults {
  PyObject *__pyx_arg_stream;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_nproc;
};

/* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */
struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit {
  PyObject_HEAD
  PyObject *__pyx_v_imp;
  PyObject *__pyx_v_nClus;
  PyObject *__pyx_v_self;
};


/* "fastphase/fastphase.pyx":279
 *             cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]
 *             ### Genotypes
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )             # <<<<<<<<<<<<<<
 *             for geno in self.genotypes.keys():
 *                 vit = imp[geno][0]
 */
struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *__pyx_outer_scope;
  PyObject *__pyx_v_geno;
  PyObject *__pyx_v_ijump;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_AddObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
#define __Pyx_BufPtrStrided3d(type, buf, i0, s0, i1, s1, i2, s2) (type)((char*)buf + i0 * s0 + i1 * s1 + i2 * s2)
/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* ToPyCTupleUtility.proto */
static PyObject* __pyx_convert__to_py___pyx_ctuple_int__and_int(__pyx_ctuple_int__and_int);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_long(npy_long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'hapflk.fastphase' */
static PyTypeObject *__pyx_ptype_6hapflk_9fastphase___pyx_scope_struct__optimfit = 0;
static PyTypeObject *__pyx_ptype_6hapflk_9fastphase___pyx_scope_struct_1_genexpr = 0;
static PyArrayObject *__pyx_f_6hapflk_9fastphase_calc_cost_matrix_haplo_tot(PyArrayObject *, int, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_f_6hapflk_9fastphase_calc_cost_matrix_geno_tot(PyArrayObject *, int, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_6hapflk_9fastphase_hap_p_all(int, double, double); /*proto*/
static double __pyx_f_6hapflk_9fastphase_gen_p_geno(int, double, double, double); /*proto*/
static double __pyx_f_6hapflk_9fastphase_myPow10(int); /*proto*/
static int __pyx_f_6hapflk_9fastphase_argmax(PyArrayObject *, int); /*proto*/
static int __pyx_f_6hapflk_9fastphase_pair2idx(int, int, int); /*proto*/
static __pyx_ctuple_int__and_int __pyx_f_6hapflk_9fastphase_idx2pair(int, int); /*proto*/
static double __pyx_f_6hapflk_9fastphase_genprG(double, double, int); /*proto*/
static double __pyx_f_6hapflk_9fastphase_probJ(int, int, double); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_genViterbi(PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_genCalc(PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_6hapflk_9fastphase_happrG(double, int); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_hapViterbi(PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_hapCalc(PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int_t = { "int_t", NULL, sizeof(__pyx_t_5numpy_int_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int_t), 0 };
#define __Pyx_MODULE_NAME "hapflk.fastphase"
extern int __pyx_module_is_main_hapflk__fastphase;
int __pyx_module_is_main_hapflk__fastphase = 0;

/* Implementation of 'hapflk.fastphase' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_EM[] = "EM";
static const char __pyx_k_ID[] = "ID";
static const char __pyx_k_aa[] = "aa";
static const char __pyx_k_gg[] = "gg";
static const char __pyx_k_hh[] = "hh";
static const char __pyx_k_it[] = "it";
static const char __pyx_k_jm[] = "jm";
static const char __pyx_k_k1[] = "k1";
static const char __pyx_k_k2[] = "k2";
static const char __pyx_k_kp[] = "kp";
static const char __pyx_k_nK[] = "nK";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_pZ[] = "pZ";
static const char __pyx_k_rr[] = "rr";
static const char __pyx_k_tt[] = "tt";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_bot[] = "bot";
static const char __pyx_k_dat[] = "dat";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_fit[] = "fit";
static const char __pyx_k_gen[] = "gen";
static const char __pyx_k_hap[] = "hap";
static const char __pyx_k_iEM[] = "iEM";
static const char __pyx_k_imp[] = "imp";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_jmk[] = "jmk";
static const char __pyx_k_kp1[] = "kp1";
static const char __pyx_k_kp2[] = "kp2";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_nEM[] = "nEM";
static const char __pyx_k_par[] = "par";
static const char __pyx_k_pth[] = "pth";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_rho[] = "rho";
static const char __pyx_k_snp[] = "snp";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_top[] = "top";
static const char __pyx_k_u2p[] = "u2p";
static const char __pyx_k_u2z[] = "u2z";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_vit[] = "vit";
static const char __pyx_k_Iter[] = "*** Iter";
static const char __pyx_k_Loci[] = "# Loci";
static const char __pyx_k_Pool[] = "Pool";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_fast[] = "fast";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_flog[] = "flog";
static const char __pyx_k_geno[] = "geno";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_item[] = "item";
static const char __pyx_k_iter[] = "iter";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_nLoc[] = "nLoc";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_nhap[] = "nhap";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_pool[] = "pool";
static const char __pyx_k_prfx[] = "prfx";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_haplo[] = "haplo";
static const char __pyx_k_ijump[] = "ijump";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_nClus[] = "nClus";
static const char __pyx_k_nLoci[] = "nLoci";
static const char __pyx_k_nbest[] = "nbest";
static const char __pyx_k_niter[] = "niter";
static const char __pyx_k_nproc[] = "nproc";
static const char __pyx_k_nsamp[] = "nsamp";
static const char __pyx_k_nstep[] = "nstep";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_par_s[] = "par_s";
static const char __pyx_k_pgeno[] = "pgeno";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_probZ[] = "probZ";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_tasks[] = "tasks";
static const char __pyx_k_theta[] = "theta";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_tohap[] = "_tohap";
static const char __pyx_k_up2pz[] = "up2pz";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_curpar[] = "curpar";
static const char __pyx_k_fitter[] = "fitter";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_impute[] = "impute";
static const char __pyx_k_is_not[] = "is not";
static const char __pyx_k_lambda[] = "<lambda>";
static const char __pyx_k_loglik[] = "loglik";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_newpar[] = "newpar";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_permut[] = "permut";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "reduce";
static const char __pyx_k_rhomin[] = "rhomin";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Init_EM[] = "*** Init EM";
static const char __pyx_k_curclus[] = "curclus";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_gen2hap[] = "gen2hap";
static const char __pyx_k_gen_dat[] = "gen_dat";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_imputer[] = "imputer";
static const char __pyx_k_liktraj[] = "liktraj";
static const char __pyx_k_logLike[] = "logLike";
static const char __pyx_k_loglike[] = "loglike";
static const char __pyx_k_missing[] = "missing";
static const char __pyx_k_parList[] = "parList";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_results[] = "results";
static const char __pyx_k_threads[] = "# threads ";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_viterbi[] = "viterbi";
static const char __pyx_k_alpha_up[] = "alpha_up";
static const char __pyx_k_clusters[] = "# clusters ";
static const char __pyx_k_cost_mat[] = "cost_mat";
static const char __pyx_k_gLogLike[] = "gLogLike";
static const char __pyx_k_hLogLike[] = "hLogLike";
static const char __pyx_k_init_par[] = "init_par";
static const char __pyx_k_log_like[] = "log_like";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_optimfit[] = "optimfit";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_theta_up[] = "theta_up";
static const char __pyx_k_viterber[] = "viterber";
static const char __pyx_k_Genotypes[] = "# Genotypes";
static const char __pyx_k_cpu_count[] = "cpu_count";
static const char __pyx_k_fastphase[] = "fastphase";
static const char __pyx_k_fitInData[] = "fitInData";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_genotypes[] = "genotypes";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_modParams[] = "modParams";
static const char __pyx_k_terminate[] = "terminate";
static const char __pyx_k_vectorize[] = "vectorize";
static const char __pyx_k_Haplotypes[] = "# Haplotypes";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_fitOutData[] = "fitOutData";
static const char __pyx_k_haplotypes[] = "haplotypes";
static const char __pyx_k_initUpdate[] = "initUpdate";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_Imputations[] = "Imputations";
static const char __pyx_k_addGenotype[] = "addGenotype";
static const char __pyx_k_addIndivFit[] = "addIndivFit";
static const char __pyx_k_switch_pars[] = "switch_pars";
static const char __pyx_k_OPTIMFIT_RUN[] = "=== OPTIMFIT RUN ===";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_addHaplotype[] = "addHaplotype";
static const char __pyx_k_imputeInData[] = "imputeInData";
static const char __pyx_k_old_log_like[] = "old_log_like";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_fastphase_fit[] = "fastphase.fit";
static const char __pyx_k_imputeOutData[] = "imputeOutData";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_scipy_optimize[] = "scipy.optimize";
static const char __pyx_k_Initial_Viterbi[] = "Initial Viterbi";
static const char __pyx_k_fastphase_flush[] = "fastphase.flush";
static const char __pyx_k_modParams_write[] = "modParams.write";
static const char __pyx_k_multiprocessing[] = "multiprocessing";
static const char __pyx_k_fastphase___exit[] = "fastphase.__exit__";
static const char __pyx_k_fastphase___init[] = "fastphase.__init__";
static const char __pyx_k_fastphase_impute[] = "fastphase.impute";
static const char __pyx_k_fitInData___init[] = "fitInData.__init__";
static const char __pyx_k_hapflk_fastphase[] = "hapflk.fastphase";
static const char __pyx_k_modParams___init[] = "modParams.__init__";
static const char __pyx_k_modParams_update[] = "modParams.update";
static const char __pyx_k_Calculating_Costs[] = "Calculating Costs";
static const char __pyx_k_fastphase___enter[] = "fastphase.__enter__";
static const char __pyx_k_fastphase_gen2hap[] = "fastphase.gen2hap";
static const char __pyx_k_fastphase_viterbi[] = "fastphase.viterbi";
static const char __pyx_k_fitOutData___init[] = "fitOutData.__init__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_fastphase_optimfit[] = "fastphase.optimfit";
static const char __pyx_k_Wrong_Genotype_Size[] = "Wrong Genotype Size:";
static const char __pyx_k_imputeInData___init[] = "imputeInData.__init__";
static const char __pyx_k_Wrong_Haplotype_Size[] = "Wrong Haplotype Size:";
static const char __pyx_k_imputeOutData___init[] = "imputeOutData.__init__";
static const char __pyx_k_modParams_initUpdate[] = "modParams.initUpdate";
static const char __pyx_k_calc_cost_matrix_geno[] = "calc_cost_matrix_geno";
static const char __pyx_k_fastphase_addGenotype[] = "fastphase.addGenotype";
static const char __pyx_k_linear_sum_assignment[] = "linear_sum_assignment";
static const char __pyx_k_modParams_addIndivFit[] = "modParams.addIndivFit";
static const char __pyx_k_calc_cost_matrix_haplo[] = "calc_cost_matrix_haplo";
static const char __pyx_k_fastphase_addHaplotype[] = "fastphase.addHaplotype";
static const char __pyx_k_Fitting_fastphase_model[] = "Fitting fastphase model";
static const char __pyx_k_fastphase_fastphase_pyx[] = "fastphase/fastphase.pyx";
static const char __pyx_k_EM_with_switched_parameters[] = "EM with switched parameters";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_Computing_optimum_permutations[] = "Computing optimum permutations";
static const char __pyx_k_A_class_for_fastphase_model_par[] = "\n    A class for fastphase model parameters.\n    Dimensions:\n    -- number of loci: N\n    -- number of clusters: K\n    Parameters:\n    -- theta (N x K): allele frequencies in clusters at each locus\n    -- alpha (N x K): cluster weights at each locus\n    -- rho (N x 1): jump probabilities in each interval\n    ";
static const char __pyx_k_A_class_to_manipulate_and_contr[] = "\n    A class to manipulate and control a fastphase model (Scheet and Stephens, 2006)\n    Initialized with a problem size = number of loci\n\n    Usage : \n    with fastphase(nloc, nproc) as fph:\n         ... do stuff ...\n    ";
static const char __pyx_k_Module_implementing_the_LD_mode[] = "\nModule implementing the LD model of Scheet and Stephens (2006)\noriginally implemented in Paul Scheet fastPHASE software.\nCode is largely inspired by Yongtao Guan Bimbam Code and as such\nis released under the GNU General Public Licence version 2\n\nMultithread version\n";
static const char __pyx_k_Usage_with_fastphase_nloc_nproc[] = "Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_fastphase_optimfit_locals_genexp[] = "fastphase.optimfit.<locals>.genexpr";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_A_class_for_fastphase_model_par;
static PyObject *__pyx_kp_s_A_class_to_manipulate_and_contr;
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_kp_u_Calculating_Costs;
static PyObject *__pyx_kp_u_Computing_optimum_permutations;
static PyObject *__pyx_n_u_EM;
static PyObject *__pyx_kp_u_EM_with_switched_parameters;
static PyObject *__pyx_kp_u_Fitting_fastphase_model;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_kp_u_Genotypes;
static PyObject *__pyx_kp_u_Haplotypes;
static PyObject *__pyx_n_s_ID;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_Imputations;
static PyObject *__pyx_kp_u_Init_EM;
static PyObject *__pyx_kp_u_Initial_Viterbi;
static PyObject *__pyx_kp_u_Iter;
static PyObject *__pyx_kp_u_Loci;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_kp_u_OPTIMFIT_RUN;
static PyObject *__pyx_n_s_Pool;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_u_Usage_with_fastphase_nloc_nproc;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u_Wrong_Genotype_Size;
static PyObject *__pyx_kp_u_Wrong_Haplotype_Size;
static PyObject *__pyx_n_u_a;
static PyObject *__pyx_n_s_aa;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_addGenotype;
static PyObject *__pyx_n_s_addHaplotype;
static PyObject *__pyx_n_s_addIndivFit;
static PyObject *__pyx_n_s_alpha;
static PyObject *__pyx_n_s_alpha_up;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_bot;
static PyObject *__pyx_n_s_calc_cost_matrix_geno;
static PyObject *__pyx_n_s_calc_cost_matrix_haplo;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_kp_u_clusters;
static PyObject *__pyx_n_s_cost_mat;
static PyObject *__pyx_n_s_cpu_count;
static PyObject *__pyx_n_s_curclus;
static PyObject *__pyx_n_s_curpar;
static PyObject *__pyx_n_s_dat;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_fast;
static PyObject *__pyx_n_s_fastphase;
static PyObject *__pyx_n_s_fastphase___enter;
static PyObject *__pyx_n_s_fastphase___exit;
static PyObject *__pyx_n_s_fastphase___init;
static PyObject *__pyx_n_s_fastphase_addGenotype;
static PyObject *__pyx_n_s_fastphase_addHaplotype;
static PyObject *__pyx_kp_s_fastphase_fastphase_pyx;
static PyObject *__pyx_n_s_fastphase_fit;
static PyObject *__pyx_n_s_fastphase_flush;
static PyObject *__pyx_n_s_fastphase_gen2hap;
static PyObject *__pyx_n_s_fastphase_impute;
static PyObject *__pyx_n_s_fastphase_optimfit;
static PyObject *__pyx_n_s_fastphase_optimfit_locals_genexp;
static PyObject *__pyx_n_s_fastphase_viterbi;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_fit;
static PyObject *__pyx_n_s_fitInData;
static PyObject *__pyx_n_s_fitInData___init;
static PyObject *__pyx_n_s_fitOutData;
static PyObject *__pyx_n_s_fitOutData___init;
static PyObject *__pyx_n_s_fitter;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_flog;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_gLogLike;
static PyObject *__pyx_n_s_gen;
static PyObject *__pyx_n_s_gen2hap;
static PyObject *__pyx_n_s_gen_dat;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_geno;
static PyObject *__pyx_n_u_geno;
static PyObject *__pyx_n_s_genotypes;
static PyObject *__pyx_n_s_gg;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_hLogLike;
static PyObject *__pyx_n_s_hap;
static PyObject *__pyx_n_s_hapflk_fastphase;
static PyObject *__pyx_n_s_haplo;
static PyObject *__pyx_n_u_haplo;
static PyObject *__pyx_n_s_haplotypes;
static PyObject *__pyx_n_s_hh;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_iEM;
static PyObject *__pyx_n_s_ijump;
static PyObject *__pyx_n_s_imp;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_impute;
static PyObject *__pyx_n_s_imputeInData;
static PyObject *__pyx_n_s_imputeInData___init;
static PyObject *__pyx_n_s_imputeOutData;
static PyObject *__pyx_n_s_imputeOutData___init;
static PyObject *__pyx_n_s_imputer;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_initUpdate;
static PyObject *__pyx_n_s_init_par;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_kp_u_is_not;
static PyObject *__pyx_n_s_it;
static PyObject *__pyx_n_s_item;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_u_iter;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_jm;
static PyObject *__pyx_n_s_jmk;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_k1;
static PyObject *__pyx_n_s_k2;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_kp;
static PyObject *__pyx_n_s_kp1;
static PyObject *__pyx_n_s_kp2;
static PyObject *__pyx_n_s_lambda;
static PyObject *__pyx_n_s_liktraj;
static PyObject *__pyx_n_s_linear_sum_assignment;
static PyObject *__pyx_n_s_logLike;
static PyObject *__pyx_n_s_log_like;
static PyObject *__pyx_n_u_loglik;
static PyObject *__pyx_n_s_loglike;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_missing;
static PyObject *__pyx_n_s_modParams;
static PyObject *__pyx_n_s_modParams___init;
static PyObject *__pyx_n_s_modParams_addIndivFit;
static PyObject *__pyx_n_s_modParams_initUpdate;
static PyObject *__pyx_n_s_modParams_update;
static PyObject *__pyx_n_s_modParams_write;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_multiprocessing;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nClus;
static PyObject *__pyx_n_s_nEM;
static PyObject *__pyx_n_s_nK;
static PyObject *__pyx_n_s_nLoc;
static PyObject *__pyx_n_s_nLoci;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_nbest;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_newpar;
static PyObject *__pyx_n_s_nhap;
static PyObject *__pyx_n_s_niter;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nproc;
static PyObject *__pyx_n_s_nsamp;
static PyObject *__pyx_n_s_nstep;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_old_log_like;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_optimfit;
static PyObject *__pyx_n_s_pZ;
static PyObject *__pyx_n_s_par;
static PyObject *__pyx_n_s_parList;
static PyObject *__pyx_n_s_par_s;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_permut;
static PyObject *__pyx_n_s_pgeno;
static PyObject *__pyx_n_s_pool;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_prfx;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_probZ;
static PyObject *__pyx_n_s_pth;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_results;
static PyObject *__pyx_n_s_rho;
static PyObject *__pyx_n_u_rho;
static PyObject *__pyx_n_s_rhomin;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_s_rr;
static PyObject *__pyx_n_s_scipy_optimize;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_u_snp;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_switch_pars;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_u_t;
static PyObject *__pyx_n_s_tasks;
static PyObject *__pyx_n_s_terminate;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theta;
static PyObject *__pyx_n_s_theta_up;
static PyObject *__pyx_kp_u_threads;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tohap;
static PyObject *__pyx_n_s_top;
static PyObject *__pyx_n_s_tt;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_u2p;
static PyObject *__pyx_n_s_u2z;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_up2pz;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_vectorize;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_s_vit;
static PyObject *__pyx_n_s_viterber;
static PyObject *__pyx_n_s_viterbi;
static PyObject *__pyx_n_u_w;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_lambda_funcdef_6hapflk_9fastphase_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fitInData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_alpha, PyObject *__pyx_v_theta, PyObject *__pyx_v_rho, PyObject *__pyx_v_data, PyObject *__pyx_v_u2z); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_10fitOutData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_logLike, PyObject *__pyx_v_top, PyObject *__pyx_v_bot, PyObject *__pyx_v_jmk, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_12imputeInData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_parList, PyObject *__pyx_v_nLoc, PyObject *__pyx_v_name, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_13imputeOutData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pgeno, PyObject *__pyx_v_name, PyObject *__pyx_v_pZ, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nLoc, PyObject *__pyx_v_nClus, PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_theta_up); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_2initUpdate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_4addIndivFit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_t, PyObject *__pyx_v_b, PyObject *__pyx_v_j, PyObject *__pyx_v_nhap); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_6update(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_25__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_8write(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_27__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nLoci, PyObject *__pyx_v_nproc, PyObject *__pyx_v_prfx); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_2__enter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_4__exit__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_6flush(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_8addHaplotype(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ID, PyObject *__pyx_v_hap, CYTHON_UNUSED PyObject *__pyx_v_missing); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_10addGenotype(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ID, PyObject *__pyx_v_gen, CYTHON_UNUSED PyObject *__pyx_v_missing); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_12gen2hap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_gen); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_14fit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nClus, PyObject *__pyx_v_nstep, PyObject *__pyx_v_params, PyObject *__pyx_v_verbose, PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_theta_up, PyObject *__pyx_v_fast); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_8optimfit_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_16optimfit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nClus, PyObject *__pyx_v_nstep, CYTHON_UNUSED PyObject *__pyx_v_params, PyObject *__pyx_v_verbose, CYTHON_UNUSED PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_fast, PyObject *__pyx_v_nEM, PyObject *__pyx_v_niter); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_18viterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parList); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_20impute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parList); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_fitter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_2viterber(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_4imputer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_6calc_cost_matrix_haplo_tot(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_h, int __pyx_v_nK); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_8calc_cost_matrix_haplo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_10calc_cost_matrix_geno_tot(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_g, int __pyx_v_nK); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_12calc_cost_matrix_geno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_14switch_pars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_par, PyObject *__pyx_v_permut); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_16genViterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_18genCalc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg, PyObject *__pyx_v_u2p); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_20hapViterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh); /* proto */
static PyObject *__pyx_pf_6hapflk_9fastphase_22hapCalc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh, PyObject *__pyx_v_u2p); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct__optimfit(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_0_01;
static PyObject *__pyx_float_0_98;
static PyObject *__pyx_float_1eneg_6;
static PyObject *__pyx_float_0_001;
static PyObject *__pyx_float_0_999;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_1000;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_slice_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__78;
/* Late includes */

/* "fastphase/fastphase.pyx":125
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 * 
 * _tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))             # <<<<<<<<<<<<<<
 * 
 * class fastphase():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_24lambda(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_24lambda = {"lambda", (PyCFunction)__pyx_pw_6hapflk_9fastphase_24lambda, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_24lambda(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_6hapflk_9fastphase_lambda(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_6hapflk_9fastphase_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_x, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
    goto __pyx_L4_next_or;
  } else {
  }
  if (!-1L) {
  } else {
    __pyx_t_2 = __Pyx_PyInt_From_long(-1L); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_x, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hapflk.fastphase.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":25
 * 
 * class fitInData():
 *     def __init__(self,type,alpha,theta,rho,data,u2z):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.alpha=alpha
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fitInData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fitInData_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fitInData_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fitInData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_type = 0;
  PyObject *__pyx_v_alpha = 0;
  PyObject *__pyx_v_theta = 0;
  PyObject *__pyx_v_rho = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_u2z = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_type,&__pyx_n_s_alpha,&__pyx_n_s_theta,&__pyx_n_s_rho,&__pyx_n_s_data,&__pyx_n_s_u2z,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 1); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 2); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 3); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rho)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 4); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 5); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u2z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 6); __PYX_ERR(0, 25, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 25, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_self = values[0];
    __pyx_v_type = values[1];
    __pyx_v_alpha = values[2];
    __pyx_v_theta = values[3];
    __pyx_v_rho = values[4];
    __pyx_v_data = values[5];
    __pyx_v_u2z = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 25, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fitInData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fitInData___init__(__pyx_self, __pyx_v_self, __pyx_v_type, __pyx_v_alpha, __pyx_v_theta, __pyx_v_rho, __pyx_v_data, __pyx_v_u2z);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fitInData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_alpha, PyObject *__pyx_v_theta, PyObject *__pyx_v_rho, PyObject *__pyx_v_data, PyObject *__pyx_v_u2z) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":26
 * class fitInData():
 *     def __init__(self,type,alpha,theta,rho,data,u2z):
 *         self.type=type             # <<<<<<<<<<<<<<
 *         self.alpha=alpha
 *         self.theta=theta
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_type, __pyx_v_type) < 0) __PYX_ERR(0, 26, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":27
 *     def __init__(self,type,alpha,theta,rho,data,u2z):
 *         self.type=type
 *         self.alpha=alpha             # <<<<<<<<<<<<<<
 *         self.theta=theta
 *         self.rho=rho
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_alpha, __pyx_v_alpha) < 0) __PYX_ERR(0, 27, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":28
 *         self.type=type
 *         self.alpha=alpha
 *         self.theta=theta             # <<<<<<<<<<<<<<
 *         self.rho=rho
 *         self.data=data
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_theta, __pyx_v_theta) < 0) __PYX_ERR(0, 28, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":29
 *         self.alpha=alpha
 *         self.theta=theta
 *         self.rho=rho             # <<<<<<<<<<<<<<
 *         self.data=data
 *         self.up2pz=u2z
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_rho, __pyx_v_rho) < 0) __PYX_ERR(0, 29, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":30
 *         self.theta=theta
 *         self.rho=rho
 *         self.data=data             # <<<<<<<<<<<<<<
 *         self.up2pz=u2z
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_data, __pyx_v_data) < 0) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":31
 *         self.rho=rho
 *         self.data=data
 *         self.up2pz=u2z             # <<<<<<<<<<<<<<
 * 
 * class fitOutData():
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_up2pz, __pyx_v_u2z) < 0) __PYX_ERR(0, 31, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":25
 * 
 * class fitInData():
 *     def __init__(self,type,alpha,theta,rho,data,u2z):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.alpha=alpha
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fitInData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":34
 * 
 * class fitOutData():
 *     def __init__(self,logLike,top,bot,jmk,val):             # <<<<<<<<<<<<<<
 *         self.logLike=logLike
 *         self.top=top
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_10fitOutData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_10fitOutData_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_10fitOutData_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_10fitOutData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_logLike = 0;
  PyObject *__pyx_v_top = 0;
  PyObject *__pyx_v_bot = 0;
  PyObject *__pyx_v_jmk = 0;
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_logLike,&__pyx_n_s_top,&__pyx_n_s_bot,&__pyx_n_s_jmk,&__pyx_n_s_val,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logLike)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 1); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_top)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 2); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bot)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 3); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jmk)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 4); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_val)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 5); __PYX_ERR(0, 34, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 34, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_self = values[0];
    __pyx_v_logLike = values[1];
    __pyx_v_top = values[2];
    __pyx_v_bot = values[3];
    __pyx_v_jmk = values[4];
    __pyx_v_val = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 34, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fitOutData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_10fitOutData___init__(__pyx_self, __pyx_v_self, __pyx_v_logLike, __pyx_v_top, __pyx_v_bot, __pyx_v_jmk, __pyx_v_val);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_10fitOutData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_logLike, PyObject *__pyx_v_top, PyObject *__pyx_v_bot, PyObject *__pyx_v_jmk, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":35
 * class fitOutData():
 *     def __init__(self,logLike,top,bot,jmk,val):
 *         self.logLike=logLike             # <<<<<<<<<<<<<<
 *         self.top=top
 *         self.bot=bot
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_logLike, __pyx_v_logLike) < 0) __PYX_ERR(0, 35, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":36
 *     def __init__(self,logLike,top,bot,jmk,val):
 *         self.logLike=logLike
 *         self.top=top             # <<<<<<<<<<<<<<
 *         self.bot=bot
 *         self.jmk=jmk
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_top, __pyx_v_top) < 0) __PYX_ERR(0, 36, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":37
 *         self.logLike=logLike
 *         self.top=top
 *         self.bot=bot             # <<<<<<<<<<<<<<
 *         self.jmk=jmk
 *         self.val=val
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bot, __pyx_v_bot) < 0) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":38
 *         self.top=top
 *         self.bot=bot
 *         self.jmk=jmk             # <<<<<<<<<<<<<<
 *         self.val=val
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_jmk, __pyx_v_jmk) < 0) __PYX_ERR(0, 38, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":39
 *         self.bot=bot
 *         self.jmk=jmk
 *         self.val=val             # <<<<<<<<<<<<<<
 * 
 * class imputeInData():
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_val, __pyx_v_val) < 0) __PYX_ERR(0, 39, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":34
 * 
 * class fitOutData():
 *     def __init__(self,logLike,top,bot,jmk,val):             # <<<<<<<<<<<<<<
 *         self.logLike=logLike
 *         self.top=top
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fitOutData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":42
 * 
 * class imputeInData():
 *     def __init__(self,type,parList,nLoc,name,data):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.parList=parList
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_12imputeInData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_12imputeInData_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_12imputeInData_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_12imputeInData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_type = 0;
  PyObject *__pyx_v_parList = 0;
  PyObject *__pyx_v_nLoc = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_type,&__pyx_n_s_parList,&__pyx_n_s_nLoc,&__pyx_n_s_name,&__pyx_n_s_data,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 1); __PYX_ERR(0, 42, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 2); __PYX_ERR(0, 42, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nLoc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 3); __PYX_ERR(0, 42, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 4); __PYX_ERR(0, 42, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, 5); __PYX_ERR(0, 42, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 42, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_self = values[0];
    __pyx_v_type = values[1];
    __pyx_v_parList = values[2];
    __pyx_v_nLoc = values[3];
    __pyx_v_name = values[4];
    __pyx_v_data = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 42, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.imputeInData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_12imputeInData___init__(__pyx_self, __pyx_v_self, __pyx_v_type, __pyx_v_parList, __pyx_v_nLoc, __pyx_v_name, __pyx_v_data);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_12imputeInData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_type, PyObject *__pyx_v_parList, PyObject *__pyx_v_nLoc, PyObject *__pyx_v_name, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":43
 * class imputeInData():
 *     def __init__(self,type,parList,nLoc,name,data):
 *         self.type=type             # <<<<<<<<<<<<<<
 *         self.parList=parList
 *         self.name=name
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_type, __pyx_v_type) < 0) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":44
 *     def __init__(self,type,parList,nLoc,name,data):
 *         self.type=type
 *         self.parList=parList             # <<<<<<<<<<<<<<
 *         self.name=name
 *         self.nLoc=nLoc
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parList, __pyx_v_parList) < 0) __PYX_ERR(0, 44, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":45
 *         self.type=type
 *         self.parList=parList
 *         self.name=name             # <<<<<<<<<<<<<<
 *         self.nLoc=nLoc
 *         self.data=data
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 45, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":46
 *         self.parList=parList
 *         self.name=name
 *         self.nLoc=nLoc             # <<<<<<<<<<<<<<
 *         self.data=data
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nLoc, __pyx_v_nLoc) < 0) __PYX_ERR(0, 46, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":47
 *         self.name=name
 *         self.nLoc=nLoc
 *         self.data=data             # <<<<<<<<<<<<<<
 * 
 * class imputeOutData():
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_data, __pyx_v_data) < 0) __PYX_ERR(0, 47, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":42
 * 
 * class imputeInData():
 *     def __init__(self,type,parList,nLoc,name,data):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.parList=parList
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hapflk.fastphase.imputeInData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":50
 * 
 * class imputeOutData():
 *     def __init__(self,pgeno,name,pZ,path=None):             # <<<<<<<<<<<<<<
 *         self.name=name
 *         self.pgeno=pgeno
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_13imputeOutData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_13imputeOutData_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_13imputeOutData_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_13imputeOutData_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_pgeno = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_pZ = 0;
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_pgeno,&__pyx_n_s_name,&__pyx_n_s_pZ,&__pyx_n_s_path,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pgeno)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 1); __PYX_ERR(0, 50, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 2); __PYX_ERR(0, 50, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pZ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 3); __PYX_ERR(0, 50, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 50, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_pgeno = values[1];
    __pyx_v_name = values[2];
    __pyx_v_pZ = values[3];
    __pyx_v_path = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 50, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.imputeOutData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_13imputeOutData___init__(__pyx_self, __pyx_v_self, __pyx_v_pgeno, __pyx_v_name, __pyx_v_pZ, __pyx_v_path);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_13imputeOutData___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pgeno, PyObject *__pyx_v_name, PyObject *__pyx_v_pZ, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":51
 * class imputeOutData():
 *     def __init__(self,pgeno,name,pZ,path=None):
 *         self.name=name             # <<<<<<<<<<<<<<
 *         self.pgeno=pgeno
 *         self.pZ=pZ
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 51, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":52
 *     def __init__(self,pgeno,name,pZ,path=None):
 *         self.name=name
 *         self.pgeno=pgeno             # <<<<<<<<<<<<<<
 *         self.pZ=pZ
 *         self.path=path
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pgeno, __pyx_v_pgeno) < 0) __PYX_ERR(0, 52, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":53
 *         self.name=name
 *         self.pgeno=pgeno
 *         self.pZ=pZ             # <<<<<<<<<<<<<<
 *         self.path=path
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pZ, __pyx_v_pZ) < 0) __PYX_ERR(0, 53, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":54
 *         self.pgeno=pgeno
 *         self.pZ=pZ
 *         self.path=path             # <<<<<<<<<<<<<<
 * 
 * class modParams():
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_path, __pyx_v_path) < 0) __PYX_ERR(0, 54, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":50
 * 
 * class imputeOutData():
 *     def __init__(self,pgeno,name,pZ,path=None):             # <<<<<<<<<<<<<<
 *         self.name=name
 *         self.pgeno=pgeno
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hapflk.fastphase.imputeOutData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":67
 *     -- rho (N x 1): jump probabilities in each interval
 *     '''
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):             # <<<<<<<<<<<<<<
 *         self.nLoc=nLoc
 *         self.nClus=nClus
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9modParams_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9modParams_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nLoc = 0;
  PyObject *__pyx_v_nClus = 0;
  PyObject *__pyx_v_rhomin = 0;
  PyObject *__pyx_v_alpha_up = 0;
  PyObject *__pyx_v_theta_up = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nLoc,&__pyx_n_s_nClus,&__pyx_n_s_rhomin,&__pyx_n_s_alpha_up,&__pyx_n_s_theta_up,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[3] = ((PyObject *)((PyObject*)__pyx_float_1eneg_6));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    values[5] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nLoc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, 1); __PYX_ERR(0, 67, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nClus)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, 2); __PYX_ERR(0, 67, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rhomin);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alpha_up);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta_up);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 67, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_nLoc = values[1];
    __pyx_v_nClus = values[2];
    __pyx_v_rhomin = values[3];
    __pyx_v_alpha_up = values[4];
    __pyx_v_theta_up = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.modParams.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9modParams___init__(__pyx_self, __pyx_v_self, __pyx_v_nLoc, __pyx_v_nClus, __pyx_v_rhomin, __pyx_v_alpha_up, __pyx_v_theta_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nLoc, PyObject *__pyx_v_nClus, PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_theta_up) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":68
 *     '''
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):
 *         self.nLoc=nLoc             # <<<<<<<<<<<<<<
 *         self.nClus=nClus
 *         self.theta=0.98*np.random.random((nLoc,nClus))+0.01 # avoid bounds 0 and 1
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nLoc, __pyx_v_nLoc) < 0) __PYX_ERR(0, 68, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":69
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):
 *         self.nLoc=nLoc
 *         self.nClus=nClus             # <<<<<<<<<<<<<<
 *         self.theta=0.98*np.random.random((nLoc,nClus))+0.01 # avoid bounds 0 and 1
 *         self.rho=np.ones((nLoc,1))/1000
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nClus, __pyx_v_nClus) < 0) __PYX_ERR(0, 69, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":70
 *         self.nLoc=nLoc
 *         self.nClus=nClus
 *         self.theta=0.98*np.random.random((nLoc,nClus))+0.01 # avoid bounds 0 and 1             # <<<<<<<<<<<<<<
 *         self.rho=np.ones((nLoc,1))/1000
 *         ##self.alpha=np.random.mtrand.dirichlet(np.ones(nClus),nLoc) # make sure sum(alpha_is)=1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_random); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_random); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nClus);
  __Pyx_GIVEREF(__pyx_v_nClus);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_nClus);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_float_0_98, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyFloat_AddObjC(__pyx_t_2, __pyx_float_0_01, 0.01, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_theta, __pyx_t_1) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":71
 *         self.nClus=nClus
 *         self.theta=0.98*np.random.random((nLoc,nClus))+0.01 # avoid bounds 0 and 1
 *         self.rho=np.ones((nLoc,1))/1000             # <<<<<<<<<<<<<<
 *         ##self.alpha=np.random.mtrand.dirichlet(np.ones(nClus),nLoc) # make sure sum(alpha_is)=1
 *         self.alpha=1.0/nClus*np.ones((nLoc,nClus))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_rho, __pyx_t_3) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":73
 *         self.rho=np.ones((nLoc,1))/1000
 *         ##self.alpha=np.random.mtrand.dirichlet(np.ones(nClus),nLoc) # make sure sum(alpha_is)=1
 *         self.alpha=1.0/nClus*np.ones((nLoc,nClus))             # <<<<<<<<<<<<<<
 *         self.rhomin=rhomin
 *         self.alpha_up = alpha_up
 */
  __pyx_t_3 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_0, __pyx_v_nClus, 1.0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nClus);
  __Pyx_GIVEREF(__pyx_v_nClus);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nClus);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_alpha, __pyx_t_4) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":74
 *         ##self.alpha=np.random.mtrand.dirichlet(np.ones(nClus),nLoc) # make sure sum(alpha_is)=1
 *         self.alpha=1.0/nClus*np.ones((nLoc,nClus))
 *         self.rhomin=rhomin             # <<<<<<<<<<<<<<
 *         self.alpha_up = alpha_up
 *         self.theta_up = theta_up
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_rhomin, __pyx_v_rhomin) < 0) __PYX_ERR(0, 74, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":75
 *         self.alpha=1.0/nClus*np.ones((nLoc,nClus))
 *         self.rhomin=rhomin
 *         self.alpha_up = alpha_up             # <<<<<<<<<<<<<<
 *         self.theta_up = theta_up
 *         self.loglike = 0
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_alpha_up, __pyx_v_alpha_up) < 0) __PYX_ERR(0, 75, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":76
 *         self.rhomin=rhomin
 *         self.alpha_up = alpha_up
 *         self.theta_up = theta_up             # <<<<<<<<<<<<<<
 *         self.loglike = 0
 *     def initUpdate(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_theta_up, __pyx_v_theta_up) < 0) __PYX_ERR(0, 76, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":77
 *         self.alpha_up = alpha_up
 *         self.theta_up = theta_up
 *         self.loglike = 0             # <<<<<<<<<<<<<<
 *     def initUpdate(self):
 *         self.top=np.zeros((self.nLoc,self.nClus))
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_loglike, __pyx_int_0) < 0) __PYX_ERR(0, 77, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":67
 *     -- rho (N x 1): jump probabilities in each interval
 *     '''
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):             # <<<<<<<<<<<<<<
 *         self.nLoc=nLoc
 *         self.nClus=nClus
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hapflk.fastphase.modParams.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":78
 *         self.theta_up = theta_up
 *         self.loglike = 0
 *     def initUpdate(self):             # <<<<<<<<<<<<<<
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_3initUpdate(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9modParams_3initUpdate = {"initUpdate", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9modParams_3initUpdate, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_3initUpdate(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("initUpdate (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9modParams_2initUpdate(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_2initUpdate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("initUpdate", 0);

  /* "fastphase/fastphase.pyx":79
 *         self.loglike = 0
 *     def initUpdate(self):
 *         self.top=np.zeros((self.nLoc,self.nClus))             # <<<<<<<<<<<<<<
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 *         self.jmk=np.zeros((self.nLoc,self.nClus))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_top, __pyx_t_1) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":80
 *     def initUpdate(self):
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))             # <<<<<<<<<<<<<<
 *         self.jmk=np.zeros((self.nLoc,self.nClus))
 *         self.jm=np.zeros((self.nLoc,1))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bot, __pyx_t_1) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":81
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 *         self.jmk=np.zeros((self.nLoc,self.nClus))             # <<<<<<<<<<<<<<
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_jmk, __pyx_t_1) < 0) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":82
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 *         self.jmk=np.zeros((self.nLoc,self.nClus))
 *         self.jm=np.zeros((self.nLoc,1))             # <<<<<<<<<<<<<<
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_1);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_jm, __pyx_t_1) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":83
 *         self.jmk=np.zeros((self.nLoc,self.nClus))
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0             # <<<<<<<<<<<<<<
 *     def addIndivFit(self,t,b,j,nhap):
 *         self.top += t
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nhap, __pyx_float_0_0) < 0) __PYX_ERR(0, 83, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":78
 *         self.theta_up = theta_up
 *         self.loglike = 0
 *     def initUpdate(self):             # <<<<<<<<<<<<<<
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hapflk.fastphase.modParams.initUpdate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":84
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):             # <<<<<<<<<<<<<<
 *         self.top += t
 *         self.bot += b
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_5addIndivFit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9modParams_5addIndivFit = {"addIndivFit", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9modParams_5addIndivFit, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_5addIndivFit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_t = 0;
  PyObject *__pyx_v_b = 0;
  PyObject *__pyx_v_j = 0;
  PyObject *__pyx_v_nhap = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addIndivFit (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_t,&__pyx_n_s_b,&__pyx_n_s_j,&__pyx_n_s_nhap,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addIndivFit", 1, 5, 5, 1); __PYX_ERR(0, 84, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addIndivFit", 1, 5, 5, 2); __PYX_ERR(0, 84, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addIndivFit", 1, 5, 5, 3); __PYX_ERR(0, 84, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nhap)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addIndivFit", 1, 5, 5, 4); __PYX_ERR(0, 84, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addIndivFit") < 0)) __PYX_ERR(0, 84, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_self = values[0];
    __pyx_v_t = values[1];
    __pyx_v_b = values[2];
    __pyx_v_j = values[3];
    __pyx_v_nhap = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addIndivFit", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 84, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.modParams.addIndivFit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9modParams_4addIndivFit(__pyx_self, __pyx_v_self, __pyx_v_t, __pyx_v_b, __pyx_v_j, __pyx_v_nhap);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_4addIndivFit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_t, PyObject *__pyx_v_b, PyObject *__pyx_v_j, PyObject *__pyx_v_nhap) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("addIndivFit", 0);

  /* "fastphase/fastphase.pyx":85
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):
 *         self.top += t             # <<<<<<<<<<<<<<
 *         self.bot += b
 *         self.jmk += j
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_top); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_v_t); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_top, __pyx_t_2) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":86
 *     def addIndivFit(self,t,b,j,nhap):
 *         self.top += t
 *         self.bot += b             # <<<<<<<<<<<<<<
 *         self.jmk += j
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bot); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_v_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bot, __pyx_t_1) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":87
 *         self.top += t
 *         self.bot += b
 *         self.jmk += j             # <<<<<<<<<<<<<<
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_jmk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_jmk, __pyx_t_2) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":88
 *         self.bot += b
 *         self.jmk += j
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))             # <<<<<<<<<<<<<<
 *         self.nhap+=nhap
 *     def update(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_jm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sum); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_j);
  __Pyx_GIVEREF(__pyx_v_j);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_j);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_axis, __pyx_int_1) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_7, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_7, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_jm, __pyx_t_4) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":89
 *         self.jmk += j
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap             # <<<<<<<<<<<<<<
 *     def update(self):
 *         ''' Update parameters using top,bot,jmk jm probabilities'''
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nhap); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_v_nhap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nhap, __pyx_t_1) < 0) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":84
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):             # <<<<<<<<<<<<<<
 *         self.top += t
 *         self.bot += b
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hapflk.fastphase.modParams.addIndivFit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":90
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap
 *     def update(self):             # <<<<<<<<<<<<<<
 *         ''' Update parameters using top,bot,jmk jm probabilities'''
 *         ## rho
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_7update(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_6hapflk_9fastphase_9modParams_6update[] = " Update parameters using top,bot,jmk jm probabilities";
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9modParams_7update = {"update", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9modParams_7update, METH_O, __pyx_doc_6hapflk_9fastphase_9modParams_6update};
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_7update(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9modParams_6update(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_6update(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("update", 0);

  /* "fastphase/fastphase.pyx":94
 *         ## rho
 *         ##self.rho=self.jm/self.nhap
 *         for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *             self.rho[i,0]=self.jm[i,0]/self.nhap
 *             if self.rho[i,0]<self.rhomin:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 94, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 94, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 94, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":95
 *         ##self.rho=self.jm/self.nhap
 *         for i in range(self.nLoc):
 *             self.rho[i,0]=self.jm[i,0]/self.nhap             # <<<<<<<<<<<<<<
 *             if self.rho[i,0]<self.rhomin:
 *                 self.rho[i,0]=self.rhomin
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_jm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nhap); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
    if (unlikely(PyObject_SetItem(__pyx_t_5, __pyx_t_6, __pyx_t_2) < 0)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":96
 *         for i in range(self.nLoc):
 *             self.rho[i,0]=self.jm[i,0]/self.nhap
 *             if self.rho[i,0]<self.rhomin:             # <<<<<<<<<<<<<<
 *                 self.rho[i,0]=self.rhomin
 *             elif self.rho[i,0]>(1-self.rhomin):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rhomin); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {

      /* "fastphase/fastphase.pyx":97
 *             self.rho[i,0]=self.jm[i,0]/self.nhap
 *             if self.rho[i,0]<self.rhomin:
 *                 self.rho[i,0]=self.rhomin             # <<<<<<<<<<<<<<
 *             elif self.rho[i,0]>(1-self.rhomin):
 *                 self.rho[i,0]=1-self.rhomin
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rhomin); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 97, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 97, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
      if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_t_5, __pyx_t_2) < 0)) __PYX_ERR(0, 97, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":96
 *         for i in range(self.nLoc):
 *             self.rho[i,0]=self.jm[i,0]/self.nhap
 *             if self.rho[i,0]<self.rhomin:             # <<<<<<<<<<<<<<
 *                 self.rho[i,0]=self.rhomin
 *             elif self.rho[i,0]>(1-self.rhomin):
 */
      goto __pyx_L5;
    }

    /* "fastphase/fastphase.pyx":98
 *             if self.rho[i,0]<self.rhomin:
 *                 self.rho[i,0]=self.rhomin
 *             elif self.rho[i,0]>(1-self.rhomin):             # <<<<<<<<<<<<<<
 *                 self.rho[i,0]=1-self.rhomin
 *         ## alpha
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rhomin); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyInt_SubtractCObj(__pyx_int_1, __pyx_t_5, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "fastphase/fastphase.pyx":99
 *                 self.rho[i,0]=self.rhomin
 *             elif self.rho[i,0]>(1-self.rhomin):
 *                 self.rho[i,0]=1-self.rhomin             # <<<<<<<<<<<<<<
 *         ## alpha
 *         #self.alpha=self.jmk/self.jm
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rhomin); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyInt_SubtractCObj(__pyx_int_1, __pyx_t_5, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
      if (unlikely(PyObject_SetItem(__pyx_t_5, __pyx_t_6, __pyx_t_2) < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":98
 *             if self.rho[i,0]<self.rhomin:
 *                 self.rho[i,0]=self.rhomin
 *             elif self.rho[i,0]>(1-self.rhomin):             # <<<<<<<<<<<<<<
 *                 self.rho[i,0]=1-self.rhomin
 *         ## alpha
 */
    }
    __pyx_L5:;

    /* "fastphase/fastphase.pyx":94
 *         ## rho
 *         ##self.rho=self.jm/self.nhap
 *         for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *             self.rho[i,0]=self.jm[i,0]/self.nhap
 *             if self.rho[i,0]<self.rhomin:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":102
 *         ## alpha
 *         #self.alpha=self.jmk/self.jm
 *         if self.alpha_up:             # <<<<<<<<<<<<<<
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "fastphase/fastphase.pyx":103
 *         #self.alpha=self.jmk/self.jm
 *         if self.alpha_up:
 *             for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *                 for j in range(self.nClus):
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 103, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 103, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 103, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":104
 *         if self.alpha_up:
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):             # <<<<<<<<<<<<<<
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 *                     if self.alpha[i,j]>=0.999:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_2 = __pyx_t_6; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 104, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 104, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 104, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_9(__pyx_t_2);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 104, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "fastphase/fastphase.pyx":105
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]             # <<<<<<<<<<<<<<
 *                     if self.alpha[i,j]>=0.999:
 *                         self.alpha[i,j]=0.999
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_jmk); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_j);
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_jm); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_0);
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_j);
        if (unlikely(PyObject_SetItem(__pyx_t_11, __pyx_t_10, __pyx_t_6) < 0)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "fastphase/fastphase.pyx":106
 *                 for j in range(self.nClus):
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 *                     if self.alpha[i,j]>=0.999:             # <<<<<<<<<<<<<<
 *                         self.alpha[i,j]=0.999
 *                     elif self.alpha[i,j]<0.001:
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_j);
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyObject_RichCompare(__pyx_t_11, __pyx_float_0_999, Py_GE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_7) {

          /* "fastphase/fastphase.pyx":107
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 *                     if self.alpha[i,j]>=0.999:
 *                         self.alpha[i,j]=0.999             # <<<<<<<<<<<<<<
 *                     elif self.alpha[i,j]<0.001:
 *                         self.alpha[i,j]=0.001
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_j);
          __Pyx_GIVEREF(__pyx_v_j);
          PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_j);
          if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_t_11, __pyx_float_0_999) < 0)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "fastphase/fastphase.pyx":106
 *                 for j in range(self.nClus):
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 *                     if self.alpha[i,j]>=0.999:             # <<<<<<<<<<<<<<
 *                         self.alpha[i,j]=0.999
 *                     elif self.alpha[i,j]<0.001:
 */
          goto __pyx_L11;
        }

        /* "fastphase/fastphase.pyx":108
 *                     if self.alpha[i,j]>=0.999:
 *                         self.alpha[i,j]=0.999
 *                     elif self.alpha[i,j]<0.001:             # <<<<<<<<<<<<<<
 *                         self.alpha[i,j]=0.001
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_j);
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyObject_RichCompare(__pyx_t_6, __pyx_float_0_001, Py_LT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_7) {

          /* "fastphase/fastphase.pyx":109
 *                         self.alpha[i,j]=0.999
 *                     elif self.alpha[i,j]<0.001:
 *                         self.alpha[i,j]=0.001             # <<<<<<<<<<<<<<
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])
 *         ## theta
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 109, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 109, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_j);
          __Pyx_GIVEREF(__pyx_v_j);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_j);
          if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_t_6, __pyx_float_0_001) < 0)) __PYX_ERR(0, 109, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "fastphase/fastphase.pyx":108
 *                     if self.alpha[i,j]>=0.999:
 *                         self.alpha[i,j]=0.999
 *                     elif self.alpha[i,j]<0.001:             # <<<<<<<<<<<<<<
 *                         self.alpha[i,j]=0.001
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])
 */
        }
        __pyx_L11:;

        /* "fastphase/fastphase.pyx":104
 *         if self.alpha_up:
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):             # <<<<<<<<<<<<<<
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 *                     if self.alpha[i,j]>=0.999:
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":110
 *                     elif self.alpha[i,j]<0.001:
 *                         self.alpha[i,j]=0.001
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])             # <<<<<<<<<<<<<<
 *         ## theta
 *         if self.theta_up:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
      __Pyx_INCREF(__pyx_slice_);
      __Pyx_GIVEREF(__pyx_slice_);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_slice_);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sum); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_i);
      __Pyx_INCREF(__pyx_slice_);
      __Pyx_GIVEREF(__pyx_slice_);
      PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_slice_);
      __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_14);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_PyNumber_InPlaceDivide(__pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_t_6, __pyx_t_12) < 0)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":103
 *         #self.alpha=self.jmk/self.jm
 *         if self.alpha_up:
 *             for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *                 for j in range(self.nClus):
 *                     self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":102
 *         ## alpha
 *         #self.alpha=self.jmk/self.jm
 *         if self.alpha_up:             # <<<<<<<<<<<<<<
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 */
  }

  /* "fastphase/fastphase.pyx":112
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])
 *         ## theta
 *         if self.theta_up:             # <<<<<<<<<<<<<<
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "fastphase/fastphase.pyx":113
 *         ## theta
 *         if self.theta_up:
 *             self.theta=self.top/self.bot             # <<<<<<<<<<<<<<
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_top); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bot); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_theta, __pyx_t_6) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":114
 *         if self.theta_up:
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *                 for j in range(self.nClus):
 *                     if self.theta[i,j]>0.999:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_6 = __pyx_t_2; __Pyx_INCREF(__pyx_t_6); __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 114, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 114, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 114, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_4(__pyx_t_6);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 114, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":115
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):             # <<<<<<<<<<<<<<
 *                     if self.theta[i,j]>0.999:
 *                         self.theta[i,j]=0.999
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 115, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_9(__pyx_t_2);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 115, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "fastphase/fastphase.pyx":116
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 *                     if self.theta[i,j]>0.999:             # <<<<<<<<<<<<<<
 *                         self.theta[i,j]=0.999
 *                     elif self.theta[i,j]<0.001:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_j);
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = PyObject_RichCompare(__pyx_t_11, __pyx_float_0_999, Py_GT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (__pyx_t_7) {

          /* "fastphase/fastphase.pyx":117
 *                 for j in range(self.nClus):
 *                     if self.theta[i,j]>0.999:
 *                         self.theta[i,j]=0.999             # <<<<<<<<<<<<<<
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_j);
          __Pyx_GIVEREF(__pyx_v_j);
          PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_j);
          if (unlikely(PyObject_SetItem(__pyx_t_12, __pyx_t_11, __pyx_float_0_999) < 0)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "fastphase/fastphase.pyx":116
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):
 *                     if self.theta[i,j]>0.999:             # <<<<<<<<<<<<<<
 *                         self.theta[i,j]=0.999
 *                     elif self.theta[i,j]<0.001:
 */
          goto __pyx_L17;
        }

        /* "fastphase/fastphase.pyx":118
 *                     if self.theta[i,j]>0.999:
 *                         self.theta[i,j]=0.999
 *                     elif self.theta[i,j]<0.001:             # <<<<<<<<<<<<<<
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_j);
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = PyObject_RichCompare(__pyx_t_1, __pyx_float_0_001, Py_LT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (__pyx_t_7) {

          /* "fastphase/fastphase.pyx":119
 *                         self.theta[i,j]=0.999
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001             # <<<<<<<<<<<<<<
 *     def write(self,stream=sys.stdout):
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 119, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_j);
          __Pyx_GIVEREF(__pyx_v_j);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_j);
          if (unlikely(PyObject_SetItem(__pyx_t_12, __pyx_t_1, __pyx_float_0_001) < 0)) __PYX_ERR(0, 119, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "fastphase/fastphase.pyx":118
 *                     if self.theta[i,j]>0.999:
 *                         self.theta[i,j]=0.999
 *                     elif self.theta[i,j]<0.001:             # <<<<<<<<<<<<<<
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):
 */
        }
        __pyx_L17:;

        /* "fastphase/fastphase.pyx":115
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):
 *                 for j in range(self.nClus):             # <<<<<<<<<<<<<<
 *                     if self.theta[i,j]>0.999:
 *                         self.theta[i,j]=0.999
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":114
 *         if self.theta_up:
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *                 for j in range(self.nClus):
 *                     if self.theta[i,j]>0.999:
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":112
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])
 *         ## theta
 *         if self.theta_up:             # <<<<<<<<<<<<<<
 *             self.theta=self.top/self.bot
 *             for i in range(self.nLoc):
 */
  }

  /* "fastphase/fastphase.pyx":90
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap
 *     def update(self):             # <<<<<<<<<<<<<<
 *         ''' Update parameters using top,bot,jmk jm probabilities'''
 *         ## rho
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hapflk.fastphase.modParams.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":120
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):             # <<<<<<<<<<<<<<
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):
 */

static PyObject *__pyx_pf_6hapflk_9fastphase_25__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_stream);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_stream);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_stream);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hapflk.fastphase.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_9write(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9modParams_9write = {"write", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9modParams_9write, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9modParams_9write(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_stream = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_stream;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "write") < 0)) __PYX_ERR(0, 120, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_stream = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.modParams.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9modParams_8write(__pyx_self, __pyx_v_self, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9modParams_8write(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_7genexpr__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_k = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("write", 0);

  /* "fastphase/fastphase.pyx":121
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)             # <<<<<<<<<<<<<<
 *         for i in range(self.nLoc):
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 121, __pyx_L5_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 121, __pyx_L5_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 121, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_7genexpr__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_n_u_t, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 121, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_tuple__2, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 121, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 121, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_6);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 121, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_8genexpr1__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_n_u_a, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 121, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 121, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __pyx_t_6 = PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_file, __pyx_v_stream) < 0) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":122
 *     def write(self,stream=sys.stdout):
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_2 = __pyx_t_6; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 122, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":123
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)             # <<<<<<<<<<<<<<
 * 
 * _tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))
 */
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_i);
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_7 = __pyx_t_3; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 123, __pyx_L19_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L19_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L19_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_9(__pyx_t_7);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 123, __pyx_L19_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_k, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_round); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_theta); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_8genexpr2__pyx_v_k);
        __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_8genexpr2__pyx_v_k);
        __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        __pyx_t_14 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_14 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_13, __pyx_int_3};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_13, __pyx_int_3};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_14, __pyx_t_13);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_14, __pyx_int_3);
          __pyx_t_13 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 123, __pyx_L19_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
      goto __pyx_L22_exit_scope;
      __pyx_L19_error:;
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
      goto __pyx_L1_error;
      __pyx_L22_exit_scope:;
    } /* exit inner scope */
    __pyx_t_7 = PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_round); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rho); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_int_0);
    __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    __pyx_t_14 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_14 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_10, __pyx_int_7};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_10, __pyx_int_7};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_14, __pyx_t_10);
      __Pyx_INCREF(__pyx_int_7);
      __Pyx_GIVEREF(__pyx_int_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_14, __pyx_int_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nClus); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_1 = __pyx_t_6; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 123, __pyx_L25_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L25_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L25_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_9(__pyx_t_1);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 123, __pyx_L25_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_k, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_round); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_i);
        __Pyx_INCREF(__pyx_8genexpr3__pyx_v_k);
        __Pyx_GIVEREF(__pyx_8genexpr3__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_8genexpr3__pyx_v_k);
        __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = NULL;
        __pyx_t_14 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_14 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_12, __pyx_int_3};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_12, __pyx_int_3};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 123, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_14, __pyx_t_12);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_14, __pyx_int_3);
          __pyx_t_12 = 0;
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 123, __pyx_L25_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_k); __pyx_8genexpr3__pyx_v_k = 0;
      goto __pyx_L28_exit_scope;
      __pyx_L25_error:;
      __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_k); __pyx_8genexpr3__pyx_v_k = 0;
      goto __pyx_L1_error;
      __pyx_L28_exit_scope:;
    } /* exit inner scope */
    __pyx_t_1 = PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_file, __pyx_v_stream) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":122
 *     def write(self,stream=sys.stdout):
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):             # <<<<<<<<<<<<<<
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":120
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):             # <<<<<<<<<<<<<<
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hapflk.fastphase.modParams.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":136
 *          ... do stuff ...
 *     '''
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):             # <<<<<<<<<<<<<<
 *         assert nLoci>0
 *         self.nLoci=nLoci
 */

static PyObject *__pyx_pf_6hapflk_9fastphase_27__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_nproc);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_nproc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_nproc);
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_None));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hapflk.fastphase.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nLoci = 0;
  PyObject *__pyx_v_nproc = 0;
  PyObject *__pyx_v_prfx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nLoci,&__pyx_n_s_nproc,&__pyx_n_s_prfx,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults1 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_nproc;
    values[3] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nLoci)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nproc);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_prfx);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 136, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_nLoci = values[1];
    __pyx_v_nproc = values[2];
    __pyx_v_prfx = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase___init__(__pyx_self, __pyx_v_self, __pyx_v_nLoci, __pyx_v_nproc, __pyx_v_prfx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nLoci, PyObject *__pyx_v_nproc, PyObject *__pyx_v_prfx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fastphase/fastphase.pyx":137
 *     '''
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):
 *         assert nLoci>0             # <<<<<<<<<<<<<<
 *         self.nLoci=nLoci
 *         self.haplotypes={}
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_nLoci, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 137, __pyx_L1_error)
    }
  }
  #endif

  /* "fastphase/fastphase.pyx":138
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):
 *         assert nLoci>0
 *         self.nLoci=nLoci             # <<<<<<<<<<<<<<
 *         self.haplotypes={}
 *         self.genotypes={}
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nLoci, __pyx_v_nLoci) < 0) __PYX_ERR(0, 138, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":139
 *         assert nLoci>0
 *         self.nLoci=nLoci
 *         self.haplotypes={}             # <<<<<<<<<<<<<<
 *         self.genotypes={}
 *         self.nproc = nproc
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes, __pyx_t_1) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":140
 *         self.nLoci=nLoci
 *         self.haplotypes={}
 *         self.genotypes={}             # <<<<<<<<<<<<<<
 *         self.nproc = nproc
 *         self.pool = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_genotypes, __pyx_t_1) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":141
 *         self.haplotypes={}
 *         self.genotypes={}
 *         self.nproc = nproc             # <<<<<<<<<<<<<<
 *         self.pool = None
 *         self.prfx = prfx
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nproc, __pyx_v_nproc) < 0) __PYX_ERR(0, 141, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":142
 *         self.genotypes={}
 *         self.nproc = nproc
 *         self.pool = None             # <<<<<<<<<<<<<<
 *         self.prfx = prfx
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pool, Py_None) < 0) __PYX_ERR(0, 142, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":143
 *         self.nproc = nproc
 *         self.pool = None
 *         self.prfx = prfx             # <<<<<<<<<<<<<<
 * 
 *     def __enter__(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_prfx, __pyx_v_prfx) < 0) __PYX_ERR(0, 143, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":136
 *          ... do stuff ...
 *     '''
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):             # <<<<<<<<<<<<<<
 *         assert nLoci>0
 *         self.nLoci=nLoci
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":145
 *         self.prfx = prfx
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_3__enter__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_3__enter__ = {"__enter__", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9fastphase_3__enter__, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_3__enter__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_2__enter__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_2__enter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__enter__", 0);

  /* "fastphase/fastphase.pyx":146
 * 
 *     def __enter__(self):
 *         self.pool = Pool ( self.nproc)             # <<<<<<<<<<<<<<
 *         if self.prfx is None:
 *             self.flog = sys.stdout
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Pool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nproc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pool, __pyx_t_1) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":147
 *     def __enter__(self):
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:             # <<<<<<<<<<<<<<
 *             self.flog = sys.stdout
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_prfx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "fastphase/fastphase.pyx":148
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:
 *             self.flog = sys.stdout             # <<<<<<<<<<<<<<
 *         else:
 *             self.flog = open(self.prfx, 'w')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_flog, __pyx_t_2) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":147
 *     def __enter__(self):
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:             # <<<<<<<<<<<<<<
 *             self.flog = sys.stdout
 *         else:
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":150
 *             self.flog = sys.stdout
 *         else:
 *             self.flog = open(self.prfx, 'w')             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_prfx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_u_w);
    __Pyx_GIVEREF(__pyx_n_u_w);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_w);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_flog, __pyx_t_2) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":151
 *         else:
 *             self.flog = open(self.prfx, 'w')
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self,*args):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":145
 *         self.prfx = prfx
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.__enter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":153
 *         return self
 * 
 *     def __exit__(self,*args):             # <<<<<<<<<<<<<<
 *         self.pool.terminate()
 *         if self.prfx is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_5__exit__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_5__exit__ = {"__exit__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_5__exit__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_5__exit__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, used_pos_args, "__exit__") < 0)) __PYX_ERR(0, 153, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__exit__", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 153, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_4__exit__(__pyx_self, __pyx_v_self, __pyx_v_args);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_4__exit__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__exit__", 0);

  /* "fastphase/fastphase.pyx":154
 * 
 *     def __exit__(self,*args):
 *         self.pool.terminate()             # <<<<<<<<<<<<<<
 *         if self.prfx is not None:
 *             self.flog.close()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_terminate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":155
 *     def __exit__(self,*args):
 *         self.pool.terminate()
 *         if self.prfx is not None:             # <<<<<<<<<<<<<<
 *             self.flog.close()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_prfx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "fastphase/fastphase.pyx":156
 *         self.pool.terminate()
 *         if self.prfx is not None:
 *             self.flog.close()             # <<<<<<<<<<<<<<
 * 
 *     def flush(self):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":155
 *     def __exit__(self,*args):
 *         self.pool.terminate()
 *         if self.prfx is not None:             # <<<<<<<<<<<<<<
 *             self.flog.close()
 * 
 */
  }

  /* "fastphase/fastphase.pyx":153
 *         return self
 * 
 *     def __exit__(self,*args):             # <<<<<<<<<<<<<<
 *         self.pool.terminate()
 *         if self.prfx is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":158
 *             self.flog.close()
 * 
 *     def flush(self):             # <<<<<<<<<<<<<<
 *         '''
 *         remove data
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_7flush(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_6hapflk_9fastphase_9fastphase_6flush[] = "\n        remove data \n        ";
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_7flush = {"flush", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9fastphase_7flush, METH_O, __pyx_doc_6hapflk_9fastphase_9fastphase_6flush};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_7flush(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flush (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_6flush(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_6flush(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("flush", 0);

  /* "fastphase/fastphase.pyx":162
 *         remove data
 *         '''
 *         self.haplotypes={}             # <<<<<<<<<<<<<<
 *         self.genotypes={}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes, __pyx_t_1) < 0) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":163
 *         '''
 *         self.haplotypes={}
 *         self.genotypes={}             # <<<<<<<<<<<<<<
 * 
 *     def addHaplotype(self,ID,hap,missing=-1):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_genotypes, __pyx_t_1) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":158
 *             self.flog.close()
 * 
 *     def flush(self):             # <<<<<<<<<<<<<<
 *         '''
 *         remove data
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":165
 *         self.genotypes={}
 * 
 *     def addHaplotype(self,ID,hap,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add an haplotype to the model observations.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_9addHaplotype(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6hapflk_9fastphase_9fastphase_8addHaplotype[] = "\n        Add an haplotype to the model observations.\n        hap is a numpy array of shape (1,nLoci).\n        Values must be 0,1 or missing\n        ";
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_9addHaplotype = {"addHaplotype", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_9addHaplotype, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6hapflk_9fastphase_9fastphase_8addHaplotype};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_9addHaplotype(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ID = 0;
  PyObject *__pyx_v_hap = 0;
  CYTHON_UNUSED PyObject *__pyx_v_missing = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addHaplotype (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ID,&__pyx_n_s_hap,&__pyx_n_s_missing,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ID)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addHaplotype", 0, 3, 4, 1); __PYX_ERR(0, 165, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hap)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addHaplotype", 0, 3, 4, 2); __PYX_ERR(0, 165, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_missing);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addHaplotype") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_ID = values[1];
    __pyx_v_hap = values[2];
    __pyx_v_missing = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addHaplotype", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.addHaplotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_8addHaplotype(__pyx_self, __pyx_v_self, __pyx_v_ID, __pyx_v_hap, __pyx_v_missing);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_8addHaplotype(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ID, PyObject *__pyx_v_hap, CYTHON_UNUSED PyObject *__pyx_v_missing) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("addHaplotype", 0);

  /* "fastphase/fastphase.pyx":171
 *         Values must be 0,1 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert hap.shape[0]==self.nLoci
 *             self.haplotypes[ID]=hap
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastphase/fastphase.pyx":172
 *         '''
 *         try:
 *             assert hap.shape[0]==self.nLoci             # <<<<<<<<<<<<<<
 *             self.haplotypes[ID]=hap
 *         except AssertionError:
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_hap, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 172, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_7)) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 172, __pyx_L3_error)
        }
      }
      #endif

      /* "fastphase/fastphase.pyx":173
 *         try:
 *             assert hap.shape[0]==self.nLoci
 *             self.haplotypes[ID]=hap             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 173, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_v_ID, __pyx_v_hap) < 0)) __PYX_ERR(0, 173, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fastphase/fastphase.pyx":171
 *         Values must be 0,1 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert hap.shape[0]==self.nLoci
 *             self.haplotypes[ID]=hap
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":174
 *             assert hap.shape[0]==self.nLoci
 *             self.haplotypes[ID]=hap
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("hapflk.fastphase.fastphase.addHaplotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 174, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);

      /* "fastphase/fastphase.pyx":175
 *             self.haplotypes[ID]=hap
 *         except AssertionError:
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)             # <<<<<<<<<<<<<<
 *             raise
 *     def addGenotype(self,ID,gen,missing=-1):
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_hap, __pyx_n_s_shape); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_9, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = PyTuple_New(4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_kp_u_Wrong_Haplotype_Size);
      __Pyx_GIVEREF(__pyx_kp_u_Wrong_Haplotype_Size);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_kp_u_Wrong_Haplotype_Size);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_10);
      __Pyx_INCREF(__pyx_kp_u_is_not);
      __Pyx_GIVEREF(__pyx_kp_u_is_not);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_kp_u_is_not);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_9);
      __pyx_t_10 = 0;
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_11, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":176
 *         except AssertionError:
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise             # <<<<<<<<<<<<<<
 *     def addGenotype(self,ID,gen,missing=-1):
 *         '''
 */
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_5);
      __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
      __PYX_ERR(0, 176, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastphase/fastphase.pyx":171
 *         Values must be 0,1 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert hap.shape[0]==self.nLoci
 *             self.haplotypes[ID]=hap
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "fastphase/fastphase.pyx":165
 *         self.genotypes={}
 * 
 *     def addHaplotype(self,ID,hap,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add an haplotype to the model observations.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.addHaplotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":177
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise
 *     def addGenotype(self,ID,gen,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add a genotype to the model observations.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_11addGenotype(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6hapflk_9fastphase_9fastphase_10addGenotype[] = "\n        Add a genotype to the model observations.\n        gen is a numpy array of shape (1,nLoci).\n        Values must be 0,1,2 or missing\n        ";
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_11addGenotype = {"addGenotype", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_11addGenotype, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6hapflk_9fastphase_9fastphase_10addGenotype};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_11addGenotype(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ID = 0;
  PyObject *__pyx_v_gen = 0;
  CYTHON_UNUSED PyObject *__pyx_v_missing = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addGenotype (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ID,&__pyx_n_s_gen,&__pyx_n_s_missing,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ID)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addGenotype", 0, 3, 4, 1); __PYX_ERR(0, 177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gen)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addGenotype", 0, 3, 4, 2); __PYX_ERR(0, 177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_missing);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addGenotype") < 0)) __PYX_ERR(0, 177, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_ID = values[1];
    __pyx_v_gen = values[2];
    __pyx_v_missing = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addGenotype", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 177, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.addGenotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_10addGenotype(__pyx_self, __pyx_v_self, __pyx_v_ID, __pyx_v_gen, __pyx_v_missing);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_10addGenotype(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ID, PyObject *__pyx_v_gen, CYTHON_UNUSED PyObject *__pyx_v_missing) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("addGenotype", 0);

  /* "fastphase/fastphase.pyx":183
 *         Values must be 0,1,2 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert gen.shape[0]==self.nLoci
 *             self.genotypes[ID]=gen
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastphase/fastphase.pyx":184
 *         '''
 *         try:
 *             assert gen.shape[0]==self.nLoci             # <<<<<<<<<<<<<<
 *             self.genotypes[ID]=gen
 *         except AssertionError:
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gen, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 184, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 184, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_7)) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 184, __pyx_L3_error)
        }
      }
      #endif

      /* "fastphase/fastphase.pyx":185
 *         try:
 *             assert gen.shape[0]==self.nLoci
 *             self.genotypes[ID]=gen             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             print("Wrong Genotype Size:",gen.shape[0],"is not",self.nLoci)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_v_ID, __pyx_v_gen) < 0)) __PYX_ERR(0, 185, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fastphase/fastphase.pyx":183
 *         Values must be 0,1,2 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert gen.shape[0]==self.nLoci
 *             self.genotypes[ID]=gen
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":186
 *             assert gen.shape[0]==self.nLoci
 *             self.genotypes[ID]=gen
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             print("Wrong Genotype Size:",gen.shape[0],"is not",self.nLoci)
 *             raise
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("hapflk.fastphase.fastphase.addGenotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 186, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);

      /* "fastphase/fastphase.pyx":187
 *             self.genotypes[ID]=gen
 *         except AssertionError:
 *             print("Wrong Genotype Size:",gen.shape[0],"is not",self.nLoci)             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_gen, __pyx_n_s_shape); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_9, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 187, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = PyTuple_New(4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 187, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_kp_u_Wrong_Genotype_Size);
      __Pyx_GIVEREF(__pyx_kp_u_Wrong_Genotype_Size);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_kp_u_Wrong_Genotype_Size);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_10);
      __Pyx_INCREF(__pyx_kp_u_is_not);
      __Pyx_GIVEREF(__pyx_kp_u_is_not);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_kp_u_is_not);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_9);
      __pyx_t_10 = 0;
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_11, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":188
 *         except AssertionError:
 *             print("Wrong Genotype Size:",gen.shape[0],"is not",self.nLoci)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_5);
      __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
      __PYX_ERR(0, 188, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastphase/fastphase.pyx":183
 *         Values must be 0,1,2 or missing
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert gen.shape[0]==self.nLoci
 *             self.genotypes[ID]=gen
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "fastphase/fastphase.pyx":177
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise
 *     def addGenotype(self,ID,gen,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add a genotype to the model observations.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.addGenotype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":191
 * 
 *     @staticmethod
 *     def gen2hap(gen):             # <<<<<<<<<<<<<<
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_13gen2hap(PyObject *__pyx_self, PyObject *__pyx_v_gen); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_13gen2hap = {"gen2hap", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9fastphase_13gen2hap, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_13gen2hap(PyObject *__pyx_self, PyObject *__pyx_v_gen) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gen2hap (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_12gen2hap(__pyx_self, ((PyObject *)__pyx_v_gen));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_12gen2hap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_gen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("gen2hap", 0);

  /* "fastphase/fastphase.pyx":192
 *     @staticmethod
 *     def gen2hap(gen):
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)             # <<<<<<<<<<<<<<
 * 
 *     def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_tohap); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_gen);
  __Pyx_GIVEREF(__pyx_v_gen);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_gen);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":191
 * 
 *     @staticmethod
 *     def gen2hap(gen):             # <<<<<<<<<<<<<<
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.gen2hap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":194
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 *     def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Fit the model on observations with nCLus clusters using nstep EM iterations
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_15fit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6hapflk_9fastphase_9fastphase_14fit[] = "\n        Fit the model on observations with nCLus clusters using nstep EM iterations\n        Multithread version.\n        ";
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_15fit = {"fit", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_15fit, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6hapflk_9fastphase_9fastphase_14fit};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_15fit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nClus = 0;
  PyObject *__pyx_v_nstep = 0;
  PyObject *__pyx_v_params = 0;
  PyObject *__pyx_v_verbose = 0;
  PyObject *__pyx_v_rhomin = 0;
  PyObject *__pyx_v_alpha_up = 0;
  PyObject *__pyx_v_theta_up = 0;
  PyObject *__pyx_v_fast = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fit (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nClus,&__pyx_n_s_nstep,&__pyx_n_s_params,&__pyx_n_s_verbose,&__pyx_n_s_rhomin,&__pyx_n_s_alpha_up,&__pyx_n_s_theta_up,&__pyx_n_s_fast,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_20));
    values[2] = ((PyObject *)((PyObject *)__pyx_int_20));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    values[5] = ((PyObject *)((PyObject*)__pyx_float_1eneg_6));
    values[6] = ((PyObject *)((PyObject *)Py_True));
    values[7] = ((PyObject *)((PyObject *)Py_True));
    values[8] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nClus);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nstep);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rhomin);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alpha_up);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta_up);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fast);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fit") < 0)) __PYX_ERR(0, 194, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_nClus = values[1];
    __pyx_v_nstep = values[2];
    __pyx_v_params = values[3];
    __pyx_v_verbose = values[4];
    __pyx_v_rhomin = values[5];
    __pyx_v_alpha_up = values[6];
    __pyx_v_theta_up = values[7];
    __pyx_v_fast = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fit", 0, 1, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 194, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.fit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_14fit(__pyx_self, __pyx_v_self, __pyx_v_nClus, __pyx_v_nstep, __pyx_v_params, __pyx_v_verbose, __pyx_v_rhomin, __pyx_v_alpha_up, __pyx_v_theta_up, __pyx_v_fast);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_14fit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nClus, PyObject *__pyx_v_nstep, PyObject *__pyx_v_params, PyObject *__pyx_v_verbose, PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_theta_up, PyObject *__pyx_v_fast) {
  PyObject *__pyx_v_par = NULL;
  CYTHON_UNUSED long __pyx_v_old_log_like;
  PyObject *__pyx_v_iEM = NULL;
  PyObject *__pyx_v_log_like = NULL;
  PyObject *__pyx_v_tasks = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_hap = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_gen = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *(*__pyx_t_22)(PyObject *);
  __Pyx_RefNannySetupContext("fit", 0);

  /* "fastphase/fastphase.pyx":199
 *         Multithread version.
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.pool is not None
 *         except AssertionError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastphase/fastphase.pyx":200
 *         '''
 *         try:
 *             assert self.pool is not None             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 200, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = (__pyx_t_4 != Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!(__pyx_t_5 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 200, __pyx_L3_error)
        }
      }
      #endif

      /* "fastphase/fastphase.pyx":199
 *         Multithread version.
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.pool is not None
 *         except AssertionError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":201
 *         try:
 *             assert self.pool is not None
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')
 *             raise
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("hapflk.fastphase.fastphase.fit", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 201, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "fastphase/fastphase.pyx":202
 *             assert self.pool is not None
 *         except AssertionError:
 *             print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 202, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":203
 *         except AssertionError:
 *             print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if params:
 */
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_7, __pyx_t_8);
      __pyx_t_4 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; 
      __PYX_ERR(0, 203, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastphase/fastphase.pyx":199
 *         Multithread version.
 *         '''
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.pool is not None
 *         except AssertionError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "fastphase/fastphase.pyx":205
 *             raise
 * 
 *         if params:             # <<<<<<<<<<<<<<
 *             par=params
 *             par.alpha_up = alpha_up
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 205, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "fastphase/fastphase.pyx":206
 * 
 *         if params:
 *             par=params             # <<<<<<<<<<<<<<
 *             par.alpha_up = alpha_up
 *             par.theta_up = theta_up
 */
    __Pyx_INCREF(__pyx_v_params);
    __pyx_v_par = __pyx_v_params;

    /* "fastphase/fastphase.pyx":207
 *         if params:
 *             par=params
 *             par.alpha_up = alpha_up             # <<<<<<<<<<<<<<
 *             par.theta_up = theta_up
 *         else:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_par, __pyx_n_s_alpha_up, __pyx_v_alpha_up) < 0) __PYX_ERR(0, 207, __pyx_L1_error)

    /* "fastphase/fastphase.pyx":208
 *             par=params
 *             par.alpha_up = alpha_up
 *             par.theta_up = theta_up             # <<<<<<<<<<<<<<
 *         else:
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_par, __pyx_n_s_theta_up, __pyx_v_theta_up) < 0) __PYX_ERR(0, 208, __pyx_L1_error)

    /* "fastphase/fastphase.pyx":205
 *             raise
 * 
 *         if params:             # <<<<<<<<<<<<<<
 *             par=params
 *             par.alpha_up = alpha_up
 */
    goto __pyx_L11;
  }

  /* "fastphase/fastphase.pyx":210
 *             par.theta_up = theta_up
 *         else:
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)             # <<<<<<<<<<<<<<
 *         if verbose:
 *             print( 'Fitting fastphase model',file=self.flog)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_modParams); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_4, __pyx_v_nClus, __pyx_v_rhomin, __pyx_v_alpha_up, __pyx_v_theta_up};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_4, __pyx_v_nClus, __pyx_v_rhomin, __pyx_v_alpha_up, __pyx_v_theta_up};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(5+__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_6, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_nClus);
      __Pyx_GIVEREF(__pyx_v_nClus);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_6, __pyx_v_nClus);
      __Pyx_INCREF(__pyx_v_rhomin);
      __Pyx_GIVEREF(__pyx_v_rhomin);
      PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_6, __pyx_v_rhomin);
      __Pyx_INCREF(__pyx_v_alpha_up);
      __Pyx_GIVEREF(__pyx_v_alpha_up);
      PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_6, __pyx_v_alpha_up);
      __Pyx_INCREF(__pyx_v_theta_up);
      __Pyx_GIVEREF(__pyx_v_theta_up);
      PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_6, __pyx_v_theta_up);
      __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_par = __pyx_t_8;
    __pyx_t_8 = 0;
  }
  __pyx_L11:;

  /* "fastphase/fastphase.pyx":211
 *         else:
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print( 'Fitting fastphase model',file=self.flog)
 *             print( '# clusters ',nClus, file=self.flog)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 211, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "fastphase/fastphase.pyx":212
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
 *         if verbose:
 *             print( 'Fitting fastphase model',file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# clusters ',nClus, file=self.flog)
 *             print( '# threads ', self.nproc, file=self.flog)
 */
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_file, __pyx_t_7) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__5, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":213
 *         if verbose:
 *             print( 'Fitting fastphase model',file=self.flog)
 *             print( '# clusters ',nClus, file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# threads ', self.nproc, file=self.flog)
 *             print( '# Loci', self.nLoci, file=self.flog)
 */
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_kp_u_clusters);
    __Pyx_GIVEREF(__pyx_kp_u_clusters);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_u_clusters);
    __Pyx_INCREF(__pyx_v_nClus);
    __Pyx_GIVEREF(__pyx_v_nClus);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nClus);
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_file, __pyx_t_10) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "fastphase/fastphase.pyx":214
 *             print( 'Fitting fastphase model',file=self.flog)
 *             print( '# clusters ',nClus, file=self.flog)
 *             print( '# threads ', self.nproc, file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# Loci', self.nLoci, file=self.flog)
 *             print( '# Haplotypes',len(self.haplotypes), file=self.flog)
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nproc); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_kp_u_threads);
    __Pyx_GIVEREF(__pyx_kp_u_threads);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_u_threads);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_file, __pyx_t_7) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_8, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":215
 *             print( '# clusters ',nClus, file=self.flog)
 *             print( '# threads ', self.nproc, file=self.flog)
 *             print( '# Loci', self.nLoci, file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# Haplotypes',len(self.haplotypes), file=self.flog)
 *             print( '# Genotypes', len(self.genotypes), file=self.flog)
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_kp_u_Loci);
    __Pyx_GIVEREF(__pyx_kp_u_Loci);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_u_Loci);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_file, __pyx_t_8) < 0) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_10, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fastphase/fastphase.pyx":216
 *             print( '# threads ', self.nproc, file=self.flog)
 *             print( '# Loci', self.nLoci, file=self.flog)
 *             print( '# Haplotypes',len(self.haplotypes), file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# Genotypes', len(self.genotypes), file=self.flog)
 *         old_log_like=1
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_11 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_kp_u_Haplotypes);
    __Pyx_GIVEREF(__pyx_kp_u_Haplotypes);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_u_Haplotypes);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_file, __pyx_t_10) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "fastphase/fastphase.pyx":217
 *             print( '# Loci', self.nLoci, file=self.flog)
 *             print( '# Haplotypes',len(self.haplotypes), file=self.flog)
 *             print( '# Genotypes', len(self.genotypes), file=self.flog)             # <<<<<<<<<<<<<<
 *         old_log_like=1
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = PyObject_Length(__pyx_t_10); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_kp_u_Genotypes);
    __Pyx_GIVEREF(__pyx_kp_u_Genotypes);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_u_Genotypes);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_file, __pyx_t_7) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_8, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":211
 *         else:
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print( 'Fitting fastphase model',file=self.flog)
 *             print( '# clusters ',nClus, file=self.flog)
 */
  }

  /* "fastphase/fastphase.pyx":218
 *             print( '# Haplotypes',len(self.haplotypes), file=self.flog)
 *             print( '# Genotypes', len(self.genotypes), file=self.flog)
 *         old_log_like=1             # <<<<<<<<<<<<<<
 * 
 *         for iEM in range(nstep):
 */
  __pyx_v_old_log_like = 1;

  /* "fastphase/fastphase.pyx":220
 *         old_log_like=1
 * 
 *         for iEM in range(nstep):             # <<<<<<<<<<<<<<
 * 
 *             log_like=0.0
 */
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_nstep); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_10 = __pyx_t_7; __Pyx_INCREF(__pyx_t_10); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 220, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_10))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_12(__pyx_t_10);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 220, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_iEM, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":222
 *         for iEM in range(nstep):
 * 
 *             log_like=0.0             # <<<<<<<<<<<<<<
 *             par.initUpdate()
 * 
 */
    __Pyx_INCREF(__pyx_float_0_0);
    __Pyx_XDECREF_SET(__pyx_v_log_like, __pyx_float_0_0);

    /* "fastphase/fastphase.pyx":223
 * 
 *             log_like=0.0
 *             par.initUpdate()             # <<<<<<<<<<<<<<
 * 
 *             tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_initUpdate); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":225
 *             par.initUpdate()
 * 
 *             tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]             # <<<<<<<<<<<<<<
 *             if fast:
 *                 tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]
 */
    { /* enter inner scope */
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 225, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__pyx_t_4 == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
        __PYX_ERR(0, 225, __pyx_L17_error)
      }
      __pyx_t_9 = __Pyx_dict_iterator(__pyx_t_4, 0, __pyx_n_s_values, (&__pyx_t_14), (&__pyx_t_6)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_8);
      __pyx_t_8 = __pyx_t_9;
      __pyx_t_9 = 0;
      while (1) {
        __pyx_t_15 = __Pyx_dict_iter_next(__pyx_t_8, __pyx_t_14, &__pyx_t_13, NULL, &__pyx_t_9, NULL, __pyx_t_6);
        if (unlikely(__pyx_t_15 == 0)) break;
        if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_hap, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_fitInData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_19 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_19)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_19);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[7] = {__pyx_t_19, __pyx_n_u_haplo, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_8genexpr4__pyx_v_hap, __pyx_int_0};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L17_error)
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[7] = {__pyx_t_19, __pyx_n_u_haplo, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_8genexpr4__pyx_v_hap, __pyx_int_0};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L17_error)
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        } else
        #endif
        {
          __pyx_t_20 = PyTuple_New(6+__pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 225, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_20);
          if (__pyx_t_19) {
            __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_19); __pyx_t_19 = NULL;
          }
          __Pyx_INCREF(__pyx_n_u_haplo);
          __Pyx_GIVEREF(__pyx_n_u_haplo);
          PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_15, __pyx_n_u_haplo);
          __Pyx_GIVEREF(__pyx_t_16);
          PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_15, __pyx_t_16);
          __Pyx_GIVEREF(__pyx_t_17);
          PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_15, __pyx_t_17);
          __Pyx_GIVEREF(__pyx_t_18);
          PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_15, __pyx_t_18);
          __Pyx_INCREF(__pyx_8genexpr4__pyx_v_hap);
          __Pyx_GIVEREF(__pyx_8genexpr4__pyx_v_hap);
          PyTuple_SET_ITEM(__pyx_t_20, 4+__pyx_t_15, __pyx_8genexpr4__pyx_v_hap);
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_GIVEREF(__pyx_int_0);
          PyTuple_SET_ITEM(__pyx_t_20, 5+__pyx_t_15, __pyx_int_0);
          __pyx_t_16 = 0;
          __pyx_t_17 = 0;
          __pyx_t_18 = 0;
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_20, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 225, __pyx_L17_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_hap); __pyx_8genexpr4__pyx_v_hap = 0;
      goto __pyx_L20_exit_scope;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_hap); __pyx_8genexpr4__pyx_v_hap = 0;
      goto __pyx_L1_error;
      __pyx_L20_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_tasks, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":226
 * 
 *             tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]
 *             if fast:             # <<<<<<<<<<<<<<
 *                 tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]
 *             else:
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_fast); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 226, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "fastphase/fastphase.pyx":227
 *             tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]
 *             if fast:
 *                 tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]             # <<<<<<<<<<<<<<
 *             else:
 *                 tasks += [ fitInData( 'geno', par.alpha,par.theta,par.rho,gen,0) for gen in  self.genotypes.values()]
 */
      { /* enter inner scope */
        __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 227, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 227, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(__pyx_t_9 == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
          __PYX_ERR(0, 227, __pyx_L24_error)
        }
        __pyx_t_4 = __Pyx_dict_iterator(__pyx_t_9, 0, __pyx_n_s_values, (&__pyx_t_13), (&__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8);
        __pyx_t_8 = __pyx_t_4;
        __pyx_t_4 = 0;
        while (1) {
          __pyx_t_15 = __Pyx_dict_iter_next(__pyx_t_8, __pyx_t_13, &__pyx_t_14, NULL, &__pyx_t_4, NULL, __pyx_t_6);
          if (unlikely(__pyx_t_15 == 0)) break;
          if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_gen, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_fitInData); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_20);
          __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_fastphase); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_gen2hap); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_21);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_21, function);
            }
          }
          __pyx_t_16 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_19, __pyx_8genexpr5__pyx_v_gen) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_8genexpr5__pyx_v_gen);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __pyx_t_21 = NULL;
          __pyx_t_15 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_21)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_21);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
              __pyx_t_15 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[7] = {__pyx_t_21, __pyx_n_u_haplo, __pyx_t_20, __pyx_t_18, __pyx_t_17, __pyx_t_16, __pyx_int_0};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[7] = {__pyx_t_21, __pyx_n_u_haplo, __pyx_t_20, __pyx_t_18, __pyx_t_17, __pyx_t_16, __pyx_int_0};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          } else
          #endif
          {
            __pyx_t_19 = PyTuple_New(6+__pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 227, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_19);
            if (__pyx_t_21) {
              __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_21); __pyx_t_21 = NULL;
            }
            __Pyx_INCREF(__pyx_n_u_haplo);
            __Pyx_GIVEREF(__pyx_n_u_haplo);
            PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_15, __pyx_n_u_haplo);
            __Pyx_GIVEREF(__pyx_t_20);
            PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_15, __pyx_t_20);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_15, __pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_17);
            PyTuple_SET_ITEM(__pyx_t_19, 3+__pyx_t_15, __pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_19, 4+__pyx_t_15, __pyx_t_16);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_19, 5+__pyx_t_15, __pyx_int_0);
            __pyx_t_20 = 0;
            __pyx_t_18 = 0;
            __pyx_t_17 = 0;
            __pyx_t_16 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_19, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 227, __pyx_L24_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_gen); __pyx_8genexpr5__pyx_v_gen = 0;
        goto __pyx_L27_exit_scope;
        __pyx_L24_error:;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_gen); __pyx_8genexpr5__pyx_v_gen = 0;
        goto __pyx_L1_error;
        __pyx_L27_exit_scope:;
      } /* exit inner scope */
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_tasks, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_tasks, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "fastphase/fastphase.pyx":226
 * 
 *             tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]
 *             if fast:             # <<<<<<<<<<<<<<
 *                 tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]
 *             else:
 */
      goto __pyx_L21;
    }

    /* "fastphase/fastphase.pyx":229
 *                 tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]
 *             else:
 *                 tasks += [ fitInData( 'geno', par.alpha,par.theta,par.rho,gen,0) for gen in  self.genotypes.values()]             # <<<<<<<<<<<<<<
 *             results = self.pool.map( fitter, tasks)
 * 
 */
    /*else*/ {
      { /* enter inner scope */
        __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 229, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_13 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__pyx_t_4 == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
          __PYX_ERR(0, 229, __pyx_L30_error)
        }
        __pyx_t_9 = __Pyx_dict_iterator(__pyx_t_4, 0, __pyx_n_s_values, (&__pyx_t_14), (&__pyx_t_6)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_7);
        __pyx_t_7 = __pyx_t_9;
        __pyx_t_9 = 0;
        while (1) {
          __pyx_t_15 = __Pyx_dict_iter_next(__pyx_t_7, __pyx_t_14, &__pyx_t_13, NULL, &__pyx_t_9, NULL, __pyx_t_6);
          if (unlikely(__pyx_t_15 == 0)) break;
          if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_gen, __pyx_t_9);
          __pyx_t_9 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_fitInData); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_18 = NULL;
          __pyx_t_15 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_18)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_18);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_15 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[7] = {__pyx_t_18, __pyx_n_u_geno, __pyx_t_19, __pyx_t_16, __pyx_t_17, __pyx_8genexpr6__pyx_v_gen, __pyx_int_0};
            __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[7] = {__pyx_t_18, __pyx_n_u_geno, __pyx_t_19, __pyx_t_16, __pyx_t_17, __pyx_8genexpr6__pyx_v_gen, __pyx_int_0};
            __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 6+__pyx_t_15); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          {
            __pyx_t_20 = PyTuple_New(6+__pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 229, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_20);
            if (__pyx_t_18) {
              __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_18); __pyx_t_18 = NULL;
            }
            __Pyx_INCREF(__pyx_n_u_geno);
            __Pyx_GIVEREF(__pyx_n_u_geno);
            PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_15, __pyx_n_u_geno);
            __Pyx_GIVEREF(__pyx_t_19);
            PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_15, __pyx_t_19);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_15, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_15, __pyx_t_17);
            __Pyx_INCREF(__pyx_8genexpr6__pyx_v_gen);
            __Pyx_GIVEREF(__pyx_8genexpr6__pyx_v_gen);
            PyTuple_SET_ITEM(__pyx_t_20, 4+__pyx_t_15, __pyx_8genexpr6__pyx_v_gen);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_20, 5+__pyx_t_15, __pyx_int_0);
            __pyx_t_19 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_20, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 229, __pyx_L30_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_gen); __pyx_8genexpr6__pyx_v_gen = 0;
        goto __pyx_L33_exit_scope;
        __pyx_L30_error:;
        __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_gen); __pyx_8genexpr6__pyx_v_gen = 0;
        goto __pyx_L1_error;
        __pyx_L33_exit_scope:;
      } /* exit inner scope */
      __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_tasks, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_tasks, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;
    }
    __pyx_L21:;

    /* "fastphase/fastphase.pyx":230
 *             else:
 *                 tasks += [ fitInData( 'geno', par.alpha,par.theta,par.rho,gen,0) for gen in  self.genotypes.values()]
 *             results = self.pool.map( fitter, tasks)             # <<<<<<<<<<<<<<
 * 
 *             for item in results:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_map); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_fitter); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_v_tasks};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_v_tasks};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_20 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_6, __pyx_t_8);
      __Pyx_INCREF(__pyx_v_tasks);
      __Pyx_GIVEREF(__pyx_v_tasks);
      PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_6, __pyx_v_tasks);
      __pyx_t_8 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_20, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_results, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":232
 *             results = self.pool.map( fitter, tasks)
 * 
 *             for item in results:             # <<<<<<<<<<<<<<
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)
 *                 log_like += item.logLike
 */
    if (likely(PyList_CheckExact(__pyx_v_results)) || PyTuple_CheckExact(__pyx_v_results)) {
      __pyx_t_7 = __pyx_v_results; __Pyx_INCREF(__pyx_t_7); __pyx_t_14 = 0;
      __pyx_t_22 = NULL;
    } else {
      __pyx_t_14 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_results); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_22 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 232, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_22)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_14); __Pyx_INCREF(__pyx_t_9); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 232, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_7, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 232, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_14); __Pyx_INCREF(__pyx_t_9); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 232, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_7, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 232, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_22(__pyx_t_7);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 232, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":233
 * 
 *             for item in results:
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)             # <<<<<<<<<<<<<<
 *                 log_like += item.logLike
 *             if verbose:
 */
      __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_addIndivFit); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_top); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_bot); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_jmk); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_val); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_19 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_20))) {
        __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_20);
        if (likely(__pyx_t_19)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
          __Pyx_INCREF(__pyx_t_19);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_20, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_20)) {
        PyObject *__pyx_temp[5] = {__pyx_t_19, __pyx_t_8, __pyx_t_4, __pyx_t_17, __pyx_t_16};
        __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_20)) {
        PyObject *__pyx_temp[5] = {__pyx_t_19, __pyx_t_8, __pyx_t_4, __pyx_t_17, __pyx_t_16};
        __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      } else
      #endif
      {
        __pyx_t_18 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        if (__pyx_t_19) {
          __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_19); __pyx_t_19 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_6, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_6, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_17);
        PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_6, __pyx_t_17);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_18, 3+__pyx_t_6, __pyx_t_16);
        __pyx_t_8 = 0;
        __pyx_t_4 = 0;
        __pyx_t_17 = 0;
        __pyx_t_16 = 0;
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_18, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      }
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":234
 *             for item in results:
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)
 *                 log_like += item.logLike             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print( iEM, log_like, file=self.flog)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_logLike); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_20 = PyNumber_InPlaceAdd(__pyx_v_log_like, __pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF_SET(__pyx_v_log_like, __pyx_t_20);
      __pyx_t_20 = 0;

      /* "fastphase/fastphase.pyx":232
 *             results = self.pool.map( fitter, tasks)
 * 
 *             for item in results:             # <<<<<<<<<<<<<<
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)
 *                 log_like += item.logLike
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":235
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)
 *                 log_like += item.logLike
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print( iEM, log_like, file=self.flog)
 *                 self.flog.flush()
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 235, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "fastphase/fastphase.pyx":236
 *                 log_like += item.logLike
 *             if verbose:
 *                 print( iEM, log_like, file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             par.update()
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_iEM);
      __Pyx_GIVEREF(__pyx_v_iEM);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_iEM);
      __Pyx_INCREF(__pyx_v_log_like);
      __Pyx_GIVEREF(__pyx_v_log_like);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_log_like);
      __pyx_t_20 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_file, __pyx_t_9) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_7, __pyx_t_20); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":237
 *             if verbose:
 *                 print( iEM, log_like, file=self.flog)
 *                 self.flog.flush()             # <<<<<<<<<<<<<<
 *             par.update()
 *             par.loglike=log_like
 */
      __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flush); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __pyx_t_20 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_20)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_20);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_9 = (__pyx_t_20) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_20) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":235
 *                 par.addIndivFit(item.top,item.bot,item.jmk,item.val)
 *                 log_like += item.logLike
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print( iEM, log_like, file=self.flog)
 *                 self.flog.flush()
 */
    }

    /* "fastphase/fastphase.pyx":238
 *                 print( iEM, log_like, file=self.flog)
 *                 self.flog.flush()
 *             par.update()             # <<<<<<<<<<<<<<
 *             par.loglike=log_like
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_20 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_20)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_20);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_9 = (__pyx_t_20) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_20) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":239
 *                 self.flog.flush()
 *             par.update()
 *             par.loglike=log_like             # <<<<<<<<<<<<<<
 * 
 *         return par
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_par, __pyx_n_s_loglike, __pyx_v_log_like) < 0) __PYX_ERR(0, 239, __pyx_L1_error)

    /* "fastphase/fastphase.pyx":220
 *         old_log_like=1
 * 
 *         for iEM in range(nstep):             # <<<<<<<<<<<<<<
 * 
 *             log_like=0.0
 */
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "fastphase/fastphase.pyx":241
 *             par.loglike=log_like
 * 
 *         return par             # <<<<<<<<<<<<<<
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_par);
  __pyx_r = __pyx_v_par;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":194
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 *     def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Fit the model on observations with nCLus clusters using nstep EM iterations
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.fit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XDECREF(__pyx_v_iEM);
  __Pyx_XDECREF(__pyx_v_log_like);
  __Pyx_XDECREF(__pyx_v_tasks);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_hap);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_gen);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_17optimfit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_17optimfit = {"optimfit", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_17optimfit, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_17optimfit(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nClus = 0;
  PyObject *__pyx_v_nstep = 0;
  CYTHON_UNUSED PyObject *__pyx_v_params = 0;
  PyObject *__pyx_v_verbose = 0;
  CYTHON_UNUSED PyObject *__pyx_v_rhomin = 0;
  PyObject *__pyx_v_alpha_up = 0;
  PyObject *__pyx_v_fast = 0;
  PyObject *__pyx_v_nEM = 0;
  PyObject *__pyx_v_niter = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("optimfit (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nClus,&__pyx_n_s_nstep,&__pyx_n_s_params,&__pyx_n_s_verbose,&__pyx_n_s_rhomin,&__pyx_n_s_alpha_up,&__pyx_n_s_fast,&__pyx_n_s_nEM,&__pyx_n_s_niter,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_20));
    values[2] = ((PyObject *)((PyObject *)__pyx_int_10));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    values[5] = ((PyObject *)((PyObject*)__pyx_float_1eneg_6));
    values[6] = ((PyObject *)((PyObject *)Py_False));
    values[7] = ((PyObject *)((PyObject *)Py_False));
    values[8] = ((PyObject *)((PyObject *)__pyx_int_10));
    values[9] = ((PyObject *)((PyObject *)__pyx_int_5));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nClus);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nstep);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rhomin);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alpha_up);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fast);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEM);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_niter);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "optimfit") < 0)) __PYX_ERR(0, 243, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_nClus = values[1];
    __pyx_v_nstep = values[2];
    __pyx_v_params = values[3];
    __pyx_v_verbose = values[4];
    __pyx_v_rhomin = values[5];
    __pyx_v_alpha_up = values[6];
    __pyx_v_fast = values[7];
    __pyx_v_nEM = values[8];
    __pyx_v_niter = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("optimfit", 0, 1, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 243, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.optimfit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_16optimfit(__pyx_self, __pyx_v_self, __pyx_v_nClus, __pyx_v_nstep, __pyx_v_params, __pyx_v_verbose, __pyx_v_rhomin, __pyx_v_alpha_up, __pyx_v_fast, __pyx_v_nEM, __pyx_v_niter);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6hapflk_9fastphase_9fastphase_8optimfit_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "fastphase/fastphase.pyx":279
 *             cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]
 *             ### Genotypes
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )             # <<<<<<<<<<<<<<
 *             for geno in self.genotypes.keys():
 *                 vit = imp[geno][0]
 */

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_8optimfit_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct_1_genexpr(__pyx_ptype_6hapflk_9fastphase___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 279, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6hapflk_9fastphase_9fastphase_8optimfit_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_fastphase_optimfit_locals_genexp, __pyx_n_s_hapflk_fastphase); if (unlikely(!gen)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.optimfit.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6hapflk_9fastphase_9fastphase_8optimfit_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 279, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 279, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 279, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_geno);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_geno, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 279, __pyx_L1_error) }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 279, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 279, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 279, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_9(__pyx_t_5);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 279, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ijump);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ijump, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_imp)) { __Pyx_RaiseClosureNameError("imp"); __PYX_ERR(0, 279, __pyx_L1_error) }
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_imp, __pyx_cur_scope->__pyx_v_geno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_ijump, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_t_10, 0, 0, &__pyx_cur_scope->__pyx_v_ijump, &__pyx_t_6, NULL, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_nClus)) { __Pyx_RaiseClosureNameError("nClus"); __PYX_ERR(0, 279, __pyx_L1_error) }
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_nClus);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_nClus);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_nClus);
      __pyx_t_11 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_t_5);
      __pyx_cur_scope->__pyx_t_4 = __pyx_t_5;
      __pyx_cur_scope->__pyx_t_5 = __pyx_t_8;
      __pyx_cur_scope->__pyx_t_6 = __pyx_t_9;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L8_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
      __pyx_t_5 = __pyx_cur_scope->__pyx_t_4;
      __pyx_cur_scope->__pyx_t_4 = 0;
      __Pyx_XGOTREF(__pyx_t_5);
      __pyx_t_8 = __pyx_cur_scope->__pyx_t_5;
      __pyx_t_9 = __pyx_cur_scope->__pyx_t_6;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 279, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_16optimfit(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nClus, PyObject *__pyx_v_nstep, CYTHON_UNUSED PyObject *__pyx_v_params, PyObject *__pyx_v_verbose, CYTHON_UNUSED PyObject *__pyx_v_rhomin, PyObject *__pyx_v_alpha_up, PyObject *__pyx_v_fast, PyObject *__pyx_v_nEM, PyObject *__pyx_v_niter) {
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *__pyx_cur_scope;
  PyObject *__pyx_v_liktraj = NULL;
  PyObject *__pyx_v_nbest = NULL;
  PyObject *__pyx_v_curpar = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_par = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_init_par = NULL;
  PyObject *__pyx_v_it = NULL;
  PyObject *__pyx_v_cost_mat = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_gen_dat = NULL;
  PyObject *__pyx_v_geno = NULL;
  PyObject *__pyx_v_vit = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_haplo = NULL;
  PyObject *__pyx_v_permut = NULL;
  PyObject *__pyx_v_newpar = NULL;
  PyObject *__pyx_v_par_s = NULL;
  PyObject *__pyx_v_dat = NULL;
  CYTHON_UNUSED PyObject *__pyx_8genexpr7__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_ijump = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_ijump = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  long __pyx_t_19;
  __Pyx_RefNannySetupContext("optimfit", 0);
  __pyx_cur_scope = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct__optimfit(__pyx_ptype_6hapflk_9fastphase___pyx_scope_struct__optimfit, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 243, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_nClus = __pyx_v_nClus;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nClus);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nClus);

  /* "fastphase/fastphase.pyx":244
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):
 *         liktraj = []             # <<<<<<<<<<<<<<
 * 
 *         print("=== OPTIMFIT RUN ===",file=self.flog)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_liktraj = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":246
 *         liktraj = []
 * 
 *         print("=== OPTIMFIT RUN ===",file=self.flog)             # <<<<<<<<<<<<<<
 *         ## Fit nEM EM to find an initial start
 *         if verbose:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_file, __pyx_t_2) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__6, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":248
 *         print("=== OPTIMFIT RUN ===",file=self.flog)
 *         ## Fit nEM EM to find an initial start
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print("*** Init EM", 1,file=self.flog)
 *         nbest = 0
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 248, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "fastphase/fastphase.pyx":249
 *         ## Fit nEM EM to find an initial start
 *         if verbose:
 *             print("*** Init EM", 1,file=self.flog)             # <<<<<<<<<<<<<<
 *         nbest = 0
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_file, __pyx_t_1) < 0) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":248
 *         print("=== OPTIMFIT RUN ===",file=self.flog)
 *         ## Fit nEM EM to find an initial start
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print("*** Init EM", 1,file=self.flog)
 *         nbest = 0
 */
  }

  /* "fastphase/fastphase.pyx":250
 *         if verbose:
 *             print("*** Init EM", 1,file=self.flog)
 *         nbest = 0             # <<<<<<<<<<<<<<
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 *         liktraj.append((0, 0, curpar.loglike))
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_nbest = __pyx_int_0;

  /* "fastphase/fastphase.pyx":251
 *             print("*** Init EM", 1,file=self.flog)
 *         nbest = 0
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)             # <<<<<<<<<<<<<<
 *         liktraj.append((0, 0, curpar.loglike))
 *         for n in range(1, nEM):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nClus, __pyx_cur_scope->__pyx_v_nClus) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nstep, __pyx_v_nstep) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_verbose, Py_True) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_alpha_up, __pyx_v_alpha_up) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fast, __pyx_v_fast) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_curpar = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":252
 *         nbest = 0
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 *         liktraj.append((0, 0, curpar.loglike))             # <<<<<<<<<<<<<<
 *         for n in range(1, nEM):
 *             if verbose:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_curpar, __pyx_n_s_loglike); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_liktraj, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":253
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 *         liktraj.append((0, 0, curpar.loglike))
 *         for n in range(1, nEM):             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("*** Init EM", n+1,file=self.flog)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_1);
  __Pyx_INCREF(__pyx_v_nEM);
  __Pyx_GIVEREF(__pyx_v_nEM);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nEM);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_2 = __pyx_t_4; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 253, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 253, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":254
 *         liktraj.append((0, 0, curpar.loglike))
 *         for n in range(1, nEM):
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("*** Init EM", n+1,file=self.flog)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 254, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":255
 *         for n in range(1, nEM):
 *             if verbose:
 *                 print("*** Init EM", n+1,file=self.flog)             # <<<<<<<<<<<<<<
 * 
 *             par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)
 */
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_kp_u_Init_EM);
      __Pyx_GIVEREF(__pyx_kp_u_Init_EM);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Init_EM);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_file, __pyx_t_8) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "fastphase/fastphase.pyx":254
 *         liktraj.append((0, 0, curpar.loglike))
 *         for n in range(1, nEM):
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("*** Init EM", n+1,file=self.flog)
 * 
 */
    }

    /* "fastphase/fastphase.pyx":257
 *                 print("*** Init EM", n+1,file=self.flog)
 * 
 *             par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)             # <<<<<<<<<<<<<<
 *             liktraj.append((n, 0, par.loglike))
 *             if par.loglike > curpar.loglike:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fit); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_nClus, __pyx_cur_scope->__pyx_v_nClus) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_nstep, __pyx_v_nstep) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_verbose, Py_True) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_alpha_up, __pyx_v_alpha_up) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fast, __pyx_v_fast) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_par, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":258
 * 
 *             par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)
 *             liktraj.append((n, 0, par.loglike))             # <<<<<<<<<<<<<<
 *             if par.loglike > curpar.loglike:
 *                 nbest = n
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_loglike); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_n);
    __Pyx_GIVEREF(__pyx_v_n);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_n);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_liktraj, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":259
 *             par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)
 *             liktraj.append((n, 0, par.loglike))
 *             if par.loglike > curpar.loglike:             # <<<<<<<<<<<<<<
 *                 nbest = n
 *                 curpar = par
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_loglike); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_curpar, __pyx_n_s_loglike); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":260
 *             liktraj.append((n, 0, par.loglike))
 *             if par.loglike > curpar.loglike:
 *                 nbest = n             # <<<<<<<<<<<<<<
 *                 curpar = par
 *         init_par = curpar
 */
      __Pyx_INCREF(__pyx_v_n);
      __Pyx_DECREF_SET(__pyx_v_nbest, __pyx_v_n);

      /* "fastphase/fastphase.pyx":261
 *             if par.loglike > curpar.loglike:
 *                 nbest = n
 *                 curpar = par             # <<<<<<<<<<<<<<
 *         init_par = curpar
 *         ## Improve it
 */
      __Pyx_INCREF(__pyx_v_par);
      __Pyx_DECREF_SET(__pyx_v_curpar, __pyx_v_par);

      /* "fastphase/fastphase.pyx":259
 *             par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)
 *             liktraj.append((n, 0, par.loglike))
 *             if par.loglike > curpar.loglike:             # <<<<<<<<<<<<<<
 *                 nbest = n
 *                 curpar = par
 */
    }

    /* "fastphase/fastphase.pyx":253
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 *         liktraj.append((0, 0, curpar.loglike))
 *         for n in range(1, nEM):             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("*** Init EM", n+1,file=self.flog)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":262
 *                 nbest = n
 *                 curpar = par
 *         init_par = curpar             # <<<<<<<<<<<<<<
 *         ## Improve it
 *         print('Initial Viterbi',file=self.flog)
 */
  __Pyx_INCREF(__pyx_v_curpar);
  __pyx_v_init_par = __pyx_v_curpar;

  /* "fastphase/fastphase.pyx":264
 *         init_par = curpar
 *         ## Improve it
 *         print('Initial Viterbi',file=self.flog)             # <<<<<<<<<<<<<<
 *         self.flog.flush()
 *         imp = self.viterbi([curpar])
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_file, __pyx_t_8) < 0) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__8, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "fastphase/fastphase.pyx":265
 *         ## Improve it
 *         print('Initial Viterbi',file=self.flog)
 *         self.flog.flush()             # <<<<<<<<<<<<<<
 *         imp = self.viterbi([curpar])
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "fastphase/fastphase.pyx":266
 *         print('Initial Viterbi',file=self.flog)
 *         self.flog.flush()
 *         imp = self.viterbi([curpar])             # <<<<<<<<<<<<<<
 * 
 *         for it in range(niter):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_viterbi); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_curpar);
  __Pyx_GIVEREF(__pyx_v_curpar);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_curpar);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_8);
  __pyx_cur_scope->__pyx_v_imp = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "fastphase/fastphase.pyx":268
 *         imp = self.viterbi([curpar])
 * 
 *         for it in range(niter):             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("*** Iter", it+1,file=self.flog)
 */
  __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_niter); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
    __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 268, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 268, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 268, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 268, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 268, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 268, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_it, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "fastphase/fastphase.pyx":269
 * 
 *         for it in range(niter):
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("*** Iter", it+1,file=self.flog)
 *                 self.flog.flush()
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 269, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":270
 *         for it in range(niter):
 *             if verbose:
 *                 print("*** Iter", it+1,file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             ## calculate costs
 */
      __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_v_it, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_kp_u_Iter);
      __Pyx_GIVEREF(__pyx_kp_u_Iter);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Iter);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_file, __pyx_t_4) < 0) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":271
 *             if verbose:
 *                 print("*** Iter", it+1,file=self.flog)
 *                 self.flog.flush()             # <<<<<<<<<<<<<<
 *             ## calculate costs
 *             if verbose:
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":269
 * 
 *         for it in range(niter):
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("*** Iter", it+1,file=self.flog)
 *                 self.flog.flush()
 */
    }

    /* "fastphase/fastphase.pyx":273
 *                 self.flog.flush()
 *             ## calculate costs
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("Calculating Costs",file=self.flog)
 *                 self.flog.flush()
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":274
 *             ## calculate costs
 *             if verbose:
 *                 print("Calculating Costs",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             ##cost_mat = np.zeros( (self.nLoci-1, nClus, nClus), dtype=np.float)
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_file, __pyx_t_2) < 0) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__9, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":275
 *             if verbose:
 *                 print("Calculating Costs",file=self.flog)
 *                 self.flog.flush()             # <<<<<<<<<<<<<<
 *             ##cost_mat = np.zeros( (self.nLoci-1, nClus, nClus), dtype=np.float)
 *             cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_flush); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":273
 *                 self.flog.flush()
 *             ## calculate costs
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("Calculating Costs",file=self.flog)
 *                 self.flog.flush()
 */
    }

    /* "fastphase/fastphase.pyx":277
 *                 self.flog.flush()
 *             ##cost_mat = np.zeros( (self.nLoci-1, nClus, nClus), dtype=np.float)
 *             cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]             # <<<<<<<<<<<<<<
 *             ### Genotypes
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
        __pyx_t_4 = __pyx_t_8; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 277, __pyx_L14_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L14_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L14_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_10(__pyx_t_4);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 277, __pyx_L14_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_i, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nClus);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nClus);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_cur_scope->__pyx_v_nClus);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nClus);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nClus);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_cur_scope->__pyx_v_nClus);
        __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_float); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_14) < 0) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_12, __pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 277, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_i); __pyx_8genexpr7__pyx_v_i = 0;
      goto __pyx_L17_exit_scope;
      __pyx_L14_error:;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_i); __pyx_8genexpr7__pyx_v_i = 0;
      goto __pyx_L1_error;
      __pyx_L17_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_cost_mat, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":279
 *             cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]
 *             ### Genotypes
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )             # <<<<<<<<<<<<<<
 *             for geno in self.genotypes.keys():
 *                 vit = imp[geno][0]
 */
    __pyx_t_2 = __pyx_pf_6hapflk_9fastphase_9fastphase_8optimfit_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_gen_dat, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":280
 *             ### Genotypes
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )
 *             for geno in self.genotypes.keys():             # <<<<<<<<<<<<<<
 *                 vit = imp[geno][0]
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 */
    __pyx_t_9 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_t_4 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(0, 280, __pyx_L1_error)
    }
    __pyx_t_14 = __Pyx_dict_iterator(__pyx_t_4, 0, __pyx_n_s_keys, (&__pyx_t_15), (&__pyx_t_16)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_14;
    __pyx_t_14 = 0;
    while (1) {
      __pyx_t_17 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_15, &__pyx_t_9, &__pyx_t_14, NULL, NULL, __pyx_t_16);
      if (unlikely(__pyx_t_17 == 0)) break;
      if (unlikely(__pyx_t_17 == -1)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_XDECREF_SET(__pyx_v_geno, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "fastphase/fastphase.pyx":281
 *             gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )
 *             for geno in self.genotypes.keys():
 *                 vit = imp[geno][0]             # <<<<<<<<<<<<<<
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 */
      __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_imp, __pyx_v_geno); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF_SET(__pyx_v_vit, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":282
 *             for geno in self.genotypes.keys():
 *                 vit = imp[geno][0]
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]             # <<<<<<<<<<<<<<
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 *                 res = self.pool.map( calc_cost_matrix_geno, args, 100)
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 282, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
          __pyx_t_8 = __pyx_t_14; __Pyx_INCREF(__pyx_t_8); __pyx_t_18 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_18 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 282, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_10 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 282, __pyx_L22_error)
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_8))) {
              if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_14 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_18); __Pyx_INCREF(__pyx_t_14); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 282, __pyx_L22_error)
              #else
              __pyx_t_14 = PySequence_ITEM(__pyx_t_8, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              #endif
            } else {
              if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_18); __Pyx_INCREF(__pyx_t_14); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 282, __pyx_L22_error)
              #else
              __pyx_t_14 = PySequence_ITEM(__pyx_t_8, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              #endif
            }
          } else {
            __pyx_t_14 = __pyx_t_10(__pyx_t_8);
            if (unlikely(!__pyx_t_14)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 282, __pyx_L22_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_14);
          }
          __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_ijump, __pyx_t_14);
          __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_8genexpr9__pyx_v_ijump, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_12 = __Pyx_PyObject_GetSlice(__pyx_v_vit, 0, 0, &__pyx_8genexpr9__pyx_v_ijump, &__pyx_t_14, NULL, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 282, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 282, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nClus);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nClus);
          PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_cur_scope->__pyx_v_nClus);
          __pyx_t_12 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 282, __pyx_L22_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_ijump); __pyx_8genexpr9__pyx_v_ijump = 0;
        goto __pyx_L25_exit_scope;
        __pyx_L22_error:;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_ijump); __pyx_8genexpr9__pyx_v_ijump = 0;
        goto __pyx_L1_error;
        __pyx_L25_exit_scope:;
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":284
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 *                 res = self.pool.map( calc_cost_matrix_geno, args, 100)             # <<<<<<<<<<<<<<
 *                 ## " cost_mat += res "
 *                 cost_mat = list( map( add, cost_mat, res))
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_map); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_calc_cost_matrix_geno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = NULL;
      __pyx_t_17 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_17 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_t_8, __pyx_v_args, __pyx_int_100};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_t_8, __pyx_v_args, __pyx_int_100};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_12) {
          __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_12); __pyx_t_12 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_17, __pyx_t_8);
        __Pyx_INCREF(__pyx_v_args);
        __Pyx_GIVEREF(__pyx_v_args);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_17, __pyx_v_args);
        __Pyx_INCREF(__pyx_int_100);
        __Pyx_GIVEREF(__pyx_int_100);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_17, __pyx_int_100);
        __pyx_t_8 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":286
 *                 res = self.pool.map( calc_cost_matrix_geno, args, 100)
 *                 ## " cost_mat += res "
 *                 cost_mat = list( map( add, cost_mat, res))             # <<<<<<<<<<<<<<
 *             ### haplotypes
 *             for haplo in self.haplotypes.keys():
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_cost_mat);
      __Pyx_GIVEREF(__pyx_v_cost_mat);
      PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_cost_mat);
      __Pyx_INCREF(__pyx_v_res);
      __Pyx_GIVEREF(__pyx_v_res);
      PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_v_res);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_cost_mat, ((PyObject*)__pyx_t_14));
      __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":288
 *                 cost_mat = list( map( add, cost_mat, res))
 *             ### haplotypes
 *             for haplo in self.haplotypes.keys():             # <<<<<<<<<<<<<<
 *                 vit = imp[haplo][0]
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 */
    __pyx_t_15 = 0;
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (unlikely(__pyx_t_14 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(0, 288, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_dict_iterator(__pyx_t_14, 0, __pyx_n_s_keys, (&__pyx_t_9), (&__pyx_t_16)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_4;
    __pyx_t_4 = 0;
    while (1) {
      __pyx_t_17 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_9, &__pyx_t_15, &__pyx_t_4, NULL, NULL, __pyx_t_16);
      if (unlikely(__pyx_t_17 == 0)) break;
      if (unlikely(__pyx_t_17 == -1)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_haplo, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":289
 *             ### haplotypes
 *             for haplo in self.haplotypes.keys():
 *                 vit = imp[haplo][0]             # <<<<<<<<<<<<<<
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 */
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_imp, __pyx_v_haplo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_vit, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "fastphase/fastphase.pyx":290
 *             for haplo in self.haplotypes.keys():
 *                 vit = imp[haplo][0]
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]             # <<<<<<<<<<<<<<
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 *                 res = self.pool.map( calc_cost_matrix_haplo, args, 100)
 */
      { /* enter inner scope */
        __pyx_t_14 = PyList_New(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 290, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 290, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
          __pyx_t_11 = __pyx_t_4; __Pyx_INCREF(__pyx_t_11); __pyx_t_18 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_18 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 290, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 290, __pyx_L30_error)
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_11))) {
              if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_11)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_18); __Pyx_INCREF(__pyx_t_4); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 290, __pyx_L30_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_11, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_18); __Pyx_INCREF(__pyx_t_4); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 290, __pyx_L30_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_11, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_10(__pyx_t_11);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 290, __pyx_L30_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_ijump, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_9genexpr10__pyx_v_ijump, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = __Pyx_PyObject_GetSlice(__pyx_v_vit, 0, 0, &__pyx_9genexpr10__pyx_v_ijump, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 290, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nClus);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nClus);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_cur_scope->__pyx_v_nClus);
          __pyx_t_8 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_14, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 290, __pyx_L30_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_ijump); __pyx_9genexpr10__pyx_v_ijump = 0;
        goto __pyx_L33_exit_scope;
        __pyx_L30_error:;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_ijump); __pyx_9genexpr10__pyx_v_ijump = 0;
        goto __pyx_L1_error;
        __pyx_L33_exit_scope:;
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_14));
      __pyx_t_14 = 0;

      /* "fastphase/fastphase.pyx":292
 *                 args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
 *                 ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
 *                 res = self.pool.map( calc_cost_matrix_haplo, args, 100)             # <<<<<<<<<<<<<<
 *                 ## " cost_mat += res "
 *                 cost_mat = list( map( add, cost_mat, res))
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_map); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_calc_cost_matrix_haplo); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = NULL;
      __pyx_t_17 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_17 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_t_11, __pyx_v_args, __pyx_int_100};
        __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_t_11, __pyx_v_args, __pyx_int_100};
        __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_17, __pyx_t_11);
        __Pyx_INCREF(__pyx_v_args);
        __Pyx_GIVEREF(__pyx_v_args);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_17, __pyx_v_args);
        __Pyx_INCREF(__pyx_int_100);
        __Pyx_GIVEREF(__pyx_int_100);
        PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_17, __pyx_int_100);
        __pyx_t_11 = 0;
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_12, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "fastphase/fastphase.pyx":294
 *                 res = self.pool.map( calc_cost_matrix_haplo, args, 100)
 *                 ## " cost_mat += res "
 *                 cost_mat = list( map( add, cost_mat, res))             # <<<<<<<<<<<<<<
 *                 ##cost_mat += self.pool.map( calc_cost_matrix_haplo, args, 100)
 *             ## combine
 */
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_add); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_14);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14);
      __Pyx_INCREF(__pyx_v_cost_mat);
      __Pyx_GIVEREF(__pyx_v_cost_mat);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_cost_mat);
      __Pyx_INCREF(__pyx_v_res);
      __Pyx_GIVEREF(__pyx_v_res);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_res);
      __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF_SET(__pyx_v_cost_mat, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":297
 *                 ##cost_mat += self.pool.map( calc_cost_matrix_haplo, args, 100)
 *             ## combine
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("Computing optimum permutations",file=self.flog)
 *                 self.flog.flush()
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 297, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":298
 *             ## combine
 *             if verbose:
 *                 print("Computing optimum permutations",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             ##cost_mat = cost_mat.reshape( ( self.nLoci-1, nClus, nClus))
 */
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_file, __pyx_t_4) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__10, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":299
 *             if verbose:
 *                 print("Computing optimum permutations",file=self.flog)
 *                 self.flog.flush()             # <<<<<<<<<<<<<<
 *             ##cost_mat = cost_mat.reshape( ( self.nLoci-1, nClus, nClus))
 *             ##args = [ cost_mat[i] for i in range( self.nLoci-1)]
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
        }
      }
      __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":297
 *                 ##cost_mat += self.pool.map( calc_cost_matrix_haplo, args, 100)
 *             ## combine
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("Computing optimum permutations",file=self.flog)
 *                 self.flog.flush()
 */
    }

    /* "fastphase/fastphase.pyx":302
 *             ##cost_mat = cost_mat.reshape( ( self.nLoci-1, nClus, nClus))
 *             ##args = [ cost_mat[i] for i in range( self.nLoci-1)]
 *             res = np.array( self.pool.map( linear_sum_assignment, cost_mat,  100))             # <<<<<<<<<<<<<<
 *             permut = res[:,1,:]
 *             newpar = switch_pars( curpar, permut)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_map); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_linear_sum_assignment); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_8 = NULL;
    __pyx_t_16 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_16 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_t_12, __pyx_v_cost_mat, __pyx_int_100};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_t_12, __pyx_v_cost_mat, __pyx_int_100};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(3+__pyx_t_16); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_16, __pyx_t_12);
      __Pyx_INCREF(__pyx_v_cost_mat);
      __Pyx_GIVEREF(__pyx_v_cost_mat);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_16, __pyx_v_cost_mat);
      __Pyx_INCREF(__pyx_int_100);
      __Pyx_GIVEREF(__pyx_int_100);
      PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_16, __pyx_int_100);
      __pyx_t_12 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_13, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_4 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_11, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":303
 *             ##args = [ cost_mat[i] for i in range( self.nLoci-1)]
 *             res = np.array( self.pool.map( linear_sum_assignment, cost_mat,  100))
 *             permut = res[:,1,:]             # <<<<<<<<<<<<<<
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:
 */
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_res, __pyx_tuple__11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_permut, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":304
 *             res = np.array( self.pool.map( linear_sum_assignment, cost_mat,  100))
 *             permut = res[:,1,:]
 *             newpar = switch_pars( curpar, permut)             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("EM with switched parameters",file=self.flog)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_switch_pars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = NULL;
    __pyx_t_16 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_16 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_curpar, __pyx_v_permut};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_curpar, __pyx_v_permut};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_14) {
        __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_14); __pyx_t_14 = NULL;
      }
      __Pyx_INCREF(__pyx_v_curpar);
      __Pyx_GIVEREF(__pyx_v_curpar);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_16, __pyx_v_curpar);
      __Pyx_INCREF(__pyx_v_permut);
      __Pyx_GIVEREF(__pyx_v_permut);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_16, __pyx_v_permut);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_newpar, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":305
 *             permut = res[:,1,:]
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("EM with switched parameters",file=self.flog)
 *                 self.flog.flush()
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":306
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:
 *                 print("EM with switched parameters",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             if (it == niter-1): ## last iteration, longer and no imputation
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_file, __pyx_t_2) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__12, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":307
 *             if verbose:
 *                 print("EM with switched parameters",file=self.flog)
 *                 self.flog.flush()             # <<<<<<<<<<<<<<
 *             if (it == niter-1): ## last iteration, longer and no imputation
 *                 par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_flush); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":305
 *             permut = res[:,1,:]
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("EM with switched parameters",file=self.flog)
 *                 self.flog.flush()
 */
    }

    /* "fastphase/fastphase.pyx":308
 *                 print("EM with switched parameters",file=self.flog)
 *                 self.flog.flush()
 *             if (it == niter-1): ## last iteration, longer and no imputation             # <<<<<<<<<<<<<<
 *                 par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_niter, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyObject_RichCompare(__pyx_v_it, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (__pyx_t_3) {

      /* "fastphase/fastphase.pyx":309
 *                 self.flog.flush()
 *             if (it == niter-1): ## last iteration, longer and no imputation
 *                 par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)             # <<<<<<<<<<<<<<
 *             else:
 *                 par_s = self.fit( nClus = nClus, nstep=nstep, verbose=True, params=newpar, alpha_up=False, fast=fast)
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nClus, __pyx_cur_scope->__pyx_v_nClus) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_INCREF(__pyx_v_nstep);
      __pyx_t_4 = __pyx_v_nstep;
      __pyx_t_19 = 30;
      __pyx_t_13 = __Pyx_PyInt_From_long(__pyx_t_19); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_12 = PyObject_RichCompare(__pyx_t_4, __pyx_t_13, Py_GT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (__pyx_t_3) {
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_14 = __pyx_t_4;
      } else {
        __pyx_t_12 = __Pyx_PyInt_From_long(__pyx_t_19); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_14 = __pyx_t_12;
        __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nstep, __pyx_t_14) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_verbose, Py_True) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_params, __pyx_v_newpar) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_alpha_up, Py_False) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fast, __pyx_v_fast) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_par_s, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "fastphase/fastphase.pyx":308
 *                 print("EM with switched parameters",file=self.flog)
 *                 self.flog.flush()
 *             if (it == niter-1): ## last iteration, longer and no imputation             # <<<<<<<<<<<<<<
 *                 par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)
 *             else:
 */
      goto __pyx_L36;
    }

    /* "fastphase/fastphase.pyx":311
 *                 par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)
 *             else:
 *                 par_s = self.fit( nClus = nClus, nstep=nstep, verbose=True, params=newpar, alpha_up=False, fast=fast)             # <<<<<<<<<<<<<<
 *                 imp = self.viterbi([par_s])
 *             curpar = par_s
 */
    /*else*/ {
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_fit); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nClus, __pyx_cur_scope->__pyx_v_nClus) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nstep, __pyx_v_nstep) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_verbose, Py_True) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_params, __pyx_v_newpar) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_alpha_up, Py_False) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fast, __pyx_v_fast) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_par_s, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "fastphase/fastphase.pyx":312
 *             else:
 *                 par_s = self.fit( nClus = nClus, nstep=nstep, verbose=True, params=newpar, alpha_up=False, fast=fast)
 *                 imp = self.viterbi([par_s])             # <<<<<<<<<<<<<<
 *             curpar = par_s
 *             liktraj.append((nbest, it+1, curpar.loglike))
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_viterbi); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_v_par_s);
      __Pyx_GIVEREF(__pyx_v_par_s);
      PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_par_s);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_11 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_14);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_imp);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_imp, __pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      __pyx_t_11 = 0;
    }
    __pyx_L36:;

    /* "fastphase/fastphase.pyx":313
 *                 par_s = self.fit( nClus = nClus, nstep=nstep, verbose=True, params=newpar, alpha_up=False, fast=fast)
 *                 imp = self.viterbi([par_s])
 *             curpar = par_s             # <<<<<<<<<<<<<<
 *             liktraj.append((nbest, it+1, curpar.loglike))
 * 
 */
    __Pyx_INCREF(__pyx_v_par_s);
    __Pyx_DECREF_SET(__pyx_v_curpar, __pyx_v_par_s);

    /* "fastphase/fastphase.pyx":314
 *                 imp = self.viterbi([par_s])
 *             curpar = par_s
 *             liktraj.append((nbest, it+1, curpar.loglike))             # <<<<<<<<<<<<<<
 * 
 *         ## add alpha update
 */
    __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_v_it, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_curpar, __pyx_n_s_loglike); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_v_nbest);
    __Pyx_GIVEREF(__pyx_v_nbest);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_nbest);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_t_2);
    __pyx_t_11 = 0;
    __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_liktraj, __pyx_t_14); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "fastphase/fastphase.pyx":268
 *         imp = self.viterbi([curpar])
 * 
 *         for it in range(niter):             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("*** Iter", it+1,file=self.flog)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":319
 *         # par_s = self.fit(nClus = nClus, nstep = max(30,nstep), verbose = True, params = par_s, alpha_up = True, theta_up = False, fast = fast)
 *         # liktraj.append((nbest, it+2, par_s.loglike))
 *         print('EM','iter','loglik',file=self.flog)             # <<<<<<<<<<<<<<
 *         for dat in liktraj:
 *             print(*dat,file=self.flog)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_file, __pyx_t_14) < 0) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__13, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "fastphase/fastphase.pyx":320
 *         # liktraj.append((nbest, it+2, par_s.loglike))
 *         print('EM','iter','loglik',file=self.flog)
 *         for dat in liktraj:             # <<<<<<<<<<<<<<
 *             print(*dat,file=self.flog)
 *         return par_s
 */
  __pyx_t_14 = __pyx_v_liktraj; __Pyx_INCREF(__pyx_t_14); __pyx_t_6 = 0;
  for (;;) {
    if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_14)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 320, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_14, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_dat, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":321
 *         print('EM','iter','loglik',file=self.flog)
 *         for dat in liktraj:
 *             print(*dat,file=self.flog)             # <<<<<<<<<<<<<<
 *         return par_s
 * 
 */
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flog); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_file, __pyx_t_11) < 0) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "fastphase/fastphase.pyx":320
 *         # liktraj.append((nbest, it+2, par_s.loglike))
 *         print('EM','iter','loglik',file=self.flog)
 *         for dat in liktraj:             # <<<<<<<<<<<<<<
 *             print(*dat,file=self.flog)
 *         return par_s
 */
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "fastphase/fastphase.pyx":322
 *         for dat in liktraj:
 *             print(*dat,file=self.flog)
 *         return par_s             # <<<<<<<<<<<<<<
 * 
 *     def viterbi(self, parList):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_par_s)) { __Pyx_RaiseUnboundLocalError("par_s"); __PYX_ERR(0, 322, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_par_s);
  __pyx_r = __pyx_v_par_s;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.optimfit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_liktraj);
  __Pyx_XDECREF(__pyx_v_nbest);
  __Pyx_XDECREF(__pyx_v_curpar);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XDECREF(__pyx_v_init_par);
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XDECREF(__pyx_v_cost_mat);
  __Pyx_XDECREF(__pyx_v_gen_dat);
  __Pyx_XDECREF(__pyx_v_geno);
  __Pyx_XDECREF(__pyx_v_vit);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_haplo);
  __Pyx_XDECREF(__pyx_v_permut);
  __Pyx_XDECREF(__pyx_v_newpar);
  __Pyx_XDECREF(__pyx_v_par_s);
  __Pyx_XDECREF(__pyx_v_dat);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_ijump);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_ijump);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":324
 *         return par_s
 * 
 *     def viterbi(self, parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_19viterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_19viterbi = {"viterbi", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_19viterbi, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_19viterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parList = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("viterbi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("viterbi", 1, 2, 2, 1); __PYX_ERR(0, 324, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "viterbi") < 0)) __PYX_ERR(0, 324, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("viterbi", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 324, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.viterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_18viterbi(__pyx_self, __pyx_v_self, __pyx_v_parList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_18viterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parList) {
  PyObject *__pyx_v_tasks = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_Imputations = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_name = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_hap = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_name = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  __Pyx_RefNannySetupContext("viterbi", 0);

  /* "fastphase/fastphase.pyx":325
 * 
 *     def viterbi(self, parList):
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]             # <<<<<<<<<<<<<<
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( viterber, tasks)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 325, __pyx_L5_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 325, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 325, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_hap, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_imputeInData); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 325, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 325, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_n_u_haplo, __pyx_v_parList, __pyx_t_9, __pyx_9genexpr11__pyx_v_name, __pyx_9genexpr11__pyx_v_hap};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_n_u_haplo, __pyx_v_parList, __pyx_t_9, __pyx_9genexpr11__pyx_v_name, __pyx_9genexpr11__pyx_v_hap};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 325, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_n_u_haplo);
        __Pyx_GIVEREF(__pyx_n_u_haplo);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_8, __pyx_n_u_haplo);
        __Pyx_INCREF(__pyx_v_parList);
        __Pyx_GIVEREF(__pyx_v_parList);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_8, __pyx_v_parList);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_8, __pyx_t_9);
        __Pyx_INCREF(__pyx_9genexpr11__pyx_v_name);
        __Pyx_GIVEREF(__pyx_9genexpr11__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_8, __pyx_9genexpr11__pyx_v_name);
        __Pyx_INCREF(__pyx_9genexpr11__pyx_v_hap);
        __Pyx_GIVEREF(__pyx_9genexpr11__pyx_v_hap);
        PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_8, __pyx_9genexpr11__pyx_v_hap);
        __pyx_t_9 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 325, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_hap); __pyx_9genexpr11__pyx_v_hap = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_name); __pyx_9genexpr11__pyx_v_name = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_hap); __pyx_9genexpr11__pyx_v_hap = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_name); __pyx_9genexpr11__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":326
 *     def viterbi(self, parList):
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]             # <<<<<<<<<<<<<<
 *         results = self.pool.map( viterber, tasks)
 *         Imputations={}
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 326, __pyx_L11_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_3, &__pyx_t_4, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 326, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_gen, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_imputeInData); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 326, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_n_u_geno, __pyx_v_parList, __pyx_t_11, __pyx_9genexpr12__pyx_v_name, __pyx_9genexpr12__pyx_v_gen};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L11_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_n_u_geno, __pyx_v_parList, __pyx_t_11, __pyx_9genexpr12__pyx_v_name, __pyx_9genexpr12__pyx_v_gen};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L11_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 326, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_n_u_geno);
        __Pyx_GIVEREF(__pyx_n_u_geno);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_8, __pyx_n_u_geno);
        __Pyx_INCREF(__pyx_v_parList);
        __Pyx_GIVEREF(__pyx_v_parList);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_8, __pyx_v_parList);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_8, __pyx_t_11);
        __Pyx_INCREF(__pyx_9genexpr12__pyx_v_name);
        __Pyx_GIVEREF(__pyx_9genexpr12__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_8, __pyx_9genexpr12__pyx_v_name);
        __Pyx_INCREF(__pyx_9genexpr12__pyx_v_gen);
        __Pyx_GIVEREF(__pyx_9genexpr12__pyx_v_gen);
        PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_8, __pyx_9genexpr12__pyx_v_gen);
        __pyx_t_11 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 326, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_gen); __pyx_9genexpr12__pyx_v_gen = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_name); __pyx_9genexpr12__pyx_v_name = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_gen); __pyx_9genexpr12__pyx_v_gen = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_name); __pyx_9genexpr12__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_tasks, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_tasks, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":327
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( viterber, tasks)             # <<<<<<<<<<<<<<
 *         Imputations={}
 *         for item in results:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_map); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_viterber); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_v_tasks};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_v_tasks};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_tasks);
    __Pyx_GIVEREF(__pyx_v_tasks);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_v_tasks);
    __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_results = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":328
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( viterber, tasks)
 *         Imputations={}             # <<<<<<<<<<<<<<
 *         for item in results:
 *             Imputations[item.name]=item.path
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_Imputations = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":329
 *         results = self.pool.map( viterber, tasks)
 *         Imputations={}
 *         for item in results:             # <<<<<<<<<<<<<<
 *             Imputations[item.name]=item.path
 *         return Imputations
 */
  if (likely(PyList_CheckExact(__pyx_v_results)) || PyTuple_CheckExact(__pyx_v_results)) {
    __pyx_t_2 = __pyx_v_results; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_results); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 329, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 329, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 329, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 329, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 329, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_12(__pyx_t_2);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 329, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":330
 *         Imputations={}
 *         for item in results:
 *             Imputations[item.name]=item.path             # <<<<<<<<<<<<<<
 *         return Imputations
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_path); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (unlikely(PyDict_SetItem(__pyx_v_Imputations, __pyx_t_10, __pyx_t_6) < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":329
 *         results = self.pool.map( viterber, tasks)
 *         Imputations={}
 *         for item in results:             # <<<<<<<<<<<<<<
 *             Imputations[item.name]=item.path
 *         return Imputations
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":331
 *         for item in results:
 *             Imputations[item.name]=item.path
 *         return Imputations             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_Imputations);
  __pyx_r = __pyx_v_Imputations;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":324
 *         return par_s
 * 
 *     def viterbi(self, parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.viterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tasks);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_Imputations);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_name);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_hap);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_name);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":376
 *         #     tasks=[]
 * 
 *     def impute(self,parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_21impute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9fastphase_21impute = {"impute", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_9fastphase_21impute, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9fastphase_21impute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parList = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("impute (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("impute", 1, 2, 2, 1); __PYX_ERR(0, 376, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "impute") < 0)) __PYX_ERR(0, 376, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("impute", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 376, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.impute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_9fastphase_20impute(__pyx_self, __pyx_v_self, __pyx_v_parList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_9fastphase_20impute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parList) {
  PyObject *__pyx_v_tasks = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_Imputations = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_name = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_hap = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_name = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  __Pyx_RefNannySetupContext("impute", 0);

  /* "fastphase/fastphase.pyx":377
 * 
 *     def impute(self,parList):
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]             # <<<<<<<<<<<<<<
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( imputer, tasks)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_haplotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 377, __pyx_L5_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 377, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 377, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_hap, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_imputeInData); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 377, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 377, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_n_u_haplo, __pyx_v_parList, __pyx_t_9, __pyx_9genexpr13__pyx_v_name, __pyx_9genexpr13__pyx_v_hap};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_n_u_haplo, __pyx_v_parList, __pyx_t_9, __pyx_9genexpr13__pyx_v_name, __pyx_9genexpr13__pyx_v_hap};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 377, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_n_u_haplo);
        __Pyx_GIVEREF(__pyx_n_u_haplo);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_8, __pyx_n_u_haplo);
        __Pyx_INCREF(__pyx_v_parList);
        __Pyx_GIVEREF(__pyx_v_parList);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_8, __pyx_v_parList);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_8, __pyx_t_9);
        __Pyx_INCREF(__pyx_9genexpr13__pyx_v_name);
        __Pyx_GIVEREF(__pyx_9genexpr13__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_8, __pyx_9genexpr13__pyx_v_name);
        __Pyx_INCREF(__pyx_9genexpr13__pyx_v_hap);
        __Pyx_GIVEREF(__pyx_9genexpr13__pyx_v_hap);
        PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_8, __pyx_9genexpr13__pyx_v_hap);
        __pyx_t_9 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 377, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_hap); __pyx_9genexpr13__pyx_v_hap = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_name); __pyx_9genexpr13__pyx_v_name = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_hap); __pyx_9genexpr13__pyx_v_hap = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_name); __pyx_9genexpr13__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":378
 *     def impute(self,parList):
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]             # <<<<<<<<<<<<<<
 *         results = self.pool.map( imputer, tasks)
 *         Imputations={}
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_genotypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 378, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 378, __pyx_L11_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 378, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_3, &__pyx_t_4, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 378, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_gen, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_imputeInData); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 378, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_nLoci); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 378, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_n_u_geno, __pyx_v_parList, __pyx_t_11, __pyx_9genexpr14__pyx_v_name, __pyx_9genexpr14__pyx_v_gen};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 378, __pyx_L11_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_n_u_geno, __pyx_v_parList, __pyx_t_11, __pyx_9genexpr14__pyx_v_name, __pyx_9genexpr14__pyx_v_gen};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 378, __pyx_L11_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 378, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_n_u_geno);
        __Pyx_GIVEREF(__pyx_n_u_geno);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_8, __pyx_n_u_geno);
        __Pyx_INCREF(__pyx_v_parList);
        __Pyx_GIVEREF(__pyx_v_parList);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_8, __pyx_v_parList);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_8, __pyx_t_11);
        __Pyx_INCREF(__pyx_9genexpr14__pyx_v_name);
        __Pyx_GIVEREF(__pyx_9genexpr14__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_8, __pyx_9genexpr14__pyx_v_name);
        __Pyx_INCREF(__pyx_9genexpr14__pyx_v_gen);
        __Pyx_GIVEREF(__pyx_9genexpr14__pyx_v_gen);
        PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_8, __pyx_9genexpr14__pyx_v_gen);
        __pyx_t_11 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 378, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 378, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_gen); __pyx_9genexpr14__pyx_v_gen = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_name); __pyx_9genexpr14__pyx_v_name = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_gen); __pyx_9genexpr14__pyx_v_gen = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_name); __pyx_9genexpr14__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_tasks, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_tasks, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":379
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( imputer, tasks)             # <<<<<<<<<<<<<<
 *         Imputations={}
 *         for item in results:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_map); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_imputer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_v_tasks};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_v_tasks};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_tasks);
    __Pyx_GIVEREF(__pyx_v_tasks);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_v_tasks);
    __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_results = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":380
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 *         results = self.pool.map( imputer, tasks)
 *         Imputations={}             # <<<<<<<<<<<<<<
 *         for item in results:
 *             Imputations[item.name]=(item.pgeno,item.pZ,item.path)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_Imputations = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":381
 *         results = self.pool.map( imputer, tasks)
 *         Imputations={}
 *         for item in results:             # <<<<<<<<<<<<<<
 *             Imputations[item.name]=(item.pgeno,item.pZ,item.path)
 *         return Imputations
 */
  if (likely(PyList_CheckExact(__pyx_v_results)) || PyTuple_CheckExact(__pyx_v_results)) {
    __pyx_t_2 = __pyx_v_results; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_results); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 381, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 381, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 381, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_12(__pyx_t_2);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 381, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":382
 *         Imputations={}
 *         for item in results:
 *             Imputations[item.name]=(item.pgeno,item.pZ,item.path)             # <<<<<<<<<<<<<<
 *         return Imputations
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_pgeno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_pZ); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_10 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyDict_SetItem(__pyx_v_Imputations, __pyx_t_1, __pyx_t_7) < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":381
 *         results = self.pool.map( imputer, tasks)
 *         Imputations={}
 *         for item in results:             # <<<<<<<<<<<<<<
 *             Imputations[item.name]=(item.pgeno,item.pZ,item.path)
 *         return Imputations
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":383
 *         for item in results:
 *             Imputations[item.name]=(item.pgeno,item.pZ,item.path)
 *         return Imputations             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_Imputations);
  __pyx_r = __pyx_v_Imputations;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":376
 *         #     tasks=[]
 * 
 *     def impute(self,parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hapflk.fastphase.fastphase.impute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tasks);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_Imputations);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_name);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_hap);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_name);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":387
 * 
 * 
 * def fitter( item):             # <<<<<<<<<<<<<<
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_1fitter(PyObject *__pyx_self, PyObject *__pyx_v_item); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_1fitter = {"fitter", (PyCFunction)__pyx_pw_6hapflk_9fastphase_1fitter, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_1fitter(PyObject *__pyx_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fitter (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_fitter(__pyx_self, ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_fitter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_v_hLogLike = NULL;
  PyObject *__pyx_v_top = NULL;
  PyObject *__pyx_v_bot = NULL;
  PyObject *__pyx_v_jmk = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_gLogLike = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("fitter", 0);

  /* "fastphase/fastphase.pyx":389
 * def fitter( item):
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :             # <<<<<<<<<<<<<<
 *         try:
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_haplo, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":390
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 *         try:             # <<<<<<<<<<<<<<
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         except ValueError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "fastphase/fastphase.pyx":391
 *     if item.type == 'haplo' :
 *         try:
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)             # <<<<<<<<<<<<<<
 *         except ValueError:
 *             print( item.data)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_alpha); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_theta); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 391, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_rho); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 391, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 391, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __pyx_f_6hapflk_9fastphase_hapCalc(__pyx_t_1, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_int_0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 391, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
          PyObject* sequence = __pyx_t_9;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 4)) {
            if (size > 4) __Pyx_RaiseTooManyValuesError(4);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 391, __pyx_L4_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 3); 
          } else {
            __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 3); 
          }
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          #else
          {
            Py_ssize_t i;
            PyObject** temps[4] = {&__pyx_t_8,&__pyx_t_7,&__pyx_t_6,&__pyx_t_1};
            for (i=0; i < 4; i++) {
              PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 391, __pyx_L4_error)
              __Pyx_GOTREF(item);
              *(temps[i]) = item;
            }
          }
          #endif
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else {
          Py_ssize_t index = -1;
          PyObject** temps[4] = {&__pyx_t_8,&__pyx_t_7,&__pyx_t_6,&__pyx_t_1};
          __pyx_t_10 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 391, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
          for (index=0; index < 4; index++) {
            PyObject* item = __pyx_t_11(__pyx_t_10); if (unlikely(!item)) goto __pyx_L10_unpacking_failed;
            __Pyx_GOTREF(item);
            *(temps[index]) = item;
          }
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 4) < 0) __PYX_ERR(0, 391, __pyx_L4_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L11_unpacking_done;
          __pyx_L10_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 391, __pyx_L4_error)
          __pyx_L11_unpacking_done:;
        }
        __pyx_v_hLogLike = __pyx_t_8;
        __pyx_t_8 = 0;
        __pyx_v_top = __pyx_t_7;
        __pyx_t_7 = 0;
        __pyx_v_bot = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_v_jmk = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "fastphase/fastphase.pyx":390
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 *         try:             # <<<<<<<<<<<<<<
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         except ValueError:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":392
 *         try:
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         except ValueError:             # <<<<<<<<<<<<<<
 *             print( item.data)
 *             raise
 */
      __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_12) {
        __Pyx_AddTraceback("hapflk.fastphase.fitter", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 392, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);

        /* "fastphase/fastphase.pyx":393
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         except ValueError:
 *             print( item.data)             # <<<<<<<<<<<<<<
 *             raise
 *         res = fitOutData(hLogLike,top,bot,jmk,1)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 393, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 393, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "fastphase/fastphase.pyx":394
 *         except ValueError:
 *             print( item.data)
 *             raise             # <<<<<<<<<<<<<<
 *         res = fitOutData(hLogLike,top,bot,jmk,1)
 *     elif item.type == 'geno' :
 */
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_1, __pyx_t_6);
        __pyx_t_9 = 0; __pyx_t_1 = 0; __pyx_t_6 = 0; 
        __PYX_ERR(0, 394, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "fastphase/fastphase.pyx":390
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 *         try:             # <<<<<<<<<<<<<<
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         except ValueError:
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "fastphase/fastphase.pyx":395
 *             print( item.data)
 *             raise
 *         res = fitOutData(hLogLike,top,bot,jmk,1)             # <<<<<<<<<<<<<<
 *     elif item.type == 'geno' :
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_fitOutData); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_v_hLogLike, __pyx_v_top, __pyx_v_bot, __pyx_v_jmk, __pyx_int_1};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 5+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_v_hLogLike, __pyx_v_top, __pyx_v_bot, __pyx_v_jmk, __pyx_int_1};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 5+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(5+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(__pyx_v_hLogLike);
      __Pyx_GIVEREF(__pyx_v_hLogLike);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_12, __pyx_v_hLogLike);
      __Pyx_INCREF(__pyx_v_top);
      __Pyx_GIVEREF(__pyx_v_top);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_12, __pyx_v_top);
      __Pyx_INCREF(__pyx_v_bot);
      __Pyx_GIVEREF(__pyx_v_bot);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_12, __pyx_v_bot);
      __Pyx_INCREF(__pyx_v_jmk);
      __Pyx_GIVEREF(__pyx_v_jmk);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_12, __pyx_v_jmk);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_12, __pyx_int_1);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_res = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":389
 * def fitter( item):
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :             # <<<<<<<<<<<<<<
 *         try:
 *             hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":396
 *             raise
 *         res = fitOutData(hLogLike,top,bot,jmk,1)
 *     elif item.type == 'geno' :             # <<<<<<<<<<<<<<
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         res = fitOutData(gLogLike,top,bot,jmk,2)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_geno, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":397
 *         res = fitOutData(hLogLike,top,bot,jmk,1)
 *     elif item.type == 'geno' :
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)             # <<<<<<<<<<<<<<
 *         res = fitOutData(gLogLike,top,bot,jmk,2)
 *     return res
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_alpha); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_theta); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_rho); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = __pyx_f_6hapflk_9fastphase_genCalc(__pyx_t_6, __pyx_t_1, __pyx_t_8, __pyx_t_9, __pyx_int_0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 397, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 3); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 3); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_9,&__pyx_t_8,&__pyx_t_1,&__pyx_t_6};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 397, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[4] = {&__pyx_t_9,&__pyx_t_8,&__pyx_t_1,&__pyx_t_6};
      __pyx_t_10 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
      for (index=0; index < 4; index++) {
        PyObject* item = __pyx_t_11(__pyx_t_10); if (unlikely(!item)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 4) < 0) __PYX_ERR(0, 397, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 397, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __pyx_v_gLogLike = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_top = __pyx_t_8;
    __pyx_t_8 = 0;
    __pyx_v_bot = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_v_jmk = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":398
 *     elif item.type == 'geno' :
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         res = fitOutData(gLogLike,top,bot,jmk,2)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_fitOutData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[6] = {__pyx_t_1, __pyx_v_gLogLike, __pyx_v_top, __pyx_v_bot, __pyx_v_jmk, __pyx_int_2};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 5+__pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[6] = {__pyx_t_1, __pyx_v_gLogLike, __pyx_v_top, __pyx_v_bot, __pyx_v_jmk, __pyx_int_2};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 5+__pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(5+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_gLogLike);
      __Pyx_GIVEREF(__pyx_v_gLogLike);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_12, __pyx_v_gLogLike);
      __Pyx_INCREF(__pyx_v_top);
      __Pyx_GIVEREF(__pyx_v_top);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_12, __pyx_v_top);
      __Pyx_INCREF(__pyx_v_bot);
      __Pyx_GIVEREF(__pyx_v_bot);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_12, __pyx_v_bot);
      __Pyx_INCREF(__pyx_v_jmk);
      __Pyx_GIVEREF(__pyx_v_jmk);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_12, __pyx_v_jmk);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_12, __pyx_int_2);
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_res = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "fastphase/fastphase.pyx":396
 *             raise
 *         res = fitOutData(hLogLike,top,bot,jmk,1)
 *     elif item.type == 'geno' :             # <<<<<<<<<<<<<<
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         res = fitOutData(gLogLike,top,bot,jmk,2)
 */
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":399
 *         gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
 *         res = fitOutData(gLogLike,top,bot,jmk,2)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def viterber( item):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_res)) { __Pyx_RaiseUnboundLocalError("res"); __PYX_ERR(0, 399, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":387
 * 
 * 
 * def fitter( item):             # <<<<<<<<<<<<<<
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hapflk.fastphase.fitter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hLogLike);
  __Pyx_XDECREF(__pyx_v_top);
  __Pyx_XDECREF(__pyx_v_bot);
  __Pyx_XDECREF(__pyx_v_jmk);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_gLogLike);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":401
 *     return res
 * 
 * def viterber( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_3viterber(PyObject *__pyx_self, PyObject *__pyx_v_item); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_3viterber = {"viterber", (PyCFunction)__pyx_pw_6hapflk_9fastphase_3viterber, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_3viterber(PyObject *__pyx_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("viterber (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_2viterber(__pyx_self, ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_2viterber(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item) {
  CYTHON_UNUSED double __pyx_v_nsamp;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_par = NULL;
  PyObject *__pyx_v_pth = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("viterber", 0);

  /* "fastphase/fastphase.pyx":406
 *     cdef int nLoc
 *     cdef double x
 *     cdef double nsamp=100             # <<<<<<<<<<<<<<
 *     if item.type == 'haplo':
 *         path = []
 */
  __pyx_v_nsamp = 100.0;

  /* "fastphase/fastphase.pyx":407
 *     cdef double x
 *     cdef double nsamp=100
 *     if item.type == 'haplo':             # <<<<<<<<<<<<<<
 *         path = []
 *         for par in item.parList:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_haplo, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":408
 *     cdef double nsamp=100
 *     if item.type == 'haplo':
 *         path = []             # <<<<<<<<<<<<<<
 *         for par in item.parList:
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_path = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":409
 *     if item.type == 'haplo':
 *         path = []
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 409, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 409, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 409, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 409, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_par, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fastphase/fastphase.pyx":410
 *         path = []
 *         for par in item.parList:
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)             # <<<<<<<<<<<<<<
 *             path.append(pth)
 *         res = imputeOutData(None,item.name,None,path)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __pyx_f_6hapflk_9fastphase_hapViterbi(__pyx_t_1, __pyx_t_6, __pyx_t_7, __pyx_t_8, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_pth, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":411
 *         for par in item.parList:
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)             # <<<<<<<<<<<<<<
 *         res = imputeOutData(None,item.name,None,path)
 *     elif item.type == 'geno':
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_path, __pyx_v_pth); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 411, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":409
 *     if item.type == 'haplo':
 *         path = []
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastphase/fastphase.pyx":412
 *             pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)
 *         res = imputeOutData(None,item.name,None,path)             # <<<<<<<<<<<<<<
 *     elif item.type == 'geno':
 *         path = []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_imputeOutData); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = NULL;
    __pyx_t_11 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[5] = {__pyx_t_7, Py_None, __pyx_t_8, Py_None, __pyx_v_path};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[5] = {__pyx_t_7, Py_None, __pyx_t_8, Py_None, __pyx_v_path};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_11, Py_None);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_11, __pyx_t_8);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_11, Py_None);
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_11, __pyx_v_path);
      __pyx_t_8 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_res = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fastphase/fastphase.pyx":407
 *     cdef double x
 *     cdef double nsamp=100
 *     if item.type == 'haplo':             # <<<<<<<<<<<<<<
 *         path = []
 *         for par in item.parList:
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":413
 *             path.append(pth)
 *         res = imputeOutData(None,item.name,None,path)
 *     elif item.type == 'geno':             # <<<<<<<<<<<<<<
 *         path = []
 *         for par in item.parList:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_geno, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":414
 *         res = imputeOutData(None,item.name,None,path)
 *     elif item.type == 'geno':
 *         path = []             # <<<<<<<<<<<<<<
 *         for par in item.parList:
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_path = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastphase/fastphase.pyx":415
 *     elif item.type == 'geno':
 *         path = []
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_9 = __pyx_t_3; __Pyx_INCREF(__pyx_t_9); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 415, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 415, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_9, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 415, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_9, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_9);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 415, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_par, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "fastphase/fastphase.pyx":416
 *         path = []
 *         for par in item.parList:
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)             # <<<<<<<<<<<<<<
 *             path.append( pth)
 *         res = imputeOutData(None,item.name,None, path)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __pyx_f_6hapflk_9fastphase_genViterbi(__pyx_t_3, __pyx_t_6, __pyx_t_8, __pyx_t_7, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_pth, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fastphase/fastphase.pyx":417
 *         for par in item.parList:
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)             # <<<<<<<<<<<<<<
 *         res = imputeOutData(None,item.name,None, path)
 *     return res
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_path, __pyx_v_pth); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 417, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":415
 *     elif item.type == 'geno':
 *         path = []
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)
 */
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":418
 *             pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)
 *         res = imputeOutData(None,item.name,None, path)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_imputeOutData); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_11 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_11 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, Py_None, __pyx_t_7, Py_None, __pyx_v_path};
      __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, Py_None, __pyx_t_7, Py_None, __pyx_v_path};
      __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_11, Py_None);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_11, __pyx_t_7);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_11, Py_None);
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_11, __pyx_v_path);
      __pyx_t_7 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_res = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":413
 *             path.append(pth)
 *         res = imputeOutData(None,item.name,None,path)
 *     elif item.type == 'geno':             # <<<<<<<<<<<<<<
 *         path = []
 *         for par in item.parList:
 */
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":419
 *             path.append( pth)
 *         res = imputeOutData(None,item.name,None, path)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def imputer( item):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_res)) { __Pyx_RaiseUnboundLocalError("res"); __PYX_ERR(0, 419, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":401
 *     return res
 * 
 * def viterber( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hapflk.fastphase.viterber", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XDECREF(__pyx_v_pth);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":421
 *     return res
 * 
 * def imputer( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_5imputer(PyObject *__pyx_self, PyObject *__pyx_v_item); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_5imputer = {"imputer", (PyCFunction)__pyx_pw_6hapflk_9fastphase_5imputer, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_5imputer(PyObject *__pyx_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("imputer (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_4imputer(__pyx_self, ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_4imputer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_item) {
  int __pyx_v_i;
  int __pyx_v_k;
  double __pyx_v_x;
  CYTHON_UNUSED double __pyx_v_nsamp;
  PyObject *__pyx_v_pgeno = NULL;
  PyObject *__pyx_v_probZ = NULL;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_par = NULL;
  PyObject *__pyx_v_pZ = NULL;
  PyObject *__pyx_v_pth = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  long __pyx_t_10;
  long __pyx_t_11;
  int __pyx_t_12;
  long __pyx_t_13;
  long __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  double __pyx_t_18;
  double __pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  PyObject *(*__pyx_t_22)(PyObject *);
  Py_ssize_t __pyx_t_23;
  PyObject *(*__pyx_t_24)(PyObject *);
  PyObject *__pyx_t_25 = NULL;
  double __pyx_t_26;
  __Pyx_RefNannySetupContext("imputer", 0);

  /* "fastphase/fastphase.pyx":426
 *     cdef int nLoc
 *     cdef double x
 *     cdef double nsamp=100             # <<<<<<<<<<<<<<
 *     if item.type == 'haplo':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 */
  __pyx_v_nsamp = 100.0;

  /* "fastphase/fastphase.pyx":427
 *     cdef double x
 *     cdef double nsamp=100
 *     if item.type == 'haplo':             # <<<<<<<<<<<<<<
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_haplo, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":428
 *     cdef double nsamp=100
 *     if item.type == 'haplo':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)             # <<<<<<<<<<<<<<
 *         probZ=[]
 *         path = []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_pgeno = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":429
 *     if item.type == 'haplo':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]             # <<<<<<<<<<<<<<
 *         path = []
 *         x = 1.0/len(item.parList)
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_probZ = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":430
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 *         path = []             # <<<<<<<<<<<<<<
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_path = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "fastphase/fastphase.pyx":431
 *         probZ=[]
 *         path = []
 *         x = 1.0/len(item.parList)             # <<<<<<<<<<<<<<
 *         for par in item.parList:
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_t_7 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 431, __pyx_L1_error)
    }
    __pyx_v_x = (1.0 / ((double)__pyx_t_7));

    /* "fastphase/fastphase.pyx":432
 *         path = []
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             for i in range(item.nLoc):
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_1 = __pyx_t_6; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 432, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 432, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 432, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 432, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_par, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "fastphase/fastphase.pyx":433
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)             # <<<<<<<<<<<<<<
 *             for i in range(item.nLoc):
 *                 for k in range(par.nClus):
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = __pyx_f_6hapflk_9fastphase_hapCalc(__pyx_t_6, __pyx_t_4, __pyx_t_3, __pyx_t_5, __pyx_int_1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_pZ, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":434
 *         for par in item.parList:
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             for i in range(item.nLoc):             # <<<<<<<<<<<<<<
 *                 for k in range(par.nClus):
 *                     pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyInt_As_long(__pyx_t_9); if (unlikely((__pyx_t_10 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_i = __pyx_t_12;

        /* "fastphase/fastphase.pyx":435
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             for i in range(item.nLoc):
 *                 for k in range(par.nClus):             # <<<<<<<<<<<<<<
 *                     pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])
 *             probZ.append(pZ)
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nClus); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 435, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_t_9); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 435, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_14 = __pyx_t_13;
        for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
          __pyx_v_k = __pyx_t_15;

          /* "fastphase/fastphase.pyx":436
 *             for i in range(item.nLoc):
 *                 for k in range(par.nClus):
 *                     pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])             # <<<<<<<<<<<<<<
 *             probZ.append(pZ)
 *             pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_pgeno, __pyx_t_16, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
          __pyx_t_3 = 0;
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pZ, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
          __pyx_t_4 = 0;
          __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyFloat_FromDouble((__pyx_v_x * __pyx_f_6hapflk_9fastphase_hap_p_all(__pyx_t_17, __pyx_t_18, __pyx_t_19))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(__Pyx_SetItemInt(__pyx_v_pgeno, __pyx_t_16, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }

      /* "fastphase/fastphase.pyx":437
 *                 for k in range(par.nClus):
 *                     pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])
 *             probZ.append(pZ)             # <<<<<<<<<<<<<<
 *             pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_probZ, __pyx_v_pZ); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 437, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":438
 *                     pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])
 *             probZ.append(pZ)
 *             pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)             # <<<<<<<<<<<<<<
 *             path.append(pth)
 *         res = imputeOutData(pgeno,item.name,probZ,path)
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_pth, Py_None);

      /* "fastphase/fastphase.pyx":439
 *             probZ.append(pZ)
 *             pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)             # <<<<<<<<<<<<<<
 *         res = imputeOutData(pgeno,item.name,probZ,path)
 *     elif item.type == 'geno':
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_path, __pyx_v_pth); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 439, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":432
 *         path = []
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             for i in range(item.nLoc):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":440
 *             pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append(pth)
 *         res = imputeOutData(pgeno,item.name,probZ,path)             # <<<<<<<<<<<<<<
 *     elif item.type == 'geno':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_imputeOutData); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_pgeno, __pyx_t_3, __pyx_v_probZ, __pyx_v_path};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_pgeno, __pyx_t_3, __pyx_v_probZ, __pyx_v_path};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(__pyx_v_pgeno);
      __Pyx_GIVEREF(__pyx_v_pgeno);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_12, __pyx_v_pgeno);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_12, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_probZ);
      __Pyx_GIVEREF(__pyx_v_probZ);
      PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_12, __pyx_v_probZ);
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_12, __pyx_v_path);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_res = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":427
 *     cdef double x
 *     cdef double nsamp=100
 *     if item.type == 'haplo':             # <<<<<<<<<<<<<<
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":441
 *             path.append(pth)
 *         res = imputeOutData(pgeno,item.name,probZ,path)
 *     elif item.type == 'geno':             # <<<<<<<<<<<<<<
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_geno, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fastphase/fastphase.pyx":442
 *         res = imputeOutData(pgeno,item.name,probZ,path)
 *     elif item.type == 'geno':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)             # <<<<<<<<<<<<<<
 *         probZ=[]
 *         path = []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_pgeno = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":443
 *     elif item.type == 'geno':
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]             # <<<<<<<<<<<<<<
 *         path = []
 *         x = 1.0/len(item.parList)
 */
    __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_v_probZ = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":444
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 *         path = []             # <<<<<<<<<<<<<<
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:
 */
    __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_v_path = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "fastphase/fastphase.pyx":445
 *         probZ=[]
 *         path = []
 *         x = 1.0/len(item.parList)             # <<<<<<<<<<<<<<
 *         for par in item.parList:
 *             pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_t_7 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 445, __pyx_L1_error)
    }
    __pyx_v_x = (1.0 / ((double)__pyx_t_7));

    /* "fastphase/fastphase.pyx":446
 *         path = []
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             ## Calculate mean genotypes
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
      __pyx_t_1 = __pyx_t_9; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 446, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 446, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 446, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 446, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_par, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":447
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:
 *             pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)             # <<<<<<<<<<<<<<
 *             ## Calculate mean genotypes
 *             for i in range(item.nLoc):
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_rho); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_f_6hapflk_9fastphase_genCalc(__pyx_t_9, __pyx_t_5, __pyx_t_6, __pyx_t_3, __pyx_int_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_pZ, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":449
 *             pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             ## Calculate mean genotypes
 *             for i in range(item.nLoc):             # <<<<<<<<<<<<<<
 *                 for k1 in range(par.nClus):
 *                     for k2 in range(par.nClus):
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 449, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_10 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_i = __pyx_t_12;

        /* "fastphase/fastphase.pyx":450
 *             ## Calculate mean genotypes
 *             for i in range(item.nLoc):
 *                 for k1 in range(par.nClus):             # <<<<<<<<<<<<<<
 *                     for k2 in range(par.nClus):
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nClus); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
          __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_21 = 0;
          __pyx_t_22 = NULL;
        } else {
          __pyx_t_21 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 450, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_22 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 450, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        for (;;) {
          if (likely(!__pyx_t_22)) {
            if (likely(PyList_CheckExact(__pyx_t_4))) {
              if (__pyx_t_21 >= PyList_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_21); __Pyx_INCREF(__pyx_t_3); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 450, __pyx_L1_error)
              #else
              __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              if (__pyx_t_21 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_21); __Pyx_INCREF(__pyx_t_3); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 450, __pyx_L1_error)
              #else
              __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_22(__pyx_t_4);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 450, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_XDECREF_SET(__pyx_v_k1, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "fastphase/fastphase.pyx":451
 *             for i in range(item.nLoc):
 *                 for k1 in range(par.nClus):
 *                     for k2 in range(par.nClus):             # <<<<<<<<<<<<<<
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 *             probZ.append(pZ)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nClus); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
            __pyx_t_3 = __pyx_t_6; __Pyx_INCREF(__pyx_t_3); __pyx_t_23 = 0;
            __pyx_t_24 = NULL;
          } else {
            __pyx_t_23 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_24 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 451, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          for (;;) {
            if (likely(!__pyx_t_24)) {
              if (likely(PyList_CheckExact(__pyx_t_3))) {
                if (__pyx_t_23 >= PyList_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_23); __Pyx_INCREF(__pyx_t_6); __pyx_t_23++; if (unlikely(0 < 0)) __PYX_ERR(0, 451, __pyx_L1_error)
                #else
                __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_23); __pyx_t_23++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 451, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              } else {
                if (__pyx_t_23 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_23); __Pyx_INCREF(__pyx_t_6); __pyx_t_23++; if (unlikely(0 < 0)) __PYX_ERR(0, 451, __pyx_L1_error)
                #else
                __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_23); __pyx_t_23++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 451, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              }
            } else {
              __pyx_t_6 = __pyx_t_24(__pyx_t_3);
              if (unlikely(!__pyx_t_6)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 451, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_6);
            }
            __Pyx_XDECREF_SET(__pyx_v_k2, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "fastphase/fastphase.pyx":452
 *                 for k1 in range(par.nClus):
 *                     for k2 in range(par.nClus):
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])             # <<<<<<<<<<<<<<
 *             probZ.append(pZ)
 *             pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_pgeno, __pyx_t_15, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_9);
            PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
            __Pyx_INCREF(__pyx_v_k1);
            __Pyx_GIVEREF(__pyx_v_k1);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_k1);
            __Pyx_INCREF(__pyx_v_k2);
            __Pyx_GIVEREF(__pyx_v_k2);
            PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_k2);
            __pyx_t_9 = 0;
            __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_pZ, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_9); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_25);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_5);
            __Pyx_INCREF(__pyx_v_k1);
            __Pyx_GIVEREF(__pyx_v_k1);
            PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_v_k1);
            __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_t_25); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_25 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_25);
            __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_25);
            PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_25);
            __Pyx_INCREF(__pyx_v_k2);
            __Pyx_GIVEREF(__pyx_v_k2);
            PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_k2);
            __pyx_t_25 = 0;
            __pyx_t_25 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_9); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_25);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_t_25); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            __pyx_t_25 = PyFloat_FromDouble(__pyx_f_6hapflk_9fastphase_gen_p_geno(__pyx_t_16, __pyx_t_19, __pyx_t_18, __pyx_t_26)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_25);
            __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_t_6, __pyx_t_25); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (unlikely(__Pyx_SetItemInt(__pyx_v_pgeno, __pyx_t_15, __pyx_t_9, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 452, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "fastphase/fastphase.pyx":451
 *             for i in range(item.nLoc):
 *                 for k1 in range(par.nClus):
 *                     for k2 in range(par.nClus):             # <<<<<<<<<<<<<<
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 *             probZ.append(pZ)
 */
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "fastphase/fastphase.pyx":450
 *             ## Calculate mean genotypes
 *             for i in range(item.nLoc):
 *                 for k1 in range(par.nClus):             # <<<<<<<<<<<<<<
 *                     for k2 in range(par.nClus):
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 */
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }

      /* "fastphase/fastphase.pyx":453
 *                     for k2 in range(par.nClus):
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 *             probZ.append(pZ)             # <<<<<<<<<<<<<<
 *             pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_probZ, __pyx_v_pZ); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":454
 *                         pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
 *             probZ.append(pZ)
 *             pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)             # <<<<<<<<<<<<<<
 *             path.append( pth)
 *         pgeno/=len(item.parList)
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_pth, Py_None);

      /* "fastphase/fastphase.pyx":455
 *             probZ.append(pZ)
 *             pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)             # <<<<<<<<<<<<<<
 *         pgeno/=len(item.parList)
 *         res = imputeOutData(pgeno,item.name,probZ, path)
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_path, __pyx_v_pth); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 455, __pyx_L1_error)

      /* "fastphase/fastphase.pyx":446
 *         path = []
 *         x = 1.0/len(item.parList)
 *         for par in item.parList:             # <<<<<<<<<<<<<<
 *             pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)
 *             ## Calculate mean genotypes
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":456
 *             pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)
 *             path.append( pth)
 *         pgeno/=len(item.parList)             # <<<<<<<<<<<<<<
 *         res = imputeOutData(pgeno,item.name,probZ, path)
 *     return res
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_parList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyNumber_InPlaceDivide(__pyx_v_pgeno, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_pgeno, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":457
 *             path.append( pth)
 *         pgeno/=len(item.parList)
 *         res = imputeOutData(pgeno,item.name,probZ, path)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_imputeOutData); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_pgeno, __pyx_t_3, __pyx_v_probZ, __pyx_v_path};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_v_pgeno, __pyx_t_3, __pyx_v_probZ, __pyx_v_path};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_25 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(__pyx_v_pgeno);
      __Pyx_GIVEREF(__pyx_v_pgeno);
      PyTuple_SET_ITEM(__pyx_t_25, 0+__pyx_t_12, __pyx_v_pgeno);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_25, 1+__pyx_t_12, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_probZ);
      __Pyx_GIVEREF(__pyx_v_probZ);
      PyTuple_SET_ITEM(__pyx_t_25, 2+__pyx_t_12, __pyx_v_probZ);
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_25, 3+__pyx_t_12, __pyx_v_path);
      __pyx_t_3 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_25, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_res = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "fastphase/fastphase.pyx":441
 *             path.append(pth)
 *         res = imputeOutData(pgeno,item.name,probZ,path)
 *     elif item.type == 'geno':             # <<<<<<<<<<<<<<
 *         pgeno=np.zeros(item.nLoc,dtype=np.float64)
 *         probZ=[]
 */
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":458
 *         pgeno/=len(item.parList)
 *         res = imputeOutData(pgeno,item.name,probZ, path)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * ##### Cluster switch functions
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_res)) { __Pyx_RaiseUnboundLocalError("res"); __PYX_ERR(0, 458, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":421
 *     return res
 * 
 * def imputer( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_AddTraceback("hapflk.fastphase.imputer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pgeno);
  __Pyx_XDECREF(__pyx_v_probZ);
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XDECREF(__pyx_v_pZ);
  __Pyx_XDECREF(__pyx_v_pth);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":462
 * ##### Cluster switch functions
 * 
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_haplo_tot( np.ndarray[np.int32_t, ndim=1] h, int nK ):             # <<<<<<<<<<<<<<
 *     cdef int l,nL
 *     nL= h.shape[0]
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_7calc_cost_matrix_haplo_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_f_6hapflk_9fastphase_calc_cost_matrix_haplo_tot(PyArrayObject *__pyx_v_h, int __pyx_v_nK, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_l;
  int __pyx_v_nL;
  PyArrayObject *__pyx_v_res = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_h;
  __Pyx_Buffer __pyx_pybuffer_h;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_res;
  __Pyx_Buffer __pyx_pybuffer_res;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  long __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  __Pyx_RefNannySetupContext("calc_cost_matrix_haplo_tot", 0);
  __pyx_pybuffer_res.pybuffer.buf = NULL;
  __pyx_pybuffer_res.refcount = 0;
  __pyx_pybuffernd_res.data = NULL;
  __pyx_pybuffernd_res.rcbuffer = &__pyx_pybuffer_res;
  __pyx_pybuffer_h.pybuffer.buf = NULL;
  __pyx_pybuffer_h.refcount = 0;
  __pyx_pybuffernd_h.data = NULL;
  __pyx_pybuffernd_h.rcbuffer = &__pyx_pybuffer_h;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_h.rcbuffer->pybuffer, (PyObject*)__pyx_v_h, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 462, __pyx_L1_error)
  }
  __pyx_pybuffernd_h.diminfo[0].strides = __pyx_pybuffernd_h.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_h.diminfo[0].shape = __pyx_pybuffernd_h.rcbuffer->pybuffer.shape[0];

  /* "fastphase/fastphase.pyx":464
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_haplo_tot( np.ndarray[np.int32_t, ndim=1] h, int nK ):
 *     cdef int l,nL
 *     nL= h.shape[0]             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)
 * 
 */
  __pyx_v_nL = (__pyx_v_h->dimensions[0]);

  /* "fastphase/fastphase.pyx":465
 *     cdef int l,nL
 *     nL= h.shape[0]
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     for l in range(nL -1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_nL - 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_res.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_res = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_res.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 465, __pyx_L1_error)
    } else {__pyx_pybuffernd_res.diminfo[0].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_res.diminfo[0].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_res.diminfo[1].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_res.diminfo[1].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_res.diminfo[2].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_res.diminfo[2].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_res = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":467
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)
 * 
 *     for l in range(nL -1):             # <<<<<<<<<<<<<<
 *         res[l, h[l], h[l+1]] -= 1
 *     return res
 */
  __pyx_t_7 = (__pyx_v_nL - 1);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_l = __pyx_t_9;

    /* "fastphase/fastphase.pyx":468
 * 
 *     for l in range(nL -1):
 *         res[l, h[l], h[l+1]] -= 1             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_10 = __pyx_v_l;
    if (__pyx_t_10 < 0) __pyx_t_10 += __pyx_pybuffernd_h.diminfo[0].shape;
    __pyx_t_11 = (__pyx_v_l + 1);
    if (__pyx_t_11 < 0) __pyx_t_11 += __pyx_pybuffernd_h.diminfo[0].shape;
    __pyx_t_12 = __pyx_v_l;
    __pyx_t_13 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_h.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_h.diminfo[0].strides));
    __pyx_t_14 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_h.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_h.diminfo[0].strides));
    if (__pyx_t_12 < 0) __pyx_t_12 += __pyx_pybuffernd_res.diminfo[0].shape;
    if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_res.diminfo[1].shape;
    if (__pyx_t_14 < 0) __pyx_t_14 += __pyx_pybuffernd_res.diminfo[2].shape;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_14, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;
  }

  /* "fastphase/fastphase.pyx":469
 *     for l in range(nL -1):
 *         res[l, h[l], h[l+1]] -= 1
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def calc_cost_matrix_haplo( args):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_res));
  __pyx_r = ((PyArrayObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":462
 * ##### Cluster switch functions
 * 
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_haplo_tot( np.ndarray[np.int32_t, ndim=1] h, int nK ):             # <<<<<<<<<<<<<<
 *     cdef int l,nL
 *     nL= h.shape[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_h.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_res.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_haplo_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_h.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_res.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_7calc_cost_matrix_haplo_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_7calc_cost_matrix_haplo_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_h = 0;
  int __pyx_v_nK;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_cost_matrix_haplo_tot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_h,&__pyx_n_s_nK,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nK)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_cost_matrix_haplo_tot", 1, 2, 2, 1); __PYX_ERR(0, 462, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_cost_matrix_haplo_tot") < 0)) __PYX_ERR(0, 462, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_h = ((PyArrayObject *)values[0]);
    __pyx_v_nK = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nK == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_cost_matrix_haplo_tot", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_haplo_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_h), __pyx_ptype_5numpy_ndarray, 1, "h", 0))) __PYX_ERR(0, 462, __pyx_L1_error)
  __pyx_r = __pyx_pf_6hapflk_9fastphase_6calc_cost_matrix_haplo_tot(__pyx_self, __pyx_v_h, __pyx_v_nK);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_6calc_cost_matrix_haplo_tot(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_h, int __pyx_v_nK) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_h;
  __Pyx_Buffer __pyx_pybuffer_h;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("calc_cost_matrix_haplo_tot", 0);
  __pyx_pybuffer_h.pybuffer.buf = NULL;
  __pyx_pybuffer_h.refcount = 0;
  __pyx_pybuffernd_h.data = NULL;
  __pyx_pybuffernd_h.rcbuffer = &__pyx_pybuffer_h;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_h.rcbuffer->pybuffer, (PyObject*)__pyx_v_h, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 462, __pyx_L1_error)
  }
  __pyx_pybuffernd_h.diminfo[0].strides = __pyx_pybuffernd_h.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_h.diminfo[0].shape = __pyx_pybuffernd_h.rcbuffer->pybuffer.shape[0];
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6hapflk_9fastphase_calc_cost_matrix_haplo_tot(__pyx_v_h, __pyx_v_nK, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_h.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_haplo_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_h.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":471
 *     return res
 * 
 * def calc_cost_matrix_haplo( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k = vit[0]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_9calc_cost_matrix_haplo(PyObject *__pyx_self, PyObject *__pyx_v_args); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_9calc_cost_matrix_haplo = {"calc_cost_matrix_haplo", (PyCFunction)__pyx_pw_6hapflk_9fastphase_9calc_cost_matrix_haplo, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_9calc_cost_matrix_haplo(PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_cost_matrix_haplo (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_8calc_cost_matrix_haplo(__pyx_self, ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_8calc_cost_matrix_haplo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_vit = NULL;
  PyObject *__pyx_v_nK = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_kp = NULL;
  PyObject *__pyx_v_cost_mat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("calc_cost_matrix_haplo", 0);

  /* "fastphase/fastphase.pyx":472
 * 
 * def calc_cost_matrix_haplo( args):
 *     vit, nK = args             # <<<<<<<<<<<<<<
 *     k = vit[0]
 *     kp = vit[1]
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_args))) || (PyList_CheckExact(__pyx_v_args))) {
    PyObject* sequence = __pyx_v_args;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 472, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 472, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_vit = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_nK = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":473
 * def calc_cost_matrix_haplo( args):
 *     vit, nK = args
 *     k = vit[0]             # <<<<<<<<<<<<<<
 *     kp = vit[1]
 *     cost_mat = np.zeros((nK,nK))
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_vit, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_k = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":474
 *     vit, nK = args
 *     k = vit[0]
 *     kp = vit[1]             # <<<<<<<<<<<<<<
 *     cost_mat = np.zeros((nK,nK))
 *     cost_mat[k, kp] -= 1
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_vit, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_kp = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":475
 *     k = vit[0]
 *     kp = vit[1]
 *     cost_mat = np.zeros((nK,nK))             # <<<<<<<<<<<<<<
 *     cost_mat[k, kp] -= 1
 *     return cost_mat
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nK);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nK);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_cost_mat = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":476
 *     kp = vit[1]
 *     cost_mat = np.zeros((nK,nK))
 *     cost_mat[k, kp] -= 1             # <<<<<<<<<<<<<<
 *     return cost_mat
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_k);
  __Pyx_GIVEREF(__pyx_v_k);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k);
  __Pyx_INCREF(__pyx_v_kp);
  __Pyx_GIVEREF(__pyx_v_kp);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":477
 *     cost_mat = np.zeros((nK,nK))
 *     cost_mat[k, kp] -= 1
 *     return cost_mat             # <<<<<<<<<<<<<<
 * 
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_geno_tot( np.ndarray[np.int32_t, ndim=2] g, int nK ):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cost_mat);
  __pyx_r = __pyx_v_cost_mat;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":471
 *     return res
 * 
 * def calc_cost_matrix_haplo( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k = vit[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_haplo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vit);
  __Pyx_XDECREF(__pyx_v_nK);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_kp);
  __Pyx_XDECREF(__pyx_v_cost_mat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":479
 *     return cost_mat
 * 
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_geno_tot( np.ndarray[np.int32_t, ndim=2] g, int nK ):             # <<<<<<<<<<<<<<
 *     cdef int l,nL
 *     cdef int k1, k2, kp1, kp2
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_11calc_cost_matrix_geno_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyArrayObject *__pyx_f_6hapflk_9fastphase_calc_cost_matrix_geno_tot(PyArrayObject *__pyx_v_g, int __pyx_v_nK, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_l;
  int __pyx_v_nL;
  int __pyx_v_k1;
  int __pyx_v_k2;
  int __pyx_v_kp1;
  int __pyx_v_kp2;
  PyArrayObject *__pyx_v_res = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_g;
  __Pyx_Buffer __pyx_pybuffer_g;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_res;
  __Pyx_Buffer __pyx_pybuffer_res;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  long __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  int __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  __Pyx_RefNannySetupContext("calc_cost_matrix_geno_tot", 0);
  __pyx_pybuffer_res.pybuffer.buf = NULL;
  __pyx_pybuffer_res.refcount = 0;
  __pyx_pybuffernd_res.data = NULL;
  __pyx_pybuffernd_res.rcbuffer = &__pyx_pybuffer_res;
  __pyx_pybuffer_g.pybuffer.buf = NULL;
  __pyx_pybuffer_g.refcount = 0;
  __pyx_pybuffernd_g.data = NULL;
  __pyx_pybuffernd_g.rcbuffer = &__pyx_pybuffer_g;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_g.rcbuffer->pybuffer, (PyObject*)__pyx_v_g, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 479, __pyx_L1_error)
  }
  __pyx_pybuffernd_g.diminfo[0].strides = __pyx_pybuffernd_g.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_g.diminfo[0].shape = __pyx_pybuffernd_g.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_g.diminfo[1].strides = __pyx_pybuffernd_g.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_g.diminfo[1].shape = __pyx_pybuffernd_g.rcbuffer->pybuffer.shape[1];

  /* "fastphase/fastphase.pyx":482
 *     cdef int l,nL
 *     cdef int k1, k2, kp1, kp2
 *     nL= g.shape[0]             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)
 * 
 */
  __pyx_v_nL = (__pyx_v_g->dimensions[0]);

  /* "fastphase/fastphase.pyx":483
 *     cdef int k1, k2, kp1, kp2
 *     nL= g.shape[0]
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     for l in range(nL -1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_nL - 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 483, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_res.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_res = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_res.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 483, __pyx_L1_error)
    } else {__pyx_pybuffernd_res.diminfo[0].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_res.diminfo[0].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_res.diminfo[1].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_res.diminfo[1].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_res.diminfo[2].strides = __pyx_pybuffernd_res.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_res.diminfo[2].shape = __pyx_pybuffernd_res.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_res = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":485
 *     cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)
 * 
 *     for l in range(nL -1):             # <<<<<<<<<<<<<<
 *         if (k1 == kp1):
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 */
  __pyx_t_7 = (__pyx_v_nL - 1);
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_l = __pyx_t_9;

    /* "fastphase/fastphase.pyx":486
 * 
 *     for l in range(nL -1):
 *         if (k1 == kp1):             # <<<<<<<<<<<<<<
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *                 res[l, k1, kp1] -= 1
 */
    __pyx_t_10 = ((__pyx_v_k1 == __pyx_v_kp1) != 0);
    if (__pyx_t_10) {

      /* "fastphase/fastphase.pyx":487
 *     for l in range(nL -1):
 *         if (k1 == kp1):
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2             # <<<<<<<<<<<<<<
 *                 res[l, k1, kp1] -= 1
 *                 res[l, k2, kp2] -= 1
 */
      __pyx_t_10 = ((__pyx_v_k2 == __pyx_v_kp2) != 0);
      if (__pyx_t_10) {

        /* "fastphase/fastphase.pyx":488
 *         if (k1 == kp1):
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *                 res[l, k1, kp1] -= 1             # <<<<<<<<<<<<<<
 *                 res[l, k2, kp2] -= 1
 *             else: ## k1 == kp1 & k2 != kp2
 */
        __pyx_t_11 = __pyx_v_l;
        __pyx_t_12 = __pyx_v_k1;
        __pyx_t_13 = __pyx_v_kp1;
        if (__pyx_t_11 < 0) __pyx_t_11 += __pyx_pybuffernd_res.diminfo[0].shape;
        if (__pyx_t_12 < 0) __pyx_t_12 += __pyx_pybuffernd_res.diminfo[1].shape;
        if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_res.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_12, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_13, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

        /* "fastphase/fastphase.pyx":489
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *                 res[l, k1, kp1] -= 1
 *                 res[l, k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *             else: ## k1 == kp1 & k2 != kp2
 *                 res[l, k1, kp1] -= 1 ## diagonal
 */
        __pyx_t_14 = __pyx_v_l;
        __pyx_t_15 = __pyx_v_k2;
        __pyx_t_16 = __pyx_v_kp2;
        if (__pyx_t_14 < 0) __pyx_t_14 += __pyx_pybuffernd_res.diminfo[0].shape;
        if (__pyx_t_15 < 0) __pyx_t_15 += __pyx_pybuffernd_res.diminfo[1].shape;
        if (__pyx_t_16 < 0) __pyx_t_16 += __pyx_pybuffernd_res.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_16, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

        /* "fastphase/fastphase.pyx":487
 *     for l in range(nL -1):
 *         if (k1 == kp1):
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2             # <<<<<<<<<<<<<<
 *                 res[l, k1, kp1] -= 1
 *                 res[l, k2, kp2] -= 1
 */
        goto __pyx_L6;
      }

      /* "fastphase/fastphase.pyx":491
 *                 res[l, k2, kp2] -= 1
 *             else: ## k1 == kp1 & k2 != kp2
 *                 res[l, k1, kp1] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *                 res[l, k2, kp2] -= 1
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 */
      /*else*/ {
        __pyx_t_17 = __pyx_v_l;
        __pyx_t_18 = __pyx_v_k1;
        __pyx_t_19 = __pyx_v_kp1;
        if (__pyx_t_17 < 0) __pyx_t_17 += __pyx_pybuffernd_res.diminfo[0].shape;
        if (__pyx_t_18 < 0) __pyx_t_18 += __pyx_pybuffernd_res.diminfo[1].shape;
        if (__pyx_t_19 < 0) __pyx_t_19 += __pyx_pybuffernd_res.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_19, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

        /* "fastphase/fastphase.pyx":492
 *             else: ## k1 == kp1 & k2 != kp2
 *                 res[l, k1, kp1] -= 1 ## diagonal
 *                 res[l, k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *             res[l, k1, kp1] -= 1 ## diagonal
 */
        __pyx_t_20 = __pyx_v_l;
        __pyx_t_21 = __pyx_v_k2;
        __pyx_t_22 = __pyx_v_kp2;
        if (__pyx_t_20 < 0) __pyx_t_20 += __pyx_pybuffernd_res.diminfo[0].shape;
        if (__pyx_t_21 < 0) __pyx_t_21 += __pyx_pybuffernd_res.diminfo[1].shape;
        if (__pyx_t_22 < 0) __pyx_t_22 += __pyx_pybuffernd_res.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_22, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;
      }
      __pyx_L6:;

      /* "fastphase/fastphase.pyx":486
 * 
 *     for l in range(nL -1):
 *         if (k1 == kp1):             # <<<<<<<<<<<<<<
 *             if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *                 res[l, k1, kp1] -= 1
 */
      goto __pyx_L5;
    }

    /* "fastphase/fastphase.pyx":493
 *                 res[l, k1, kp1] -= 1 ## diagonal
 *                 res[l, k2, kp2] -= 1
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *             res[l, k1, kp1] -= 1 ## diagonal
 *             res[l, k2, kp2] -= 1
 */
    __pyx_t_10 = ((__pyx_v_k2 == __pyx_v_kp2) != 0);
    if (__pyx_t_10) {

      /* "fastphase/fastphase.pyx":494
 *                 res[l, k2, kp2] -= 1
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *             res[l, k1, kp1] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *             res[l, k2, kp2] -= 1
 *         elif ( k1 == kp2) and (k2 != kp1):
 */
      __pyx_t_23 = __pyx_v_l;
      __pyx_t_24 = __pyx_v_k1;
      __pyx_t_25 = __pyx_v_kp1;
      if (__pyx_t_23 < 0) __pyx_t_23 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_24 < 0) __pyx_t_24 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_25, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

      /* "fastphase/fastphase.pyx":495
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *             res[l, k1, kp1] -= 1 ## diagonal
 *             res[l, k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *         elif ( k1 == kp2) and (k2 != kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal
 */
      __pyx_t_26 = __pyx_v_l;
      __pyx_t_27 = __pyx_v_k2;
      __pyx_t_28 = __pyx_v_kp2;
      if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_28, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

      /* "fastphase/fastphase.pyx":493
 *                 res[l, k1, kp1] -= 1 ## diagonal
 *                 res[l, k2, kp2] -= 1
 *         elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *             res[l, k1, kp1] -= 1 ## diagonal
 *             res[l, k2, kp2] -= 1
 */
      goto __pyx_L5;
    }

    /* "fastphase/fastphase.pyx":496
 *             res[l, k1, kp1] -= 1 ## diagonal
 *             res[l, k2, kp2] -= 1
 *         elif ( k1 == kp2) and (k2 != kp1):             # <<<<<<<<<<<<<<
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[l, k2, kp1] -= 1
 */
    __pyx_t_29 = ((__pyx_v_k1 == __pyx_v_kp2) != 0);
    if (__pyx_t_29) {
    } else {
      __pyx_t_10 = __pyx_t_29;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_29 = ((__pyx_v_k2 != __pyx_v_kp1) != 0);
    __pyx_t_10 = __pyx_t_29;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_10) {

      /* "fastphase/fastphase.pyx":497
 *             res[l, k2, kp2] -= 1
 *         elif ( k1 == kp2) and (k2 != kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *             res[l, k2, kp1] -= 1
 *         elif ( k1 != kp2) and (k2 == kp1):
 */
      __pyx_t_30 = __pyx_v_l;
      __pyx_t_31 = __pyx_v_k1;
      __pyx_t_32 = __pyx_v_kp2;
      if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_31, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_32, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

      /* "fastphase/fastphase.pyx":498
 *         elif ( k1 == kp2) and (k2 != kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[l, k2, kp1] -= 1             # <<<<<<<<<<<<<<
 *         elif ( k1 != kp2) and (k2 == kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal
 */
      __pyx_t_33 = __pyx_v_l;
      __pyx_t_34 = __pyx_v_k2;
      __pyx_t_35 = __pyx_v_kp1;
      if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_34, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_35, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

      /* "fastphase/fastphase.pyx":496
 *             res[l, k1, kp1] -= 1 ## diagonal
 *             res[l, k2, kp2] -= 1
 *         elif ( k1 == kp2) and (k2 != kp1):             # <<<<<<<<<<<<<<
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[l, k2, kp1] -= 1
 */
      goto __pyx_L5;
    }

    /* "fastphase/fastphase.pyx":499
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[l, k2, kp1] -= 1
 *         elif ( k1 != kp2) and (k2 == kp1):             # <<<<<<<<<<<<<<
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[k2, kp1] -= 1
 */
    __pyx_t_29 = ((__pyx_v_k1 != __pyx_v_kp2) != 0);
    if (__pyx_t_29) {
    } else {
      __pyx_t_10 = __pyx_t_29;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_29 = ((__pyx_v_k2 == __pyx_v_kp1) != 0);
    __pyx_t_10 = __pyx_t_29;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_10) {

      /* "fastphase/fastphase.pyx":500
 *             res[l, k2, kp1] -= 1
 *         elif ( k1 != kp2) and (k2 == kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *             res[k2, kp1] -= 1
 *         else:
 */
      __pyx_t_36 = __pyx_v_l;
      __pyx_t_37 = __pyx_v_k1;
      __pyx_t_38 = __pyx_v_kp2;
      if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_37, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_38, __pyx_pybuffernd_res.diminfo[2].strides) -= 1.0;

      /* "fastphase/fastphase.pyx":501
 *         elif ( k1 != kp2) and (k2 == kp1):
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[k2, kp1] -= 1             # <<<<<<<<<<<<<<
 *         else:
 *             res[l, k1, kp1] -= 0.5
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_k2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_kp1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_res), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_5, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_res), __pyx_t_4, __pyx_t_1) < 0)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastphase/fastphase.pyx":499
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[l, k2, kp1] -= 1
 *         elif ( k1 != kp2) and (k2 == kp1):             # <<<<<<<<<<<<<<
 *             res[l, k1, kp2] -= 1 ## diagonal
 *             res[k2, kp1] -= 1
 */
      goto __pyx_L5;
    }

    /* "fastphase/fastphase.pyx":503
 *             res[k2, kp1] -= 1
 *         else:
 *             res[l, k1, kp1] -= 0.5             # <<<<<<<<<<<<<<
 *             res[l, k1, kp2] -= 0.5
 *             res[l, k2, kp1] -= 0.5
 */
    /*else*/ {
      __pyx_t_39 = __pyx_v_l;
      __pyx_t_40 = __pyx_v_k1;
      __pyx_t_41 = __pyx_v_kp1;
      if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_40, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_41, __pyx_pybuffernd_res.diminfo[2].strides) -= 0.5;

      /* "fastphase/fastphase.pyx":504
 *         else:
 *             res[l, k1, kp1] -= 0.5
 *             res[l, k1, kp2] -= 0.5             # <<<<<<<<<<<<<<
 *             res[l, k2, kp1] -= 0.5
 *             res[l, k2, kp2] -= 0.5
 */
      __pyx_t_42 = __pyx_v_l;
      __pyx_t_43 = __pyx_v_k1;
      __pyx_t_44 = __pyx_v_kp2;
      if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_44 < 0) __pyx_t_44 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_42, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_43, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_44, __pyx_pybuffernd_res.diminfo[2].strides) -= 0.5;

      /* "fastphase/fastphase.pyx":505
 *             res[l, k1, kp1] -= 0.5
 *             res[l, k1, kp2] -= 0.5
 *             res[l, k2, kp1] -= 0.5             # <<<<<<<<<<<<<<
 *             res[l, k2, kp2] -= 0.5
 *     return res
 */
      __pyx_t_45 = __pyx_v_l;
      __pyx_t_46 = __pyx_v_k2;
      __pyx_t_47 = __pyx_v_kp1;
      if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_46, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_47, __pyx_pybuffernd_res.diminfo[2].strides) -= 0.5;

      /* "fastphase/fastphase.pyx":506
 *             res[l, k1, kp2] -= 0.5
 *             res[l, k2, kp1] -= 0.5
 *             res[l, k2, kp2] -= 0.5             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
      __pyx_t_48 = __pyx_v_l;
      __pyx_t_49 = __pyx_v_k2;
      __pyx_t_50 = __pyx_v_kp2;
      if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_res.diminfo[0].shape;
      if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_res.diminfo[1].shape;
      if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_res.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_res.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_res.diminfo[0].strides, __pyx_t_49, __pyx_pybuffernd_res.diminfo[1].strides, __pyx_t_50, __pyx_pybuffernd_res.diminfo[2].strides) -= 0.5;
    }
    __pyx_L5:;
  }

  /* "fastphase/fastphase.pyx":507
 *             res[l, k2, kp1] -= 0.5
 *             res[l, k2, kp2] -= 0.5
 *     return res             # <<<<<<<<<<<<<<
 * 
 * def calc_cost_matrix_geno( args):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_res));
  __pyx_r = ((PyArrayObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":479
 *     return cost_mat
 * 
 * cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_geno_tot( np.ndarray[np.int32_t, ndim=2] g, int nK ):             # <<<<<<<<<<<<<<
 *     cdef int l,nL
 *     cdef int k1, k2, kp1, kp2
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_g.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_res.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_geno_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_g.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_res.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_11calc_cost_matrix_geno_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_11calc_cost_matrix_geno_tot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_g = 0;
  int __pyx_v_nK;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_cost_matrix_geno_tot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_g,&__pyx_n_s_nK,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nK)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_cost_matrix_geno_tot", 1, 2, 2, 1); __PYX_ERR(0, 479, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_cost_matrix_geno_tot") < 0)) __PYX_ERR(0, 479, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_g = ((PyArrayObject *)values[0]);
    __pyx_v_nK = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nK == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_cost_matrix_geno_tot", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_geno_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_g), __pyx_ptype_5numpy_ndarray, 1, "g", 0))) __PYX_ERR(0, 479, __pyx_L1_error)
  __pyx_r = __pyx_pf_6hapflk_9fastphase_10calc_cost_matrix_geno_tot(__pyx_self, __pyx_v_g, __pyx_v_nK);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_10calc_cost_matrix_geno_tot(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_g, int __pyx_v_nK) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_g;
  __Pyx_Buffer __pyx_pybuffer_g;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("calc_cost_matrix_geno_tot", 0);
  __pyx_pybuffer_g.pybuffer.buf = NULL;
  __pyx_pybuffer_g.refcount = 0;
  __pyx_pybuffernd_g.data = NULL;
  __pyx_pybuffernd_g.rcbuffer = &__pyx_pybuffer_g;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_g.rcbuffer->pybuffer, (PyObject*)__pyx_v_g, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 479, __pyx_L1_error)
  }
  __pyx_pybuffernd_g.diminfo[0].strides = __pyx_pybuffernd_g.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_g.diminfo[0].shape = __pyx_pybuffernd_g.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_g.diminfo[1].strides = __pyx_pybuffernd_g.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_g.diminfo[1].shape = __pyx_pybuffernd_g.rcbuffer->pybuffer.shape[1];
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6hapflk_9fastphase_calc_cost_matrix_geno_tot(__pyx_v_g, __pyx_v_nK, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_g.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_geno_tot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_g.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":509
 *     return res
 * 
 * def calc_cost_matrix_geno( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k1, k2 = vit[0]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_13calc_cost_matrix_geno(PyObject *__pyx_self, PyObject *__pyx_v_args); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_13calc_cost_matrix_geno = {"calc_cost_matrix_geno", (PyCFunction)__pyx_pw_6hapflk_9fastphase_13calc_cost_matrix_geno, METH_O, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_13calc_cost_matrix_geno(PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_cost_matrix_geno (wrapper)", 0);
  __pyx_r = __pyx_pf_6hapflk_9fastphase_12calc_cost_matrix_geno(__pyx_self, ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_12calc_cost_matrix_geno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_vit = NULL;
  PyObject *__pyx_v_nK = NULL;
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_kp1 = NULL;
  PyObject *__pyx_v_kp2 = NULL;
  PyObject *__pyx_v_cost_mat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("calc_cost_matrix_geno", 0);

  /* "fastphase/fastphase.pyx":510
 * 
 * def calc_cost_matrix_geno( args):
 *     vit, nK = args             # <<<<<<<<<<<<<<
 *     k1, k2 = vit[0]
 *     kp1, kp2 = vit[1]
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_args))) || (PyList_CheckExact(__pyx_v_args))) {
    PyObject* sequence = __pyx_v_args;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 510, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 510, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 510, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_vit = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_nK = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":511
 * def calc_cost_matrix_geno( args):
 *     vit, nK = args
 *     k1, k2 = vit[0]             # <<<<<<<<<<<<<<
 *     kp1, kp2 = vit[1]
 *     cost_mat = np.zeros((nK,nK))
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_vit, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 511, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_4(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_5), 2) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 511, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_v_k1 = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_k2 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":512
 *     vit, nK = args
 *     k1, k2 = vit[0]
 *     kp1, kp2 = vit[1]             # <<<<<<<<<<<<<<
 *     cost_mat = np.zeros((nK,nK))
 *     if (k1 == kp1):
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_vit, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 512, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_4(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L7_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_1 = __pyx_t_4(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L7_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_5), 2) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_unpacking_done;
    __pyx_L7_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 512, __pyx_L1_error)
    __pyx_L8_unpacking_done:;
  }
  __pyx_v_kp1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_kp2 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":513
 *     k1, k2 = vit[0]
 *     kp1, kp2 = vit[1]
 *     cost_mat = np.zeros((nK,nK))             # <<<<<<<<<<<<<<
 *     if (k1 == kp1):
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nK);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nK);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_cost_mat = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":514
 *     kp1, kp2 = vit[1]
 *     cost_mat = np.zeros((nK,nK))
 *     if (k1 == kp1):             # <<<<<<<<<<<<<<
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *             cost_mat[k1, kp1] -= 1
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k1, __pyx_v_kp1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "fastphase/fastphase.pyx":515
 *     cost_mat = np.zeros((nK,nK))
 *     if (k1 == kp1):
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2             # <<<<<<<<<<<<<<
 *             cost_mat[k1, kp1] -= 1
 *             cost_mat[k2, kp2] -= 1
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_k2, __pyx_v_kp2, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "fastphase/fastphase.pyx":516
 *     if (k1 == kp1):
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *             cost_mat[k1, kp1] -= 1             # <<<<<<<<<<<<<<
 *             cost_mat[k2, kp2] -= 1
 *         else: ## k1 == kp1 & k2 != kp2
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_k1);
      __Pyx_GIVEREF(__pyx_v_k1);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
      __Pyx_INCREF(__pyx_v_kp1);
      __Pyx_GIVEREF(__pyx_v_kp1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":517
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *             cost_mat[k1, kp1] -= 1
 *             cost_mat[k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *         else: ## k1 == kp1 & k2 != kp2
 *             cost_mat[k1, kp1] -= 1 ## diagonal
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_k2);
      __Pyx_GIVEREF(__pyx_v_k2);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
      __Pyx_INCREF(__pyx_v_kp2);
      __Pyx_GIVEREF(__pyx_v_kp2);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":515
 *     cost_mat = np.zeros((nK,nK))
 *     if (k1 == kp1):
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2             # <<<<<<<<<<<<<<
 *             cost_mat[k1, kp1] -= 1
 *             cost_mat[k2, kp2] -= 1
 */
      goto __pyx_L10;
    }

    /* "fastphase/fastphase.pyx":519
 *             cost_mat[k2, kp2] -= 1
 *         else: ## k1 == kp1 & k2 != kp2
 *             cost_mat[k1, kp1] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *             cost_mat[k2, kp2] -= 1
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 */
    /*else*/ {
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_k1);
      __Pyx_GIVEREF(__pyx_v_k1);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
      __Pyx_INCREF(__pyx_v_kp1);
      __Pyx_GIVEREF(__pyx_v_kp1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 519, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fastphase/fastphase.pyx":520
 *         else: ## k1 == kp1 & k2 != kp2
 *             cost_mat[k1, kp1] -= 1 ## diagonal
 *             cost_mat[k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_k2);
      __Pyx_GIVEREF(__pyx_v_k2);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
      __Pyx_INCREF(__pyx_v_kp2);
      __Pyx_GIVEREF(__pyx_v_kp2);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L10:;

    /* "fastphase/fastphase.pyx":514
 *     kp1, kp2 = vit[1]
 *     cost_mat = np.zeros((nK,nK))
 *     if (k1 == kp1):             # <<<<<<<<<<<<<<
 *         if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
 *             cost_mat[k1, kp1] -= 1
 */
    goto __pyx_L9;
  }

  /* "fastphase/fastphase.pyx":521
 *             cost_mat[k1, kp1] -= 1 ## diagonal
 *             cost_mat[k2, kp2] -= 1
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 *         cost_mat[k2, kp2] -= 1
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k2, __pyx_v_kp2, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "fastphase/fastphase.pyx":522
 *             cost_mat[k2, kp2] -= 1
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *         cost_mat[k1, kp1] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *         cost_mat[k2, kp2] -= 1
 *     elif ( k1 == kp2) and (k2 != kp1):
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_kp1);
    __Pyx_GIVEREF(__pyx_v_kp1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":523
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 *         cost_mat[k2, kp2] -= 1             # <<<<<<<<<<<<<<
 *     elif ( k1 == kp2) and (k2 != kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
    __Pyx_INCREF(__pyx_v_kp2);
    __Pyx_GIVEREF(__pyx_v_kp2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":521
 *             cost_mat[k1, kp1] -= 1 ## diagonal
 *             cost_mat[k2, kp2] -= 1
 *     elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 *         cost_mat[k2, kp2] -= 1
 */
    goto __pyx_L9;
  }

  /* "fastphase/fastphase.pyx":524
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 *         cost_mat[k2, kp2] -= 1
 *     elif ( k1 == kp2) and (k2 != kp1):             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k1, __pyx_v_kp2, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k2, __pyx_v_kp1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_6) {

    /* "fastphase/fastphase.pyx":525
 *         cost_mat[k2, kp2] -= 1
 *     elif ( k1 == kp2) and (k2 != kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *         cost_mat[k2, kp1] -= 1
 *     elif ( k1 != kp2) and (k2 == kp1):
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_kp2);
    __Pyx_GIVEREF(__pyx_v_kp2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":526
 *     elif ( k1 == kp2) and (k2 != kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1             # <<<<<<<<<<<<<<
 *     elif ( k1 != kp2) and (k2 == kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
    __Pyx_INCREF(__pyx_v_kp1);
    __Pyx_GIVEREF(__pyx_v_kp1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":524
 *         cost_mat[k1, kp1] -= 1 ## diagonal
 *         cost_mat[k2, kp2] -= 1
 *     elif ( k1 == kp2) and (k2 != kp1):             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 */
    goto __pyx_L9;
  }

  /* "fastphase/fastphase.pyx":527
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 *     elif ( k1 != kp2) and (k2 == kp1):             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k1, __pyx_v_kp2, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k2, __pyx_v_kp1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_6) {

    /* "fastphase/fastphase.pyx":528
 *         cost_mat[k2, kp1] -= 1
 *     elif ( k1 != kp2) and (k2 == kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal             # <<<<<<<<<<<<<<
 *         cost_mat[k2, kp1] -= 1
 *     else:
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_kp2);
    __Pyx_GIVEREF(__pyx_v_kp2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":529
 *     elif ( k1 != kp2) and (k2 == kp1):
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1             # <<<<<<<<<<<<<<
 *     else:
 *         cost_mat[ k1, kp1] -= 0.5
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
    __Pyx_INCREF(__pyx_v_kp1);
    __Pyx_GIVEREF(__pyx_v_kp1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":527
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 *     elif ( k1 != kp2) and (k2 == kp1):             # <<<<<<<<<<<<<<
 *         cost_mat[k1, kp2] -= 1 ## diagonal
 *         cost_mat[k2, kp1] -= 1
 */
    goto __pyx_L9;
  }

  /* "fastphase/fastphase.pyx":531
 *         cost_mat[k2, kp1] -= 1
 *     else:
 *         cost_mat[ k1, kp1] -= 0.5             # <<<<<<<<<<<<<<
 *         cost_mat[ k1, kp2] -= 0.5
 *         cost_mat[ k2, kp1] -= 0.5
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_kp1);
    __Pyx_GIVEREF(__pyx_v_kp1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_t_3, __pyx_float_0_5, 0.5, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":532
 *     else:
 *         cost_mat[ k1, kp1] -= 0.5
 *         cost_mat[ k1, kp2] -= 0.5             # <<<<<<<<<<<<<<
 *         cost_mat[ k2, kp1] -= 0.5
 *         cost_mat[ k2, kp2] -= 0.5
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_kp2);
    __Pyx_GIVEREF(__pyx_v_kp2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyFloat_SubtractObjC(__pyx_t_1, __pyx_float_0_5, 0.5, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":533
 *         cost_mat[ k1, kp1] -= 0.5
 *         cost_mat[ k1, kp2] -= 0.5
 *         cost_mat[ k2, kp1] -= 0.5             # <<<<<<<<<<<<<<
 *         cost_mat[ k2, kp2] -= 0.5
 *     return cost_mat
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
    __Pyx_INCREF(__pyx_v_kp1);
    __Pyx_GIVEREF(__pyx_v_kp1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_t_3, __pyx_float_0_5, 0.5, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fastphase/fastphase.pyx":534
 *         cost_mat[ k1, kp2] -= 0.5
 *         cost_mat[ k2, kp1] -= 0.5
 *         cost_mat[ k2, kp2] -= 0.5             # <<<<<<<<<<<<<<
 *     return cost_mat
 * 
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k2);
    __Pyx_INCREF(__pyx_v_kp2);
    __Pyx_GIVEREF(__pyx_v_kp2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_kp2);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_cost_mat, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyFloat_SubtractObjC(__pyx_t_1, __pyx_float_0_5, 0.5, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_cost_mat, __pyx_t_2, __pyx_t_3) < 0)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L9:;

  /* "fastphase/fastphase.pyx":535
 *         cost_mat[ k2, kp1] -= 0.5
 *         cost_mat[ k2, kp2] -= 0.5
 *     return cost_mat             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cost_mat);
  __pyx_r = __pyx_v_cost_mat;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":509
 *     return res
 * 
 * def calc_cost_matrix_geno( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k1, k2 = vit[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hapflk.fastphase.calc_cost_matrix_geno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vit);
  __Pyx_XDECREF(__pyx_v_nK);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_kp1);
  __Pyx_XDECREF(__pyx_v_kp2);
  __Pyx_XDECREF(__pyx_v_cost_mat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":538
 * 
 * 
 * def switch_pars(par, permut):             # <<<<<<<<<<<<<<
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_15switch_pars(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6hapflk_9fastphase_15switch_pars = {"switch_pars", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_15switch_pars, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6hapflk_9fastphase_15switch_pars(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_par = 0;
  PyObject *__pyx_v_permut = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("switch_pars (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_par,&__pyx_n_s_permut,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_par)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_permut)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("switch_pars", 1, 2, 2, 1); __PYX_ERR(0, 538, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "switch_pars") < 0)) __PYX_ERR(0, 538, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_par = values[0];
    __pyx_v_permut = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("switch_pars", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 538, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.switch_pars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_14switch_pars(__pyx_self, __pyx_v_par, __pyx_v_permut);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_14switch_pars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_par, PyObject *__pyx_v_permut) {
  PyObject *__pyx_v_newpar = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_curclus = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("switch_pars", 0);

  /* "fastphase/fastphase.pyx":539
 * 
 * def switch_pars(par, permut):
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)             # <<<<<<<<<<<<<<
 *     for i in range(par.nClus):
 *         newpar.theta[0,i]=par.theta[0,i]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_modParams); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nClus); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_alpha_up); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_alpha_up, __pyx_t_2) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_newpar = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":540
 * def switch_pars(par, permut):
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):             # <<<<<<<<<<<<<<
 *         newpar.theta[0,i]=par.theta[0,i]
 *         curclus = i
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nClus); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 540, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastphase/fastphase.pyx":541
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):
 *         newpar.theta[0,i]=par.theta[0,i]             # <<<<<<<<<<<<<<
 *         curclus = i
 *         for j in range(1, par.nLoc):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_i);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_newpar, __pyx_n_s_theta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_i);
    if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_t_3, __pyx_t_1) < 0)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastphase/fastphase.pyx":542
 *     for i in range(par.nClus):
 *         newpar.theta[0,i]=par.theta[0,i]
 *         curclus = i             # <<<<<<<<<<<<<<
 *         for j in range(1, par.nLoc):
 *             curclus =  permut[ j-1, curclus]
 */
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_v_curclus, __pyx_v_i);

    /* "fastphase/fastphase.pyx":543
 *         newpar.theta[0,i]=par.theta[0,i]
 *         curclus = i
 *         for j in range(1, par.nLoc):             # <<<<<<<<<<<<<<
 *             curclus =  permut[ j-1, curclus]
 *             newpar.theta[ j,i] = par.theta[ j, curclus]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 543, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 543, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 543, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 543, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fastphase/fastphase.pyx":544
 *         curclus = i
 *         for j in range(1, par.nLoc):
 *             curclus =  permut[ j-1, curclus]             # <<<<<<<<<<<<<<
 *             newpar.theta[ j,i] = par.theta[ j, curclus]
 *     return newpar
 */
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_j, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_curclus);
      __Pyx_GIVEREF(__pyx_v_curclus);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_curclus);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_permut, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_curclus, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fastphase/fastphase.pyx":545
 *         for j in range(1, par.nLoc):
 *             curclus =  permut[ j-1, curclus]
 *             newpar.theta[ j,i] = par.theta[ j, curclus]             # <<<<<<<<<<<<<<
 *     return newpar
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_theta); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_j);
      __Pyx_GIVEREF(__pyx_v_j);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_j);
      __Pyx_INCREF(__pyx_v_curclus);
      __Pyx_GIVEREF(__pyx_v_curclus);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_curclus);
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_newpar, __pyx_n_s_theta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_j);
      __Pyx_GIVEREF(__pyx_v_j);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_j);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_i);
      if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_t_1, __pyx_t_9) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastphase/fastphase.pyx":543
 *         newpar.theta[0,i]=par.theta[0,i]
 *         curclus = i
 *         for j in range(1, par.nLoc):             # <<<<<<<<<<<<<<
 *             curclus =  permut[ j-1, curclus]
 *             newpar.theta[ j,i] = par.theta[ j, curclus]
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastphase/fastphase.pyx":540
 * def switch_pars(par, permut):
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):             # <<<<<<<<<<<<<<
 *         newpar.theta[0,i]=par.theta[0,i]
 *         curclus = i
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":546
 *             curclus =  permut[ j-1, curclus]
 *             newpar.theta[ j,i] = par.theta[ j, curclus]
 *     return newpar             # <<<<<<<<<<<<<<
 * 
 * ############################### Calculations #################################
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_newpar);
  __pyx_r = __pyx_v_newpar;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":538
 * 
 * 
 * def switch_pars(par, permut):             # <<<<<<<<<<<<<<
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hapflk.fastphase.switch_pars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_newpar);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_curclus);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":550
 * ############################### Calculations #################################
 * 
 * cdef double hap_p_all(int i, double pz, double theta):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     if i<0:
 */

static double __pyx_f_6hapflk_9fastphase_hap_p_all(int __pyx_v_i, double __pyx_v_pz, double __pyx_v_theta) {
  double __pyx_v_rez;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("hap_p_all", 0);

  /* "fastphase/fastphase.pyx":552
 * cdef double hap_p_all(int i, double pz, double theta):
 *     cdef double rez
 *     if i<0:             # <<<<<<<<<<<<<<
 *         rez=pz*theta
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "fastphase/fastphase.pyx":553
 *     cdef double rez
 *     if i<0:
 *         rez=pz*theta             # <<<<<<<<<<<<<<
 *     else:
 *         rez=i*pz
 */
    __pyx_v_rez = (__pyx_v_pz * __pyx_v_theta);

    /* "fastphase/fastphase.pyx":552
 * cdef double hap_p_all(int i, double pz, double theta):
 *     cdef double rez
 *     if i<0:             # <<<<<<<<<<<<<<
 *         rez=pz*theta
 *     else:
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":555
 *         rez=pz*theta
 *     else:
 *         rez=i*pz             # <<<<<<<<<<<<<<
 *     return rez
 * 
 */
  /*else*/ {
    __pyx_v_rez = (__pyx_v_i * __pyx_v_pz);
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":556
 *     else:
 *         rez=i*pz
 *     return rez             # <<<<<<<<<<<<<<
 * 
 * cdef double gen_p_geno(int i, double pz, double theta1, double theta2):
 */
  __pyx_r = __pyx_v_rez;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":550
 * ############################### Calculations #################################
 * 
 * cdef double hap_p_all(int i, double pz, double theta):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     if i<0:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":558
 *     return rez
 * 
 * cdef double gen_p_geno(int i, double pz, double theta1, double theta2):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
 */

static double __pyx_f_6hapflk_9fastphase_gen_p_geno(int __pyx_v_i, double __pyx_v_pz, double __pyx_v_theta1, double __pyx_v_theta2) {
  double __pyx_v_rez;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("gen_p_geno", 0);

  /* "fastphase/fastphase.pyx":561
 *     cdef double rez
 *     ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
 *     if i<0:             # <<<<<<<<<<<<<<
 *         rez=pz*(theta1+theta2)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "fastphase/fastphase.pyx":562
 *     ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
 *     if i<0:
 *         rez=pz*(theta1+theta2)             # <<<<<<<<<<<<<<
 *     else:
 *         rez=i*pz
 */
    __pyx_v_rez = (__pyx_v_pz * (__pyx_v_theta1 + __pyx_v_theta2));

    /* "fastphase/fastphase.pyx":561
 *     cdef double rez
 *     ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
 *     if i<0:             # <<<<<<<<<<<<<<
 *         rez=pz*(theta1+theta2)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":564
 *         rez=pz*(theta1+theta2)
 *     else:
 *         rez=i*pz             # <<<<<<<<<<<<<<
 *         #rez=pz*(theta1+theta2)
 *     return rez
 */
  /*else*/ {
    __pyx_v_rez = (__pyx_v_i * __pyx_v_pz);
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":566
 *         rez=i*pz
 *         #rez=pz*(theta1+theta2)
 *     return rez             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_rez;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":558
 *     return rez
 * 
 * cdef double gen_p_geno(int i, double pz, double theta1, double theta2):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":569
 * 
 * 
 * cdef double myPow10(int x):             # <<<<<<<<<<<<<<
 *     cdef double rez=1.0
 *     cdef int i
 */

static double __pyx_f_6hapflk_9fastphase_myPow10(int __pyx_v_x) {
  double __pyx_v_rez;
  CYTHON_UNUSED int __pyx_v_i;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("myPow10", 0);

  /* "fastphase/fastphase.pyx":570
 * 
 * cdef double myPow10(int x):
 *     cdef double rez=1.0             # <<<<<<<<<<<<<<
 *     cdef int i
 *     if x==0:
 */
  __pyx_v_rez = 1.0;

  /* "fastphase/fastphase.pyx":572
 *     cdef double rez=1.0
 *     cdef int i
 *     if x==0:             # <<<<<<<<<<<<<<
 *         return rez
 *     elif x<0:
 */
  __pyx_t_1 = ((__pyx_v_x == 0) != 0);
  if (__pyx_t_1) {

    /* "fastphase/fastphase.pyx":573
 *     cdef int i
 *     if x==0:
 *         return rez             # <<<<<<<<<<<<<<
 *     elif x<0:
 *         for i in range(-x):
 */
    __pyx_r = __pyx_v_rez;
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":572
 *     cdef double rez=1.0
 *     cdef int i
 *     if x==0:             # <<<<<<<<<<<<<<
 *         return rez
 *     elif x<0:
 */
  }

  /* "fastphase/fastphase.pyx":574
 *     if x==0:
 *         return rez
 *     elif x<0:             # <<<<<<<<<<<<<<
 *         for i in range(-x):
 *             rez*=0.1
 */
  __pyx_t_1 = ((__pyx_v_x < 0) != 0);
  if (__pyx_t_1) {

    /* "fastphase/fastphase.pyx":575
 *         return rez
 *     elif x<0:
 *         for i in range(-x):             # <<<<<<<<<<<<<<
 *             rez*=0.1
 *     else:
 */
    __pyx_t_2 = (-__pyx_v_x);
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "fastphase/fastphase.pyx":576
 *     elif x<0:
 *         for i in range(-x):
 *             rez*=0.1             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(x):
 */
      __pyx_v_rez = (__pyx_v_rez * 0.1);
    }

    /* "fastphase/fastphase.pyx":574
 *     if x==0:
 *         return rez
 *     elif x<0:             # <<<<<<<<<<<<<<
 *         for i in range(-x):
 *             rez*=0.1
 */
    goto __pyx_L3;
  }

  /* "fastphase/fastphase.pyx":578
 *             rez*=0.1
 *     else:
 *         for i in range(x):             # <<<<<<<<<<<<<<
 *             rez*=10
 *     return rez
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_x;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "fastphase/fastphase.pyx":579
 *     else:
 *         for i in range(x):
 *             rez*=10             # <<<<<<<<<<<<<<
 *     return rez
 * 
 */
      __pyx_v_rez = (__pyx_v_rez * 10.0);
    }
  }
  __pyx_L3:;

  /* "fastphase/fastphase.pyx":580
 *         for i in range(x):
 *             rez*=10
 *     return rez             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_rez;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":569
 * 
 * 
 * cdef double myPow10(int x):             # <<<<<<<<<<<<<<
 *     cdef double rez=1.0
 *     cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":583
 * 
 * 
 * cdef int argmax( np.ndarray[ np.float64_t, ndim=1] a, int size):             # <<<<<<<<<<<<<<
 *     cdef int i, best_i
 *     cdef double v, best_v
 */

static int __pyx_f_6hapflk_9fastphase_argmax(PyArrayObject *__pyx_v_a, int __pyx_v_size) {
  int __pyx_v_i;
  int __pyx_v_best_i;
  double __pyx_v_best_v;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_a;
  __Pyx_Buffer __pyx_pybuffer_a;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  __Pyx_RefNannySetupContext("argmax", 0);
  __pyx_pybuffer_a.pybuffer.buf = NULL;
  __pyx_pybuffer_a.refcount = 0;
  __pyx_pybuffernd_a.data = NULL;
  __pyx_pybuffernd_a.rcbuffer = &__pyx_pybuffer_a;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_a, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 583, __pyx_L1_error)
  }
  __pyx_pybuffernd_a.diminfo[0].strides = __pyx_pybuffernd_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_a.diminfo[0].shape = __pyx_pybuffernd_a.rcbuffer->pybuffer.shape[0];

  /* "fastphase/fastphase.pyx":587
 *     cdef double v, best_v
 * 
 *     best_v = a[ 0 ]             # <<<<<<<<<<<<<<
 *     best_i = 0
 *     for i in range(1, size):
 */
  __pyx_t_1 = 0;
  if (__pyx_t_1 < 0) __pyx_t_1 += __pyx_pybuffernd_a.diminfo[0].shape;
  __pyx_v_best_v = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_a.rcbuffer->pybuffer.buf, __pyx_t_1, __pyx_pybuffernd_a.diminfo[0].strides));

  /* "fastphase/fastphase.pyx":588
 * 
 *     best_v = a[ 0 ]
 *     best_i = 0             # <<<<<<<<<<<<<<
 *     for i in range(1, size):
 *         if a[ i ] > best_v:
 */
  __pyx_v_best_i = 0;

  /* "fastphase/fastphase.pyx":589
 *     best_v = a[ 0 ]
 *     best_i = 0
 *     for i in range(1, size):             # <<<<<<<<<<<<<<
 *         if a[ i ] > best_v:
 *             best_i = i
 */
  __pyx_t_2 = __pyx_v_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 1; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fastphase/fastphase.pyx":590
 *     best_i = 0
 *     for i in range(1, size):
 *         if a[ i ] > best_v:             # <<<<<<<<<<<<<<
 *             best_i = i
 *             best_v = a[ i ]
 */
    __pyx_t_5 = __pyx_v_i;
    if (__pyx_t_5 < 0) __pyx_t_5 += __pyx_pybuffernd_a.diminfo[0].shape;
    __pyx_t_6 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_a.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_a.diminfo[0].strides)) > __pyx_v_best_v) != 0);
    if (__pyx_t_6) {

      /* "fastphase/fastphase.pyx":591
 *     for i in range(1, size):
 *         if a[ i ] > best_v:
 *             best_i = i             # <<<<<<<<<<<<<<
 *             best_v = a[ i ]
 *     return best_i
 */
      __pyx_v_best_i = __pyx_v_i;

      /* "fastphase/fastphase.pyx":592
 *         if a[ i ] > best_v:
 *             best_i = i
 *             best_v = a[ i ]             # <<<<<<<<<<<<<<
 *     return best_i
 * 
 */
      __pyx_t_7 = __pyx_v_i;
      if (__pyx_t_7 < 0) __pyx_t_7 += __pyx_pybuffernd_a.diminfo[0].shape;
      __pyx_v_best_v = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_a.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_a.diminfo[0].strides));

      /* "fastphase/fastphase.pyx":590
 *     best_i = 0
 *     for i in range(1, size):
 *         if a[ i ] > best_v:             # <<<<<<<<<<<<<<
 *             best_i = i
 *             best_v = a[ i ]
 */
    }
  }

  /* "fastphase/fastphase.pyx":593
 *             best_i = i
 *             best_v = a[ i ]
 *     return best_i             # <<<<<<<<<<<<<<
 * 
 * cdef int pair2idx( int k1, int k2, int nK):
 */
  __pyx_r = __pyx_v_best_i;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":583
 * 
 * 
 * cdef int argmax( np.ndarray[ np.float64_t, ndim=1] a, int size):             # <<<<<<<<<<<<<<
 *     cdef int i, best_i
 *     cdef double v, best_v
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_a.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("hapflk.fastphase.argmax", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_a.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":595
 *     return best_i
 * 
 * cdef int pair2idx( int k1, int k2, int nK):             # <<<<<<<<<<<<<<
 *     cdef int k, l
 *     k = min( k1, k2)
 */

static int __pyx_f_6hapflk_9fastphase_pair2idx(int __pyx_v_k1, int __pyx_v_k2, int __pyx_v_nK) {
  int __pyx_v_k;
  int __pyx_v_l;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("pair2idx", 0);

  /* "fastphase/fastphase.pyx":597
 * cdef int pair2idx( int k1, int k2, int nK):
 *     cdef int k, l
 *     k = min( k1, k2)             # <<<<<<<<<<<<<<
 *     l = max( k1, k2)
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)
 */
  __pyx_t_1 = __pyx_v_k2;
  __pyx_t_2 = __pyx_v_k1;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_k = __pyx_t_3;

  /* "fastphase/fastphase.pyx":598
 *     cdef int k, l
 *     k = min( k1, k2)
 *     l = max( k1, k2)             # <<<<<<<<<<<<<<
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)
 * 
 */
  __pyx_t_3 = __pyx_v_k2;
  __pyx_t_1 = __pyx_v_k1;
  if (((__pyx_t_3 > __pyx_t_1) != 0)) {
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  __pyx_v_l = __pyx_t_2;

  /* "fastphase/fastphase.pyx":599
 *     k = min( k1, k2)
 *     l = max( k1, k2)
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)             # <<<<<<<<<<<<<<
 * 
 * cdef ( int, int) idx2pair( int idx, int nK):
 */
  __pyx_r = (((__pyx_v_nK * __pyx_v_k) - __Pyx_div_long((__pyx_v_k * (__pyx_v_k - 1)), 2)) + (__pyx_v_l - __pyx_v_k));
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":595
 *     return best_i
 * 
 * cdef int pair2idx( int k1, int k2, int nK):             # <<<<<<<<<<<<<<
 *     cdef int k, l
 *     k = min( k1, k2)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":601
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)
 * 
 * cdef ( int, int) idx2pair( int idx, int nK):             # <<<<<<<<<<<<<<
 *     cdef int k, l, nelem, curk
 *     nelem = 0
 */

static __pyx_ctuple_int__and_int __pyx_f_6hapflk_9fastphase_idx2pair(int __pyx_v_idx, int __pyx_v_nK) {
  int __pyx_v_k;
  int __pyx_v_l;
  int __pyx_v_nelem;
  int __pyx_v_curk;
  __pyx_ctuple_int__and_int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __pyx_ctuple_int__and_int __pyx_t_5;
  __Pyx_RefNannySetupContext("idx2pair", 0);

  /* "fastphase/fastphase.pyx":603
 * cdef ( int, int) idx2pair( int idx, int nK):
 *     cdef int k, l, nelem, curk
 *     nelem = 0             # <<<<<<<<<<<<<<
 *     for curk in range(nK):
 *         if ( idx - nelem) < ( nK - curk):
 */
  __pyx_v_nelem = 0;

  /* "fastphase/fastphase.pyx":604
 *     cdef int k, l, nelem, curk
 *     nelem = 0
 *     for curk in range(nK):             # <<<<<<<<<<<<<<
 *         if ( idx - nelem) < ( nK - curk):
 *             l = curk + (idx - nelem)
 */
  __pyx_t_1 = __pyx_v_nK;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_curk = __pyx_t_3;

    /* "fastphase/fastphase.pyx":605
 *     nelem = 0
 *     for curk in range(nK):
 *         if ( idx - nelem) < ( nK - curk):             # <<<<<<<<<<<<<<
 *             l = curk + (idx - nelem)
 *             k = curk
 */
    __pyx_t_4 = (((__pyx_v_idx - __pyx_v_nelem) < (__pyx_v_nK - __pyx_v_curk)) != 0);
    if (__pyx_t_4) {

      /* "fastphase/fastphase.pyx":606
 *     for curk in range(nK):
 *         if ( idx - nelem) < ( nK - curk):
 *             l = curk + (idx - nelem)             # <<<<<<<<<<<<<<
 *             k = curk
 *             break
 */
      __pyx_v_l = (__pyx_v_curk + (__pyx_v_idx - __pyx_v_nelem));

      /* "fastphase/fastphase.pyx":607
 *         if ( idx - nelem) < ( nK - curk):
 *             l = curk + (idx - nelem)
 *             k = curk             # <<<<<<<<<<<<<<
 *             break
 *         else:
 */
      __pyx_v_k = __pyx_v_curk;

      /* "fastphase/fastphase.pyx":608
 *             l = curk + (idx - nelem)
 *             k = curk
 *             break             # <<<<<<<<<<<<<<
 *         else:
 *             nelem += nK - curk
 */
      goto __pyx_L4_break;

      /* "fastphase/fastphase.pyx":605
 *     nelem = 0
 *     for curk in range(nK):
 *         if ( idx - nelem) < ( nK - curk):             # <<<<<<<<<<<<<<
 *             l = curk + (idx - nelem)
 *             k = curk
 */
    }

    /* "fastphase/fastphase.pyx":610
 *             break
 *         else:
 *             nelem += nK - curk             # <<<<<<<<<<<<<<
 *     return (k, l)
 * 
 */
    /*else*/ {
      __pyx_v_nelem = (__pyx_v_nelem + (__pyx_v_nK - __pyx_v_curk));
    }
  }
  __pyx_L4_break:;

  /* "fastphase/fastphase.pyx":611
 *         else:
 *             nelem += nK - curk
 *     return (k, l)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5.f0 = __pyx_v_k;
  __pyx_t_5.f1 = __pyx_v_l;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":601
 *     return  nK * k - ( k * ( k -1))//2  + ( l - k)
 * 
 * cdef ( int, int) idx2pair( int idx, int nK):             # <<<<<<<<<<<<<<
 *     cdef int k, l, nelem, curk
 *     nelem = 0
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":616
 * ##### Genotype Calculations
 * 
 * cdef double genprG(double t1, double t2, int g):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     if g==0:
 */

static double __pyx_f_6hapflk_9fastphase_genprG(double __pyx_v_t1, double __pyx_v_t2, int __pyx_v_g) {
  double __pyx_v_rez;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genprG", 0);

  /* "fastphase/fastphase.pyx":618
 * cdef double genprG(double t1, double t2, int g):
 *     cdef double rez
 *     if g==0:             # <<<<<<<<<<<<<<
 *         rez=(1-t1)*(1-t2)
 *     elif g==1:
 */
  switch (__pyx_v_g) {
    case 0:

    /* "fastphase/fastphase.pyx":619
 *     cdef double rez
 *     if g==0:
 *         rez=(1-t1)*(1-t2)             # <<<<<<<<<<<<<<
 *     elif g==1:
 *         rez=t1+t2-2*t1*t2
 */
    __pyx_v_rez = ((1.0 - __pyx_v_t1) * (1.0 - __pyx_v_t2));

    /* "fastphase/fastphase.pyx":618
 * cdef double genprG(double t1, double t2, int g):
 *     cdef double rez
 *     if g==0:             # <<<<<<<<<<<<<<
 *         rez=(1-t1)*(1-t2)
 *     elif g==1:
 */
    break;
    case 1:

    /* "fastphase/fastphase.pyx":621
 *         rez=(1-t1)*(1-t2)
 *     elif g==1:
 *         rez=t1+t2-2*t1*t2             # <<<<<<<<<<<<<<
 *     elif g==2:
 *         rez=t1*t2
 */
    __pyx_v_rez = ((__pyx_v_t1 + __pyx_v_t2) - ((2.0 * __pyx_v_t1) * __pyx_v_t2));

    /* "fastphase/fastphase.pyx":620
 *     if g==0:
 *         rez=(1-t1)*(1-t2)
 *     elif g==1:             # <<<<<<<<<<<<<<
 *         rez=t1+t2-2*t1*t2
 *     elif g==2:
 */
    break;
    case 2:

    /* "fastphase/fastphase.pyx":623
 *         rez=t1+t2-2*t1*t2
 *     elif g==2:
 *         rez=t1*t2             # <<<<<<<<<<<<<<
 *     else:
 *         rez=1
 */
    __pyx_v_rez = (__pyx_v_t1 * __pyx_v_t2);

    /* "fastphase/fastphase.pyx":622
 *     elif g==1:
 *         rez=t1+t2-2*t1*t2
 *     elif g==2:             # <<<<<<<<<<<<<<
 *         rez=t1*t2
 *     else:
 */
    break;
    default:

    /* "fastphase/fastphase.pyx":625
 *         rez=t1*t2
 *     else:
 *         rez=1             # <<<<<<<<<<<<<<
 *     return rez
 * 
 */
    __pyx_v_rez = 1.0;
    break;
  }

  /* "fastphase/fastphase.pyx":626
 *     else:
 *         rez=1
 *     return rez             # <<<<<<<<<<<<<<
 * 
 * cdef double probJ(int m,int s, double rho):
 */
  __pyx_r = __pyx_v_rez;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":616
 * ##### Genotype Calculations
 * 
 * cdef double genprG(double t1, double t2, int g):             # <<<<<<<<<<<<<<
 *     cdef double rez
 *     if g==0:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":628
 *     return rez
 * 
 * cdef double probJ(int m,int s, double rho):             # <<<<<<<<<<<<<<
 *     if s==0:
 *         return (1-rho)*(1-rho)
 */

static double __pyx_f_6hapflk_9fastphase_probJ(CYTHON_UNUSED int __pyx_v_m, int __pyx_v_s, double __pyx_v_rho) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probJ", 0);

  /* "fastphase/fastphase.pyx":629
 * 
 * cdef double probJ(int m,int s, double rho):
 *     if s==0:             # <<<<<<<<<<<<<<
 *         return (1-rho)*(1-rho)
 *     elif s==1:
 */
  switch (__pyx_v_s) {
    case 0:

    /* "fastphase/fastphase.pyx":630
 * cdef double probJ(int m,int s, double rho):
 *     if s==0:
 *         return (1-rho)*(1-rho)             # <<<<<<<<<<<<<<
 *     elif s==1:
 *         return 2*(1-rho)*rho
 */
    __pyx_r = ((1.0 - __pyx_v_rho) * (1.0 - __pyx_v_rho));
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":629
 * 
 * cdef double probJ(int m,int s, double rho):
 *     if s==0:             # <<<<<<<<<<<<<<
 *         return (1-rho)*(1-rho)
 *     elif s==1:
 */
    break;
    case 1:

    /* "fastphase/fastphase.pyx":632
 *         return (1-rho)*(1-rho)
 *     elif s==1:
 *         return 2*(1-rho)*rho             # <<<<<<<<<<<<<<
 *     elif s==2:
 *         return rho*rho
 */
    __pyx_r = ((2.0 * (1.0 - __pyx_v_rho)) * __pyx_v_rho);
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":631
 *     if s==0:
 *         return (1-rho)*(1-rho)
 *     elif s==1:             # <<<<<<<<<<<<<<
 *         return 2*(1-rho)*rho
 *     elif s==2:
 */
    break;
    case 2:

    /* "fastphase/fastphase.pyx":634
 *         return 2*(1-rho)*rho
 *     elif s==2:
 *         return rho*rho             # <<<<<<<<<<<<<<
 * 
 * cpdef genViterbi( aa, tt, rr, gg):
 */
    __pyx_r = (__pyx_v_rho * __pyx_v_rho);
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":633
 *     elif s==1:
 *         return 2*(1-rho)*rho
 *     elif s==2:             # <<<<<<<<<<<<<<
 *         return rho*rho
 * 
 */
    break;
    default: break;
  }

  /* "fastphase/fastphase.pyx":628
 *     return rez
 * 
 * cdef double probJ(int m,int s, double rho):             # <<<<<<<<<<<<<<
 *     if s==0:
 *         return (1-rho)*(1-rho)
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":636
 *         return rho*rho
 * 
 * cpdef genViterbi( aa, tt, rr, gg):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_17genViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_genViterbi(PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_alpha = 0;
  PyArrayObject *__pyx_v_theta = 0;
  PyArrayObject *__pyx_v_rho = 0;
  PyArrayObject *__pyx_v_gen = 0;
  int __pyx_v_nLoc;
  int __pyx_v_nK;
  int __pyx_v_ikp;
  int __pyx_v_prev_kp;
  int __pyx_v_npairs;
  int __pyx_v_k1;
  int __pyx_v_k2;
  int __pyx_v_kp1;
  int __pyx_v_kp2;
  int __pyx_v_m;
  PyArrayObject *__pyx_v_delta = 0;
  PyArrayObject *__pyx_v_psi = 0;
  PyArrayObject *__pyx_v_soluce = 0;
  PyArrayObject *__pyx_v_tempVal = 0;
  int __pyx_9genexpr15__pyx_v_ikp;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_alpha;
  __Pyx_Buffer __pyx_pybuffer_alpha;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_delta;
  __Pyx_Buffer __pyx_pybuffer_delta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gen;
  __Pyx_Buffer __pyx_pybuffer_gen;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_psi;
  __Pyx_Buffer __pyx_pybuffer_psi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rho;
  __Pyx_Buffer __pyx_pybuffer_rho;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_soluce;
  __Pyx_Buffer __pyx_pybuffer_soluce;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tempVal;
  __Pyx_Buffer __pyx_pybuffer_tempVal;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_theta;
  __Pyx_Buffer __pyx_pybuffer_theta;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  __pyx_ctuple_int__and_int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  int __pyx_t_41;
  int __pyx_t_42;
  int __pyx_t_43;
  int __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  Py_ssize_t __pyx_t_51;
  Py_ssize_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  Py_ssize_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  int __pyx_t_64;
  Py_ssize_t __pyx_t_65;
  Py_ssize_t __pyx_t_66;
  Py_ssize_t __pyx_t_67;
  Py_ssize_t __pyx_t_68;
  Py_ssize_t __pyx_t_69;
  Py_ssize_t __pyx_t_70;
  Py_ssize_t __pyx_t_71;
  Py_ssize_t __pyx_t_72;
  Py_ssize_t __pyx_t_73;
  Py_ssize_t __pyx_t_74;
  Py_ssize_t __pyx_t_75;
  Py_ssize_t __pyx_t_76;
  Py_ssize_t __pyx_t_77;
  Py_ssize_t __pyx_t_78;
  Py_ssize_t __pyx_t_79;
  Py_ssize_t __pyx_t_80;
  Py_ssize_t __pyx_t_81;
  Py_ssize_t __pyx_t_82;
  Py_ssize_t __pyx_t_83;
  Py_ssize_t __pyx_t_84;
  Py_ssize_t __pyx_t_85;
  Py_ssize_t __pyx_t_86;
  Py_ssize_t __pyx_t_87;
  Py_ssize_t __pyx_t_88;
  Py_ssize_t __pyx_t_89;
  Py_ssize_t __pyx_t_90;
  Py_ssize_t __pyx_t_91;
  Py_ssize_t __pyx_t_92;
  Py_ssize_t __pyx_t_93;
  Py_ssize_t __pyx_t_94;
  Py_ssize_t __pyx_t_95;
  Py_ssize_t __pyx_t_96;
  Py_ssize_t __pyx_t_97;
  Py_ssize_t __pyx_t_98;
  PyObject *(*__pyx_t_99)(PyObject *);
  __Pyx_RefNannySetupContext("genViterbi", 0);
  __pyx_pybuffer_alpha.pybuffer.buf = NULL;
  __pyx_pybuffer_alpha.refcount = 0;
  __pyx_pybuffernd_alpha.data = NULL;
  __pyx_pybuffernd_alpha.rcbuffer = &__pyx_pybuffer_alpha;
  __pyx_pybuffer_theta.pybuffer.buf = NULL;
  __pyx_pybuffer_theta.refcount = 0;
  __pyx_pybuffernd_theta.data = NULL;
  __pyx_pybuffernd_theta.rcbuffer = &__pyx_pybuffer_theta;
  __pyx_pybuffer_rho.pybuffer.buf = NULL;
  __pyx_pybuffer_rho.refcount = 0;
  __pyx_pybuffernd_rho.data = NULL;
  __pyx_pybuffernd_rho.rcbuffer = &__pyx_pybuffer_rho;
  __pyx_pybuffer_gen.pybuffer.buf = NULL;
  __pyx_pybuffer_gen.refcount = 0;
  __pyx_pybuffernd_gen.data = NULL;
  __pyx_pybuffernd_gen.rcbuffer = &__pyx_pybuffer_gen;
  __pyx_pybuffer_delta.pybuffer.buf = NULL;
  __pyx_pybuffer_delta.refcount = 0;
  __pyx_pybuffernd_delta.data = NULL;
  __pyx_pybuffernd_delta.rcbuffer = &__pyx_pybuffer_delta;
  __pyx_pybuffer_psi.pybuffer.buf = NULL;
  __pyx_pybuffer_psi.refcount = 0;
  __pyx_pybuffernd_psi.data = NULL;
  __pyx_pybuffernd_psi.rcbuffer = &__pyx_pybuffer_psi;
  __pyx_pybuffer_soluce.pybuffer.buf = NULL;
  __pyx_pybuffer_soluce.refcount = 0;
  __pyx_pybuffernd_soluce.data = NULL;
  __pyx_pybuffernd_soluce.rcbuffer = &__pyx_pybuffer_soluce;
  __pyx_pybuffer_tempVal.pybuffer.buf = NULL;
  __pyx_pybuffer_tempVal.refcount = 0;
  __pyx_pybuffernd_tempVal.data = NULL;
  __pyx_pybuffernd_tempVal.rcbuffer = &__pyx_pybuffer_tempVal;

  /* "fastphase/fastphase.pyx":637
 * 
 * cpdef genViterbi( aa, tt, rr, gg):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 */
  if (!(likely(((__pyx_v_aa) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_aa, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 637, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_aa;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_alpha = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 637, __pyx_L1_error)
    } else {__pyx_pybuffernd_alpha.diminfo[0].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_alpha.diminfo[0].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_alpha.diminfo[1].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_alpha.diminfo[1].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_alpha = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":638
 * cpdef genViterbi( aa, tt, rr, gg):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] gen = gg
 */
  if (!(likely(((__pyx_v_tt) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_tt, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 638, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_tt;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_theta.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_theta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 638, __pyx_L1_error)
    } else {__pyx_pybuffernd_theta.diminfo[0].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_theta.diminfo[0].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_theta.diminfo[1].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_theta.diminfo[1].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_theta = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":639
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t, ndim=1] gen = gg
 * 
 */
  if (!(likely(((__pyx_v_rr) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_rr, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 639, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_rr;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rho.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_rho = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 639, __pyx_L1_error)
    } else {__pyx_pybuffernd_rho.diminfo[0].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rho.diminfo[0].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_rho.diminfo[1].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_rho.diminfo[1].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_rho = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":640
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] gen = gg             # <<<<<<<<<<<<<<
 * 
 *     ## cython declarations
 */
  if (!(likely(((__pyx_v_gg) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_gg, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 640, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_gg;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gen.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gen = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 640, __pyx_L1_error)
    } else {__pyx_pybuffernd_gen.diminfo[0].strides = __pyx_pybuffernd_gen.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gen.diminfo[0].shape = __pyx_pybuffernd_gen.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_v_gen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":647
 *     cdef double best_v
 * 
 *     nLoc = alpha.shape[0]             # <<<<<<<<<<<<<<
 *     nK = alpha.shape[1]
 *     npairs = nK * ( nK + 1)//2
 */
  __pyx_v_nLoc = (__pyx_v_alpha->dimensions[0]);

  /* "fastphase/fastphase.pyx":648
 * 
 *     nLoc = alpha.shape[0]
 *     nK = alpha.shape[1]             # <<<<<<<<<<<<<<
 *     npairs = nK * ( nK + 1)//2
 * 
 */
  __pyx_v_nK = (__pyx_v_alpha->dimensions[1]);

  /* "fastphase/fastphase.pyx":649
 *     nLoc = alpha.shape[0]
 *     nK = alpha.shape[1]
 *     npairs = nK * ( nK + 1)//2             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((npairs, nLoc), dtype=np.float64)
 */
  __pyx_v_npairs = __Pyx_div_long((__pyx_v_nK * (__pyx_v_nK + 1)), 2);

  /* "fastphase/fastphase.pyx":651
 *     npairs = nK * ( nK + 1)//2
 * 
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((npairs, nLoc), dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((npairs, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.zeros(nLoc, dtype= np.int)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_npairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 651, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_delta.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_delta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 651, __pyx_L1_error)
    } else {__pyx_pybuffernd_delta.diminfo[0].strides = __pyx_pybuffernd_delta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_delta.diminfo[0].shape = __pyx_pybuffernd_delta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_delta.diminfo[1].strides = __pyx_pybuffernd_delta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_delta.diminfo[1].shape = __pyx_pybuffernd_delta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_delta = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":652
 * 
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((npairs, nLoc), dtype=np.float64)
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((npairs, nLoc), dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.zeros(nLoc, dtype= np.int)
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( npairs, dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_npairs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 652, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_psi.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_psi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 652, __pyx_L1_error)
    } else {__pyx_pybuffernd_psi.diminfo[0].strides = __pyx_pybuffernd_psi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_psi.diminfo[0].shape = __pyx_pybuffernd_psi.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_psi.diminfo[1].strides = __pyx_pybuffernd_psi.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_psi.diminfo[1].shape = __pyx_pybuffernd_psi.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_psi = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":653
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((npairs, nLoc), dtype=np.float64)
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((npairs, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.zeros(nLoc, dtype= np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( npairs, dtype=np.float64)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 653, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_soluce = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 653, __pyx_L1_error)
    } else {__pyx_pybuffernd_soluce.diminfo[0].strides = __pyx_pybuffernd_soluce.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_soluce.diminfo[0].shape = __pyx_pybuffernd_soluce.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_soluce = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":654
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((npairs, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.zeros(nLoc, dtype= np.int)
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( npairs, dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_npairs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 654, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tempVal = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 654, __pyx_L1_error)
    } else {__pyx_pybuffernd_tempVal.diminfo[0].strides = __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tempVal.diminfo[0].shape = __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_tempVal = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":658
 * 
 *     ## initialization
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         ikp = pair2idx( k1, k1, nK)
 *         delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))
 */
  __pyx_t_10 = __pyx_v_nK;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_k1 = __pyx_t_12;

    /* "fastphase/fastphase.pyx":659
 *     ## initialization
 *     for k1 in range(nK):
 *         ikp = pair2idx( k1, k1, nK)             # <<<<<<<<<<<<<<
 *         delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))
 *         for k2 in range( k1+1 , nK):
 */
    __pyx_v_ikp = __pyx_f_6hapflk_9fastphase_pair2idx(__pyx_v_k1, __pyx_v_k1, __pyx_v_nK);

    /* "fastphase/fastphase.pyx":660
 *     for k1 in range(nK):
 *         ikp = pair2idx( k1, k1, nK)
 *         delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))             # <<<<<<<<<<<<<<
 *         for k2 in range( k1+1 , nK):
 *             ikp = pair2idx( k1, k2, nK)
 */
    __pyx_t_13 = 0;
    __pyx_t_14 = __pyx_v_k1;
    if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_alpha.diminfo[0].shape;
    if (__pyx_t_14 < 0) __pyx_t_14 += __pyx_pybuffernd_alpha.diminfo[1].shape;
    __pyx_t_15 = 0;
    __pyx_t_16 = __pyx_v_k1;
    if (__pyx_t_15 < 0) __pyx_t_15 += __pyx_pybuffernd_alpha.diminfo[0].shape;
    if (__pyx_t_16 < 0) __pyx_t_16 += __pyx_pybuffernd_alpha.diminfo[1].shape;
    __pyx_t_17 = 0;
    __pyx_t_18 = __pyx_v_k1;
    if (__pyx_t_17 < 0) __pyx_t_17 += __pyx_pybuffernd_theta.diminfo[0].shape;
    if (__pyx_t_18 < 0) __pyx_t_18 += __pyx_pybuffernd_theta.diminfo[1].shape;
    __pyx_t_19 = 0;
    __pyx_t_20 = __pyx_v_k1;
    if (__pyx_t_19 < 0) __pyx_t_19 += __pyx_pybuffernd_theta.diminfo[0].shape;
    if (__pyx_t_20 < 0) __pyx_t_20 += __pyx_pybuffernd_theta.diminfo[1].shape;
    __pyx_t_21 = 0;
    if (__pyx_t_21 < 0) __pyx_t_21 += __pyx_pybuffernd_gen.diminfo[0].shape;
    __pyx_t_22 = __pyx_v_ikp;
    __pyx_t_23 = 0;
    if (__pyx_t_22 < 0) __pyx_t_22 += __pyx_pybuffernd_delta.diminfo[0].shape;
    if (__pyx_t_23 < 0) __pyx_t_23 += __pyx_pybuffernd_delta.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_delta.diminfo[1].strides) = ((log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_alpha.diminfo[1].strides))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + log(__pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_20, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_gen.diminfo[0].strides)))));

    /* "fastphase/fastphase.pyx":661
 *         ikp = pair2idx( k1, k1, nK)
 *         delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))
 *         for k2 in range( k1+1 , nK):             # <<<<<<<<<<<<<<
 *             ikp = pair2idx( k1, k2, nK)
 *             delta[ikp,0] = log(2) + log( alpha[ 0, k1]) + log( alpha[ 0, k2]) +log( genprG( theta[ 0, k1], theta[ 0, k2], gen[0]))
 */
    __pyx_t_24 = __pyx_v_nK;
    __pyx_t_25 = __pyx_t_24;
    for (__pyx_t_26 = (__pyx_v_k1 + 1); __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
      __pyx_v_k2 = __pyx_t_26;

      /* "fastphase/fastphase.pyx":662
 *         delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))
 *         for k2 in range( k1+1 , nK):
 *             ikp = pair2idx( k1, k2, nK)             # <<<<<<<<<<<<<<
 *             delta[ikp,0] = log(2) + log( alpha[ 0, k1]) + log( alpha[ 0, k2]) +log( genprG( theta[ 0, k1], theta[ 0, k2], gen[0]))
 * 
 */
      __pyx_v_ikp = __pyx_f_6hapflk_9fastphase_pair2idx(__pyx_v_k1, __pyx_v_k2, __pyx_v_nK);

      /* "fastphase/fastphase.pyx":663
 *         for k2 in range( k1+1 , nK):
 *             ikp = pair2idx( k1, k2, nK)
 *             delta[ikp,0] = log(2) + log( alpha[ 0, k1]) + log( alpha[ 0, k2]) +log( genprG( theta[ 0, k1], theta[ 0, k2], gen[0]))             # <<<<<<<<<<<<<<
 * 
 *     ## recursion
 */
      __pyx_t_27 = 0;
      __pyx_t_28 = __pyx_v_k1;
      if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_29 = 0;
      __pyx_t_30 = __pyx_v_k2;
      if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_31 = 0;
      __pyx_t_32 = __pyx_v_k1;
      if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_33 = 0;
      __pyx_t_34 = __pyx_v_k2;
      if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_35 = 0;
      if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_gen.diminfo[0].shape;
      __pyx_t_36 = __pyx_v_ikp;
      __pyx_t_37 = 0;
      if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_delta.diminfo[0].shape;
      if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_delta.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_37, __pyx_pybuffernd_delta.diminfo[1].strides) = (((log(2.0) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_28, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_30, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + log(__pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_32, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_34, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_gen.diminfo[0].strides)))));
    }
  }

  /* "fastphase/fastphase.pyx":666
 * 
 *     ## recursion
 *     for m in range(1, nLoc):             # <<<<<<<<<<<<<<
 *         for ikp in range(npairs):
 *             k1, k2 = idx2pair( ikp, nK)
 */
  __pyx_t_10 = __pyx_v_nLoc;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 1; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_m = __pyx_t_12;

    /* "fastphase/fastphase.pyx":667
 *     ## recursion
 *     for m in range(1, nLoc):
 *         for ikp in range(npairs):             # <<<<<<<<<<<<<<
 *             k1, k2 = idx2pair( ikp, nK)
 *             for prev_kp in range(npairs):
 */
    __pyx_t_24 = __pyx_v_npairs;
    __pyx_t_25 = __pyx_t_24;
    for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
      __pyx_v_ikp = __pyx_t_26;

      /* "fastphase/fastphase.pyx":668
 *     for m in range(1, nLoc):
 *         for ikp in range(npairs):
 *             k1, k2 = idx2pair( ikp, nK)             # <<<<<<<<<<<<<<
 *             for prev_kp in range(npairs):
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 */
      __pyx_t_38 = __pyx_f_6hapflk_9fastphase_idx2pair(__pyx_v_ikp, __pyx_v_nK);
      __pyx_t_39 = __pyx_t_38.f0;
      __pyx_t_40 = __pyx_t_38.f1;
      __pyx_v_k1 = __pyx_t_39;
      __pyx_v_k2 = __pyx_t_40;

      /* "fastphase/fastphase.pyx":669
 *         for ikp in range(npairs):
 *             k1, k2 = idx2pair( ikp, nK)
 *             for prev_kp in range(npairs):             # <<<<<<<<<<<<<<
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 *                 if (k1 == kp1):
 */
      __pyx_t_40 = __pyx_v_npairs;
      __pyx_t_39 = __pyx_t_40;
      for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_39; __pyx_t_41+=1) {
        __pyx_v_prev_kp = __pyx_t_41;

        /* "fastphase/fastphase.pyx":670
 *             k1, k2 = idx2pair( ikp, nK)
 *             for prev_kp in range(npairs):
 *                 kp1, kp2 = idx2pair( prev_kp, nK)             # <<<<<<<<<<<<<<
 *                 if (k1 == kp1):
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2
 */
        __pyx_t_38 = __pyx_f_6hapflk_9fastphase_idx2pair(__pyx_v_prev_kp, __pyx_v_nK);
        __pyx_t_42 = __pyx_t_38.f0;
        __pyx_t_43 = __pyx_t_38.f1;
        __pyx_v_kp1 = __pyx_t_42;
        __pyx_v_kp2 = __pyx_t_43;

        /* "fastphase/fastphase.pyx":671
 *             for prev_kp in range(npairs):
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 *                 if (k1 == kp1):             # <<<<<<<<<<<<<<
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
 */
        __pyx_t_44 = ((__pyx_v_k1 == __pyx_v_kp1) != 0);
        if (__pyx_t_44) {

          /* "fastphase/fastphase.pyx":672
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 *                 if (k1 == kp1):
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
 *                     else: ## k1 == kp1 & k2 != kp2
 */
          __pyx_t_44 = ((__pyx_v_k2 == __pyx_v_kp2) != 0);
          if (__pyx_t_44) {

            /* "fastphase/fastphase.pyx":673
 *                 if (k1 == kp1):
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]             # <<<<<<<<<<<<<<
 *                     else: ## k1 == kp1 & k2 != kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]
 */
            __pyx_t_45 = __pyx_v_m;
            __pyx_t_46 = 0;
            if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_rho.diminfo[0].shape;
            if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_rho.diminfo[1].shape;
            __pyx_t_47 = __pyx_v_prev_kp;
            __pyx_t_48 = (__pyx_v_m - 1);
            if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_delta.diminfo[0].shape;
            if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_delta.diminfo[1].shape;
            __pyx_t_49 = __pyx_v_prev_kp;
            if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_tempVal.diminfo[0].strides) = (log(__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 0, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_46, __pyx_pybuffernd_rho.diminfo[1].strides)))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_48, __pyx_pybuffernd_delta.diminfo[1].strides)));

            /* "fastphase/fastphase.pyx":672
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 *                 if (k1 == kp1):
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
 *                     else: ## k1 == kp1 & k2 != kp2
 */
            goto __pyx_L14;
          }

          /* "fastphase/fastphase.pyx":675
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
 *                     else: ## k1 == kp1 & k2 != kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]             # <<<<<<<<<<<<<<
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 */
          /*else*/ {
            __pyx_t_50 = __pyx_v_m;
            __pyx_t_51 = 0;
            if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_rho.diminfo[0].shape;
            if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_rho.diminfo[1].shape;
            __pyx_t_52 = __pyx_v_m;
            __pyx_t_53 = __pyx_v_k2;
            if (__pyx_t_52 < 0) __pyx_t_52 += __pyx_pybuffernd_alpha.diminfo[0].shape;
            if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_alpha.diminfo[1].shape;
            __pyx_t_54 = __pyx_v_prev_kp;
            __pyx_t_55 = (__pyx_v_m - 1);
            if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_delta.diminfo[0].shape;
            if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_delta.diminfo[1].shape;
            __pyx_t_56 = __pyx_v_prev_kp;
            if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_56, __pyx_pybuffernd_tempVal.diminfo[0].strides) = ((log(__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_51, __pyx_pybuffernd_rho.diminfo[1].strides)))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_53, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_55, __pyx_pybuffernd_delta.diminfo[1].strides)));
          }
          __pyx_L14:;

          /* "fastphase/fastphase.pyx":671
 *             for prev_kp in range(npairs):
 *                 kp1, kp2 = idx2pair( prev_kp, nK)
 *                 if (k1 == kp1):             # <<<<<<<<<<<<<<
 *                     if (k2 == kp2): ## k1 == kp1 & k2 == kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
 */
          goto __pyx_L13;
        }

        /* "fastphase/fastphase.pyx":676
 *                     else: ## k1 == kp1 & k2 != kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing
 */
        __pyx_t_44 = ((__pyx_v_k2 == __pyx_v_kp2) != 0);
        if (__pyx_t_44) {

          /* "fastphase/fastphase.pyx":677
 *                         tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]             # <<<<<<<<<<<<<<
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]
 */
          __pyx_t_57 = __pyx_v_m;
          __pyx_t_58 = 0;
          if (__pyx_t_57 < 0) __pyx_t_57 += __pyx_pybuffernd_rho.diminfo[0].shape;
          if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_rho.diminfo[1].shape;
          __pyx_t_59 = __pyx_v_m;
          __pyx_t_60 = __pyx_v_k1;
          if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_alpha.diminfo[0].shape;
          if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_alpha.diminfo[1].shape;
          __pyx_t_61 = __pyx_v_prev_kp;
          __pyx_t_62 = (__pyx_v_m - 1);
          if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_delta.diminfo[0].shape;
          if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_delta.diminfo[1].shape;
          __pyx_t_63 = __pyx_v_prev_kp;
          if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_63, __pyx_pybuffernd_tempVal.diminfo[0].strides) = ((log(__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_57, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_58, __pyx_pybuffernd_rho.diminfo[1].strides)))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_59, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_60, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_61, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_62, __pyx_pybuffernd_delta.diminfo[1].strides)));

          /* "fastphase/fastphase.pyx":676
 *                     else: ## k1 == kp1 & k2 != kp2
 *                         tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2             # <<<<<<<<<<<<<<
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing
 */
          goto __pyx_L13;
        }

        /* "fastphase/fastphase.pyx":678
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing             # <<<<<<<<<<<<<<
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]
 *                 else: ## k1 !=kp1 & k2 != kp2
 */
        __pyx_t_64 = ((__pyx_v_k1 == __pyx_v_kp2) != 0);
        if (!__pyx_t_64) {
        } else {
          __pyx_t_44 = __pyx_t_64;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_64 = ((__pyx_v_k2 == __pyx_v_kp1) != 0);
        __pyx_t_44 = __pyx_t_64;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_44) {

          /* "fastphase/fastphase.pyx":679
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]             # <<<<<<<<<<<<<<
 *                 else: ## k1 !=kp1 & k2 != kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 2, rho[m, 0])) + log( alpha[ m, k1]) + alpha[m, k2] +  delta[ prev_kp, m-1]
 */
          __pyx_t_65 = __pyx_v_m;
          __pyx_t_66 = 0;
          if (__pyx_t_65 < 0) __pyx_t_65 += __pyx_pybuffernd_rho.diminfo[0].shape;
          if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_rho.diminfo[1].shape;
          __pyx_t_67 = __pyx_v_m;
          __pyx_t_68 = __pyx_v_k1;
          if (__pyx_t_67 < 0) __pyx_t_67 += __pyx_pybuffernd_alpha.diminfo[0].shape;
          if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_alpha.diminfo[1].shape;
          __pyx_t_69 = __pyx_v_prev_kp;
          __pyx_t_70 = (__pyx_v_m - 1);
          if (__pyx_t_69 < 0) __pyx_t_69 += __pyx_pybuffernd_delta.diminfo[0].shape;
          if (__pyx_t_70 < 0) __pyx_t_70 += __pyx_pybuffernd_delta.diminfo[1].shape;
          __pyx_t_71 = __pyx_v_prev_kp;
          if (__pyx_t_71 < 0) __pyx_t_71 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_tempVal.diminfo[0].strides) = ((log(__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_65, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_66, __pyx_pybuffernd_rho.diminfo[1].strides)))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_67, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_68, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_69, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_70, __pyx_pybuffernd_delta.diminfo[1].strides)));

          /* "fastphase/fastphase.pyx":678
 *                 elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
 *                 elif ( k1 == kp2) or (k2 == kp1): ## crossing             # <<<<<<<<<<<<<<
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]
 *                 else: ## k1 !=kp1 & k2 != kp2
 */
          goto __pyx_L13;
        }

        /* "fastphase/fastphase.pyx":681
 *                     tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]
 *                 else: ## k1 !=kp1 & k2 != kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 2, rho[m, 0])) + log( alpha[ m, k1]) + alpha[m, k2] +  delta[ prev_kp, m-1]             # <<<<<<<<<<<<<<
 *             psi[ ikp, m] = argmax(tempVal, npairs)
 *             delta[ ikp, m] = tempVal[ psi[ ikp, m]] + log( genprG( theta[ m, k1], theta[ m, k2], gen[m]))
 */
        /*else*/ {
          __pyx_t_72 = __pyx_v_m;
          __pyx_t_73 = 0;
          if (__pyx_t_72 < 0) __pyx_t_72 += __pyx_pybuffernd_rho.diminfo[0].shape;
          if (__pyx_t_73 < 0) __pyx_t_73 += __pyx_pybuffernd_rho.diminfo[1].shape;
          __pyx_t_74 = __pyx_v_m;
          __pyx_t_75 = __pyx_v_k1;
          if (__pyx_t_74 < 0) __pyx_t_74 += __pyx_pybuffernd_alpha.diminfo[0].shape;
          if (__pyx_t_75 < 0) __pyx_t_75 += __pyx_pybuffernd_alpha.diminfo[1].shape;
          __pyx_t_76 = __pyx_v_m;
          __pyx_t_77 = __pyx_v_k2;
          if (__pyx_t_76 < 0) __pyx_t_76 += __pyx_pybuffernd_alpha.diminfo[0].shape;
          if (__pyx_t_77 < 0) __pyx_t_77 += __pyx_pybuffernd_alpha.diminfo[1].shape;
          __pyx_t_78 = __pyx_v_prev_kp;
          __pyx_t_79 = (__pyx_v_m - 1);
          if (__pyx_t_78 < 0) __pyx_t_78 += __pyx_pybuffernd_delta.diminfo[0].shape;
          if (__pyx_t_79 < 0) __pyx_t_79 += __pyx_pybuffernd_delta.diminfo[1].shape;
          __pyx_t_80 = __pyx_v_prev_kp;
          if (__pyx_t_80 < 0) __pyx_t_80 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_80, __pyx_pybuffernd_tempVal.diminfo[0].strides) = (((log(__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 2, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_72, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_73, __pyx_pybuffernd_rho.diminfo[1].strides)))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_74, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_75, __pyx_pybuffernd_alpha.diminfo[1].strides)))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_76, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_77, __pyx_pybuffernd_alpha.diminfo[1].strides))) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_78, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_79, __pyx_pybuffernd_delta.diminfo[1].strides)));
        }
        __pyx_L13:;
      }

      /* "fastphase/fastphase.pyx":682
 *                 else: ## k1 !=kp1 & k2 != kp2
 *                     tempVal[ prev_kp] = log(probJ(m, 2, rho[m, 0])) + log( alpha[ m, k1]) + alpha[m, k2] +  delta[ prev_kp, m-1]
 *             psi[ ikp, m] = argmax(tempVal, npairs)             # <<<<<<<<<<<<<<
 *             delta[ ikp, m] = tempVal[ psi[ ikp, m]] + log( genprG( theta[ m, k1], theta[ m, k2], gen[m]))
 * 
 */
      __pyx_t_81 = __pyx_v_ikp;
      __pyx_t_82 = __pyx_v_m;
      if (__pyx_t_81 < 0) __pyx_t_81 += __pyx_pybuffernd_psi.diminfo[0].shape;
      if (__pyx_t_82 < 0) __pyx_t_82 += __pyx_pybuffernd_psi.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_81, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_82, __pyx_pybuffernd_psi.diminfo[1].strides) = __pyx_f_6hapflk_9fastphase_argmax(((PyArrayObject *)__pyx_v_tempVal), __pyx_v_npairs);

      /* "fastphase/fastphase.pyx":683
 *                     tempVal[ prev_kp] = log(probJ(m, 2, rho[m, 0])) + log( alpha[ m, k1]) + alpha[m, k2] +  delta[ prev_kp, m-1]
 *             psi[ ikp, m] = argmax(tempVal, npairs)
 *             delta[ ikp, m] = tempVal[ psi[ ikp, m]] + log( genprG( theta[ m, k1], theta[ m, k2], gen[m]))             # <<<<<<<<<<<<<<
 * 
 *     ## termination
 */
      __pyx_t_83 = __pyx_v_ikp;
      __pyx_t_84 = __pyx_v_m;
      if (__pyx_t_83 < 0) __pyx_t_83 += __pyx_pybuffernd_psi.diminfo[0].shape;
      if (__pyx_t_84 < 0) __pyx_t_84 += __pyx_pybuffernd_psi.diminfo[1].shape;
      __pyx_t_85 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_83, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_84, __pyx_pybuffernd_psi.diminfo[1].strides));
      if (__pyx_t_85 < 0) __pyx_t_85 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
      __pyx_t_86 = __pyx_v_m;
      __pyx_t_87 = __pyx_v_k1;
      if (__pyx_t_86 < 0) __pyx_t_86 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_87 < 0) __pyx_t_87 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_88 = __pyx_v_m;
      __pyx_t_89 = __pyx_v_k2;
      if (__pyx_t_88 < 0) __pyx_t_88 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_89 < 0) __pyx_t_89 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_90 = __pyx_v_m;
      if (__pyx_t_90 < 0) __pyx_t_90 += __pyx_pybuffernd_gen.diminfo[0].shape;
      __pyx_t_91 = __pyx_v_ikp;
      __pyx_t_92 = __pyx_v_m;
      if (__pyx_t_91 < 0) __pyx_t_91 += __pyx_pybuffernd_delta.diminfo[0].shape;
      if (__pyx_t_92 < 0) __pyx_t_92 += __pyx_pybuffernd_delta.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_91, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_92, __pyx_pybuffernd_delta.diminfo[1].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_85, __pyx_pybuffernd_tempVal.diminfo[0].strides)) + log(__pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_86, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_87, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_89, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_90, __pyx_pybuffernd_gen.diminfo[0].strides)))));
    }
  }

  /* "fastphase/fastphase.pyx":686
 * 
 *     ## termination
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], npairs)             # <<<<<<<<<<<<<<
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 */
  __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_nLoc - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_slice_);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_delta), __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 686, __pyx_L1_error)
  __pyx_t_93 = (__pyx_v_nLoc - 1);
  if (__pyx_t_93 < 0) __pyx_t_93 += __pyx_pybuffernd_soluce.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_soluce.diminfo[0].strides) = __pyx_f_6hapflk_9fastphase_argmax(((PyArrayObject *)__pyx_t_4), __pyx_v_npairs);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":687
 *     ## termination
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], npairs)
 *     for m in range( nLoc - 2, -1, -1):             # <<<<<<<<<<<<<<
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 *     return [ idx2pair(ikp, nK) for ikp in soluce]
 */
  for (__pyx_t_10 = (__pyx_v_nLoc - 2); __pyx_t_10 > -1; __pyx_t_10-=1) {
    __pyx_v_m = __pyx_t_10;

    /* "fastphase/fastphase.pyx":688
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], npairs)
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]             # <<<<<<<<<<<<<<
 *     return [ idx2pair(ikp, nK) for ikp in soluce]
 * 
 */
    __pyx_t_94 = (__pyx_v_m + 1);
    if (__pyx_t_94 < 0) __pyx_t_94 += __pyx_pybuffernd_soluce.diminfo[0].shape;
    __pyx_t_95 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_94, __pyx_pybuffernd_soluce.diminfo[0].strides));
    __pyx_t_96 = (__pyx_v_m + 1);
    if (__pyx_t_95 < 0) __pyx_t_95 += __pyx_pybuffernd_psi.diminfo[0].shape;
    if (__pyx_t_96 < 0) __pyx_t_96 += __pyx_pybuffernd_psi.diminfo[1].shape;
    __pyx_t_97 = __pyx_v_m;
    if (__pyx_t_97 < 0) __pyx_t_97 += __pyx_pybuffernd_soluce.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_97, __pyx_pybuffernd_soluce.diminfo[0].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_95, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_96, __pyx_pybuffernd_psi.diminfo[1].strides));
  }

  /* "fastphase/fastphase.pyx":689
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 *     return [ idx2pair(ikp, nK) for ikp in soluce]             # <<<<<<<<<<<<<<
 * 
 * cpdef genCalc(aa,tt,rr,gg,u2p):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_soluce))) || PyTuple_CheckExact(((PyObject *)__pyx_v_soluce))) {
      __pyx_t_5 = ((PyObject *)__pyx_v_soluce); __Pyx_INCREF(__pyx_t_5); __pyx_t_98 = 0;
      __pyx_t_99 = NULL;
    } else {
      __pyx_t_98 = -1; __pyx_t_5 = PyObject_GetIter(((PyObject *)__pyx_v_soluce)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_99 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 689, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_99)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_98 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_98); __Pyx_INCREF(__pyx_t_3); __pyx_t_98++; if (unlikely(0 < 0)) __PYX_ERR(0, 689, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_98); __pyx_t_98++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 689, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_98 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_98); __Pyx_INCREF(__pyx_t_3); __pyx_t_98++; if (unlikely(0 < 0)) __PYX_ERR(0, 689, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_98); __pyx_t_98++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 689, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_99(__pyx_t_5);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 689, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_9genexpr15__pyx_v_ikp = __pyx_t_10;
      __pyx_t_3 = __pyx_convert__to_py___pyx_ctuple_int__and_int(__pyx_f_6hapflk_9fastphase_idx2pair(__pyx_9genexpr15__pyx_v_ikp, __pyx_v_nK)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } /* exit inner scope */
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":636
 *         return rho*rho
 * 
 * cpdef genViterbi( aa, tt, rr, gg):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_delta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gen.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_psi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.genViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_delta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gen.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_psi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_alpha);
  __Pyx_XDECREF((PyObject *)__pyx_v_theta);
  __Pyx_XDECREF((PyObject *)__pyx_v_rho);
  __Pyx_XDECREF((PyObject *)__pyx_v_gen);
  __Pyx_XDECREF((PyObject *)__pyx_v_delta);
  __Pyx_XDECREF((PyObject *)__pyx_v_psi);
  __Pyx_XDECREF((PyObject *)__pyx_v_soluce);
  __Pyx_XDECREF((PyObject *)__pyx_v_tempVal);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_17genViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_17genViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aa = 0;
  PyObject *__pyx_v_tt = 0;
  PyObject *__pyx_v_rr = 0;
  PyObject *__pyx_v_gg = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genViterbi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aa,&__pyx_n_s_tt,&__pyx_n_s_rr,&__pyx_n_s_gg,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aa)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genViterbi", 1, 4, 4, 1); __PYX_ERR(0, 636, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genViterbi", 1, 4, 4, 2); __PYX_ERR(0, 636, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gg)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genViterbi", 1, 4, 4, 3); __PYX_ERR(0, 636, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "genViterbi") < 0)) __PYX_ERR(0, 636, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_aa = values[0];
    __pyx_v_tt = values[1];
    __pyx_v_rr = values[2];
    __pyx_v_gg = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("genViterbi", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 636, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.genViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_16genViterbi(__pyx_self, __pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_gg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_16genViterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("genViterbi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6hapflk_9fastphase_genViterbi(__pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_gg, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.genViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":691
 *     return [ idx2pair(ikp, nK) for ikp in soluce]
 * 
 * cpdef genCalc(aa,tt,rr,gg,u2p):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha=aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_19genCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_genCalc(PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg, PyObject *__pyx_v_u2p, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_alpha = 0;
  PyArrayObject *__pyx_v_theta = 0;
  PyArrayObject *__pyx_v_rho = 0;
  PyArrayObject *__pyx_v_gen = 0;
  int __pyx_v_up2pz;
  int __pyx_v_tScale;
  double __pyx_v_dummy;
  double __pyx_v_t1;
  double __pyx_v_t2;
  double __pyx_v_temp;
  double __pyx_v_tScaleTemp;
  double __pyx_v_normC;
  PyObject *__pyx_v_nK = 0;
  int __pyx_v_k;
  int __pyx_v_k1;
  int __pyx_v_k2;
  int __pyx_v_m;
  double __pyx_v_logLikelihood;
  PyObject *__pyx_v_nLoc = NULL;
  PyArrayObject *__pyx_v_betaScale = 0;
  PyArrayObject *__pyx_v_tSumk = 0;
  PyArrayObject *__pyx_v_tDoubleSum = 0;
  PyArrayObject *__pyx_v_mBeta = 0;
  PyArrayObject *__pyx_v_mPhi = 0;
  PyArrayObject *__pyx_v_phiScale = 0;
  PyArrayObject *__pyx_v_probZ = 0;
  PyArrayObject *__pyx_v_jmk = 0;
  PyArrayObject *__pyx_v_top = 0;
  PyArrayObject *__pyx_v_bot = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_alpha;
  __Pyx_Buffer __pyx_pybuffer_alpha;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_betaScale;
  __Pyx_Buffer __pyx_pybuffer_betaScale;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bot;
  __Pyx_Buffer __pyx_pybuffer_bot;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gen;
  __Pyx_Buffer __pyx_pybuffer_gen;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_jmk;
  __Pyx_Buffer __pyx_pybuffer_jmk;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mBeta;
  __Pyx_Buffer __pyx_pybuffer_mBeta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mPhi;
  __Pyx_Buffer __pyx_pybuffer_mPhi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phiScale;
  __Pyx_Buffer __pyx_pybuffer_phiScale;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_probZ;
  __Pyx_Buffer __pyx_pybuffer_probZ;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rho;
  __Pyx_Buffer __pyx_pybuffer_rho;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tDoubleSum;
  __Pyx_Buffer __pyx_pybuffer_tDoubleSum;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tSumk;
  __Pyx_Buffer __pyx_pybuffer_tSumk;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_theta;
  __Pyx_Buffer __pyx_pybuffer_theta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_top;
  __Pyx_Buffer __pyx_pybuffer_top;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  long __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  long __pyx_t_14;
  int __pyx_t_15;
  double __pyx_t_16;
  double __pyx_t_17;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  long __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  Py_ssize_t __pyx_t_51;
  Py_ssize_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  Py_ssize_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  Py_ssize_t __pyx_t_64;
  Py_ssize_t __pyx_t_65;
  Py_ssize_t __pyx_t_66;
  int __pyx_t_67;
  Py_ssize_t __pyx_t_68;
  Py_ssize_t __pyx_t_69;
  Py_ssize_t __pyx_t_70;
  Py_ssize_t __pyx_t_71;
  Py_ssize_t __pyx_t_72;
  Py_ssize_t __pyx_t_73;
  Py_ssize_t __pyx_t_74;
  Py_ssize_t __pyx_t_75;
  Py_ssize_t __pyx_t_76;
  Py_ssize_t __pyx_t_77;
  Py_ssize_t __pyx_t_78;
  Py_ssize_t __pyx_t_79;
  Py_ssize_t __pyx_t_80;
  Py_ssize_t __pyx_t_81;
  PyArrayObject *__pyx_t_82 = NULL;
  PyArrayObject *__pyx_t_83 = NULL;
  Py_ssize_t __pyx_t_84;
  Py_ssize_t __pyx_t_85;
  Py_ssize_t __pyx_t_86;
  Py_ssize_t __pyx_t_87;
  Py_ssize_t __pyx_t_88;
  Py_ssize_t __pyx_t_89;
  Py_ssize_t __pyx_t_90;
  Py_ssize_t __pyx_t_91;
  Py_ssize_t __pyx_t_92;
  Py_ssize_t __pyx_t_93;
  Py_ssize_t __pyx_t_94;
  Py_ssize_t __pyx_t_95;
  Py_ssize_t __pyx_t_96;
  Py_ssize_t __pyx_t_97;
  Py_ssize_t __pyx_t_98;
  Py_ssize_t __pyx_t_99;
  Py_ssize_t __pyx_t_100;
  Py_ssize_t __pyx_t_101;
  Py_ssize_t __pyx_t_102;
  Py_ssize_t __pyx_t_103;
  Py_ssize_t __pyx_t_104;
  Py_ssize_t __pyx_t_105;
  Py_ssize_t __pyx_t_106;
  Py_ssize_t __pyx_t_107;
  Py_ssize_t __pyx_t_108;
  Py_ssize_t __pyx_t_109;
  Py_ssize_t __pyx_t_110;
  Py_ssize_t __pyx_t_111;
  Py_ssize_t __pyx_t_112;
  Py_ssize_t __pyx_t_113;
  Py_ssize_t __pyx_t_114;
  Py_ssize_t __pyx_t_115;
  Py_ssize_t __pyx_t_116;
  Py_ssize_t __pyx_t_117;
  Py_ssize_t __pyx_t_118;
  Py_ssize_t __pyx_t_119;
  Py_ssize_t __pyx_t_120;
  Py_ssize_t __pyx_t_121;
  Py_ssize_t __pyx_t_122;
  Py_ssize_t __pyx_t_123;
  Py_ssize_t __pyx_t_124;
  Py_ssize_t __pyx_t_125;
  Py_ssize_t __pyx_t_126;
  long __pyx_t_127;
  Py_ssize_t __pyx_t_128;
  Py_ssize_t __pyx_t_129;
  Py_ssize_t __pyx_t_130;
  Py_ssize_t __pyx_t_131;
  Py_ssize_t __pyx_t_132;
  Py_ssize_t __pyx_t_133;
  Py_ssize_t __pyx_t_134;
  Py_ssize_t __pyx_t_135;
  Py_ssize_t __pyx_t_136;
  Py_ssize_t __pyx_t_137;
  Py_ssize_t __pyx_t_138;
  Py_ssize_t __pyx_t_139;
  Py_ssize_t __pyx_t_140;
  Py_ssize_t __pyx_t_141;
  Py_ssize_t __pyx_t_142;
  Py_ssize_t __pyx_t_143;
  Py_ssize_t __pyx_t_144;
  Py_ssize_t __pyx_t_145;
  Py_ssize_t __pyx_t_146;
  Py_ssize_t __pyx_t_147;
  Py_ssize_t __pyx_t_148;
  Py_ssize_t __pyx_t_149;
  Py_ssize_t __pyx_t_150;
  Py_ssize_t __pyx_t_151;
  Py_ssize_t __pyx_t_152;
  Py_ssize_t __pyx_t_153;
  Py_ssize_t __pyx_t_154;
  Py_ssize_t __pyx_t_155;
  Py_ssize_t __pyx_t_156;
  Py_ssize_t __pyx_t_157;
  Py_ssize_t __pyx_t_158;
  Py_ssize_t __pyx_t_159;
  Py_ssize_t __pyx_t_160;
  Py_ssize_t __pyx_t_161;
  Py_ssize_t __pyx_t_162;
  Py_ssize_t __pyx_t_163;
  Py_ssize_t __pyx_t_164;
  Py_ssize_t __pyx_t_165;
  Py_ssize_t __pyx_t_166;
  Py_ssize_t __pyx_t_167;
  Py_ssize_t __pyx_t_168;
  Py_ssize_t __pyx_t_169;
  Py_ssize_t __pyx_t_170;
  Py_ssize_t __pyx_t_171;
  Py_ssize_t __pyx_t_172;
  Py_ssize_t __pyx_t_173;
  Py_ssize_t __pyx_t_174;
  Py_ssize_t __pyx_t_175;
  Py_ssize_t __pyx_t_176;
  Py_ssize_t __pyx_t_177;
  Py_ssize_t __pyx_t_178;
  Py_ssize_t __pyx_t_179;
  Py_ssize_t __pyx_t_180;
  Py_ssize_t __pyx_t_181;
  Py_ssize_t __pyx_t_182;
  Py_ssize_t __pyx_t_183;
  Py_ssize_t __pyx_t_184;
  Py_ssize_t __pyx_t_185;
  Py_ssize_t __pyx_t_186;
  Py_ssize_t __pyx_t_187;
  Py_ssize_t __pyx_t_188;
  Py_ssize_t __pyx_t_189;
  Py_ssize_t __pyx_t_190;
  Py_ssize_t __pyx_t_191;
  Py_ssize_t __pyx_t_192;
  PyArrayObject *__pyx_t_193 = NULL;
  PyArrayObject *__pyx_t_194 = NULL;
  PyArrayObject *__pyx_t_195 = NULL;
  PyArrayObject *__pyx_t_196 = NULL;
  Py_ssize_t __pyx_t_197;
  Py_ssize_t __pyx_t_198;
  Py_ssize_t __pyx_t_199;
  Py_ssize_t __pyx_t_200;
  Py_ssize_t __pyx_t_201;
  Py_ssize_t __pyx_t_202;
  Py_ssize_t __pyx_t_203;
  Py_ssize_t __pyx_t_204;
  Py_ssize_t __pyx_t_205;
  Py_ssize_t __pyx_t_206;
  Py_ssize_t __pyx_t_207;
  Py_ssize_t __pyx_t_208;
  Py_ssize_t __pyx_t_209;
  Py_ssize_t __pyx_t_210;
  Py_ssize_t __pyx_t_211;
  Py_ssize_t __pyx_t_212;
  Py_ssize_t __pyx_t_213;
  Py_ssize_t __pyx_t_214;
  Py_ssize_t __pyx_t_215;
  Py_ssize_t __pyx_t_216;
  Py_ssize_t __pyx_t_217;
  Py_ssize_t __pyx_t_218;
  Py_ssize_t __pyx_t_219;
  Py_ssize_t __pyx_t_220;
  Py_ssize_t __pyx_t_221;
  Py_ssize_t __pyx_t_222;
  Py_ssize_t __pyx_t_223;
  Py_ssize_t __pyx_t_224;
  Py_ssize_t __pyx_t_225;
  Py_ssize_t __pyx_t_226;
  Py_ssize_t __pyx_t_227;
  Py_ssize_t __pyx_t_228;
  Py_ssize_t __pyx_t_229;
  Py_ssize_t __pyx_t_230;
  Py_ssize_t __pyx_t_231;
  Py_ssize_t __pyx_t_232;
  Py_ssize_t __pyx_t_233;
  Py_ssize_t __pyx_t_234;
  Py_ssize_t __pyx_t_235;
  Py_ssize_t __pyx_t_236;
  Py_ssize_t __pyx_t_237;
  __pyx_t_5numpy_float64_t __pyx_t_238;
  Py_ssize_t __pyx_t_239;
  Py_ssize_t __pyx_t_240;
  Py_ssize_t __pyx_t_241;
  Py_ssize_t __pyx_t_242;
  Py_ssize_t __pyx_t_243;
  Py_ssize_t __pyx_t_244;
  Py_ssize_t __pyx_t_245;
  Py_ssize_t __pyx_t_246;
  Py_ssize_t __pyx_t_247;
  Py_ssize_t __pyx_t_248;
  Py_ssize_t __pyx_t_249;
  Py_ssize_t __pyx_t_250;
  Py_ssize_t __pyx_t_251;
  Py_ssize_t __pyx_t_252;
  Py_ssize_t __pyx_t_253;
  Py_ssize_t __pyx_t_254;
  Py_ssize_t __pyx_t_255;
  Py_ssize_t __pyx_t_256;
  Py_ssize_t __pyx_t_257;
  Py_ssize_t __pyx_t_258;
  Py_ssize_t __pyx_t_259;
  Py_ssize_t __pyx_t_260;
  Py_ssize_t __pyx_t_261;
  Py_ssize_t __pyx_t_262;
  Py_ssize_t __pyx_t_263;
  Py_ssize_t __pyx_t_264;
  Py_ssize_t __pyx_t_265;
  Py_ssize_t __pyx_t_266;
  Py_ssize_t __pyx_t_267;
  Py_ssize_t __pyx_t_268;
  Py_ssize_t __pyx_t_269;
  Py_ssize_t __pyx_t_270;
  Py_ssize_t __pyx_t_271;
  Py_ssize_t __pyx_t_272;
  Py_ssize_t __pyx_t_273;
  Py_ssize_t __pyx_t_274;
  Py_ssize_t __pyx_t_275;
  Py_ssize_t __pyx_t_276;
  Py_ssize_t __pyx_t_277;
  Py_ssize_t __pyx_t_278;
  Py_ssize_t __pyx_t_279;
  Py_ssize_t __pyx_t_280;
  Py_ssize_t __pyx_t_281;
  Py_ssize_t __pyx_t_282;
  Py_ssize_t __pyx_t_283;
  Py_ssize_t __pyx_t_284;
  Py_ssize_t __pyx_t_285;
  Py_ssize_t __pyx_t_286;
  Py_ssize_t __pyx_t_287;
  Py_ssize_t __pyx_t_288;
  Py_ssize_t __pyx_t_289;
  __Pyx_RefNannySetupContext("genCalc", 0);
  __pyx_pybuffer_alpha.pybuffer.buf = NULL;
  __pyx_pybuffer_alpha.refcount = 0;
  __pyx_pybuffernd_alpha.data = NULL;
  __pyx_pybuffernd_alpha.rcbuffer = &__pyx_pybuffer_alpha;
  __pyx_pybuffer_theta.pybuffer.buf = NULL;
  __pyx_pybuffer_theta.refcount = 0;
  __pyx_pybuffernd_theta.data = NULL;
  __pyx_pybuffernd_theta.rcbuffer = &__pyx_pybuffer_theta;
  __pyx_pybuffer_rho.pybuffer.buf = NULL;
  __pyx_pybuffer_rho.refcount = 0;
  __pyx_pybuffernd_rho.data = NULL;
  __pyx_pybuffernd_rho.rcbuffer = &__pyx_pybuffer_rho;
  __pyx_pybuffer_gen.pybuffer.buf = NULL;
  __pyx_pybuffer_gen.refcount = 0;
  __pyx_pybuffernd_gen.data = NULL;
  __pyx_pybuffernd_gen.rcbuffer = &__pyx_pybuffer_gen;
  __pyx_pybuffer_betaScale.pybuffer.buf = NULL;
  __pyx_pybuffer_betaScale.refcount = 0;
  __pyx_pybuffernd_betaScale.data = NULL;
  __pyx_pybuffernd_betaScale.rcbuffer = &__pyx_pybuffer_betaScale;
  __pyx_pybuffer_tSumk.pybuffer.buf = NULL;
  __pyx_pybuffer_tSumk.refcount = 0;
  __pyx_pybuffernd_tSumk.data = NULL;
  __pyx_pybuffernd_tSumk.rcbuffer = &__pyx_pybuffer_tSumk;
  __pyx_pybuffer_tDoubleSum.pybuffer.buf = NULL;
  __pyx_pybuffer_tDoubleSum.refcount = 0;
  __pyx_pybuffernd_tDoubleSum.data = NULL;
  __pyx_pybuffernd_tDoubleSum.rcbuffer = &__pyx_pybuffer_tDoubleSum;
  __pyx_pybuffer_mBeta.pybuffer.buf = NULL;
  __pyx_pybuffer_mBeta.refcount = 0;
  __pyx_pybuffernd_mBeta.data = NULL;
  __pyx_pybuffernd_mBeta.rcbuffer = &__pyx_pybuffer_mBeta;
  __pyx_pybuffer_mPhi.pybuffer.buf = NULL;
  __pyx_pybuffer_mPhi.refcount = 0;
  __pyx_pybuffernd_mPhi.data = NULL;
  __pyx_pybuffernd_mPhi.rcbuffer = &__pyx_pybuffer_mPhi;
  __pyx_pybuffer_phiScale.pybuffer.buf = NULL;
  __pyx_pybuffer_phiScale.refcount = 0;
  __pyx_pybuffernd_phiScale.data = NULL;
  __pyx_pybuffernd_phiScale.rcbuffer = &__pyx_pybuffer_phiScale;
  __pyx_pybuffer_probZ.pybuffer.buf = NULL;
  __pyx_pybuffer_probZ.refcount = 0;
  __pyx_pybuffernd_probZ.data = NULL;
  __pyx_pybuffernd_probZ.rcbuffer = &__pyx_pybuffer_probZ;
  __pyx_pybuffer_jmk.pybuffer.buf = NULL;
  __pyx_pybuffer_jmk.refcount = 0;
  __pyx_pybuffernd_jmk.data = NULL;
  __pyx_pybuffernd_jmk.rcbuffer = &__pyx_pybuffer_jmk;
  __pyx_pybuffer_top.pybuffer.buf = NULL;
  __pyx_pybuffer_top.refcount = 0;
  __pyx_pybuffernd_top.data = NULL;
  __pyx_pybuffernd_top.rcbuffer = &__pyx_pybuffer_top;
  __pyx_pybuffer_bot.pybuffer.buf = NULL;
  __pyx_pybuffer_bot.refcount = 0;
  __pyx_pybuffernd_bot.data = NULL;
  __pyx_pybuffernd_bot.rcbuffer = &__pyx_pybuffer_bot;

  /* "fastphase/fastphase.pyx":692
 * 
 * cpdef genCalc(aa,tt,rr,gg,u2p):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha=aa             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho=rr
 */
  if (!(likely(((__pyx_v_aa) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_aa, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 692, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_aa;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_alpha = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 692, __pyx_L1_error)
    } else {__pyx_pybuffernd_alpha.diminfo[0].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_alpha.diminfo[0].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_alpha.diminfo[1].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_alpha.diminfo[1].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_alpha = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":693
 * cpdef genCalc(aa,tt,rr,gg,u2p):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha=aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] rho=rr
 *     cdef np.ndarray[np.int_t, ndim=1] gen=gg
 */
  if (!(likely(((__pyx_v_tt) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_tt, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 693, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_tt;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_theta.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_theta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 693, __pyx_L1_error)
    } else {__pyx_pybuffernd_theta.diminfo[0].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_theta.diminfo[0].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_theta.diminfo[1].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_theta.diminfo[1].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_theta = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":694
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha=aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho=rr             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t, ndim=1] gen=gg
 *     cdef int up2pz=u2p
 */
  if (!(likely(((__pyx_v_rr) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_rr, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 694, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_rr;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rho.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_rho = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 694, __pyx_L1_error)
    } else {__pyx_pybuffernd_rho.diminfo[0].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rho.diminfo[0].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_rho.diminfo[1].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_rho.diminfo[1].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_rho = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":695
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho=rr
 *     cdef np.ndarray[np.int_t, ndim=1] gen=gg             # <<<<<<<<<<<<<<
 *     cdef int up2pz=u2p
 *     ## implementation comment: TODO = try to minimize the memory footprint
 */
  if (!(likely(((__pyx_v_gg) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_gg, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 695, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_gg;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gen.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gen = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 695, __pyx_L1_error)
    } else {__pyx_pybuffernd_gen.diminfo[0].strides = __pyx_pybuffernd_gen.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gen.diminfo[0].shape = __pyx_pybuffernd_gen.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_v_gen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":696
 *     cdef np.ndarray[np.float64_t, ndim=2] rho=rr
 *     cdef np.ndarray[np.int_t, ndim=1] gen=gg
 *     cdef int up2pz=u2p             # <<<<<<<<<<<<<<
 *     ## implementation comment: TODO = try to minimize the memory footprint
 *     ## of big Nloc*nK*nK matrices
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_u2p); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L1_error)
  __pyx_v_up2pz = __pyx_t_2;

  /* "fastphase/fastphase.pyx":710
 *     ## end cython declarations
 * 
 *     nLoc=alpha.shape[0]             # <<<<<<<<<<<<<<
 *     nK=alpha.shape[1]
 *     ##
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_alpha->dimensions[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nLoc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":711
 * 
 *     nLoc=alpha.shape[0]
 *     nK=alpha.shape[1]             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute backward probabilities
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_alpha->dimensions[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nK = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":715
 *     ## compute backward probabilities
 *     ##
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] tSumk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=1] tDoubleSum=np.zeros(nLoc,dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nLoc);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_6);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_betaScale = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 715, __pyx_L1_error)
    } else {__pyx_pybuffernd_betaScale.diminfo[0].strides = __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_betaScale.diminfo[0].shape = __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_betaScale = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "fastphase/fastphase.pyx":716
 *     ##
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
 *     cdef np.ndarray[np.float64_t,ndim=2] tSumk=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=1] tDoubleSum=np.zeros(nLoc,dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=3] mBeta=np.zeros((nLoc,nK,nK),dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_nK);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 716, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tSumk.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_tSumk = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 716, __pyx_L1_error)
    } else {__pyx_pybuffernd_tSumk.diminfo[0].strides = __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tSumk.diminfo[0].shape = __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tSumk.diminfo[1].strides = __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tSumk.diminfo[1].shape = __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_tSumk = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":717
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
 *     cdef np.ndarray[np.float64_t,ndim=2] tSumk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=1] tDoubleSum=np.zeros(nLoc,dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=3] mBeta=np.zeros((nLoc,nK,nK),dtype=np.float64)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nLoc);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 717, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tDoubleSum = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 717, __pyx_L1_error)
    } else {__pyx_pybuffernd_tDoubleSum.diminfo[0].strides = __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tDoubleSum.diminfo[0].shape = __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_tDoubleSum = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":718
 *     cdef np.ndarray[np.float64_t,ndim=2] tSumk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=1] tDoubleSum=np.zeros(nLoc,dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=3] mBeta=np.zeros((nLoc,nK,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     for k1 in range(nK):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_nK);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_nK);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 718, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_mBeta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 718, __pyx_L1_error)
    } else {__pyx_pybuffernd_mBeta.diminfo[0].strides = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mBeta.diminfo[0].shape = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mBeta.diminfo[1].strides = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mBeta.diminfo[1].shape = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mBeta.diminfo[2].strides = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mBeta.diminfo[2].shape = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_mBeta = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":720
 *     cdef np.ndarray[np.float64_t,ndim=3] mBeta=np.zeros((nLoc,nK,nK),dtype=np.float64)
 * 
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         for k2 in range(nK):
 *             mBeta[nLoc-1,k1,k2]=1
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_k1 = __pyx_t_2;

    /* "fastphase/fastphase.pyx":721
 * 
 *     for k1 in range(nK):
 *         for k2 in range(nK):             # <<<<<<<<<<<<<<
 *             mBeta[nLoc-1,k1,k2]=1
 *     ## calc the marginal sum at locus M as in appendix A
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k2 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":722
 *     for k1 in range(nK):
 *         for k2 in range(nK):
 *             mBeta[nLoc-1,k1,k2]=1             # <<<<<<<<<<<<<<
 *     ## calc the marginal sum at locus M as in appendix A
 *     for k1 in range(nK):
 */
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_k2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
      __pyx_t_4 = 0;
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_mBeta), __pyx_t_1, __pyx_int_1) < 0)) __PYX_ERR(0, 722, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "fastphase/fastphase.pyx":724
 *             mBeta[nLoc-1,k1,k2]=1
 *     ## calc the marginal sum at locus M as in appendix A
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         for k2 in range(nK):
 *             tSumk[nLoc-1,k1] += genprG(theta[nLoc-1,k1],theta[nLoc-1,k2],gen[nLoc-1])*mBeta[nLoc-1,k1,k2]*alpha[nLoc-1,k2]
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 724, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_k1 = __pyx_t_2;

    /* "fastphase/fastphase.pyx":725
 *     ## calc the marginal sum at locus M as in appendix A
 *     for k1 in range(nK):
 *         for k2 in range(nK):             # <<<<<<<<<<<<<<
 *             tSumk[nLoc-1,k1] += genprG(theta[nLoc-1,k1],theta[nLoc-1,k2],gen[nLoc-1])*mBeta[nLoc-1,k1,k2]*alpha[nLoc-1,k2]
 *         tDoubleSum[nLoc-1]+=tSumk[nLoc-1,k1]*alpha[nLoc-1,k1]
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k2 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":726
 *     for k1 in range(nK):
 *         for k2 in range(nK):
 *             tSumk[nLoc-1,k1] += genprG(theta[nLoc-1,k1],theta[nLoc-1,k2],gen[nLoc-1])*mBeta[nLoc-1,k1,k2]*alpha[nLoc-1,k2]             # <<<<<<<<<<<<<<
 *         tDoubleSum[nLoc-1]+=tSumk[nLoc-1,k1]*alpha[nLoc-1,k1]
 *     ## recurrent calculations backward
 */
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_tSumk), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
      __pyx_t_1 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_theta), __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_k2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
      __pyx_t_4 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_theta), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_gen), __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_18 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_18 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6hapflk_9fastphase_genprG(__pyx_t_16, __pyx_t_17, __pyx_t_18)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_19 = __Pyx_PyInt_From_int(__pyx_v_k2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_20 = PyTuple_New(3); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_19);
      PyTuple_SET_ITEM(__pyx_t_20, 2, __pyx_t_19);
      __pyx_t_6 = 0;
      __pyx_t_4 = 0;
      __pyx_t_19 = 0;
      __pyx_t_19 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_mBeta), __pyx_t_20); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __pyx_t_20 = PyNumber_Multiply(__pyx_t_1, __pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __pyx_t_19 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_k2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_19);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_19);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
      __pyx_t_19 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_alpha), __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Multiply(__pyx_t_20, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_tSumk), __pyx_t_3, __pyx_t_1) < 0)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "fastphase/fastphase.pyx":727
 *         for k2 in range(nK):
 *             tSumk[nLoc-1,k1] += genprG(theta[nLoc-1,k1],theta[nLoc-1,k2],gen[nLoc-1])*mBeta[nLoc-1,k1,k2]*alpha[nLoc-1,k2]
 *         tDoubleSum[nLoc-1]+=tSumk[nLoc-1,k1]*alpha[nLoc-1,k1]             # <<<<<<<<<<<<<<
 *     ## recurrent calculations backward
 *     for m in range(nLoc-1,0,-1):
 */
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_tDoubleSum), __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_5);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_tSumk), __pyx_t_20); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_k1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_4);
    __pyx_t_20 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_alpha), __pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_19 = PyNumber_Multiply(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_tDoubleSum), __pyx_t_3, __pyx_t_4) < 0)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "fastphase/fastphase.pyx":729
 *         tDoubleSum[nLoc-1]+=tSumk[nLoc-1,k1]*alpha[nLoc-1,k1]
 *     ## recurrent calculations backward
 *     for m in range(nLoc-1,0,-1):             # <<<<<<<<<<<<<<
 *         ## these loops could be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):
 */
  __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (__pyx_t_2 = __pyx_t_11; __pyx_t_2 > 0; __pyx_t_2-=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":731
 *     for m in range(nLoc-1,0,-1):
 *         ## these loops could be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):             # <<<<<<<<<<<<<<
 *             for k2 in range(k1,nK):
 *                 temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])
 */
    __pyx_t_12 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_12 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 731, __pyx_L1_error)
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_13; __pyx_t_15+=1) {
      __pyx_v_k1 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":732
 *         ## these loops could be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):
 *             for k2 in range(k1,nK):             # <<<<<<<<<<<<<<
 *                 temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])
 *                 temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]
 */
      __pyx_t_14 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_14 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
      __pyx_t_21 = __pyx_t_14;
      for (__pyx_t_18 = __pyx_v_k1; __pyx_t_18 < __pyx_t_21; __pyx_t_18+=1) {
        __pyx_v_k2 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":733
 *         for k1 in range(nK):
 *             for k2 in range(k1,nK):
 *                 temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])             # <<<<<<<<<<<<<<
 *                 temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]
 *                 mBeta[m-1,k1,k2]=temp+probJ(m,2,rho[m,0])*tDoubleSum[m]
 */
        __pyx_t_22 = __pyx_v_m;
        __pyx_t_23 = 0;
        if (__pyx_t_22 < 0) __pyx_t_22 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_23 < 0) __pyx_t_23 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_24 = __pyx_v_m;
        __pyx_t_25 = __pyx_v_k1;
        if (__pyx_t_24 < 0) __pyx_t_24 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        __pyx_t_26 = __pyx_v_m;
        __pyx_t_27 = __pyx_v_k2;
        if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        __pyx_v_temp = ((0.5 * __pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_rho.diminfo[1].strides)))) * ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_25, __pyx_pybuffernd_tSumk.diminfo[1].strides)) + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_tSumk.diminfo[1].strides))));

        /* "fastphase/fastphase.pyx":734
 *             for k2 in range(k1,nK):
 *                 temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])
 *                 temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]             # <<<<<<<<<<<<<<
 *                 mBeta[m-1,k1,k2]=temp+probJ(m,2,rho[m,0])*tDoubleSum[m]
 *                 mBeta[m-1,k2,k1]=mBeta[m-1,k1,k2]
 */
        __pyx_t_28 = __pyx_v_m;
        __pyx_t_29 = 0;
        if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_30 = __pyx_v_m;
        __pyx_t_31 = __pyx_v_k1;
        if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_32 = __pyx_v_m;
        __pyx_t_33 = __pyx_v_k2;
        if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_34 = __pyx_v_m;
        if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_gen.diminfo[0].shape;
        __pyx_t_35 = __pyx_v_m;
        __pyx_t_36 = __pyx_v_k1;
        __pyx_t_37 = __pyx_v_k2;
        if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        __pyx_v_temp = (__pyx_v_temp + ((__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 0, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_rho.diminfo[1].strides))) * __pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_31, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_33, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_gen.diminfo[0].strides)))) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_36, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_37, __pyx_pybuffernd_mBeta.diminfo[2].strides))));

        /* "fastphase/fastphase.pyx":735
 *                 temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])
 *                 temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]
 *                 mBeta[m-1,k1,k2]=temp+probJ(m,2,rho[m,0])*tDoubleSum[m]             # <<<<<<<<<<<<<<
 *                 mBeta[m-1,k2,k1]=mBeta[m-1,k1,k2]
 *         ## marginal sum and real sum for loc m-1
 */
        __pyx_t_38 = __pyx_v_m;
        __pyx_t_39 = 0;
        if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_40 = __pyx_v_m;
        if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
        __pyx_t_41 = (__pyx_v_m - 1);
        __pyx_t_42 = __pyx_v_k1;
        __pyx_t_43 = __pyx_v_k2;
        if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_43, __pyx_pybuffernd_mBeta.diminfo[2].strides) = (__pyx_v_temp + (__pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 2, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_39, __pyx_pybuffernd_rho.diminfo[1].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_40, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides))));

        /* "fastphase/fastphase.pyx":736
 *                 temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]
 *                 mBeta[m-1,k1,k2]=temp+probJ(m,2,rho[m,0])*tDoubleSum[m]
 *                 mBeta[m-1,k2,k1]=mBeta[m-1,k1,k2]             # <<<<<<<<<<<<<<
 *         ## marginal sum and real sum for loc m-1
 *         for k1 in range(nK):
 */
        __pyx_t_44 = (__pyx_v_m - 1);
        __pyx_t_45 = __pyx_v_k1;
        __pyx_t_46 = __pyx_v_k2;
        if (__pyx_t_44 < 0) __pyx_t_44 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        __pyx_t_47 = (__pyx_v_m - 1);
        __pyx_t_48 = __pyx_v_k2;
        __pyx_t_49 = __pyx_v_k1;
        if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_48, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_49, __pyx_pybuffernd_mBeta.diminfo[2].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_44, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_45, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_46, __pyx_pybuffernd_mBeta.diminfo[2].strides));
      }
    }

    /* "fastphase/fastphase.pyx":738
 *                 mBeta[m-1,k2,k1]=mBeta[m-1,k1,k2]
 *         ## marginal sum and real sum for loc m-1
 *         for k1 in range(nK):             # <<<<<<<<<<<<<<
 *             for k2 in range(nK):
 *                 tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]
 */
    __pyx_t_12 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_12 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L1_error)
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_13; __pyx_t_15+=1) {
      __pyx_v_k1 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":739
 *         ## marginal sum and real sum for loc m-1
 *         for k1 in range(nK):
 *             for k2 in range(nK):             # <<<<<<<<<<<<<<
 *                 tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]
 *             tDoubleSum[m-1]+=tSumk[m-1,k1]*alpha[m-1,k1]
 */
      __pyx_t_14 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_14 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L1_error)
      __pyx_t_21 = __pyx_t_14;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_21; __pyx_t_18+=1) {
        __pyx_v_k2 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":740
 *         for k1 in range(nK):
 *             for k2 in range(nK):
 *                 tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]             # <<<<<<<<<<<<<<
 *             tDoubleSum[m-1]+=tSumk[m-1,k1]*alpha[m-1,k1]
 *         tScaleTemp=np.sum(mBeta[m-1])
 */
        __pyx_t_50 = (__pyx_v_m - 1);
        __pyx_t_51 = __pyx_v_k1;
        if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_52 = (__pyx_v_m - 1);
        __pyx_t_53 = __pyx_v_k2;
        if (__pyx_t_52 < 0) __pyx_t_52 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_54 = (__pyx_v_m - 1);
        if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_gen.diminfo[0].shape;
        __pyx_t_55 = (__pyx_v_m - 1);
        __pyx_t_56 = __pyx_v_k1;
        __pyx_t_57 = __pyx_v_k2;
        if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_57 < 0) __pyx_t_57 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        __pyx_t_58 = (__pyx_v_m - 1);
        __pyx_t_59 = __pyx_v_k2;
        if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_60 = (__pyx_v_m - 1);
        __pyx_t_61 = __pyx_v_k1;
        if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_60, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_61, __pyx_pybuffernd_tSumk.diminfo[1].strides) += ((__pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_51, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_53, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_gen.diminfo[0].strides))) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_55, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_56, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_57, __pyx_pybuffernd_mBeta.diminfo[2].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_59, __pyx_pybuffernd_alpha.diminfo[1].strides)));
      }

      /* "fastphase/fastphase.pyx":741
 *             for k2 in range(nK):
 *                 tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]
 *             tDoubleSum[m-1]+=tSumk[m-1,k1]*alpha[m-1,k1]             # <<<<<<<<<<<<<<
 *         tScaleTemp=np.sum(mBeta[m-1])
 *         ## tScaleTemp=0 a gerer <---
 */
      __pyx_t_62 = (__pyx_v_m - 1);
      __pyx_t_63 = __pyx_v_k1;
      if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
      if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
      __pyx_t_64 = (__pyx_v_m - 1);
      __pyx_t_65 = __pyx_v_k1;
      if (__pyx_t_64 < 0) __pyx_t_64 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_65 < 0) __pyx_t_65 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_66 = (__pyx_v_m - 1);
      if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) += ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_62, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_63, __pyx_pybuffernd_tSumk.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_65, __pyx_pybuffernd_alpha.diminfo[1].strides)));
    }

    /* "fastphase/fastphase.pyx":742
 *                 tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]
 *             tDoubleSum[m-1]+=tSumk[m-1,k1]*alpha[m-1,k1]
 *         tScaleTemp=np.sum(mBeta[m-1])             # <<<<<<<<<<<<<<
 *         ## tScaleTemp=0 a gerer <---
 *         if tScaleTemp<1e-20:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sum); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_12 = (__pyx_v_m - 1);
    __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_mBeta), __pyx_t_12, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_19);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_19, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tScaleTemp = __pyx_t_17;

    /* "fastphase/fastphase.pyx":744
 *         tScaleTemp=np.sum(mBeta[m-1])
 *         ## tScaleTemp=0 a gerer <---
 *         if tScaleTemp<1e-20:             # <<<<<<<<<<<<<<
 *             tScale=20
 *         else:
 */
    __pyx_t_67 = ((__pyx_v_tScaleTemp < 1e-20) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":745
 *         ## tScaleTemp=0 a gerer <---
 *         if tScaleTemp<1e-20:
 *             tScale=20             # <<<<<<<<<<<<<<
 *         else:
 *             tScale=int(-log(tScaleTemp)/log(10))
 */
      __pyx_v_tScale = 20;

      /* "fastphase/fastphase.pyx":744
 *         tScaleTemp=np.sum(mBeta[m-1])
 *         ## tScaleTemp=0 a gerer <---
 *         if tScaleTemp<1e-20:             # <<<<<<<<<<<<<<
 *             tScale=20
 *         else:
 */
      goto __pyx_L21;
    }

    /* "fastphase/fastphase.pyx":747
 *             tScale=20
 *         else:
 *             tScale=int(-log(tScaleTemp)/log(10))             # <<<<<<<<<<<<<<
 *         if tScale <=0:
 *             tScale=0
 */
    /*else*/ {
      __pyx_t_17 = (-log(__pyx_v_tScaleTemp));
      __pyx_t_16 = log(10.0);
      if (unlikely(__pyx_t_16 == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 747, __pyx_L1_error)
      }
      __pyx_v_tScale = ((int)(__pyx_t_17 / __pyx_t_16));
    }
    __pyx_L21:;

    /* "fastphase/fastphase.pyx":748
 *         else:
 *             tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale <=0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 */
    __pyx_t_67 = ((__pyx_v_tScale <= 0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":749
 *             tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale <=0:
 *             tScale=0             # <<<<<<<<<<<<<<
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale>0:
 */
      __pyx_v_tScale = 0;

      /* "fastphase/fastphase.pyx":748
 *         else:
 *             tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale <=0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 */
    }

    /* "fastphase/fastphase.pyx":750
 *         if tScale <=0:
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale             # <<<<<<<<<<<<<<
 *         if tScale>0:
 *             dummy=myPow10(tScale)
 */
    __pyx_t_68 = __pyx_v_m;
    if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    __pyx_t_69 = (__pyx_v_m - 1);
    if (__pyx_t_69 < 0) __pyx_t_69 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_69, __pyx_pybuffernd_betaScale.diminfo[0].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_betaScale.diminfo[0].strides)) + __pyx_v_tScale);

    /* "fastphase/fastphase.pyx":751
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale>0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tDoubleSum[m-1]*=dummy
 */
    __pyx_t_67 = ((__pyx_v_tScale > 0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":752
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale>0:
 *             dummy=myPow10(tScale)             # <<<<<<<<<<<<<<
 *             tDoubleSum[m-1]*=dummy
 *             for k1 in range(nK):
 */
      __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_v_tScale);

      /* "fastphase/fastphase.pyx":753
 *         if tScale>0:
 *             dummy=myPow10(tScale)
 *             tDoubleSum[m-1]*=dummy             # <<<<<<<<<<<<<<
 *             for k1 in range(nK):
 *                 tSumk[m-1,k1]*=dummy
 */
      __pyx_t_70 = (__pyx_v_m - 1);
      if (__pyx_t_70 < 0) __pyx_t_70 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_70, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) *= __pyx_v_dummy;

      /* "fastphase/fastphase.pyx":754
 *             dummy=myPow10(tScale)
 *             tDoubleSum[m-1]*=dummy
 *             for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                 tSumk[m-1,k1]*=dummy
 *                 mBeta[m-1,k1,k1]*=dummy
 */
      __pyx_t_12 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_12 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L1_error)
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_13; __pyx_t_15+=1) {
        __pyx_v_k1 = __pyx_t_15;

        /* "fastphase/fastphase.pyx":755
 *             tDoubleSum[m-1]*=dummy
 *             for k1 in range(nK):
 *                 tSumk[m-1,k1]*=dummy             # <<<<<<<<<<<<<<
 *                 mBeta[m-1,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):
 */
        __pyx_t_71 = (__pyx_v_m - 1);
        __pyx_t_72 = __pyx_v_k1;
        if (__pyx_t_71 < 0) __pyx_t_71 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_72 < 0) __pyx_t_72 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_72, __pyx_pybuffernd_tSumk.diminfo[1].strides) *= __pyx_v_dummy;

        /* "fastphase/fastphase.pyx":756
 *             for k1 in range(nK):
 *                 tSumk[m-1,k1]*=dummy
 *                 mBeta[m-1,k1,k1]*=dummy             # <<<<<<<<<<<<<<
 *                 for k2 in range(k1+1,nK):
 *                     mBeta[m-1,k1,k2]*=dummy
 */
        __pyx_t_73 = (__pyx_v_m - 1);
        __pyx_t_74 = __pyx_v_k1;
        __pyx_t_75 = __pyx_v_k1;
        if (__pyx_t_73 < 0) __pyx_t_73 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_74 < 0) __pyx_t_74 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_75 < 0) __pyx_t_75 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_73, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_74, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_75, __pyx_pybuffernd_mBeta.diminfo[2].strides) *= __pyx_v_dummy;

        /* "fastphase/fastphase.pyx":757
 *                 tSumk[m-1,k1]*=dummy
 *                 mBeta[m-1,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):             # <<<<<<<<<<<<<<
 *                     mBeta[m-1,k1,k2]*=dummy
 *                     mBeta[m-1,k2,k1]*=dummy
 */
        __pyx_t_14 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_14 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L1_error)
        __pyx_t_21 = __pyx_t_14;
        for (__pyx_t_18 = (__pyx_v_k1 + 1); __pyx_t_18 < __pyx_t_21; __pyx_t_18+=1) {
          __pyx_v_k2 = __pyx_t_18;

          /* "fastphase/fastphase.pyx":758
 *                 mBeta[m-1,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):
 *                     mBeta[m-1,k1,k2]*=dummy             # <<<<<<<<<<<<<<
 *                     mBeta[m-1,k2,k1]*=dummy
 *     ##
 */
          __pyx_t_76 = (__pyx_v_m - 1);
          __pyx_t_77 = __pyx_v_k1;
          __pyx_t_78 = __pyx_v_k2;
          if (__pyx_t_76 < 0) __pyx_t_76 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
          if (__pyx_t_77 < 0) __pyx_t_77 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
          if (__pyx_t_78 < 0) __pyx_t_78 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_76, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_77, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_78, __pyx_pybuffernd_mBeta.diminfo[2].strides) *= __pyx_v_dummy;

          /* "fastphase/fastphase.pyx":759
 *                 for k2 in range(k1+1,nK):
 *                     mBeta[m-1,k1,k2]*=dummy
 *                     mBeta[m-1,k2,k1]*=dummy             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute forward probabilities
 */
          __pyx_t_79 = (__pyx_v_m - 1);
          __pyx_t_80 = __pyx_v_k2;
          __pyx_t_81 = __pyx_v_k1;
          if (__pyx_t_79 < 0) __pyx_t_79 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
          if (__pyx_t_80 < 0) __pyx_t_80 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
          if (__pyx_t_81 < 0) __pyx_t_81 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_79, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_80, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_81, __pyx_pybuffernd_mBeta.diminfo[2].strides) *= __pyx_v_dummy;
        }
      }

      /* "fastphase/fastphase.pyx":751
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale>0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tDoubleSum[m-1]*=dummy
 */
    }
  }

  /* "fastphase/fastphase.pyx":763
 *     ## compute forward probabilities
 *     ##
 *     cdef np.ndarray[np.float64_t,ndim=3] mPhi=np.zeros((nLoc,nK,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
 *     ## at locus 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_nK);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_nK);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_t_82 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer, (PyObject*)__pyx_t_82, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_mPhi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 763, __pyx_L1_error)
    } else {__pyx_pybuffernd_mPhi.diminfo[0].strides = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mPhi.diminfo[0].shape = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mPhi.diminfo[1].strides = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mPhi.diminfo[1].shape = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mPhi.diminfo[2].strides = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mPhi.diminfo[2].shape = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_82 = 0;
  __pyx_v_mPhi = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":764
 *     ##
 *     cdef np.ndarray[np.float64_t,ndim=3] mPhi=np.zeros((nLoc,nK,nK),dtype=np.float64)
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)             # <<<<<<<<<<<<<<
 *     ## at locus 0
 *     for k1 in range(nK):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nLoc);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_np); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 764, __pyx_L1_error)
  __pyx_t_83 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer, (PyObject*)__pyx_t_83, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_phiScale = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 764, __pyx_L1_error)
    } else {__pyx_pybuffernd_phiScale.diminfo[0].strides = __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phiScale.diminfo[0].shape = __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_83 = 0;
  __pyx_v_phiScale = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":766
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
 *     ## at locus 0
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         for k2 in range(k1,nK):
 *             mPhi[0,k1,k2]=alpha[0,k1]*alpha[0,k2]*genprG(theta[0,k1],theta[0,k2],gen[0])
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_k1 = __pyx_t_2;

    /* "fastphase/fastphase.pyx":767
 *     ## at locus 0
 *     for k1 in range(nK):
 *         for k2 in range(k1,nK):             # <<<<<<<<<<<<<<
 *             mPhi[0,k1,k2]=alpha[0,k1]*alpha[0,k2]*genprG(theta[0,k1],theta[0,k2],gen[0])
 *             mPhi[0,k2,k1]=mPhi[0,k1,k2]
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = __pyx_v_k1; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k2 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":768
 *     for k1 in range(nK):
 *         for k2 in range(k1,nK):
 *             mPhi[0,k1,k2]=alpha[0,k1]*alpha[0,k2]*genprG(theta[0,k1],theta[0,k2],gen[0])             # <<<<<<<<<<<<<<
 *             mPhi[0,k2,k1]=mPhi[0,k1,k2]
 *     ## calc the marginal sum at locus 0 (appx A)
 */
      __pyx_t_84 = 0;
      __pyx_t_85 = __pyx_v_k1;
      if (__pyx_t_84 < 0) __pyx_t_84 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_85 < 0) __pyx_t_85 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_86 = 0;
      __pyx_t_87 = __pyx_v_k2;
      if (__pyx_t_86 < 0) __pyx_t_86 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_87 < 0) __pyx_t_87 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_88 = 0;
      __pyx_t_89 = __pyx_v_k1;
      if (__pyx_t_88 < 0) __pyx_t_88 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_89 < 0) __pyx_t_89 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_90 = 0;
      __pyx_t_91 = __pyx_v_k2;
      if (__pyx_t_90 < 0) __pyx_t_90 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_91 < 0) __pyx_t_91 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_92 = 0;
      if (__pyx_t_92 < 0) __pyx_t_92 += __pyx_pybuffernd_gen.diminfo[0].shape;
      __pyx_t_93 = 0;
      __pyx_t_94 = __pyx_v_k1;
      __pyx_t_95 = __pyx_v_k2;
      if (__pyx_t_93 < 0) __pyx_t_93 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_94 < 0) __pyx_t_94 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      if (__pyx_t_95 < 0) __pyx_t_95 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_94, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_95, __pyx_pybuffernd_mPhi.diminfo[2].strides) = (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_84, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_85, __pyx_pybuffernd_alpha.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_86, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_87, __pyx_pybuffernd_alpha.diminfo[1].strides))) * __pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_89, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_90, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_91, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_92, __pyx_pybuffernd_gen.diminfo[0].strides))));

      /* "fastphase/fastphase.pyx":769
 *         for k2 in range(k1,nK):
 *             mPhi[0,k1,k2]=alpha[0,k1]*alpha[0,k2]*genprG(theta[0,k1],theta[0,k2],gen[0])
 *             mPhi[0,k2,k1]=mPhi[0,k1,k2]             # <<<<<<<<<<<<<<
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tDoubleSum[0]=0
 */
      __pyx_t_96 = 0;
      __pyx_t_97 = __pyx_v_k1;
      __pyx_t_98 = __pyx_v_k2;
      if (__pyx_t_96 < 0) __pyx_t_96 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_97 < 0) __pyx_t_97 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      if (__pyx_t_98 < 0) __pyx_t_98 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
      __pyx_t_99 = 0;
      __pyx_t_100 = __pyx_v_k2;
      __pyx_t_101 = __pyx_v_k1;
      if (__pyx_t_99 < 0) __pyx_t_99 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_100 < 0) __pyx_t_100 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      if (__pyx_t_101 < 0) __pyx_t_101 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
      *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_99, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_100, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_101, __pyx_pybuffernd_mPhi.diminfo[2].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_96, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_97, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_98, __pyx_pybuffernd_mPhi.diminfo[2].strides));
    }
  }

  /* "fastphase/fastphase.pyx":771
 *             mPhi[0,k2,k1]=mPhi[0,k1,k2]
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tDoubleSum[0]=0             # <<<<<<<<<<<<<<
 *     for k1 in range(nK):
 *         tSumk[0,k1]=0
 */
  __pyx_t_102 = 0;
  if (__pyx_t_102 < 0) __pyx_t_102 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_102, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) = 0.0;

  /* "fastphase/fastphase.pyx":772
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tDoubleSum[0]=0
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         tSumk[0,k1]=0
 *         for k2 in range(nK):
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 772, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_k1 = __pyx_t_2;

    /* "fastphase/fastphase.pyx":773
 *     tDoubleSum[0]=0
 *     for k1 in range(nK):
 *         tSumk[0,k1]=0             # <<<<<<<<<<<<<<
 *         for k2 in range(nK):
 *             tSumk[0,k1]+=mPhi[0,k1,k2]
 */
    __pyx_t_103 = 0;
    __pyx_t_104 = __pyx_v_k1;
    if (__pyx_t_103 < 0) __pyx_t_103 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
    if (__pyx_t_104 < 0) __pyx_t_104 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_103, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_104, __pyx_pybuffernd_tSumk.diminfo[1].strides) = 0.0;

    /* "fastphase/fastphase.pyx":774
 *     for k1 in range(nK):
 *         tSumk[0,k1]=0
 *         for k2 in range(nK):             # <<<<<<<<<<<<<<
 *             tSumk[0,k1]+=mPhi[0,k1,k2]
 *         tDoubleSum[0]+=tSumk[0,k1]
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k2 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":775
 *         tSumk[0,k1]=0
 *         for k2 in range(nK):
 *             tSumk[0,k1]+=mPhi[0,k1,k2]             # <<<<<<<<<<<<<<
 *         tDoubleSum[0]+=tSumk[0,k1]
 *     tScale=0
 */
      __pyx_t_105 = 0;
      __pyx_t_106 = __pyx_v_k1;
      __pyx_t_107 = __pyx_v_k2;
      if (__pyx_t_105 < 0) __pyx_t_105 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_106 < 0) __pyx_t_106 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      if (__pyx_t_107 < 0) __pyx_t_107 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
      __pyx_t_108 = 0;
      __pyx_t_109 = __pyx_v_k1;
      if (__pyx_t_108 < 0) __pyx_t_108 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
      if (__pyx_t_109 < 0) __pyx_t_109 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_108, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_109, __pyx_pybuffernd_tSumk.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_105, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_106, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_107, __pyx_pybuffernd_mPhi.diminfo[2].strides));
    }

    /* "fastphase/fastphase.pyx":776
 *         for k2 in range(nK):
 *             tSumk[0,k1]+=mPhi[0,k1,k2]
 *         tDoubleSum[0]+=tSumk[0,k1]             # <<<<<<<<<<<<<<
 *     tScale=0
 *     if tDoubleSum[0] != 0:
 */
    __pyx_t_110 = 0;
    __pyx_t_111 = __pyx_v_k1;
    if (__pyx_t_110 < 0) __pyx_t_110 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
    if (__pyx_t_111 < 0) __pyx_t_111 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
    __pyx_t_112 = 0;
    if (__pyx_t_112 < 0) __pyx_t_112 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_112, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) += (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_110, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_111, __pyx_pybuffernd_tSumk.diminfo[1].strides));
  }

  /* "fastphase/fastphase.pyx":777
 *             tSumk[0,k1]+=mPhi[0,k1,k2]
 *         tDoubleSum[0]+=tSumk[0,k1]
 *     tScale=0             # <<<<<<<<<<<<<<
 *     if tDoubleSum[0] != 0:
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 */
  __pyx_v_tScale = 0;

  /* "fastphase/fastphase.pyx":778
 *         tDoubleSum[0]+=tSumk[0,k1]
 *     tScale=0
 *     if tDoubleSum[0] != 0:             # <<<<<<<<<<<<<<
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 *         if tScale <0:
 */
  __pyx_t_113 = 0;
  if (__pyx_t_113 < 0) __pyx_t_113 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
  __pyx_t_67 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_113, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides)) != 0.0) != 0);
  if (__pyx_t_67) {

    /* "fastphase/fastphase.pyx":779
 *     tScale=0
 *     if tDoubleSum[0] != 0:
 *         tScale=int(-log(tDoubleSum[0])/log(10))             # <<<<<<<<<<<<<<
 *         if tScale <0:
 *             tScale=0
 */
    __pyx_t_114 = 0;
    if (__pyx_t_114 < 0) __pyx_t_114 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
    __pyx_t_16 = (-log((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_114, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides))));
    __pyx_t_17 = log(10.0);
    if (unlikely(__pyx_t_17 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 779, __pyx_L1_error)
    }
    __pyx_v_tScale = ((int)(__pyx_t_16 / __pyx_t_17));

    /* "fastphase/fastphase.pyx":780
 *     if tDoubleSum[0] != 0:
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 *         if tScale <0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         phiScale[0]=tScale
 */
    __pyx_t_67 = ((__pyx_v_tScale < 0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":781
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 *         if tScale <0:
 *             tScale=0             # <<<<<<<<<<<<<<
 *         phiScale[0]=tScale
 *         if tScale>0:
 */
      __pyx_v_tScale = 0;

      /* "fastphase/fastphase.pyx":780
 *     if tDoubleSum[0] != 0:
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 *         if tScale <0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         phiScale[0]=tScale
 */
    }

    /* "fastphase/fastphase.pyx":782
 *         if tScale <0:
 *             tScale=0
 *         phiScale[0]=tScale             # <<<<<<<<<<<<<<
 *         if tScale>0:
 *             dummy=myPow10(tScale)
 */
    __pyx_t_115 = 0;
    if (__pyx_t_115 < 0) __pyx_t_115 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_115, __pyx_pybuffernd_phiScale.diminfo[0].strides) = __pyx_v_tScale;

    /* "fastphase/fastphase.pyx":783
 *             tScale=0
 *         phiScale[0]=tScale
 *         if tScale>0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             for k1 in range(nK):
 */
    __pyx_t_67 = ((__pyx_v_tScale > 0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":784
 *         phiScale[0]=tScale
 *         if tScale>0:
 *             dummy=myPow10(tScale)             # <<<<<<<<<<<<<<
 *             for k1 in range(nK):
 *                 tSumk[0,k1]*=dummy
 */
      __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_v_tScale);

      /* "fastphase/fastphase.pyx":785
 *         if tScale>0:
 *             dummy=myPow10(tScale)
 *             for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                 tSumk[0,k1]*=dummy
 *                 mPhi[0,k1,k1]*=dummy
 */
      __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 785, __pyx_L1_error)
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
        __pyx_v_k1 = __pyx_t_2;

        /* "fastphase/fastphase.pyx":786
 *             dummy=myPow10(tScale)
 *             for k1 in range(nK):
 *                 tSumk[0,k1]*=dummy             # <<<<<<<<<<<<<<
 *                 mPhi[0,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):
 */
        __pyx_t_116 = 0;
        __pyx_t_117 = __pyx_v_k1;
        if (__pyx_t_116 < 0) __pyx_t_116 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_117 < 0) __pyx_t_117 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_116, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_117, __pyx_pybuffernd_tSumk.diminfo[1].strides) *= __pyx_v_dummy;

        /* "fastphase/fastphase.pyx":787
 *             for k1 in range(nK):
 *                 tSumk[0,k1]*=dummy
 *                 mPhi[0,k1,k1]*=dummy             # <<<<<<<<<<<<<<
 *                 for k2 in range(k1+1,nK):
 *                     mPhi[0,k1,k2]*=dummy
 */
        __pyx_t_118 = 0;
        __pyx_t_119 = __pyx_v_k1;
        __pyx_t_120 = __pyx_v_k1;
        if (__pyx_t_118 < 0) __pyx_t_118 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_119 < 0) __pyx_t_119 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_120 < 0) __pyx_t_120 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_118, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_119, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_120, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;

        /* "fastphase/fastphase.pyx":788
 *                 tSumk[0,k1]*=dummy
 *                 mPhi[0,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):             # <<<<<<<<<<<<<<
 *                     mPhi[0,k1,k2]*=dummy
 *                     mPhi[0,k2,k1]*=dummy
 */
        __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L1_error)
        __pyx_t_14 = __pyx_t_13;
        for (__pyx_t_15 = (__pyx_v_k1 + 1); __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
          __pyx_v_k2 = __pyx_t_15;

          /* "fastphase/fastphase.pyx":789
 *                 mPhi[0,k1,k1]*=dummy
 *                 for k2 in range(k1+1,nK):
 *                     mPhi[0,k1,k2]*=dummy             # <<<<<<<<<<<<<<
 *                     mPhi[0,k2,k1]*=dummy
 *     # do the reccurence
 */
          __pyx_t_121 = 0;
          __pyx_t_122 = __pyx_v_k1;
          __pyx_t_123 = __pyx_v_k2;
          if (__pyx_t_121 < 0) __pyx_t_121 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
          if (__pyx_t_122 < 0) __pyx_t_122 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
          if (__pyx_t_123 < 0) __pyx_t_123 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_121, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_122, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_123, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;

          /* "fastphase/fastphase.pyx":790
 *                 for k2 in range(k1+1,nK):
 *                     mPhi[0,k1,k2]*=dummy
 *                     mPhi[0,k2,k1]*=dummy             # <<<<<<<<<<<<<<
 *     # do the reccurence
 *     for m in range(nLoc-1):
 */
          __pyx_t_124 = 0;
          __pyx_t_125 = __pyx_v_k2;
          __pyx_t_126 = __pyx_v_k1;
          if (__pyx_t_124 < 0) __pyx_t_124 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
          if (__pyx_t_125 < 0) __pyx_t_125 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
          if (__pyx_t_126 < 0) __pyx_t_126 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_124, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_125, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_126, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;
        }
      }

      /* "fastphase/fastphase.pyx":783
 *             tScale=0
 *         phiScale[0]=tScale
 *         if tScale>0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             for k1 in range(nK):
 */
    }

    /* "fastphase/fastphase.pyx":778
 *         tDoubleSum[0]+=tSumk[0,k1]
 *     tScale=0
 *     if tDoubleSum[0] != 0:             # <<<<<<<<<<<<<<
 *         tScale=int(-log(tDoubleSum[0])/log(10))
 *         if tScale <0:
 */
  }

  /* "fastphase/fastphase.pyx":792
 *                     mPhi[0,k2,k1]*=dummy
 *     # do the reccurence
 *     for m in range(nLoc-1):             # <<<<<<<<<<<<<<
 *         ## This loop can be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):
 */
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":794
 *     for m in range(nLoc-1):
 *         ## This loop can be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):             # <<<<<<<<<<<<<<
 *             for k2 in range(k1,nK):
 *                 temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k1 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":795
 *         ## This loop can be parallelized across cluster pairs #CUDA
 *         for k1 in range(nK):
 *             for k2 in range(k1,nK):             # <<<<<<<<<<<<<<
 *                 temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]
 *                 temp*=0.5*probJ(m+1,1,rho[m+1,0])
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 795, __pyx_L1_error)
      __pyx_t_127 = __pyx_t_21;
      for (__pyx_t_18 = __pyx_v_k1; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
        __pyx_v_k2 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":796
 *         for k1 in range(nK):
 *             for k2 in range(k1,nK):
 *                 temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]             # <<<<<<<<<<<<<<
 *                 temp*=0.5*probJ(m+1,1,rho[m+1,0])
 *                 temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]
 */
        __pyx_t_128 = (__pyx_v_m + 1);
        __pyx_t_129 = __pyx_v_k1;
        if (__pyx_t_128 < 0) __pyx_t_128 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_129 < 0) __pyx_t_129 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_130 = __pyx_v_m;
        __pyx_t_131 = __pyx_v_k2;
        if (__pyx_t_130 < 0) __pyx_t_130 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_131 < 0) __pyx_t_131 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        __pyx_t_132 = (__pyx_v_m + 1);
        __pyx_t_133 = __pyx_v_k2;
        if (__pyx_t_132 < 0) __pyx_t_132 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_133 < 0) __pyx_t_133 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_134 = __pyx_v_m;
        __pyx_t_135 = __pyx_v_k1;
        if (__pyx_t_134 < 0) __pyx_t_134 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_135 < 0) __pyx_t_135 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        __pyx_v_temp = (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_128, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_129, __pyx_pybuffernd_alpha.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_130, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_131, __pyx_pybuffernd_tSumk.diminfo[1].strides))) + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_132, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_133, __pyx_pybuffernd_alpha.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_134, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_135, __pyx_pybuffernd_tSumk.diminfo[1].strides))));

        /* "fastphase/fastphase.pyx":797
 *             for k2 in range(k1,nK):
 *                 temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]
 *                 temp*=0.5*probJ(m+1,1,rho[m+1,0])             # <<<<<<<<<<<<<<
 *                 temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]
 *                 temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]
 */
        __pyx_t_136 = (__pyx_v_m + 1);
        __pyx_t_137 = 0;
        if (__pyx_t_136 < 0) __pyx_t_136 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_137 < 0) __pyx_t_137 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_v_temp = (__pyx_v_temp * (0.5 * __pyx_f_6hapflk_9fastphase_probJ((__pyx_v_m + 1), 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_136, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_137, __pyx_pybuffernd_rho.diminfo[1].strides)))));

        /* "fastphase/fastphase.pyx":798
 *                 temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]
 *                 temp*=0.5*probJ(m+1,1,rho[m+1,0])
 *                 temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]             # <<<<<<<<<<<<<<
 *                 temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]
 *                 mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])
 */
        __pyx_t_138 = (__pyx_v_m + 1);
        __pyx_t_139 = 0;
        if (__pyx_t_138 < 0) __pyx_t_138 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_139 < 0) __pyx_t_139 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_140 = __pyx_v_m;
        __pyx_t_141 = __pyx_v_k1;
        __pyx_t_142 = __pyx_v_k2;
        if (__pyx_t_140 < 0) __pyx_t_140 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_141 < 0) __pyx_t_141 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_142 < 0) __pyx_t_142 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        __pyx_v_temp = (__pyx_v_temp + (__pyx_f_6hapflk_9fastphase_probJ((__pyx_v_m + 1), 0, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_138, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_139, __pyx_pybuffernd_rho.diminfo[1].strides))) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_140, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_141, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_142, __pyx_pybuffernd_mPhi.diminfo[2].strides))));

        /* "fastphase/fastphase.pyx":799
 *                 temp*=0.5*probJ(m+1,1,rho[m+1,0])
 *                 temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]
 *                 temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]             # <<<<<<<<<<<<<<
 *                 mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])
 *                 mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]
 */
        __pyx_t_143 = (__pyx_v_m + 1);
        __pyx_t_144 = 0;
        if (__pyx_t_143 < 0) __pyx_t_143 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_144 < 0) __pyx_t_144 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_145 = (__pyx_v_m + 1);
        __pyx_t_146 = __pyx_v_k1;
        if (__pyx_t_145 < 0) __pyx_t_145 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_146 < 0) __pyx_t_146 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_147 = (__pyx_v_m + 1);
        __pyx_t_148 = __pyx_v_k2;
        if (__pyx_t_147 < 0) __pyx_t_147 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_148 < 0) __pyx_t_148 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_149 = __pyx_v_m;
        if (__pyx_t_149 < 0) __pyx_t_149 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
        __pyx_v_temp = (__pyx_v_temp + (((__pyx_f_6hapflk_9fastphase_probJ((__pyx_v_m + 1), 2, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_143, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_144, __pyx_pybuffernd_rho.diminfo[1].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_145, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_146, __pyx_pybuffernd_alpha.diminfo[1].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_147, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_148, __pyx_pybuffernd_alpha.diminfo[1].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_149, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides))));

        /* "fastphase/fastphase.pyx":800
 *                 temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]
 *                 temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]
 *                 mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])             # <<<<<<<<<<<<<<
 *                 mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]
 *         tDoubleSum[m+1]=0
 */
        __pyx_t_150 = (__pyx_v_m + 1);
        __pyx_t_151 = __pyx_v_k1;
        if (__pyx_t_150 < 0) __pyx_t_150 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_151 < 0) __pyx_t_151 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_152 = (__pyx_v_m + 1);
        __pyx_t_153 = __pyx_v_k2;
        if (__pyx_t_152 < 0) __pyx_t_152 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_153 < 0) __pyx_t_153 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_154 = (__pyx_v_m + 1);
        if (__pyx_t_154 < 0) __pyx_t_154 += __pyx_pybuffernd_gen.diminfo[0].shape;
        __pyx_t_155 = (__pyx_v_m + 1);
        __pyx_t_156 = __pyx_v_k1;
        __pyx_t_157 = __pyx_v_k2;
        if (__pyx_t_155 < 0) __pyx_t_155 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_156 < 0) __pyx_t_156 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_157 < 0) __pyx_t_157 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_155, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_156, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_157, __pyx_pybuffernd_mPhi.diminfo[2].strides) = (__pyx_v_temp * __pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_150, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_151, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_152, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_153, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_154, __pyx_pybuffernd_gen.diminfo[0].strides))));

        /* "fastphase/fastphase.pyx":801
 *                 temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]
 *                 mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])
 *                 mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]             # <<<<<<<<<<<<<<
 *         tDoubleSum[m+1]=0
 *         for k1 in range(nK):
 */
        __pyx_t_158 = (__pyx_v_m + 1);
        __pyx_t_159 = __pyx_v_k1;
        __pyx_t_160 = __pyx_v_k2;
        if (__pyx_t_158 < 0) __pyx_t_158 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_159 < 0) __pyx_t_159 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_160 < 0) __pyx_t_160 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        __pyx_t_161 = (__pyx_v_m + 1);
        __pyx_t_162 = __pyx_v_k2;
        __pyx_t_163 = __pyx_v_k1;
        if (__pyx_t_161 < 0) __pyx_t_161 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_162 < 0) __pyx_t_162 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_163 < 0) __pyx_t_163 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_161, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_162, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_163, __pyx_pybuffernd_mPhi.diminfo[2].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_158, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_159, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_160, __pyx_pybuffernd_mPhi.diminfo[2].strides));
      }
    }

    /* "fastphase/fastphase.pyx":802
 *                 mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])
 *                 mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]
 *         tDoubleSum[m+1]=0             # <<<<<<<<<<<<<<
 *         for k1 in range(nK):
 *             tSumk[m+1,k1]=0
 */
    __pyx_t_164 = (__pyx_v_m + 1);
    if (__pyx_t_164 < 0) __pyx_t_164 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_164, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) = 0.0;

    /* "fastphase/fastphase.pyx":803
 *                 mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]
 *         tDoubleSum[m+1]=0
 *         for k1 in range(nK):             # <<<<<<<<<<<<<<
 *             tSumk[m+1,k1]=0
 *             for k2 in range(nK):
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 803, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k1 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":804
 *         tDoubleSum[m+1]=0
 *         for k1 in range(nK):
 *             tSumk[m+1,k1]=0             # <<<<<<<<<<<<<<
 *             for k2 in range(nK):
 *                 tSumk[m+1,k1]+=mPhi[m+1,k1,k2]
 */
      __pyx_t_165 = (__pyx_v_m + 1);
      __pyx_t_166 = __pyx_v_k1;
      if (__pyx_t_165 < 0) __pyx_t_165 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
      if (__pyx_t_166 < 0) __pyx_t_166 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_165, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_166, __pyx_pybuffernd_tSumk.diminfo[1].strides) = 0.0;

      /* "fastphase/fastphase.pyx":805
 *         for k1 in range(nK):
 *             tSumk[m+1,k1]=0
 *             for k2 in range(nK):             # <<<<<<<<<<<<<<
 *                 tSumk[m+1,k1]+=mPhi[m+1,k1,k2]
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 805, __pyx_L1_error)
      __pyx_t_127 = __pyx_t_21;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
        __pyx_v_k2 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":806
 *             tSumk[m+1,k1]=0
 *             for k2 in range(nK):
 *                 tSumk[m+1,k1]+=mPhi[m+1,k1,k2]             # <<<<<<<<<<<<<<
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]
 *         tScale=0
 */
        __pyx_t_167 = (__pyx_v_m + 1);
        __pyx_t_168 = __pyx_v_k1;
        __pyx_t_169 = __pyx_v_k2;
        if (__pyx_t_167 < 0) __pyx_t_167 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_168 < 0) __pyx_t_168 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        if (__pyx_t_169 < 0) __pyx_t_169 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
        __pyx_t_170 = (__pyx_v_m + 1);
        __pyx_t_171 = __pyx_v_k1;
        if (__pyx_t_170 < 0) __pyx_t_170 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_171 < 0) __pyx_t_171 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_170, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_171, __pyx_pybuffernd_tSumk.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_167, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_168, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_169, __pyx_pybuffernd_mPhi.diminfo[2].strides));
      }

      /* "fastphase/fastphase.pyx":807
 *             for k2 in range(nK):
 *                 tSumk[m+1,k1]+=mPhi[m+1,k1,k2]
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]             # <<<<<<<<<<<<<<
 *         tScale=0
 *         if tDoubleSum[m+1]<=0:
 */
      __pyx_t_172 = (__pyx_v_m + 1);
      __pyx_t_173 = __pyx_v_k1;
      if (__pyx_t_172 < 0) __pyx_t_172 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
      if (__pyx_t_173 < 0) __pyx_t_173 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
      __pyx_t_174 = (__pyx_v_m + 1);
      if (__pyx_t_174 < 0) __pyx_t_174 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_174, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) += (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_172, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_173, __pyx_pybuffernd_tSumk.diminfo[1].strides));
    }

    /* "fastphase/fastphase.pyx":808
 *                 tSumk[m+1,k1]+=mPhi[m+1,k1,k2]
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]
 *         tScale=0             # <<<<<<<<<<<<<<
 *         if tDoubleSum[m+1]<=0:
 *             phiScale[m+1]=phiScale[m]
 */
    __pyx_v_tScale = 0;

    /* "fastphase/fastphase.pyx":809
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]
 *         tScale=0
 *         if tDoubleSum[m+1]<=0:             # <<<<<<<<<<<<<<
 *             phiScale[m+1]=phiScale[m]
 *         else:
 */
    __pyx_t_175 = (__pyx_v_m + 1);
    if (__pyx_t_175 < 0) __pyx_t_175 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
    __pyx_t_67 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_175, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides)) <= 0.0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":810
 *         tScale=0
 *         if tDoubleSum[m+1]<=0:
 *             phiScale[m+1]=phiScale[m]             # <<<<<<<<<<<<<<
 *         else:
 *             tScale=int(-log(tDoubleSum[m+1])/log(10))
 */
      __pyx_t_176 = __pyx_v_m;
      if (__pyx_t_176 < 0) __pyx_t_176 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      __pyx_t_177 = (__pyx_v_m + 1);
      if (__pyx_t_177 < 0) __pyx_t_177 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_177, __pyx_pybuffernd_phiScale.diminfo[0].strides) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_176, __pyx_pybuffernd_phiScale.diminfo[0].strides));

      /* "fastphase/fastphase.pyx":809
 *             tDoubleSum[m+1]+=tSumk[m+1,k1]
 *         tScale=0
 *         if tDoubleSum[m+1]<=0:             # <<<<<<<<<<<<<<
 *             phiScale[m+1]=phiScale[m]
 *         else:
 */
      goto __pyx_L53;
    }

    /* "fastphase/fastphase.pyx":812
 *             phiScale[m+1]=phiScale[m]
 *         else:
 *             tScale=int(-log(tDoubleSum[m+1])/log(10))             # <<<<<<<<<<<<<<
 *             if tScale<0:
 *                 tScale=0
 */
    /*else*/ {
      __pyx_t_178 = (__pyx_v_m + 1);
      if (__pyx_t_178 < 0) __pyx_t_178 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
      __pyx_t_17 = (-log((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_178, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides))));
      __pyx_t_16 = log(10.0);
      if (unlikely(__pyx_t_16 == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 812, __pyx_L1_error)
      }
      __pyx_v_tScale = ((int)(__pyx_t_17 / __pyx_t_16));

      /* "fastphase/fastphase.pyx":813
 *         else:
 *             tScale=int(-log(tDoubleSum[m+1])/log(10))
 *             if tScale<0:             # <<<<<<<<<<<<<<
 *                 tScale=0
 *             phiScale[m+1]=phiScale[m]+tScale
 */
      __pyx_t_67 = ((__pyx_v_tScale < 0) != 0);
      if (__pyx_t_67) {

        /* "fastphase/fastphase.pyx":814
 *             tScale=int(-log(tDoubleSum[m+1])/log(10))
 *             if tScale<0:
 *                 tScale=0             # <<<<<<<<<<<<<<
 *             phiScale[m+1]=phiScale[m]+tScale
 *             if tScale>0:
 */
        __pyx_v_tScale = 0;

        /* "fastphase/fastphase.pyx":813
 *         else:
 *             tScale=int(-log(tDoubleSum[m+1])/log(10))
 *             if tScale<0:             # <<<<<<<<<<<<<<
 *                 tScale=0
 *             phiScale[m+1]=phiScale[m]+tScale
 */
      }

      /* "fastphase/fastphase.pyx":815
 *             if tScale<0:
 *                 tScale=0
 *             phiScale[m+1]=phiScale[m]+tScale             # <<<<<<<<<<<<<<
 *             if tScale>0:
 *                 dummy=myPow10(tScale)
 */
      __pyx_t_179 = __pyx_v_m;
      if (__pyx_t_179 < 0) __pyx_t_179 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      __pyx_t_180 = (__pyx_v_m + 1);
      if (__pyx_t_180 < 0) __pyx_t_180 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_180, __pyx_pybuffernd_phiScale.diminfo[0].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_179, __pyx_pybuffernd_phiScale.diminfo[0].strides)) + __pyx_v_tScale);

      /* "fastphase/fastphase.pyx":816
 *                 tScale=0
 *             phiScale[m+1]=phiScale[m]+tScale
 *             if tScale>0:             # <<<<<<<<<<<<<<
 *                 dummy=myPow10(tScale)
 *                 tDoubleSum[m+1]*=dummy
 */
      __pyx_t_67 = ((__pyx_v_tScale > 0) != 0);
      if (__pyx_t_67) {

        /* "fastphase/fastphase.pyx":817
 *             phiScale[m+1]=phiScale[m]+tScale
 *             if tScale>0:
 *                 dummy=myPow10(tScale)             # <<<<<<<<<<<<<<
 *                 tDoubleSum[m+1]*=dummy
 *                 for k1 in range(nK):
 */
        __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_v_tScale);

        /* "fastphase/fastphase.pyx":818
 *             if tScale>0:
 *                 dummy=myPow10(tScale)
 *                 tDoubleSum[m+1]*=dummy             # <<<<<<<<<<<<<<
 *                 for k1 in range(nK):
 *                     tSumk[m+1,k1]*=dummy
 */
        __pyx_t_181 = (__pyx_v_m + 1);
        if (__pyx_t_181 < 0) __pyx_t_181 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_181, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides) *= __pyx_v_dummy;

        /* "fastphase/fastphase.pyx":819
 *                 dummy=myPow10(tScale)
 *                 tDoubleSum[m+1]*=dummy
 *                 for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                     tSumk[m+1,k1]*=dummy
 *                     mPhi[m+1,k1,k1]*=dummy
 */
        __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 819, __pyx_L1_error)
        __pyx_t_14 = __pyx_t_13;
        for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
          __pyx_v_k1 = __pyx_t_15;

          /* "fastphase/fastphase.pyx":820
 *                 tDoubleSum[m+1]*=dummy
 *                 for k1 in range(nK):
 *                     tSumk[m+1,k1]*=dummy             # <<<<<<<<<<<<<<
 *                     mPhi[m+1,k1,k1]*=dummy
 *                     for k2 in range(k1+1,nK):
 */
          __pyx_t_182 = (__pyx_v_m + 1);
          __pyx_t_183 = __pyx_v_k1;
          if (__pyx_t_182 < 0) __pyx_t_182 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
          if (__pyx_t_183 < 0) __pyx_t_183 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
          *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_182, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_183, __pyx_pybuffernd_tSumk.diminfo[1].strides) *= __pyx_v_dummy;

          /* "fastphase/fastphase.pyx":821
 *                 for k1 in range(nK):
 *                     tSumk[m+1,k1]*=dummy
 *                     mPhi[m+1,k1,k1]*=dummy             # <<<<<<<<<<<<<<
 *                     for k2 in range(k1+1,nK):
 *                         mPhi[m+1,k1,k2]*=dummy
 */
          __pyx_t_184 = (__pyx_v_m + 1);
          __pyx_t_185 = __pyx_v_k1;
          __pyx_t_186 = __pyx_v_k1;
          if (__pyx_t_184 < 0) __pyx_t_184 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
          if (__pyx_t_185 < 0) __pyx_t_185 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
          if (__pyx_t_186 < 0) __pyx_t_186 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_184, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_185, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_186, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;

          /* "fastphase/fastphase.pyx":822
 *                     tSumk[m+1,k1]*=dummy
 *                     mPhi[m+1,k1,k1]*=dummy
 *                     for k2 in range(k1+1,nK):             # <<<<<<<<<<<<<<
 *                         mPhi[m+1,k1,k2]*=dummy
 *                         mPhi[m+1,k2,k1]*=dummy
 */
          __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)
          __pyx_t_127 = __pyx_t_21;
          for (__pyx_t_18 = (__pyx_v_k1 + 1); __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
            __pyx_v_k2 = __pyx_t_18;

            /* "fastphase/fastphase.pyx":823
 *                     mPhi[m+1,k1,k1]*=dummy
 *                     for k2 in range(k1+1,nK):
 *                         mPhi[m+1,k1,k2]*=dummy             # <<<<<<<<<<<<<<
 *                         mPhi[m+1,k2,k1]*=dummy
 *     ## end mPhi
 */
            __pyx_t_187 = (__pyx_v_m + 1);
            __pyx_t_188 = __pyx_v_k1;
            __pyx_t_189 = __pyx_v_k2;
            if (__pyx_t_187 < 0) __pyx_t_187 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
            if (__pyx_t_188 < 0) __pyx_t_188 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
            if (__pyx_t_189 < 0) __pyx_t_189 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
            *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_187, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_188, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_189, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;

            /* "fastphase/fastphase.pyx":824
 *                     for k2 in range(k1+1,nK):
 *                         mPhi[m+1,k1,k2]*=dummy
 *                         mPhi[m+1,k2,k1]*=dummy             # <<<<<<<<<<<<<<
 *     ## end mPhi
 *     logLikelihood=log(tDoubleSum[nLoc-1])/log(10)-phiScale[nLoc-1]
 */
            __pyx_t_190 = (__pyx_v_m + 1);
            __pyx_t_191 = __pyx_v_k2;
            __pyx_t_192 = __pyx_v_k1;
            if (__pyx_t_190 < 0) __pyx_t_190 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
            if (__pyx_t_191 < 0) __pyx_t_191 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
            if (__pyx_t_192 < 0) __pyx_t_192 += __pyx_pybuffernd_mPhi.diminfo[2].shape;
            *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_190, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_191, __pyx_pybuffernd_mPhi.diminfo[1].strides, __pyx_t_192, __pyx_pybuffernd_mPhi.diminfo[2].strides) *= __pyx_v_dummy;
          }
        }

        /* "fastphase/fastphase.pyx":816
 *                 tScale=0
 *             phiScale[m+1]=phiScale[m]+tScale
 *             if tScale>0:             # <<<<<<<<<<<<<<
 *                 dummy=myPow10(tScale)
 *                 tDoubleSum[m+1]*=dummy
 */
      }
    }
    __pyx_L53:;
  }

  /* "fastphase/fastphase.pyx":826
 *                         mPhi[m+1,k2,k1]*=dummy
 *     ## end mPhi
 *     logLikelihood=log(tDoubleSum[nLoc-1])/log(10)-phiScale[nLoc-1]             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute Individual Contribution top,bottom,jmk
 */
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_tDoubleSum), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_17 = log(__pyx_t_16);
  __pyx_t_16 = log(10.0);
  if (unlikely(__pyx_t_16 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 826, __pyx_L1_error)
  }
  __pyx_t_4 = PyFloat_FromDouble((__pyx_t_17 / __pyx_t_16)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_phiScale), __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_16 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_16 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_logLikelihood = __pyx_t_16;

  /* "fastphase/fastphase.pyx":831
 *     ##
 *     # calc ProbZ
 *     cdef np.ndarray[np.float64_t,ndim=3] probZ=mPhi*mBeta             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 */
  __pyx_t_1 = PyNumber_Multiply(((PyObject *)__pyx_v_mPhi), ((PyObject *)__pyx_v_mBeta)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 831, __pyx_L1_error)
  __pyx_t_193 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer, (PyObject*)__pyx_t_193, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_probZ = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 831, __pyx_L1_error)
    } else {__pyx_pybuffernd_probZ.diminfo[0].strides = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_probZ.diminfo[0].shape = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_probZ.diminfo[1].strides = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_probZ.diminfo[1].shape = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_probZ.diminfo[2].strides = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_probZ.diminfo[2].shape = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_193 = 0;
  __pyx_v_probZ = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":832
 *     # calc ProbZ
 *     cdef np.ndarray[np.float64_t,ndim=3] probZ=mPhi*mBeta
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nK);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_19) < 0) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_19) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_19, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 832, __pyx_L1_error)
  __pyx_t_194 = ((PyArrayObject *)__pyx_t_19);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer, (PyObject*)__pyx_t_194, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_jmk = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 832, __pyx_L1_error)
    } else {__pyx_pybuffernd_jmk.diminfo[0].strides = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_jmk.diminfo[0].shape = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_jmk.diminfo[1].strides = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_jmk.diminfo[1].shape = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_194 = 0;
  __pyx_v_jmk = ((PyArrayObject *)__pyx_t_19);
  __pyx_t_19 = 0;

  /* "fastphase/fastphase.pyx":833
 *     cdef np.ndarray[np.float64_t,ndim=3] probZ=mPhi*mBeta
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 *     ## normalize
 */
  __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_np); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_v_nK);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_19);
  __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_19); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 833, __pyx_L1_error)
  __pyx_t_195 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_top.rcbuffer->pybuffer, (PyObject*)__pyx_t_195, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_top = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_top.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 833, __pyx_L1_error)
    } else {__pyx_pybuffernd_top.diminfo[0].strides = __pyx_pybuffernd_top.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_top.diminfo[0].shape = __pyx_pybuffernd_top.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_top.diminfo[1].strides = __pyx_pybuffernd_top.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_top.diminfo[1].shape = __pyx_pybuffernd_top.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_195 = 0;
  __pyx_v_top = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":834
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     ## normalize
 *     for m in range(nLoc):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_nLoc);
  __Pyx_GIVEREF(__pyx_v_nLoc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_nLoc);
  __Pyx_INCREF(__pyx_v_nK);
  __Pyx_GIVEREF(__pyx_v_nK);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_nK);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 834, __pyx_L1_error)
  __pyx_t_196 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bot.rcbuffer->pybuffer, (PyObject*)__pyx_t_196, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_bot = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 834, __pyx_L1_error)
    } else {__pyx_pybuffernd_bot.diminfo[0].strides = __pyx_pybuffernd_bot.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bot.diminfo[0].shape = __pyx_pybuffernd_bot.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_bot.diminfo[1].strides = __pyx_pybuffernd_bot.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_bot.diminfo[1].shape = __pyx_pybuffernd_bot.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_196 = 0;
  __pyx_v_bot = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":836
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 *     ## normalize
 *     for m in range(nLoc):             # <<<<<<<<<<<<<<
 *         normC=0
 *         for k1 in range(nK):
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nLoc); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":837
 *     ## normalize
 *     for m in range(nLoc):
 *         normC=0             # <<<<<<<<<<<<<<
 *         for k1 in range(nK):
 *             normC+=probZ[m,k1,k1]
 */
    __pyx_v_normC = 0.0;

    /* "fastphase/fastphase.pyx":838
 *     for m in range(nLoc):
 *         normC=0
 *         for k1 in range(nK):             # <<<<<<<<<<<<<<
 *             normC+=probZ[m,k1,k1]
 *             for k2 in range(k1+1,nK):
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k1 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":839
 *         normC=0
 *         for k1 in range(nK):
 *             normC+=probZ[m,k1,k1]             # <<<<<<<<<<<<<<
 *             for k2 in range(k1+1,nK):
 *                 normC+=2*probZ[m,k1,k2]
 */
      __pyx_t_197 = __pyx_v_m;
      __pyx_t_198 = __pyx_v_k1;
      __pyx_t_199 = __pyx_v_k1;
      if (__pyx_t_197 < 0) __pyx_t_197 += __pyx_pybuffernd_probZ.diminfo[0].shape;
      if (__pyx_t_198 < 0) __pyx_t_198 += __pyx_pybuffernd_probZ.diminfo[1].shape;
      if (__pyx_t_199 < 0) __pyx_t_199 += __pyx_pybuffernd_probZ.diminfo[2].shape;
      __pyx_v_normC = (__pyx_v_normC + (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_197, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_198, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_199, __pyx_pybuffernd_probZ.diminfo[2].strides)));

      /* "fastphase/fastphase.pyx":840
 *         for k1 in range(nK):
 *             normC+=probZ[m,k1,k1]
 *             for k2 in range(k1+1,nK):             # <<<<<<<<<<<<<<
 *                 normC+=2*probZ[m,k1,k2]
 *         probZ[m]/=normC
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L1_error)
      __pyx_t_127 = __pyx_t_21;
      for (__pyx_t_18 = (__pyx_v_k1 + 1); __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
        __pyx_v_k2 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":841
 *             normC+=probZ[m,k1,k1]
 *             for k2 in range(k1+1,nK):
 *                 normC+=2*probZ[m,k1,k2]             # <<<<<<<<<<<<<<
 *         probZ[m]/=normC
 *     if up2pz>0:
 */
        __pyx_t_200 = __pyx_v_m;
        __pyx_t_201 = __pyx_v_k1;
        __pyx_t_202 = __pyx_v_k2;
        if (__pyx_t_200 < 0) __pyx_t_200 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_201 < 0) __pyx_t_201 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        if (__pyx_t_202 < 0) __pyx_t_202 += __pyx_pybuffernd_probZ.diminfo[2].shape;
        __pyx_v_normC = (__pyx_v_normC + (2.0 * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_200, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_201, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_202, __pyx_pybuffernd_probZ.diminfo[2].strides))));
      }
    }

    /* "fastphase/fastphase.pyx":842
 *             for k2 in range(k1+1,nK):
 *                 normC+=2*probZ[m,k1,k2]
 *         probZ[m]/=normC             # <<<<<<<<<<<<<<
 *     if up2pz>0:
 *         return probZ
 */
    __pyx_t_15 = __pyx_v_m;
    __pyx_t_5 = __Pyx_GetItemInt(((PyObject *)__pyx_v_probZ), __pyx_t_15, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_normC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyNumber_InPlaceDivide(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_probZ), __pyx_t_15, __pyx_t_4, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "fastphase/fastphase.pyx":843
 *                 normC+=2*probZ[m,k1,k2]
 *         probZ[m]/=normC
 *     if up2pz>0:             # <<<<<<<<<<<<<<
 *         return probZ
 *     # calc jmk
 */
  __pyx_t_67 = ((__pyx_v_up2pz > 0) != 0);
  if (__pyx_t_67) {

    /* "fastphase/fastphase.pyx":844
 *         probZ[m]/=normC
 *     if up2pz>0:
 *         return probZ             # <<<<<<<<<<<<<<
 *     # calc jmk
 *     for k1 in range(nK):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_probZ));
    __pyx_r = ((PyObject *)__pyx_v_probZ);
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":843
 *                 normC+=2*probZ[m,k1,k2]
 *         probZ[m]/=normC
 *     if up2pz>0:             # <<<<<<<<<<<<<<
 *         return probZ
 *     # calc jmk
 */
  }

  /* "fastphase/fastphase.pyx":846
 *         return probZ
 *     # calc jmk
 *     for k1 in range(nK):             # <<<<<<<<<<<<<<
 *         ##jmk[0,k1]=2*alpha[0,k1]
 *         jmk[0, k1] = probZ[0,k1,k1]
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_k1 = __pyx_t_2;

    /* "fastphase/fastphase.pyx":848
 *     for k1 in range(nK):
 *         ##jmk[0,k1]=2*alpha[0,k1]
 *         jmk[0, k1] = probZ[0,k1,k1]             # <<<<<<<<<<<<<<
 *         for k2 in range(nK):
 *             jmk[0, k1] += probZ[0, k2, k1]
 */
    __pyx_t_203 = 0;
    __pyx_t_204 = __pyx_v_k1;
    __pyx_t_205 = __pyx_v_k1;
    if (__pyx_t_203 < 0) __pyx_t_203 += __pyx_pybuffernd_probZ.diminfo[0].shape;
    if (__pyx_t_204 < 0) __pyx_t_204 += __pyx_pybuffernd_probZ.diminfo[1].shape;
    if (__pyx_t_205 < 0) __pyx_t_205 += __pyx_pybuffernd_probZ.diminfo[2].shape;
    __pyx_t_206 = 0;
    __pyx_t_207 = __pyx_v_k1;
    if (__pyx_t_206 < 0) __pyx_t_206 += __pyx_pybuffernd_jmk.diminfo[0].shape;
    if (__pyx_t_207 < 0) __pyx_t_207 += __pyx_pybuffernd_jmk.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_206, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_207, __pyx_pybuffernd_jmk.diminfo[1].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_203, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_204, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_205, __pyx_pybuffernd_probZ.diminfo[2].strides));

    /* "fastphase/fastphase.pyx":849
 *         ##jmk[0,k1]=2*alpha[0,k1]
 *         jmk[0, k1] = probZ[0,k1,k1]
 *         for k2 in range(nK):             # <<<<<<<<<<<<<<
 *             jmk[0, k1] += probZ[0, k2, k1]
 * 
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k2 = __pyx_t_15;

      /* "fastphase/fastphase.pyx":850
 *         jmk[0, k1] = probZ[0,k1,k1]
 *         for k2 in range(nK):
 *             jmk[0, k1] += probZ[0, k2, k1]             # <<<<<<<<<<<<<<
 * 
 *     for m in range(1,nLoc):
 */
      __pyx_t_208 = 0;
      __pyx_t_209 = __pyx_v_k2;
      __pyx_t_210 = __pyx_v_k1;
      if (__pyx_t_208 < 0) __pyx_t_208 += __pyx_pybuffernd_probZ.diminfo[0].shape;
      if (__pyx_t_209 < 0) __pyx_t_209 += __pyx_pybuffernd_probZ.diminfo[1].shape;
      if (__pyx_t_210 < 0) __pyx_t_210 += __pyx_pybuffernd_probZ.diminfo[2].shape;
      __pyx_t_211 = 0;
      __pyx_t_212 = __pyx_v_k1;
      if (__pyx_t_211 < 0) __pyx_t_211 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_212 < 0) __pyx_t_212 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_211, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_212, __pyx_pybuffernd_jmk.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_208, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_209, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_210, __pyx_pybuffernd_probZ.diminfo[2].strides));
    }
  }

  /* "fastphase/fastphase.pyx":852
 *             jmk[0, k1] += probZ[0, k2, k1]
 * 
 *     for m in range(1,nLoc):             # <<<<<<<<<<<<<<
 *         dummy = myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nLoc); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 1; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":853
 * 
 *     for m in range(1,nLoc):
 *         dummy = myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])             # <<<<<<<<<<<<<<
 *         for k in range(nK):
 *             for k1 in range(nK):
 */
    __pyx_t_213 = (__pyx_v_m - 1);
    if (__pyx_t_213 < 0) __pyx_t_213 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    __pyx_t_214 = __pyx_v_m;
    if (__pyx_t_214 < 0) __pyx_t_214 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    __pyx_t_4 = __Pyx_PyInt_From_npy_long(((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_213, __pyx_pybuffernd_phiScale.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_214, __pyx_pybuffernd_betaScale.diminfo[0].strides)))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_phiScale), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_t_15);

    /* "fastphase/fastphase.pyx":854
 *     for m in range(1,nLoc):
 *         dummy = myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             for k1 in range(nK):
 *                 temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 854, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k = __pyx_t_15;

      /* "fastphase/fastphase.pyx":855
 *         dummy = myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):
 *             for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                 temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])
 *                 temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L1_error)
      __pyx_t_127 = __pyx_t_21;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
        __pyx_v_k1 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":856
 *         for k in range(nK):
 *             for k1 in range(nK):
 *                 temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])             # <<<<<<<<<<<<<<
 *                 temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]
 *                 jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]
 */
        __pyx_t_215 = (__pyx_v_m - 1);
        __pyx_t_216 = __pyx_v_k1;
        if (__pyx_t_215 < 0) __pyx_t_215 += __pyx_pybuffernd_tSumk.diminfo[0].shape;
        if (__pyx_t_216 < 0) __pyx_t_216 += __pyx_pybuffernd_tSumk.diminfo[1].shape;
        __pyx_t_217 = __pyx_v_m;
        __pyx_t_218 = 0;
        if (__pyx_t_217 < 0) __pyx_t_217 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_218 < 0) __pyx_t_218 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_v_temp = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSumk.rcbuffer->pybuffer.buf, __pyx_t_215, __pyx_pybuffernd_tSumk.diminfo[0].strides, __pyx_t_216, __pyx_pybuffernd_tSumk.diminfo[1].strides)) * __pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 1, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_217, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_218, __pyx_pybuffernd_rho.diminfo[1].strides))));

        /* "fastphase/fastphase.pyx":857
 *             for k1 in range(nK):
 *                 temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])
 *                 temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]             # <<<<<<<<<<<<<<
 *                 jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]
 *             jmk[m,k] *= alpha[m,k]
 */
        __pyx_t_219 = __pyx_v_m;
        __pyx_t_220 = 0;
        if (__pyx_t_219 < 0) __pyx_t_219 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_220 < 0) __pyx_t_220 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_221 = (__pyx_v_m - 1);
        if (__pyx_t_221 < 0) __pyx_t_221 += __pyx_pybuffernd_tDoubleSum.diminfo[0].shape;
        __pyx_t_222 = __pyx_v_m;
        __pyx_t_223 = __pyx_v_k1;
        if (__pyx_t_222 < 0) __pyx_t_222 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_223 < 0) __pyx_t_223 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_v_temp = (__pyx_v_temp + (((2.0 * __pyx_f_6hapflk_9fastphase_probJ(__pyx_v_m, 2, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_219, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_220, __pyx_pybuffernd_rho.diminfo[1].strides)))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer.buf, __pyx_t_221, __pyx_pybuffernd_tDoubleSum.diminfo[0].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_222, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_223, __pyx_pybuffernd_alpha.diminfo[1].strides))));

        /* "fastphase/fastphase.pyx":858
 *                 temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])
 *                 temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]
 *                 jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]             # <<<<<<<<<<<<<<
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tDoubleSum[nLoc-1]
 */
        __pyx_t_224 = __pyx_v_m;
        __pyx_t_225 = __pyx_v_k;
        if (__pyx_t_224 < 0) __pyx_t_224 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_225 < 0) __pyx_t_225 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_226 = __pyx_v_m;
        __pyx_t_227 = __pyx_v_k1;
        if (__pyx_t_226 < 0) __pyx_t_226 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_227 < 0) __pyx_t_227 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_228 = __pyx_v_m;
        if (__pyx_t_228 < 0) __pyx_t_228 += __pyx_pybuffernd_gen.diminfo[0].shape;
        __pyx_t_229 = __pyx_v_m;
        __pyx_t_230 = __pyx_v_k;
        __pyx_t_231 = __pyx_v_k1;
        if (__pyx_t_229 < 0) __pyx_t_229 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_230 < 0) __pyx_t_230 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        if (__pyx_t_231 < 0) __pyx_t_231 += __pyx_pybuffernd_mBeta.diminfo[2].shape;
        __pyx_t_232 = __pyx_v_m;
        __pyx_t_233 = __pyx_v_k;
        if (__pyx_t_232 < 0) __pyx_t_232 += __pyx_pybuffernd_jmk.diminfo[0].shape;
        if (__pyx_t_233 < 0) __pyx_t_233 += __pyx_pybuffernd_jmk.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_232, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_233, __pyx_pybuffernd_jmk.diminfo[1].strides) += ((__pyx_v_temp * __pyx_f_6hapflk_9fastphase_genprG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_224, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_225, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_226, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_227, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_228, __pyx_pybuffernd_gen.diminfo[0].strides)))) * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_229, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_230, __pyx_pybuffernd_mBeta.diminfo[1].strides, __pyx_t_231, __pyx_pybuffernd_mBeta.diminfo[2].strides)));
      }

      /* "fastphase/fastphase.pyx":859
 *                 temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]
 *                 jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]
 *             jmk[m,k] *= alpha[m,k]             # <<<<<<<<<<<<<<
 *             jmk[m,k] /= tDoubleSum[nLoc-1]
 *             jmk[m,k] /= dummy
 */
      __pyx_t_234 = __pyx_v_m;
      __pyx_t_235 = __pyx_v_k;
      if (__pyx_t_234 < 0) __pyx_t_234 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_235 < 0) __pyx_t_235 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_236 = __pyx_v_m;
      __pyx_t_237 = __pyx_v_k;
      if (__pyx_t_236 < 0) __pyx_t_236 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_237 < 0) __pyx_t_237 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_236, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_237, __pyx_pybuffernd_jmk.diminfo[1].strides) *= (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_234, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_235, __pyx_pybuffernd_alpha.diminfo[1].strides));

      /* "fastphase/fastphase.pyx":860
 *                 jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tDoubleSum[nLoc-1]             # <<<<<<<<<<<<<<
 *             jmk[m,k] /= dummy
 *     # calc top,bottom
 */
      __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_nLoc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_tDoubleSum), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_238 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_238 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 860, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_239 = __pyx_v_m;
      __pyx_t_240 = __pyx_v_k;
      if (__pyx_t_239 < 0) __pyx_t_239 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_240 < 0) __pyx_t_240 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_239, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_240, __pyx_pybuffernd_jmk.diminfo[1].strides) /= __pyx_t_238;

      /* "fastphase/fastphase.pyx":861
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tDoubleSum[nLoc-1]
 *             jmk[m,k] /= dummy             # <<<<<<<<<<<<<<
 *     # calc top,bottom
 *     for m in range(nLoc):
 */
      __pyx_t_241 = __pyx_v_m;
      __pyx_t_242 = __pyx_v_k;
      if (__pyx_t_241 < 0) __pyx_t_241 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_242 < 0) __pyx_t_242 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_241, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_242, __pyx_pybuffernd_jmk.diminfo[1].strides) /= __pyx_v_dummy;
    }
  }

  /* "fastphase/fastphase.pyx":863
 *             jmk[m,k] /= dummy
 *     # calc top,bottom
 *     for m in range(nLoc):             # <<<<<<<<<<<<<<
 *         ## bottom
 *         for k in range(nK):
 */
  __pyx_t_11 = __Pyx_PyInt_As_long(__pyx_v_nLoc); if (unlikely((__pyx_t_11 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 863, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":865
 *     for m in range(nLoc):
 *         ## bottom
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             bot[m,k]+=probZ[m,k,k]
 *             for k1 in range(nK):
 */
    __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_k = __pyx_t_15;

      /* "fastphase/fastphase.pyx":866
 *         ## bottom
 *         for k in range(nK):
 *             bot[m,k]+=probZ[m,k,k]             # <<<<<<<<<<<<<<
 *             for k1 in range(nK):
 *                 bot[m,k]+=probZ[m,k1,k]
 */
      __pyx_t_243 = __pyx_v_m;
      __pyx_t_244 = __pyx_v_k;
      __pyx_t_245 = __pyx_v_k;
      if (__pyx_t_243 < 0) __pyx_t_243 += __pyx_pybuffernd_probZ.diminfo[0].shape;
      if (__pyx_t_244 < 0) __pyx_t_244 += __pyx_pybuffernd_probZ.diminfo[1].shape;
      if (__pyx_t_245 < 0) __pyx_t_245 += __pyx_pybuffernd_probZ.diminfo[2].shape;
      __pyx_t_246 = __pyx_v_m;
      __pyx_t_247 = __pyx_v_k;
      if (__pyx_t_246 < 0) __pyx_t_246 += __pyx_pybuffernd_bot.diminfo[0].shape;
      if (__pyx_t_247 < 0) __pyx_t_247 += __pyx_pybuffernd_bot.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf, __pyx_t_246, __pyx_pybuffernd_bot.diminfo[0].strides, __pyx_t_247, __pyx_pybuffernd_bot.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_243, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_244, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_245, __pyx_pybuffernd_probZ.diminfo[2].strides));

      /* "fastphase/fastphase.pyx":867
 *         for k in range(nK):
 *             bot[m,k]+=probZ[m,k,k]
 *             for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                 bot[m,k]+=probZ[m,k1,k]
 *         ## top
 */
      __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 867, __pyx_L1_error)
      __pyx_t_127 = __pyx_t_21;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
        __pyx_v_k1 = __pyx_t_18;

        /* "fastphase/fastphase.pyx":868
 *             bot[m,k]+=probZ[m,k,k]
 *             for k1 in range(nK):
 *                 bot[m,k]+=probZ[m,k1,k]             # <<<<<<<<<<<<<<
 *         ## top
 *         if gen[m]==0:
 */
        __pyx_t_248 = __pyx_v_m;
        __pyx_t_249 = __pyx_v_k1;
        __pyx_t_250 = __pyx_v_k;
        if (__pyx_t_248 < 0) __pyx_t_248 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_249 < 0) __pyx_t_249 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        if (__pyx_t_250 < 0) __pyx_t_250 += __pyx_pybuffernd_probZ.diminfo[2].shape;
        __pyx_t_251 = __pyx_v_m;
        __pyx_t_252 = __pyx_v_k;
        if (__pyx_t_251 < 0) __pyx_t_251 += __pyx_pybuffernd_bot.diminfo[0].shape;
        if (__pyx_t_252 < 0) __pyx_t_252 += __pyx_pybuffernd_bot.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf, __pyx_t_251, __pyx_pybuffernd_bot.diminfo[0].strides, __pyx_t_252, __pyx_pybuffernd_bot.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_248, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_249, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_250, __pyx_pybuffernd_probZ.diminfo[2].strides));
      }
    }

    /* "fastphase/fastphase.pyx":870
 *                 bot[m,k]+=probZ[m,k1,k]
 *         ## top
 *         if gen[m]==0:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k]=0
 */
    __pyx_t_253 = __pyx_v_m;
    if (__pyx_t_253 < 0) __pyx_t_253 += __pyx_pybuffernd_gen.diminfo[0].shape;
    __pyx_t_67 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_253, __pyx_pybuffernd_gen.diminfo[0].strides)) == 0) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":871
 *         ## top
 *         if gen[m]==0:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 top[m,k]=0
 *         elif gen[m]==1:
 */
      __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 871, __pyx_L1_error)
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "fastphase/fastphase.pyx":872
 *         if gen[m]==0:
 *             for k in range(nK):
 *                 top[m,k]=0             # <<<<<<<<<<<<<<
 *         elif gen[m]==1:
 *             for k in range(nK):
 */
        __pyx_t_254 = __pyx_v_m;
        __pyx_t_255 = __pyx_v_k;
        if (__pyx_t_254 < 0) __pyx_t_254 += __pyx_pybuffernd_top.diminfo[0].shape;
        if (__pyx_t_255 < 0) __pyx_t_255 += __pyx_pybuffernd_top.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_254, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_255, __pyx_pybuffernd_top.diminfo[1].strides) = 0.0;
      }

      /* "fastphase/fastphase.pyx":870
 *                 bot[m,k]+=probZ[m,k1,k]
 *         ## top
 *         if gen[m]==0:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k]=0
 */
      goto __pyx_L83;
    }

    /* "fastphase/fastphase.pyx":873
 *             for k in range(nK):
 *                 top[m,k]=0
 *         elif gen[m]==1:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 for k1 in range(nK):
 */
    __pyx_t_256 = __pyx_v_m;
    if (__pyx_t_256 < 0) __pyx_t_256 += __pyx_pybuffernd_gen.diminfo[0].shape;
    __pyx_t_67 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_256, __pyx_pybuffernd_gen.diminfo[0].strides)) == 1) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":874
 *                 top[m,k]=0
 *         elif gen[m]==1:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 for k1 in range(nK):
 *                     t1=theta[m,k]*(1-theta[m,k1])
 */
      __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 874, __pyx_L1_error)
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "fastphase/fastphase.pyx":875
 *         elif gen[m]==1:
 *             for k in range(nK):
 *                 for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                     t1=theta[m,k]*(1-theta[m,k1])
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])
 */
        __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 875, __pyx_L1_error)
        __pyx_t_127 = __pyx_t_21;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
          __pyx_v_k1 = __pyx_t_18;

          /* "fastphase/fastphase.pyx":876
 *             for k in range(nK):
 *                 for k1 in range(nK):
 *                     t1=theta[m,k]*(1-theta[m,k1])             # <<<<<<<<<<<<<<
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])
 *                     if (k == k1):
 */
          __pyx_t_257 = __pyx_v_m;
          __pyx_t_258 = __pyx_v_k;
          if (__pyx_t_257 < 0) __pyx_t_257 += __pyx_pybuffernd_theta.diminfo[0].shape;
          if (__pyx_t_258 < 0) __pyx_t_258 += __pyx_pybuffernd_theta.diminfo[1].shape;
          __pyx_t_259 = __pyx_v_m;
          __pyx_t_260 = __pyx_v_k1;
          if (__pyx_t_259 < 0) __pyx_t_259 += __pyx_pybuffernd_theta.diminfo[0].shape;
          if (__pyx_t_260 < 0) __pyx_t_260 += __pyx_pybuffernd_theta.diminfo[1].shape;
          __pyx_v_t1 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_257, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_258, __pyx_pybuffernd_theta.diminfo[1].strides)) * (1.0 - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_259, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_260, __pyx_pybuffernd_theta.diminfo[1].strides))));

          /* "fastphase/fastphase.pyx":877
 *                 for k1 in range(nK):
 *                     t1=theta[m,k]*(1-theta[m,k1])
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])             # <<<<<<<<<<<<<<
 *                     if (k == k1):
 *                         top[m,k] += 2*probZ[m,k,k1]*t1/t2
 */
          __pyx_t_261 = __pyx_v_m;
          __pyx_t_262 = __pyx_v_k1;
          if (__pyx_t_261 < 0) __pyx_t_261 += __pyx_pybuffernd_theta.diminfo[0].shape;
          if (__pyx_t_262 < 0) __pyx_t_262 += __pyx_pybuffernd_theta.diminfo[1].shape;
          __pyx_t_263 = __pyx_v_m;
          __pyx_t_264 = __pyx_v_k;
          if (__pyx_t_263 < 0) __pyx_t_263 += __pyx_pybuffernd_theta.diminfo[0].shape;
          if (__pyx_t_264 < 0) __pyx_t_264 += __pyx_pybuffernd_theta.diminfo[1].shape;
          __pyx_v_t2 = (__pyx_v_t1 + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_261, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_262, __pyx_pybuffernd_theta.diminfo[1].strides)) * (1.0 - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_263, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_264, __pyx_pybuffernd_theta.diminfo[1].strides)))));

          /* "fastphase/fastphase.pyx":878
 *                     t1=theta[m,k]*(1-theta[m,k1])
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])
 *                     if (k == k1):             # <<<<<<<<<<<<<<
 *                         top[m,k] += 2*probZ[m,k,k1]*t1/t2
 *                     else:
 */
          __pyx_t_67 = ((__pyx_v_k == __pyx_v_k1) != 0);
          if (__pyx_t_67) {

            /* "fastphase/fastphase.pyx":879
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])
 *                     if (k == k1):
 *                         top[m,k] += 2*probZ[m,k,k1]*t1/t2             # <<<<<<<<<<<<<<
 *                     else:
 *                         top[m,k] += probZ[m,k,k1]*t1/t2
 */
            __pyx_t_265 = __pyx_v_m;
            __pyx_t_266 = __pyx_v_k;
            __pyx_t_267 = __pyx_v_k1;
            if (__pyx_t_265 < 0) __pyx_t_265 += __pyx_pybuffernd_probZ.diminfo[0].shape;
            if (__pyx_t_266 < 0) __pyx_t_266 += __pyx_pybuffernd_probZ.diminfo[1].shape;
            if (__pyx_t_267 < 0) __pyx_t_267 += __pyx_pybuffernd_probZ.diminfo[2].shape;
            __pyx_t_238 = ((2.0 * (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_265, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_266, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_267, __pyx_pybuffernd_probZ.diminfo[2].strides))) * __pyx_v_t1);
            if (unlikely(__pyx_v_t2 == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 879, __pyx_L1_error)
            }
            __pyx_t_268 = __pyx_v_m;
            __pyx_t_269 = __pyx_v_k;
            if (__pyx_t_268 < 0) __pyx_t_268 += __pyx_pybuffernd_top.diminfo[0].shape;
            if (__pyx_t_269 < 0) __pyx_t_269 += __pyx_pybuffernd_top.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_268, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_269, __pyx_pybuffernd_top.diminfo[1].strides) += (__pyx_t_238 / ((__pyx_t_5numpy_float64_t)__pyx_v_t2));

            /* "fastphase/fastphase.pyx":878
 *                     t1=theta[m,k]*(1-theta[m,k1])
 *                     t2=t1+theta[m,k1]*(1-theta[m,k])
 *                     if (k == k1):             # <<<<<<<<<<<<<<
 *                         top[m,k] += 2*probZ[m,k,k1]*t1/t2
 *                     else:
 */
            goto __pyx_L90;
          }

          /* "fastphase/fastphase.pyx":881
 *                         top[m,k] += 2*probZ[m,k,k1]*t1/t2
 *                     else:
 *                         top[m,k] += probZ[m,k,k1]*t1/t2             # <<<<<<<<<<<<<<
 *         elif gen[m]==2:
 *             for k in range(nK):
 */
          /*else*/ {
            __pyx_t_270 = __pyx_v_m;
            __pyx_t_271 = __pyx_v_k;
            __pyx_t_272 = __pyx_v_k1;
            if (__pyx_t_270 < 0) __pyx_t_270 += __pyx_pybuffernd_probZ.diminfo[0].shape;
            if (__pyx_t_271 < 0) __pyx_t_271 += __pyx_pybuffernd_probZ.diminfo[1].shape;
            if (__pyx_t_272 < 0) __pyx_t_272 += __pyx_pybuffernd_probZ.diminfo[2].shape;
            __pyx_t_238 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_270, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_271, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_272, __pyx_pybuffernd_probZ.diminfo[2].strides)) * __pyx_v_t1);
            if (unlikely(__pyx_v_t2 == 0)) {
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              __PYX_ERR(0, 881, __pyx_L1_error)
            }
            __pyx_t_273 = __pyx_v_m;
            __pyx_t_274 = __pyx_v_k;
            if (__pyx_t_273 < 0) __pyx_t_273 += __pyx_pybuffernd_top.diminfo[0].shape;
            if (__pyx_t_274 < 0) __pyx_t_274 += __pyx_pybuffernd_top.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_273, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_274, __pyx_pybuffernd_top.diminfo[1].strides) += (__pyx_t_238 / ((__pyx_t_5numpy_float64_t)__pyx_v_t2));
          }
          __pyx_L90:;
        }
      }

      /* "fastphase/fastphase.pyx":873
 *             for k in range(nK):
 *                 top[m,k]=0
 *         elif gen[m]==1:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 for k1 in range(nK):
 */
      goto __pyx_L83;
    }

    /* "fastphase/fastphase.pyx":882
 *                     else:
 *                         top[m,k] += probZ[m,k,k1]*t1/t2
 *         elif gen[m]==2:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k] += probZ[m,k,k]
 */
    __pyx_t_275 = __pyx_v_m;
    if (__pyx_t_275 < 0) __pyx_t_275 += __pyx_pybuffernd_gen.diminfo[0].shape;
    __pyx_t_67 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_gen.rcbuffer->pybuffer.buf, __pyx_t_275, __pyx_pybuffernd_gen.diminfo[0].strides)) == 2) != 0);
    if (__pyx_t_67) {

      /* "fastphase/fastphase.pyx":883
 *                         top[m,k] += probZ[m,k,k1]*t1/t2
 *         elif gen[m]==2:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 top[m,k] += probZ[m,k,k]
 *                 for k1 in range(nK):
 */
      __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 883, __pyx_L1_error)
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "fastphase/fastphase.pyx":884
 *         elif gen[m]==2:
 *             for k in range(nK):
 *                 top[m,k] += probZ[m,k,k]             # <<<<<<<<<<<<<<
 *                 for k1 in range(nK):
 *                     top[m,k] += probZ[m,k,k1]
 */
        __pyx_t_276 = __pyx_v_m;
        __pyx_t_277 = __pyx_v_k;
        __pyx_t_278 = __pyx_v_k;
        if (__pyx_t_276 < 0) __pyx_t_276 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_277 < 0) __pyx_t_277 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        if (__pyx_t_278 < 0) __pyx_t_278 += __pyx_pybuffernd_probZ.diminfo[2].shape;
        __pyx_t_279 = __pyx_v_m;
        __pyx_t_280 = __pyx_v_k;
        if (__pyx_t_279 < 0) __pyx_t_279 += __pyx_pybuffernd_top.diminfo[0].shape;
        if (__pyx_t_280 < 0) __pyx_t_280 += __pyx_pybuffernd_top.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_279, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_280, __pyx_pybuffernd_top.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_276, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_277, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_278, __pyx_pybuffernd_probZ.diminfo[2].strides));

        /* "fastphase/fastphase.pyx":885
 *             for k in range(nK):
 *                 top[m,k] += probZ[m,k,k]
 *                 for k1 in range(nK):             # <<<<<<<<<<<<<<
 *                     top[m,k] += probZ[m,k,k1]
 *         else:
 */
        __pyx_t_21 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_21 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 885, __pyx_L1_error)
        __pyx_t_127 = __pyx_t_21;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_127; __pyx_t_18+=1) {
          __pyx_v_k1 = __pyx_t_18;

          /* "fastphase/fastphase.pyx":886
 *                 top[m,k] += probZ[m,k,k]
 *                 for k1 in range(nK):
 *                     top[m,k] += probZ[m,k,k1]             # <<<<<<<<<<<<<<
 *         else:
 *             for k in range(nK):
 */
          __pyx_t_281 = __pyx_v_m;
          __pyx_t_282 = __pyx_v_k;
          __pyx_t_283 = __pyx_v_k1;
          if (__pyx_t_281 < 0) __pyx_t_281 += __pyx_pybuffernd_probZ.diminfo[0].shape;
          if (__pyx_t_282 < 0) __pyx_t_282 += __pyx_pybuffernd_probZ.diminfo[1].shape;
          if (__pyx_t_283 < 0) __pyx_t_283 += __pyx_pybuffernd_probZ.diminfo[2].shape;
          __pyx_t_284 = __pyx_v_m;
          __pyx_t_285 = __pyx_v_k;
          if (__pyx_t_284 < 0) __pyx_t_284 += __pyx_pybuffernd_top.diminfo[0].shape;
          if (__pyx_t_285 < 0) __pyx_t_285 += __pyx_pybuffernd_top.diminfo[1].shape;
          *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_284, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_285, __pyx_pybuffernd_top.diminfo[1].strides) += (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_281, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_282, __pyx_pybuffernd_probZ.diminfo[1].strides, __pyx_t_283, __pyx_pybuffernd_probZ.diminfo[2].strides));
        }
      }

      /* "fastphase/fastphase.pyx":882
 *                     else:
 *                         top[m,k] += probZ[m,k,k1]*t1/t2
 *         elif gen[m]==2:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k] += probZ[m,k,k]
 */
      goto __pyx_L83;
    }

    /* "fastphase/fastphase.pyx":888
 *                     top[m,k] += probZ[m,k,k1]
 *         else:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 top[m,k]=0
 *                 bot[m,k]=0
 */
    /*else*/ {
      __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_nK); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L1_error)
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "fastphase/fastphase.pyx":889
 *         else:
 *             for k in range(nK):
 *                 top[m,k]=0             # <<<<<<<<<<<<<<
 *                 bot[m,k]=0
 *     return logLikelihood,top,bot,jmk
 */
        __pyx_t_286 = __pyx_v_m;
        __pyx_t_287 = __pyx_v_k;
        if (__pyx_t_286 < 0) __pyx_t_286 += __pyx_pybuffernd_top.diminfo[0].shape;
        if (__pyx_t_287 < 0) __pyx_t_287 += __pyx_pybuffernd_top.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_286, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_287, __pyx_pybuffernd_top.diminfo[1].strides) = 0.0;

        /* "fastphase/fastphase.pyx":890
 *             for k in range(nK):
 *                 top[m,k]=0
 *                 bot[m,k]=0             # <<<<<<<<<<<<<<
 *     return logLikelihood,top,bot,jmk
 * 
 */
        __pyx_t_288 = __pyx_v_m;
        __pyx_t_289 = __pyx_v_k;
        if (__pyx_t_288 < 0) __pyx_t_288 += __pyx_pybuffernd_bot.diminfo[0].shape;
        if (__pyx_t_289 < 0) __pyx_t_289 += __pyx_pybuffernd_bot.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf, __pyx_t_288, __pyx_pybuffernd_bot.diminfo[0].strides, __pyx_t_289, __pyx_pybuffernd_bot.diminfo[1].strides) = 0.0;
      }
    }
    __pyx_L83:;
  }

  /* "fastphase/fastphase.pyx":891
 *                 top[m,k]=0
 *                 bot[m,k]=0
 *     return logLikelihood,top,bot,jmk             # <<<<<<<<<<<<<<
 * 
 * #### Haplotype Calculations
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_logLikelihood); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_v_top));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_top));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_top));
  __Pyx_INCREF(((PyObject *)__pyx_v_bot));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bot));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_bot));
  __Pyx_INCREF(((PyObject *)__pyx_v_jmk));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_jmk));
  PyTuple_SET_ITEM(__pyx_t_3, 3, ((PyObject *)__pyx_v_jmk));
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":691
 *     return [ idx2pair(ikp, nK) for ikp in soluce]
 * 
 * cpdef genCalc(aa,tt,rr,gg,u2p):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha=aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta=tt
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bot.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gen.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tSumk.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_top.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.genCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bot.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gen.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tDoubleSum.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tSumk.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_top.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_alpha);
  __Pyx_XDECREF((PyObject *)__pyx_v_theta);
  __Pyx_XDECREF((PyObject *)__pyx_v_rho);
  __Pyx_XDECREF((PyObject *)__pyx_v_gen);
  __Pyx_XDECREF(__pyx_v_nK);
  __Pyx_XDECREF(__pyx_v_nLoc);
  __Pyx_XDECREF((PyObject *)__pyx_v_betaScale);
  __Pyx_XDECREF((PyObject *)__pyx_v_tSumk);
  __Pyx_XDECREF((PyObject *)__pyx_v_tDoubleSum);
  __Pyx_XDECREF((PyObject *)__pyx_v_mBeta);
  __Pyx_XDECREF((PyObject *)__pyx_v_mPhi);
  __Pyx_XDECREF((PyObject *)__pyx_v_phiScale);
  __Pyx_XDECREF((PyObject *)__pyx_v_probZ);
  __Pyx_XDECREF((PyObject *)__pyx_v_jmk);
  __Pyx_XDECREF((PyObject *)__pyx_v_top);
  __Pyx_XDECREF((PyObject *)__pyx_v_bot);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_19genCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_19genCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aa = 0;
  PyObject *__pyx_v_tt = 0;
  PyObject *__pyx_v_rr = 0;
  PyObject *__pyx_v_gg = 0;
  PyObject *__pyx_v_u2p = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genCalc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aa,&__pyx_n_s_tt,&__pyx_n_s_rr,&__pyx_n_s_gg,&__pyx_n_s_u2p,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aa)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genCalc", 1, 5, 5, 1); __PYX_ERR(0, 691, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genCalc", 1, 5, 5, 2); __PYX_ERR(0, 691, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gg)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genCalc", 1, 5, 5, 3); __PYX_ERR(0, 691, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u2p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("genCalc", 1, 5, 5, 4); __PYX_ERR(0, 691, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "genCalc") < 0)) __PYX_ERR(0, 691, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_aa = values[0];
    __pyx_v_tt = values[1];
    __pyx_v_rr = values[2];
    __pyx_v_gg = values[3];
    __pyx_v_u2p = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("genCalc", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 691, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.genCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_18genCalc(__pyx_self, __pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_gg, __pyx_v_u2p);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_18genCalc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_gg, PyObject *__pyx_v_u2p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("genCalc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6hapflk_9fastphase_genCalc(__pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_gg, __pyx_v_u2p, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.genCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":895
 * #### Haplotype Calculations
 * 
 * cdef double happrG(double t,int s):             # <<<<<<<<<<<<<<
 *     if s==0:
 *         return 1-t
 */

static double __pyx_f_6hapflk_9fastphase_happrG(double __pyx_v_t, int __pyx_v_s) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("happrG", 0);

  /* "fastphase/fastphase.pyx":896
 * 
 * cdef double happrG(double t,int s):
 *     if s==0:             # <<<<<<<<<<<<<<
 *         return 1-t
 *     elif s==1:
 */
  switch (__pyx_v_s) {
    case 0:

    /* "fastphase/fastphase.pyx":897
 * cdef double happrG(double t,int s):
 *     if s==0:
 *         return 1-t             # <<<<<<<<<<<<<<
 *     elif s==1:
 *         return t
 */
    __pyx_r = (1.0 - __pyx_v_t);
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":896
 * 
 * cdef double happrG(double t,int s):
 *     if s==0:             # <<<<<<<<<<<<<<
 *         return 1-t
 *     elif s==1:
 */
    break;
    case 1:

    /* "fastphase/fastphase.pyx":899
 *         return 1-t
 *     elif s==1:
 *         return t             # <<<<<<<<<<<<<<
 *     else:
 *         return 1
 */
    __pyx_r = __pyx_v_t;
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":898
 *     if s==0:
 *         return 1-t
 *     elif s==1:             # <<<<<<<<<<<<<<
 *         return t
 *     else:
 */
    break;
    default:

    /* "fastphase/fastphase.pyx":901
 *         return t
 *     else:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 * cpdef hapViterbi( aa, tt, rr, hh):
 */
    __pyx_r = 1.0;
    goto __pyx_L0;
    break;
  }

  /* "fastphase/fastphase.pyx":895
 * #### Haplotype Calculations
 * 
 * cdef double happrG(double t,int s):             # <<<<<<<<<<<<<<
 *     if s==0:
 *         return 1-t
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":903
 *         return 1
 * 
 * cpdef hapViterbi( aa, tt, rr, hh):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_21hapViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_hapViterbi(PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_alpha = 0;
  PyArrayObject *__pyx_v_theta = 0;
  PyArrayObject *__pyx_v_rho = 0;
  PyArrayObject *__pyx_v_hap = 0;
  int __pyx_v_nLoc;
  int __pyx_v_nK;
  int __pyx_v_k;
  int __pyx_v_prev_k;
  int __pyx_v_m;
  PyArrayObject *__pyx_v_delta = 0;
  PyArrayObject *__pyx_v_psi = 0;
  PyArrayObject *__pyx_v_soluce = 0;
  PyArrayObject *__pyx_v_tempVal = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_alpha;
  __Pyx_Buffer __pyx_pybuffer_alpha;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_delta;
  __Pyx_Buffer __pyx_pybuffer_delta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hap;
  __Pyx_Buffer __pyx_pybuffer_hap;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_psi;
  __Pyx_Buffer __pyx_pybuffer_psi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rho;
  __Pyx_Buffer __pyx_pybuffer_rho;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_soluce;
  __Pyx_Buffer __pyx_pybuffer_soluce;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tempVal;
  __Pyx_Buffer __pyx_pybuffer_tempVal;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_theta;
  __Pyx_Buffer __pyx_pybuffer_theta;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  int __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  int __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  Py_ssize_t __pyx_t_51;
  Py_ssize_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  __Pyx_RefNannySetupContext("hapViterbi", 0);
  __pyx_pybuffer_alpha.pybuffer.buf = NULL;
  __pyx_pybuffer_alpha.refcount = 0;
  __pyx_pybuffernd_alpha.data = NULL;
  __pyx_pybuffernd_alpha.rcbuffer = &__pyx_pybuffer_alpha;
  __pyx_pybuffer_theta.pybuffer.buf = NULL;
  __pyx_pybuffer_theta.refcount = 0;
  __pyx_pybuffernd_theta.data = NULL;
  __pyx_pybuffernd_theta.rcbuffer = &__pyx_pybuffer_theta;
  __pyx_pybuffer_rho.pybuffer.buf = NULL;
  __pyx_pybuffer_rho.refcount = 0;
  __pyx_pybuffernd_rho.data = NULL;
  __pyx_pybuffernd_rho.rcbuffer = &__pyx_pybuffer_rho;
  __pyx_pybuffer_hap.pybuffer.buf = NULL;
  __pyx_pybuffer_hap.refcount = 0;
  __pyx_pybuffernd_hap.data = NULL;
  __pyx_pybuffernd_hap.rcbuffer = &__pyx_pybuffer_hap;
  __pyx_pybuffer_delta.pybuffer.buf = NULL;
  __pyx_pybuffer_delta.refcount = 0;
  __pyx_pybuffernd_delta.data = NULL;
  __pyx_pybuffernd_delta.rcbuffer = &__pyx_pybuffer_delta;
  __pyx_pybuffer_psi.pybuffer.buf = NULL;
  __pyx_pybuffer_psi.refcount = 0;
  __pyx_pybuffernd_psi.data = NULL;
  __pyx_pybuffernd_psi.rcbuffer = &__pyx_pybuffer_psi;
  __pyx_pybuffer_soluce.pybuffer.buf = NULL;
  __pyx_pybuffer_soluce.refcount = 0;
  __pyx_pybuffernd_soluce.data = NULL;
  __pyx_pybuffernd_soluce.rcbuffer = &__pyx_pybuffer_soluce;
  __pyx_pybuffer_tempVal.pybuffer.buf = NULL;
  __pyx_pybuffer_tempVal.refcount = 0;
  __pyx_pybuffernd_tempVal.data = NULL;
  __pyx_pybuffernd_tempVal.rcbuffer = &__pyx_pybuffer_tempVal;

  /* "fastphase/fastphase.pyx":904
 * 
 * cpdef hapViterbi( aa, tt, rr, hh):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 */
  if (!(likely(((__pyx_v_aa) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_aa, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 904, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_aa;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_alpha = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 904, __pyx_L1_error)
    } else {__pyx_pybuffernd_alpha.diminfo[0].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_alpha.diminfo[0].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_alpha.diminfo[1].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_alpha.diminfo[1].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_alpha = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":905
 * cpdef hapViterbi( aa, tt, rr, hh):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh
 */
  if (!(likely(((__pyx_v_tt) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_tt, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 905, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_tt;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_theta.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_theta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 905, __pyx_L1_error)
    } else {__pyx_pybuffernd_theta.diminfo[0].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_theta.diminfo[0].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_theta.diminfo[1].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_theta.diminfo[1].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_theta = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":906
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh
 * 
 */
  if (!(likely(((__pyx_v_rr) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_rr, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_rr;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rho.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_rho = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 906, __pyx_L1_error)
    } else {__pyx_pybuffernd_rho.diminfo[0].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rho.diminfo[0].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_rho.diminfo[1].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_rho.diminfo[1].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_rho = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":907
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh             # <<<<<<<<<<<<<<
 * 
 *     ## cython declarations
 */
  if (!(likely(((__pyx_v_hh) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_hh, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 907, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_hh;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hap.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_hap = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 907, __pyx_L1_error)
    } else {__pyx_pybuffernd_hap.diminfo[0].strides = __pyx_pybuffernd_hap.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hap.diminfo[0].shape = __pyx_pybuffernd_hap.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_v_hap = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":913
 *     cdef int k, prev_k, best_k, m
 *     cdef double best_v
 *     nLoc = alpha.shape[0]             # <<<<<<<<<<<<<<
 *     nK = alpha.shape[1]
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)
 */
  __pyx_v_nLoc = (__pyx_v_alpha->dimensions[0]);

  /* "fastphase/fastphase.pyx":914
 *     cdef double best_v
 *     nLoc = alpha.shape[0]
 *     nK = alpha.shape[1]             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)
 */
  __pyx_v_nK = (__pyx_v_alpha->dimensions[1]);

  /* "fastphase/fastphase.pyx":915
 *     nLoc = alpha.shape[0]
 *     nK = alpha.shape[1]
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.empty(nLoc, dtype= np.int)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 915, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_delta.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_delta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 915, __pyx_L1_error)
    } else {__pyx_pybuffernd_delta.diminfo[0].strides = __pyx_pybuffernd_delta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_delta.diminfo[0].shape = __pyx_pybuffernd_delta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_delta.diminfo[1].strides = __pyx_pybuffernd_delta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_delta.diminfo[1].shape = __pyx_pybuffernd_delta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_delta = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":916
 *     nK = alpha.shape[1]
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.empty(nLoc, dtype= np.int)
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( nK, dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 916, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_psi.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_psi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 916, __pyx_L1_error)
    } else {__pyx_pybuffernd_psi.diminfo[0].strides = __pyx_pybuffernd_psi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_psi.diminfo[0].shape = __pyx_pybuffernd_psi.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_psi.diminfo[1].strides = __pyx_pybuffernd_psi.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_psi.diminfo[1].shape = __pyx_pybuffernd_psi.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_psi = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":917
 *     cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.empty(nLoc, dtype= np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( nK, dtype=np.float64)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_soluce = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 917, __pyx_L1_error)
    } else {__pyx_pybuffernd_soluce.diminfo[0].strides = __pyx_pybuffernd_soluce.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_soluce.diminfo[0].shape = __pyx_pybuffernd_soluce.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_soluce = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":918
 *     cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)
 *     cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.empty(nLoc, dtype= np.int)
 *     cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( nK, dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     ## initialization
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 918, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tempVal = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 918, __pyx_L1_error)
    } else {__pyx_pybuffernd_tempVal.diminfo[0].strides = __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tempVal.diminfo[0].shape = __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_tempVal = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":921
 * 
 *     ## initialization
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         delta[k,0] = log(alpha[ 0, k]) + log(happrG( theta[ 0, k], hap[0]))
 * 
 */
  __pyx_t_10 = __pyx_v_nK;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_k = __pyx_t_12;

    /* "fastphase/fastphase.pyx":922
 *     ## initialization
 *     for k in range(nK):
 *         delta[k,0] = log(alpha[ 0, k]) + log(happrG( theta[ 0, k], hap[0]))             # <<<<<<<<<<<<<<
 * 
 *     ## recursion
 */
    __pyx_t_13 = 0;
    __pyx_t_14 = __pyx_v_k;
    if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_alpha.diminfo[0].shape;
    if (__pyx_t_14 < 0) __pyx_t_14 += __pyx_pybuffernd_alpha.diminfo[1].shape;
    __pyx_t_15 = 0;
    __pyx_t_16 = __pyx_v_k;
    if (__pyx_t_15 < 0) __pyx_t_15 += __pyx_pybuffernd_theta.diminfo[0].shape;
    if (__pyx_t_16 < 0) __pyx_t_16 += __pyx_pybuffernd_theta.diminfo[1].shape;
    __pyx_t_17 = 0;
    if (__pyx_t_17 < 0) __pyx_t_17 += __pyx_pybuffernd_hap.diminfo[0].shape;
    __pyx_t_18 = __pyx_v_k;
    __pyx_t_19 = 0;
    if (__pyx_t_18 < 0) __pyx_t_18 += __pyx_pybuffernd_delta.diminfo[0].shape;
    if (__pyx_t_19 < 0) __pyx_t_19 += __pyx_pybuffernd_delta.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_delta.diminfo[1].strides) = (log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_alpha.diminfo[1].strides))) + log(__pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_hap.diminfo[0].strides)))));
  }

  /* "fastphase/fastphase.pyx":925
 * 
 *     ## recursion
 *     for m in range(1, nLoc):             # <<<<<<<<<<<<<<
 *         for k in range(nK):
 *             for prev_k in range(nK):
 */
  __pyx_t_10 = __pyx_v_nLoc;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 1; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_m = __pyx_t_12;

    /* "fastphase/fastphase.pyx":926
 *     ## recursion
 *     for m in range(1, nLoc):
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             for prev_k in range(nK):
 *                 ## jump
 */
    __pyx_t_20 = __pyx_v_nK;
    __pyx_t_21 = __pyx_t_20;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
      __pyx_v_k = __pyx_t_22;

      /* "fastphase/fastphase.pyx":927
 *     for m in range(1, nLoc):
 *         for k in range(nK):
 *             for prev_k in range(nK):             # <<<<<<<<<<<<<<
 *                 ## jump
 *                 tempVal[ prev_k] = log(rho[ m, 0])+log( alpha[ m, k])
 */
      __pyx_t_23 = __pyx_v_nK;
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_24; __pyx_t_25+=1) {
        __pyx_v_prev_k = __pyx_t_25;

        /* "fastphase/fastphase.pyx":929
 *             for prev_k in range(nK):
 *                 ## jump
 *                 tempVal[ prev_k] = log(rho[ m, 0])+log( alpha[ m, k])             # <<<<<<<<<<<<<<
 *                 if k == prev_k:
 *                     ## no jump
 */
        __pyx_t_26 = __pyx_v_m;
        __pyx_t_27 = 0;
        if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_rho.diminfo[0].shape;
        if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_rho.diminfo[1].shape;
        __pyx_t_28 = __pyx_v_m;
        __pyx_t_29 = __pyx_v_k;
        if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_alpha.diminfo[0].shape;
        if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_alpha.diminfo[1].shape;
        __pyx_t_30 = __pyx_v_prev_k;
        if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_tempVal.diminfo[0].strides) = (log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_rho.diminfo[1].strides))) + log((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_alpha.diminfo[1].strides))));

        /* "fastphase/fastphase.pyx":930
 *                 ## jump
 *                 tempVal[ prev_k] = log(rho[ m, 0])+log( alpha[ m, k])
 *                 if k == prev_k:             # <<<<<<<<<<<<<<
 *                     ## no jump
 *                     tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))
 */
        __pyx_t_31 = ((__pyx_v_k == __pyx_v_prev_k) != 0);
        if (__pyx_t_31) {

          /* "fastphase/fastphase.pyx":932
 *                 if k == prev_k:
 *                     ## no jump
 *                     tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))             # <<<<<<<<<<<<<<
 *                 tempVal[ prev_k] += delta[prev_k, m-1]
 *                 psi[ k, m] = argmax(tempVal, nK)
 */
          __pyx_t_32 = __pyx_v_m;
          __pyx_t_33 = 0;
          if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_rho.diminfo[0].shape;
          if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_rho.diminfo[1].shape;
          __pyx_t_34 = __pyx_v_m;
          __pyx_t_35 = __pyx_v_k;
          if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_alpha.diminfo[0].shape;
          if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_alpha.diminfo[1].shape;
          __pyx_t_36 = __pyx_v_m;
          __pyx_t_37 = 0;
          if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_rho.diminfo[0].shape;
          if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_rho.diminfo[1].shape;
          __pyx_t_38 = __pyx_v_prev_k;
          if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_tempVal.diminfo[0].strides) = log((((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_33, __pyx_pybuffernd_rho.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_35, __pyx_pybuffernd_alpha.diminfo[1].strides))) + (1.0 - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_37, __pyx_pybuffernd_rho.diminfo[1].strides)))));

          /* "fastphase/fastphase.pyx":930
 *                 ## jump
 *                 tempVal[ prev_k] = log(rho[ m, 0])+log( alpha[ m, k])
 *                 if k == prev_k:             # <<<<<<<<<<<<<<
 *                     ## no jump
 *                     tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))
 */
        }

        /* "fastphase/fastphase.pyx":933
 *                     ## no jump
 *                     tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))
 *                 tempVal[ prev_k] += delta[prev_k, m-1]             # <<<<<<<<<<<<<<
 *                 psi[ k, m] = argmax(tempVal, nK)
 *                 delta[ k, m] = tempVal[ psi[ k, m]] +log( happrG( theta[ m, k], hap[ m]))
 */
        __pyx_t_39 = __pyx_v_prev_k;
        __pyx_t_40 = (__pyx_v_m - 1);
        if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_delta.diminfo[0].shape;
        if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_delta.diminfo[1].shape;
        __pyx_t_41 = __pyx_v_prev_k;
        if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_tempVal.diminfo[0].strides) += (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_40, __pyx_pybuffernd_delta.diminfo[1].strides));

        /* "fastphase/fastphase.pyx":934
 *                     tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))
 *                 tempVal[ prev_k] += delta[prev_k, m-1]
 *                 psi[ k, m] = argmax(tempVal, nK)             # <<<<<<<<<<<<<<
 *                 delta[ k, m] = tempVal[ psi[ k, m]] +log( happrG( theta[ m, k], hap[ m]))
 * 
 */
        __pyx_t_42 = __pyx_v_k;
        __pyx_t_43 = __pyx_v_m;
        if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_psi.diminfo[0].shape;
        if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_psi.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_42, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_43, __pyx_pybuffernd_psi.diminfo[1].strides) = __pyx_f_6hapflk_9fastphase_argmax(((PyArrayObject *)__pyx_v_tempVal), __pyx_v_nK);

        /* "fastphase/fastphase.pyx":935
 *                 tempVal[ prev_k] += delta[prev_k, m-1]
 *                 psi[ k, m] = argmax(tempVal, nK)
 *                 delta[ k, m] = tempVal[ psi[ k, m]] +log( happrG( theta[ m, k], hap[ m]))             # <<<<<<<<<<<<<<
 * 
 *     ## termination
 */
        __pyx_t_44 = __pyx_v_k;
        __pyx_t_45 = __pyx_v_m;
        if (__pyx_t_44 < 0) __pyx_t_44 += __pyx_pybuffernd_psi.diminfo[0].shape;
        if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_psi.diminfo[1].shape;
        __pyx_t_46 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_44, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_45, __pyx_pybuffernd_psi.diminfo[1].strides));
        if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_tempVal.diminfo[0].shape;
        __pyx_t_47 = __pyx_v_m;
        __pyx_t_48 = __pyx_v_k;
        if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_theta.diminfo[0].shape;
        if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_theta.diminfo[1].shape;
        __pyx_t_49 = __pyx_v_m;
        if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_hap.diminfo[0].shape;
        __pyx_t_50 = __pyx_v_k;
        __pyx_t_51 = __pyx_v_m;
        if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_delta.diminfo[0].shape;
        if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_delta.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_delta.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_delta.diminfo[0].strides, __pyx_t_51, __pyx_pybuffernd_delta.diminfo[1].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tempVal.rcbuffer->pybuffer.buf, __pyx_t_46, __pyx_pybuffernd_tempVal.diminfo[0].strides)) + log(__pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_48, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_hap.diminfo[0].strides)))));
      }
    }
  }

  /* "fastphase/fastphase.pyx":938
 * 
 *     ## termination
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], nK)             # <<<<<<<<<<<<<<
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 */
  __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_nLoc - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_slice_);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_delta), __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 938, __pyx_L1_error)
  __pyx_t_52 = (__pyx_v_nLoc - 1);
  if (__pyx_t_52 < 0) __pyx_t_52 += __pyx_pybuffernd_soluce.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_soluce.diminfo[0].strides) = __pyx_f_6hapflk_9fastphase_argmax(((PyArrayObject *)__pyx_t_4), __pyx_v_nK);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":939
 *     ## termination
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], nK)
 *     for m in range( nLoc - 2, -1, -1):             # <<<<<<<<<<<<<<
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 *     return soluce
 */
  for (__pyx_t_10 = (__pyx_v_nLoc - 2); __pyx_t_10 > -1; __pyx_t_10-=1) {
    __pyx_v_m = __pyx_t_10;

    /* "fastphase/fastphase.pyx":940
 *     soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], nK)
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]             # <<<<<<<<<<<<<<
 *     return soluce
 * 
 */
    __pyx_t_53 = (__pyx_v_m + 1);
    if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_soluce.diminfo[0].shape;
    __pyx_t_54 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_soluce.diminfo[0].strides));
    __pyx_t_55 = (__pyx_v_m + 1);
    if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_psi.diminfo[0].shape;
    if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_psi.diminfo[1].shape;
    __pyx_t_56 = __pyx_v_m;
    if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_soluce.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_soluce.rcbuffer->pybuffer.buf, __pyx_t_56, __pyx_pybuffernd_soluce.diminfo[0].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_psi.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_psi.diminfo[0].strides, __pyx_t_55, __pyx_pybuffernd_psi.diminfo[1].strides));
  }

  /* "fastphase/fastphase.pyx":941
 *     for m in range( nLoc - 2, -1, -1):
 *         soluce[ m ] = psi[ soluce[ m+1 ], m+1]
 *     return soluce             # <<<<<<<<<<<<<<
 * 
 * cpdef hapCalc(aa,tt,rr,hh,u2p):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_soluce));
  __pyx_r = ((PyObject *)__pyx_v_soluce);
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":903
 *         return 1
 * 
 * cpdef hapViterbi( aa, tt, rr, hh):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_delta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hap.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_psi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.hapViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_delta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hap.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_psi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_soluce.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tempVal.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_alpha);
  __Pyx_XDECREF((PyObject *)__pyx_v_theta);
  __Pyx_XDECREF((PyObject *)__pyx_v_rho);
  __Pyx_XDECREF((PyObject *)__pyx_v_hap);
  __Pyx_XDECREF((PyObject *)__pyx_v_delta);
  __Pyx_XDECREF((PyObject *)__pyx_v_psi);
  __Pyx_XDECREF((PyObject *)__pyx_v_soluce);
  __Pyx_XDECREF((PyObject *)__pyx_v_tempVal);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_21hapViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_21hapViterbi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aa = 0;
  PyObject *__pyx_v_tt = 0;
  PyObject *__pyx_v_rr = 0;
  PyObject *__pyx_v_hh = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hapViterbi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aa,&__pyx_n_s_tt,&__pyx_n_s_rr,&__pyx_n_s_hh,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aa)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapViterbi", 1, 4, 4, 1); __PYX_ERR(0, 903, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapViterbi", 1, 4, 4, 2); __PYX_ERR(0, 903, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapViterbi", 1, 4, 4, 3); __PYX_ERR(0, 903, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "hapViterbi") < 0)) __PYX_ERR(0, 903, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_aa = values[0];
    __pyx_v_tt = values[1];
    __pyx_v_rr = values[2];
    __pyx_v_hh = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hapViterbi", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 903, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.hapViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_20hapViterbi(__pyx_self, __pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_hh);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_20hapViterbi(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("hapViterbi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6hapflk_9fastphase_hapViterbi(__pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_hh, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.hapViterbi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastphase/fastphase.pyx":943
 *     return soluce
 * 
 * cpdef hapCalc(aa,tt,rr,hh,u2p):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

static PyObject *__pyx_pw_6hapflk_9fastphase_23hapCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6hapflk_9fastphase_hapCalc(PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh, PyObject *__pyx_v_u2p, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_alpha = 0;
  PyArrayObject *__pyx_v_theta = 0;
  PyArrayObject *__pyx_v_rho = 0;
  PyArrayObject *__pyx_v_hap = 0;
  int __pyx_v_up2pz;
  int __pyx_v_nLoc;
  int __pyx_v_nK;
  int __pyx_v_tScale;
  int __pyx_v_k;
  int __pyx_v_m;
  double __pyx_v_dummy;
  double __pyx_v_tScaleTemp;
  double __pyx_v_temp;
  double __pyx_v_normC;
  double __pyx_v_logLikelihood;
  PyArrayObject *__pyx_v_betaScale = 0;
  PyArrayObject *__pyx_v_tSum = 0;
  PyArrayObject *__pyx_v_mBeta = 0;
  PyArrayObject *__pyx_v_mPhi = 0;
  PyArrayObject *__pyx_v_phiScale = 0;
  PyArrayObject *__pyx_v_probZ = 0;
  PyArrayObject *__pyx_v_jmk = 0;
  PyArrayObject *__pyx_v_top = 0;
  PyArrayObject *__pyx_v_bot = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_alpha;
  __Pyx_Buffer __pyx_pybuffer_alpha;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_betaScale;
  __Pyx_Buffer __pyx_pybuffer_betaScale;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bot;
  __Pyx_Buffer __pyx_pybuffer_bot;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hap;
  __Pyx_Buffer __pyx_pybuffer_hap;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_jmk;
  __Pyx_Buffer __pyx_pybuffer_jmk;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mBeta;
  __Pyx_Buffer __pyx_pybuffer_mBeta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mPhi;
  __Pyx_Buffer __pyx_pybuffer_mPhi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phiScale;
  __Pyx_Buffer __pyx_pybuffer_phiScale;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_probZ;
  __Pyx_Buffer __pyx_pybuffer_probZ;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rho;
  __Pyx_Buffer __pyx_pybuffer_rho;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tSum;
  __Pyx_Buffer __pyx_pybuffer_tSum;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_theta;
  __Pyx_Buffer __pyx_pybuffer_theta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_top;
  __Pyx_Buffer __pyx_pybuffer_top;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  double __pyx_t_43;
  double __pyx_t_44;
  int __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  PyArrayObject *__pyx_t_51 = NULL;
  PyArrayObject *__pyx_t_52 = NULL;
  Py_ssize_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  Py_ssize_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  long __pyx_t_64;
  long __pyx_t_65;
  Py_ssize_t __pyx_t_66;
  Py_ssize_t __pyx_t_67;
  Py_ssize_t __pyx_t_68;
  Py_ssize_t __pyx_t_69;
  Py_ssize_t __pyx_t_70;
  Py_ssize_t __pyx_t_71;
  Py_ssize_t __pyx_t_72;
  Py_ssize_t __pyx_t_73;
  Py_ssize_t __pyx_t_74;
  Py_ssize_t __pyx_t_75;
  Py_ssize_t __pyx_t_76;
  Py_ssize_t __pyx_t_77;
  Py_ssize_t __pyx_t_78;
  Py_ssize_t __pyx_t_79;
  Py_ssize_t __pyx_t_80;
  Py_ssize_t __pyx_t_81;
  Py_ssize_t __pyx_t_82;
  Py_ssize_t __pyx_t_83;
  Py_ssize_t __pyx_t_84;
  Py_ssize_t __pyx_t_85;
  Py_ssize_t __pyx_t_86;
  Py_ssize_t __pyx_t_87;
  Py_ssize_t __pyx_t_88;
  Py_ssize_t __pyx_t_89;
  Py_ssize_t __pyx_t_90;
  Py_ssize_t __pyx_t_91;
  Py_ssize_t __pyx_t_92;
  Py_ssize_t __pyx_t_93;
  Py_ssize_t __pyx_t_94;
  PyArrayObject *__pyx_t_95 = NULL;
  PyArrayObject *__pyx_t_96 = NULL;
  PyArrayObject *__pyx_t_97 = NULL;
  PyArrayObject *__pyx_t_98 = NULL;
  int __pyx_t_99;
  int __pyx_t_100;
  Py_ssize_t __pyx_t_101;
  Py_ssize_t __pyx_t_102;
  Py_ssize_t __pyx_t_103;
  Py_ssize_t __pyx_t_104;
  Py_ssize_t __pyx_t_105;
  Py_ssize_t __pyx_t_106;
  Py_ssize_t __pyx_t_107;
  Py_ssize_t __pyx_t_108;
  Py_ssize_t __pyx_t_109;
  Py_ssize_t __pyx_t_110;
  Py_ssize_t __pyx_t_111;
  Py_ssize_t __pyx_t_112;
  Py_ssize_t __pyx_t_113;
  Py_ssize_t __pyx_t_114;
  Py_ssize_t __pyx_t_115;
  Py_ssize_t __pyx_t_116;
  Py_ssize_t __pyx_t_117;
  Py_ssize_t __pyx_t_118;
  Py_ssize_t __pyx_t_119;
  Py_ssize_t __pyx_t_120;
  Py_ssize_t __pyx_t_121;
  Py_ssize_t __pyx_t_122;
  Py_ssize_t __pyx_t_123;
  Py_ssize_t __pyx_t_124;
  Py_ssize_t __pyx_t_125;
  Py_ssize_t __pyx_t_126;
  Py_ssize_t __pyx_t_127;
  Py_ssize_t __pyx_t_128;
  Py_ssize_t __pyx_t_129;
  Py_ssize_t __pyx_t_130;
  Py_ssize_t __pyx_t_131;
  Py_ssize_t __pyx_t_132;
  Py_ssize_t __pyx_t_133;
  Py_ssize_t __pyx_t_134;
  Py_ssize_t __pyx_t_135;
  Py_ssize_t __pyx_t_136;
  Py_ssize_t __pyx_t_137;
  Py_ssize_t __pyx_t_138;
  Py_ssize_t __pyx_t_139;
  Py_ssize_t __pyx_t_140;
  Py_ssize_t __pyx_t_141;
  Py_ssize_t __pyx_t_142;
  Py_ssize_t __pyx_t_143;
  Py_ssize_t __pyx_t_144;
  __Pyx_RefNannySetupContext("hapCalc", 0);
  __pyx_pybuffer_alpha.pybuffer.buf = NULL;
  __pyx_pybuffer_alpha.refcount = 0;
  __pyx_pybuffernd_alpha.data = NULL;
  __pyx_pybuffernd_alpha.rcbuffer = &__pyx_pybuffer_alpha;
  __pyx_pybuffer_theta.pybuffer.buf = NULL;
  __pyx_pybuffer_theta.refcount = 0;
  __pyx_pybuffernd_theta.data = NULL;
  __pyx_pybuffernd_theta.rcbuffer = &__pyx_pybuffer_theta;
  __pyx_pybuffer_rho.pybuffer.buf = NULL;
  __pyx_pybuffer_rho.refcount = 0;
  __pyx_pybuffernd_rho.data = NULL;
  __pyx_pybuffernd_rho.rcbuffer = &__pyx_pybuffer_rho;
  __pyx_pybuffer_hap.pybuffer.buf = NULL;
  __pyx_pybuffer_hap.refcount = 0;
  __pyx_pybuffernd_hap.data = NULL;
  __pyx_pybuffernd_hap.rcbuffer = &__pyx_pybuffer_hap;
  __pyx_pybuffer_betaScale.pybuffer.buf = NULL;
  __pyx_pybuffer_betaScale.refcount = 0;
  __pyx_pybuffernd_betaScale.data = NULL;
  __pyx_pybuffernd_betaScale.rcbuffer = &__pyx_pybuffer_betaScale;
  __pyx_pybuffer_tSum.pybuffer.buf = NULL;
  __pyx_pybuffer_tSum.refcount = 0;
  __pyx_pybuffernd_tSum.data = NULL;
  __pyx_pybuffernd_tSum.rcbuffer = &__pyx_pybuffer_tSum;
  __pyx_pybuffer_mBeta.pybuffer.buf = NULL;
  __pyx_pybuffer_mBeta.refcount = 0;
  __pyx_pybuffernd_mBeta.data = NULL;
  __pyx_pybuffernd_mBeta.rcbuffer = &__pyx_pybuffer_mBeta;
  __pyx_pybuffer_mPhi.pybuffer.buf = NULL;
  __pyx_pybuffer_mPhi.refcount = 0;
  __pyx_pybuffernd_mPhi.data = NULL;
  __pyx_pybuffernd_mPhi.rcbuffer = &__pyx_pybuffer_mPhi;
  __pyx_pybuffer_phiScale.pybuffer.buf = NULL;
  __pyx_pybuffer_phiScale.refcount = 0;
  __pyx_pybuffernd_phiScale.data = NULL;
  __pyx_pybuffernd_phiScale.rcbuffer = &__pyx_pybuffer_phiScale;
  __pyx_pybuffer_probZ.pybuffer.buf = NULL;
  __pyx_pybuffer_probZ.refcount = 0;
  __pyx_pybuffernd_probZ.data = NULL;
  __pyx_pybuffernd_probZ.rcbuffer = &__pyx_pybuffer_probZ;
  __pyx_pybuffer_jmk.pybuffer.buf = NULL;
  __pyx_pybuffer_jmk.refcount = 0;
  __pyx_pybuffernd_jmk.data = NULL;
  __pyx_pybuffernd_jmk.rcbuffer = &__pyx_pybuffer_jmk;
  __pyx_pybuffer_top.pybuffer.buf = NULL;
  __pyx_pybuffer_top.refcount = 0;
  __pyx_pybuffernd_top.data = NULL;
  __pyx_pybuffernd_top.rcbuffer = &__pyx_pybuffer_top;
  __pyx_pybuffer_bot.pybuffer.buf = NULL;
  __pyx_pybuffer_bot.refcount = 0;
  __pyx_pybuffernd_bot.data = NULL;
  __pyx_pybuffernd_bot.rcbuffer = &__pyx_pybuffer_bot;

  /* "fastphase/fastphase.pyx":944
 * 
 * cpdef hapCalc(aa,tt,rr,hh,u2p):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 */
  if (!(likely(((__pyx_v_aa) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_aa, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 944, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_aa;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_alpha = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 944, __pyx_L1_error)
    } else {__pyx_pybuffernd_alpha.diminfo[0].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_alpha.diminfo[0].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_alpha.diminfo[1].strides = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_alpha.diminfo[1].shape = __pyx_pybuffernd_alpha.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_alpha = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":945
 * cpdef hapCalc(aa,tt,rr,hh,u2p):
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh
 */
  if (!(likely(((__pyx_v_tt) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_tt, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 945, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_tt;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_theta.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_theta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 945, __pyx_L1_error)
    } else {__pyx_pybuffernd_theta.diminfo[0].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_theta.diminfo[0].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_theta.diminfo[1].strides = __pyx_pybuffernd_theta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_theta.diminfo[1].shape = __pyx_pybuffernd_theta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_theta = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":946
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh
 *     cdef int up2pz=u2p
 */
  if (!(likely(((__pyx_v_rr) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_rr, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 946, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_rr;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rho.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_rho = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 946, __pyx_L1_error)
    } else {__pyx_pybuffernd_rho.diminfo[0].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rho.diminfo[0].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_rho.diminfo[1].strides = __pyx_pybuffernd_rho.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_rho.diminfo[1].shape = __pyx_pybuffernd_rho.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_rho = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":947
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh             # <<<<<<<<<<<<<<
 *     cdef int up2pz=u2p
 *     ## cython declarations
 */
  if (!(likely(((__pyx_v_hh) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_hh, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 947, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_hh;
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hap.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_hap = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 947, __pyx_L1_error)
    } else {__pyx_pybuffernd_hap.diminfo[0].strides = __pyx_pybuffernd_hap.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hap.diminfo[0].shape = __pyx_pybuffernd_hap.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_v_hap = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":948
 *     cdef np.ndarray[np.float64_t, ndim=2] rho = rr
 *     cdef np.ndarray[np.int_t, ndim=1] hap = hh
 *     cdef int up2pz=u2p             # <<<<<<<<<<<<<<
 *     ## cython declarations
 *     cdef int nLoc,nK,tScale
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_u2p); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L1_error)
  __pyx_v_up2pz = __pyx_t_2;

  /* "fastphase/fastphase.pyx":956
 *     cdef double logLikelihood
 *     ## end cython declarations
 *     nLoc=alpha.shape[0]             # <<<<<<<<<<<<<<
 *     nK=alpha.shape[1]
 *     ##
 */
  __pyx_v_nLoc = (__pyx_v_alpha->dimensions[0]);

  /* "fastphase/fastphase.pyx":957
 *     ## end cython declarations
 *     nLoc=alpha.shape[0]
 *     nK=alpha.shape[1]             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute backward probabilities
 */
  __pyx_v_nK = (__pyx_v_alpha->dimensions[1]);

  /* "fastphase/fastphase.pyx":961
 *     ## compute backward probabilities
 *     ##
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=1] tSum=np.zeros(nLoc,dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] mBeta=np.zeros((nLoc,nK),dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 961, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_6);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_betaScale = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 961, __pyx_L1_error)
    } else {__pyx_pybuffernd_betaScale.diminfo[0].strides = __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_betaScale.diminfo[0].shape = __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_betaScale = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "fastphase/fastphase.pyx":962
 *     ##
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
 *     cdef np.ndarray[np.float64_t,ndim=1] tSum=np.zeros(nLoc,dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] mBeta=np.zeros((nLoc,nK),dtype=np.float64)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 962, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tSum.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tSum = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 962, __pyx_L1_error)
    } else {__pyx_pybuffernd_tSum.diminfo[0].strides = __pyx_pybuffernd_tSum.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tSum.diminfo[0].shape = __pyx_pybuffernd_tSum.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_tSum = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":963
 *     cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
 *     cdef np.ndarray[np.float64_t,ndim=1] tSum=np.zeros(nLoc,dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] mBeta=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     for k in range(nK):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 963, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_mBeta = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 963, __pyx_L1_error)
    } else {__pyx_pybuffernd_mBeta.diminfo[0].strides = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mBeta.diminfo[0].shape = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mBeta.diminfo[1].strides = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mBeta.diminfo[1].shape = __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_mBeta = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":965
 *     cdef np.ndarray[np.float64_t,ndim=2] mBeta=np.zeros((nLoc,nK),dtype=np.float64)
 * 
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         mBeta[nLoc-1,k]=1
 * 
 */
  __pyx_t_2 = __pyx_v_nK;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_k = __pyx_t_11;

    /* "fastphase/fastphase.pyx":966
 * 
 *     for k in range(nK):
 *         mBeta[nLoc-1,k]=1             # <<<<<<<<<<<<<<
 * 
 *     for k in range(nK):
 */
    __pyx_t_12 = (__pyx_v_nLoc - 1);
    __pyx_t_13 = __pyx_v_k;
    if (__pyx_t_12 < 0) __pyx_t_12 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
    if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_mBeta.diminfo[1].strides) = 1.0;
  }

  /* "fastphase/fastphase.pyx":968
 *         mBeta[nLoc-1,k]=1
 * 
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         tSum[nLoc-1] += happrG(theta[nLoc-1,k],hap[nLoc-1])*alpha[nLoc-1,k]
 *     for m in range(nLoc-1,0,-1):
 */
  __pyx_t_2 = __pyx_v_nK;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_k = __pyx_t_11;

    /* "fastphase/fastphase.pyx":969
 * 
 *     for k in range(nK):
 *         tSum[nLoc-1] += happrG(theta[nLoc-1,k],hap[nLoc-1])*alpha[nLoc-1,k]             # <<<<<<<<<<<<<<
 *     for m in range(nLoc-1,0,-1):
 *         tScaleTemp=0
 */
    __pyx_t_14 = (__pyx_v_nLoc - 1);
    __pyx_t_15 = __pyx_v_k;
    if (__pyx_t_14 < 0) __pyx_t_14 += __pyx_pybuffernd_theta.diminfo[0].shape;
    if (__pyx_t_15 < 0) __pyx_t_15 += __pyx_pybuffernd_theta.diminfo[1].shape;
    __pyx_t_16 = (__pyx_v_nLoc - 1);
    if (__pyx_t_16 < 0) __pyx_t_16 += __pyx_pybuffernd_hap.diminfo[0].shape;
    __pyx_t_17 = (__pyx_v_nLoc - 1);
    __pyx_t_18 = __pyx_v_k;
    if (__pyx_t_17 < 0) __pyx_t_17 += __pyx_pybuffernd_alpha.diminfo[0].shape;
    if (__pyx_t_18 < 0) __pyx_t_18 += __pyx_pybuffernd_alpha.diminfo[1].shape;
    __pyx_t_19 = (__pyx_v_nLoc - 1);
    if (__pyx_t_19 < 0) __pyx_t_19 += __pyx_pybuffernd_tSum.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_tSum.diminfo[0].strides) += (__pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_hap.diminfo[0].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_alpha.diminfo[1].strides)));
  }

  /* "fastphase/fastphase.pyx":970
 *     for k in range(nK):
 *         tSum[nLoc-1] += happrG(theta[nLoc-1,k],hap[nLoc-1])*alpha[nLoc-1,k]
 *     for m in range(nLoc-1,0,-1):             # <<<<<<<<<<<<<<
 *         tScaleTemp=0
 *         ## this loop can be parallelized across clusters #CUDA
 */
  for (__pyx_t_2 = (__pyx_v_nLoc - 1); __pyx_t_2 > 0; __pyx_t_2-=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":971
 *         tSum[nLoc-1] += happrG(theta[nLoc-1,k],hap[nLoc-1])*alpha[nLoc-1,k]
 *     for m in range(nLoc-1,0,-1):
 *         tScaleTemp=0             # <<<<<<<<<<<<<<
 *         ## this loop can be parallelized across clusters #CUDA
 *         for k in range(nK):
 */
    __pyx_v_tScaleTemp = 0.0;

    /* "fastphase/fastphase.pyx":973
 *         tScaleTemp=0
 *         ## this loop can be parallelized across clusters #CUDA
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             temp=(1.0-rho[m,0])*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             mBeta[m-1,k]=temp+rho[m,0]*tSum[m]
 */
    __pyx_t_10 = __pyx_v_nK;
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_11; __pyx_t_20+=1) {
      __pyx_v_k = __pyx_t_20;

      /* "fastphase/fastphase.pyx":974
 *         ## this loop can be parallelized across clusters #CUDA
 *         for k in range(nK):
 *             temp=(1.0-rho[m,0])*happrG(theta[m,k],hap[m])*mBeta[m,k]             # <<<<<<<<<<<<<<
 *             mBeta[m-1,k]=temp+rho[m,0]*tSum[m]
 *             tSum[m-1]+=happrG(theta[m-1,k],hap[m-1])*mBeta[m-1,k]*alpha[m-1,k]
 */
      __pyx_t_21 = __pyx_v_m;
      __pyx_t_22 = 0;
      if (__pyx_t_21 < 0) __pyx_t_21 += __pyx_pybuffernd_rho.diminfo[0].shape;
      if (__pyx_t_22 < 0) __pyx_t_22 += __pyx_pybuffernd_rho.diminfo[1].shape;
      __pyx_t_23 = __pyx_v_m;
      __pyx_t_24 = __pyx_v_k;
      if (__pyx_t_23 < 0) __pyx_t_23 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_24 < 0) __pyx_t_24 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_25 = __pyx_v_m;
      if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_hap.diminfo[0].shape;
      __pyx_t_26 = __pyx_v_m;
      __pyx_t_27 = __pyx_v_k;
      if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
      if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
      __pyx_v_temp = (((1.0 - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_rho.diminfo[1].strides))) * __pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_hap.diminfo[0].strides)))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_mBeta.diminfo[1].strides)));

      /* "fastphase/fastphase.pyx":975
 *         for k in range(nK):
 *             temp=(1.0-rho[m,0])*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             mBeta[m-1,k]=temp+rho[m,0]*tSum[m]             # <<<<<<<<<<<<<<
 *             tSum[m-1]+=happrG(theta[m-1,k],hap[m-1])*mBeta[m-1,k]*alpha[m-1,k]
 *             tScaleTemp+=mBeta[m-1,k]
 */
      __pyx_t_28 = __pyx_v_m;
      __pyx_t_29 = 0;
      if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_rho.diminfo[0].shape;
      if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_rho.diminfo[1].shape;
      __pyx_t_30 = __pyx_v_m;
      if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      __pyx_t_31 = (__pyx_v_m - 1);
      __pyx_t_32 = __pyx_v_k;
      if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
      if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_32, __pyx_pybuffernd_mBeta.diminfo[1].strides) = (__pyx_v_temp + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_rho.diminfo[1].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_tSum.diminfo[0].strides))));

      /* "fastphase/fastphase.pyx":976
 *             temp=(1.0-rho[m,0])*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             mBeta[m-1,k]=temp+rho[m,0]*tSum[m]
 *             tSum[m-1]+=happrG(theta[m-1,k],hap[m-1])*mBeta[m-1,k]*alpha[m-1,k]             # <<<<<<<<<<<<<<
 *             tScaleTemp+=mBeta[m-1,k]
 *         # if np.isnan(tScaleTemp):
 */
      __pyx_t_33 = (__pyx_v_m - 1);
      __pyx_t_34 = __pyx_v_k;
      if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_35 = (__pyx_v_m - 1);
      if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_hap.diminfo[0].shape;
      __pyx_t_36 = (__pyx_v_m - 1);
      __pyx_t_37 = __pyx_v_k;
      if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
      if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
      __pyx_t_38 = (__pyx_v_m - 1);
      __pyx_t_39 = __pyx_v_k;
      if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_40 = (__pyx_v_m - 1);
      if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_40, __pyx_pybuffernd_tSum.diminfo[0].strides) += ((__pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_34, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_hap.diminfo[0].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_37, __pyx_pybuffernd_mBeta.diminfo[1].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_39, __pyx_pybuffernd_alpha.diminfo[1].strides)));

      /* "fastphase/fastphase.pyx":977
 *             mBeta[m-1,k]=temp+rho[m,0]*tSum[m]
 *             tSum[m-1]+=happrG(theta[m-1,k],hap[m-1])*mBeta[m-1,k]*alpha[m-1,k]
 *             tScaleTemp+=mBeta[m-1,k]             # <<<<<<<<<<<<<<
 *         # if np.isnan(tScaleTemp):
 *         #     print mBeta[m-1]
 */
      __pyx_t_41 = (__pyx_v_m - 1);
      __pyx_t_42 = __pyx_v_k;
      if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
      if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
      __pyx_v_tScaleTemp = (__pyx_v_tScaleTemp + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_mBeta.diminfo[1].strides)));
    }

    /* "fastphase/fastphase.pyx":986
 *         #     print hap[m]
 *         #     print 'BLA'
 *         tScale=int(-log(tScaleTemp)/log(10))             # <<<<<<<<<<<<<<
 *         if tScale<0:
 *             tScale=0
 */
    __pyx_t_43 = (-log(__pyx_v_tScaleTemp));
    __pyx_t_44 = log(10.0);
    if (unlikely(__pyx_t_44 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 986, __pyx_L1_error)
    }
    __pyx_v_tScale = ((int)(__pyx_t_43 / __pyx_t_44));

    /* "fastphase/fastphase.pyx":987
 *         #     print 'BLA'
 *         tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale<0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 */
    __pyx_t_45 = ((__pyx_v_tScale < 0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":988
 *         tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale<0:
 *             tScale=0             # <<<<<<<<<<<<<<
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale >0:
 */
      __pyx_v_tScale = 0;

      /* "fastphase/fastphase.pyx":987
 *         #     print 'BLA'
 *         tScale=int(-log(tScaleTemp)/log(10))
 *         if tScale<0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 */
    }

    /* "fastphase/fastphase.pyx":989
 *         if tScale<0:
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale             # <<<<<<<<<<<<<<
 *         if tScale >0:
 *             dummy=myPow10(tScale)
 */
    __pyx_t_46 = __pyx_v_m;
    if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    __pyx_t_47 = (__pyx_v_m - 1);
    if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_betaScale.diminfo[0].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_46, __pyx_pybuffernd_betaScale.diminfo[0].strides)) + __pyx_v_tScale);

    /* "fastphase/fastphase.pyx":990
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale >0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tSum[m-1]*=dummy
 */
    __pyx_t_45 = ((__pyx_v_tScale > 0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":991
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale >0:
 *             dummy=myPow10(tScale)             # <<<<<<<<<<<<<<
 *             tSum[m-1]*=dummy
 *             for k in range(nK):
 */
      __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_v_tScale);

      /* "fastphase/fastphase.pyx":992
 *         if tScale >0:
 *             dummy=myPow10(tScale)
 *             tSum[m-1]*=dummy             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 mBeta[m-1,k]*=dummy
 */
      __pyx_t_48 = (__pyx_v_m - 1);
      if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_tSum.diminfo[0].strides) *= __pyx_v_dummy;

      /* "fastphase/fastphase.pyx":993
 *             dummy=myPow10(tScale)
 *             tSum[m-1]*=dummy
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 mBeta[m-1,k]*=dummy
 *     ##
 */
      __pyx_t_10 = __pyx_v_nK;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_11; __pyx_t_20+=1) {
        __pyx_v_k = __pyx_t_20;

        /* "fastphase/fastphase.pyx":994
 *             tSum[m-1]*=dummy
 *             for k in range(nK):
 *                 mBeta[m-1,k]*=dummy             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute forward probabilities
 */
        __pyx_t_49 = (__pyx_v_m - 1);
        __pyx_t_50 = __pyx_v_k;
        if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
        if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_50, __pyx_pybuffernd_mBeta.diminfo[1].strides) *= __pyx_v_dummy;
      }

      /* "fastphase/fastphase.pyx":990
 *             tScale=0
 *         betaScale[m-1]=betaScale[m]+tScale
 *         if tScale >0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tSum[m-1]*=dummy
 */
    }
  }

  /* "fastphase/fastphase.pyx":998
 *     ## compute forward probabilities
 *     ##
 *     cdef np.ndarray[np.float64_t,ndim=2] mPhi=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
 *     for k in range(nK):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 998, __pyx_L1_error)
  __pyx_t_51 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer, (PyObject*)__pyx_t_51, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_mPhi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 998, __pyx_L1_error)
    } else {__pyx_pybuffernd_mPhi.diminfo[0].strides = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mPhi.diminfo[0].shape = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mPhi.diminfo[1].strides = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mPhi.diminfo[1].shape = __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_51 = 0;
  __pyx_v_mPhi = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":999
 *     ##
 *     cdef np.ndarray[np.float64_t,ndim=2] mPhi=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)             # <<<<<<<<<<<<<<
 *     for k in range(nK):
 *         mPhi[0,k]=alpha[0,k]*happrG(theta[0,k],hap[0])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 999, __pyx_L1_error)
  __pyx_t_52 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer, (PyObject*)__pyx_t_52, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_phiScale = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 999, __pyx_L1_error)
    } else {__pyx_pybuffernd_phiScale.diminfo[0].strides = __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phiScale.diminfo[0].shape = __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_52 = 0;
  __pyx_v_phiScale = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":1000
 *     cdef np.ndarray[np.float64_t,ndim=2] mPhi=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         mPhi[0,k]=alpha[0,k]*happrG(theta[0,k],hap[0])
 *     ## calc the marginal sum at locus 0 (appx A)
 */
  __pyx_t_2 = __pyx_v_nK;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_k = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1001
 *     cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
 *     for k in range(nK):
 *         mPhi[0,k]=alpha[0,k]*happrG(theta[0,k],hap[0])             # <<<<<<<<<<<<<<
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tSum[0]=0
 */
    __pyx_t_53 = 0;
    __pyx_t_54 = __pyx_v_k;
    if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_alpha.diminfo[0].shape;
    if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_alpha.diminfo[1].shape;
    __pyx_t_55 = 0;
    __pyx_t_56 = __pyx_v_k;
    if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_theta.diminfo[0].shape;
    if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_theta.diminfo[1].shape;
    __pyx_t_57 = 0;
    if (__pyx_t_57 < 0) __pyx_t_57 += __pyx_pybuffernd_hap.diminfo[0].shape;
    __pyx_t_58 = 0;
    __pyx_t_59 = __pyx_v_k;
    if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
    if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_59, __pyx_pybuffernd_mPhi.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_54, __pyx_pybuffernd_alpha.diminfo[1].strides)) * __pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_55, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_56, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_57, __pyx_pybuffernd_hap.diminfo[0].strides))));
  }

  /* "fastphase/fastphase.pyx":1003
 *         mPhi[0,k]=alpha[0,k]*happrG(theta[0,k],hap[0])
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tSum[0]=0             # <<<<<<<<<<<<<<
 *     for k in range(nK):
 *         tSum[0]+=mPhi[0,k]
 */
  __pyx_t_60 = 0;
  if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_tSum.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_60, __pyx_pybuffernd_tSum.diminfo[0].strides) = 0.0;

  /* "fastphase/fastphase.pyx":1004
 *     ## calc the marginal sum at locus 0 (appx A)
 *     tSum[0]=0
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         tSum[0]+=mPhi[0,k]
 *     ## calc Phi
 */
  __pyx_t_2 = __pyx_v_nK;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_k = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1005
 *     tSum[0]=0
 *     for k in range(nK):
 *         tSum[0]+=mPhi[0,k]             # <<<<<<<<<<<<<<
 *     ## calc Phi
 *     for m in range(nLoc-1):
 */
    __pyx_t_61 = 0;
    __pyx_t_62 = __pyx_v_k;
    if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
    if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
    __pyx_t_63 = 0;
    if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_tSum.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_63, __pyx_pybuffernd_tSum.diminfo[0].strides) += (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_61, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_62, __pyx_pybuffernd_mPhi.diminfo[1].strides));
  }

  /* "fastphase/fastphase.pyx":1007
 *         tSum[0]+=mPhi[0,k]
 *     ## calc Phi
 *     for m in range(nLoc-1):             # <<<<<<<<<<<<<<
 *         tSum[m+1]=0
 *         ## this loop could be parallelized across clusters #CUDA
 */
  __pyx_t_64 = (__pyx_v_nLoc - 1);
  __pyx_t_65 = __pyx_t_64;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_65; __pyx_t_2+=1) {
    __pyx_v_m = __pyx_t_2;

    /* "fastphase/fastphase.pyx":1008
 *     ## calc Phi
 *     for m in range(nLoc-1):
 *         tSum[m+1]=0             # <<<<<<<<<<<<<<
 *         ## this loop could be parallelized across clusters #CUDA
 *         for k in range(nK):
 */
    __pyx_t_66 = (__pyx_v_m + 1);
    if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_tSum.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_tSum.diminfo[0].strides) = 0.0;

    /* "fastphase/fastphase.pyx":1010
 *         tSum[m+1]=0
 *         ## this loop could be parallelized across clusters #CUDA
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             temp=(1-rho[m+1,0])*mPhi[m,k]+rho[m+1,0]*alpha[m+1,k]*tSum[m]
 *             mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])
 */
    __pyx_t_10 = __pyx_v_nK;
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_11; __pyx_t_20+=1) {
      __pyx_v_k = __pyx_t_20;

      /* "fastphase/fastphase.pyx":1011
 *         ## this loop could be parallelized across clusters #CUDA
 *         for k in range(nK):
 *             temp=(1-rho[m+1,0])*mPhi[m,k]+rho[m+1,0]*alpha[m+1,k]*tSum[m]             # <<<<<<<<<<<<<<
 *             mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])
 *             tSum[m+1]+=mPhi[m+1,k]
 */
      __pyx_t_67 = (__pyx_v_m + 1);
      __pyx_t_68 = 0;
      if (__pyx_t_67 < 0) __pyx_t_67 += __pyx_pybuffernd_rho.diminfo[0].shape;
      if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_rho.diminfo[1].shape;
      __pyx_t_69 = __pyx_v_m;
      __pyx_t_70 = __pyx_v_k;
      if (__pyx_t_69 < 0) __pyx_t_69 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_70 < 0) __pyx_t_70 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      __pyx_t_71 = (__pyx_v_m + 1);
      __pyx_t_72 = 0;
      if (__pyx_t_71 < 0) __pyx_t_71 += __pyx_pybuffernd_rho.diminfo[0].shape;
      if (__pyx_t_72 < 0) __pyx_t_72 += __pyx_pybuffernd_rho.diminfo[1].shape;
      __pyx_t_73 = (__pyx_v_m + 1);
      __pyx_t_74 = __pyx_v_k;
      if (__pyx_t_73 < 0) __pyx_t_73 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_74 < 0) __pyx_t_74 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_75 = __pyx_v_m;
      if (__pyx_t_75 < 0) __pyx_t_75 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      __pyx_v_temp = (((1.0 - (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_67, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_68, __pyx_pybuffernd_rho.diminfo[1].strides))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_69, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_70, __pyx_pybuffernd_mPhi.diminfo[1].strides))) + (((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_72, __pyx_pybuffernd_rho.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_73, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_74, __pyx_pybuffernd_alpha.diminfo[1].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_75, __pyx_pybuffernd_tSum.diminfo[0].strides))));

      /* "fastphase/fastphase.pyx":1012
 *         for k in range(nK):
 *             temp=(1-rho[m+1,0])*mPhi[m,k]+rho[m+1,0]*alpha[m+1,k]*tSum[m]
 *             mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])             # <<<<<<<<<<<<<<
 *             tSum[m+1]+=mPhi[m+1,k]
 *         tScale=0
 */
      __pyx_t_76 = (__pyx_v_m + 1);
      __pyx_t_77 = __pyx_v_k;
      if (__pyx_t_76 < 0) __pyx_t_76 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_77 < 0) __pyx_t_77 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_78 = (__pyx_v_m + 1);
      if (__pyx_t_78 < 0) __pyx_t_78 += __pyx_pybuffernd_hap.diminfo[0].shape;
      __pyx_t_79 = (__pyx_v_m + 1);
      __pyx_t_80 = __pyx_v_k;
      if (__pyx_t_79 < 0) __pyx_t_79 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_80 < 0) __pyx_t_80 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_79, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_80, __pyx_pybuffernd_mPhi.diminfo[1].strides) = (__pyx_v_temp * __pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_76, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_77, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_78, __pyx_pybuffernd_hap.diminfo[0].strides))));

      /* "fastphase/fastphase.pyx":1013
 *             temp=(1-rho[m+1,0])*mPhi[m,k]+rho[m+1,0]*alpha[m+1,k]*tSum[m]
 *             mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])
 *             tSum[m+1]+=mPhi[m+1,k]             # <<<<<<<<<<<<<<
 *         tScale=0
 *         if tSum[m+1] < 0:
 */
      __pyx_t_81 = (__pyx_v_m + 1);
      __pyx_t_82 = __pyx_v_k;
      if (__pyx_t_81 < 0) __pyx_t_81 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
      if (__pyx_t_82 < 0) __pyx_t_82 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
      __pyx_t_83 = (__pyx_v_m + 1);
      if (__pyx_t_83 < 0) __pyx_t_83 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_83, __pyx_pybuffernd_tSum.diminfo[0].strides) += (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_81, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_82, __pyx_pybuffernd_mPhi.diminfo[1].strides));
    }

    /* "fastphase/fastphase.pyx":1014
 *             mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])
 *             tSum[m+1]+=mPhi[m+1,k]
 *         tScale=0             # <<<<<<<<<<<<<<
 *         if tSum[m+1] < 0:
 *             phiScale[m+1]=phiScale[m]
 */
    __pyx_v_tScale = 0;

    /* "fastphase/fastphase.pyx":1015
 *             tSum[m+1]+=mPhi[m+1,k]
 *         tScale=0
 *         if tSum[m+1] < 0:             # <<<<<<<<<<<<<<
 *             phiScale[m+1]=phiScale[m]
 *         else:
 */
    __pyx_t_84 = (__pyx_v_m + 1);
    if (__pyx_t_84 < 0) __pyx_t_84 += __pyx_pybuffernd_tSum.diminfo[0].shape;
    __pyx_t_45 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_84, __pyx_pybuffernd_tSum.diminfo[0].strides)) < 0.0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":1016
 *         tScale=0
 *         if tSum[m+1] < 0:
 *             phiScale[m+1]=phiScale[m]             # <<<<<<<<<<<<<<
 *         else:
 *             tScale=int(-log(tSum[m+1])/log(10))
 */
      __pyx_t_85 = __pyx_v_m;
      if (__pyx_t_85 < 0) __pyx_t_85 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      __pyx_t_86 = (__pyx_v_m + 1);
      if (__pyx_t_86 < 0) __pyx_t_86 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_86, __pyx_pybuffernd_phiScale.diminfo[0].strides) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_85, __pyx_pybuffernd_phiScale.diminfo[0].strides));

      /* "fastphase/fastphase.pyx":1015
 *             tSum[m+1]+=mPhi[m+1,k]
 *         tScale=0
 *         if tSum[m+1] < 0:             # <<<<<<<<<<<<<<
 *             phiScale[m+1]=phiScale[m]
 *         else:
 */
      goto __pyx_L23;
    }

    /* "fastphase/fastphase.pyx":1018
 *             phiScale[m+1]=phiScale[m]
 *         else:
 *             tScale=int(-log(tSum[m+1])/log(10))             # <<<<<<<<<<<<<<
 *         if tScale < 0:
 *             tScale=0
 */
    /*else*/ {
      __pyx_t_87 = (__pyx_v_m + 1);
      if (__pyx_t_87 < 0) __pyx_t_87 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      __pyx_t_44 = (-log((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_87, __pyx_pybuffernd_tSum.diminfo[0].strides))));
      __pyx_t_43 = log(10.0);
      if (unlikely(__pyx_t_43 == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 1018, __pyx_L1_error)
      }
      __pyx_v_tScale = ((int)(__pyx_t_44 / __pyx_t_43));
    }
    __pyx_L23:;

    /* "fastphase/fastphase.pyx":1019
 *         else:
 *             tScale=int(-log(tSum[m+1])/log(10))
 *         if tScale < 0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         phiScale[m+1]=phiScale[m]+tScale
 */
    __pyx_t_45 = ((__pyx_v_tScale < 0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":1020
 *             tScale=int(-log(tSum[m+1])/log(10))
 *         if tScale < 0:
 *             tScale=0             # <<<<<<<<<<<<<<
 *         phiScale[m+1]=phiScale[m]+tScale
 *         if tScale > 0:
 */
      __pyx_v_tScale = 0;

      /* "fastphase/fastphase.pyx":1019
 *         else:
 *             tScale=int(-log(tSum[m+1])/log(10))
 *         if tScale < 0:             # <<<<<<<<<<<<<<
 *             tScale=0
 *         phiScale[m+1]=phiScale[m]+tScale
 */
    }

    /* "fastphase/fastphase.pyx":1021
 *         if tScale < 0:
 *             tScale=0
 *         phiScale[m+1]=phiScale[m]+tScale             # <<<<<<<<<<<<<<
 *         if tScale > 0:
 *             dummy=myPow10(tScale)
 */
    __pyx_t_88 = __pyx_v_m;
    if (__pyx_t_88 < 0) __pyx_t_88 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    __pyx_t_89 = (__pyx_v_m + 1);
    if (__pyx_t_89 < 0) __pyx_t_89 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_89, __pyx_pybuffernd_phiScale.diminfo[0].strides) = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_phiScale.diminfo[0].strides)) + __pyx_v_tScale);

    /* "fastphase/fastphase.pyx":1022
 *             tScale=0
 *         phiScale[m+1]=phiScale[m]+tScale
 *         if tScale > 0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tSum[m+1]*=dummy
 */
    __pyx_t_45 = ((__pyx_v_tScale > 0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":1023
 *         phiScale[m+1]=phiScale[m]+tScale
 *         if tScale > 0:
 *             dummy=myPow10(tScale)             # <<<<<<<<<<<<<<
 *             tSum[m+1]*=dummy
 *             for k in range(nK):
 */
      __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10(__pyx_v_tScale);

      /* "fastphase/fastphase.pyx":1024
 *         if tScale > 0:
 *             dummy=myPow10(tScale)
 *             tSum[m+1]*=dummy             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 mPhi[m+1,k]*=dummy
 */
      __pyx_t_90 = (__pyx_v_m + 1);
      if (__pyx_t_90 < 0) __pyx_t_90 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_90, __pyx_pybuffernd_tSum.diminfo[0].strides) *= __pyx_v_dummy;

      /* "fastphase/fastphase.pyx":1025
 *             dummy=myPow10(tScale)
 *             tSum[m+1]*=dummy
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 mPhi[m+1,k]*=dummy
 *     ## end calc Phi
 */
      __pyx_t_10 = __pyx_v_nK;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_11; __pyx_t_20+=1) {
        __pyx_v_k = __pyx_t_20;

        /* "fastphase/fastphase.pyx":1026
 *             tSum[m+1]*=dummy
 *             for k in range(nK):
 *                 mPhi[m+1,k]*=dummy             # <<<<<<<<<<<<<<
 *     ## end calc Phi
 *     logLikelihood=log(tSum[nLoc-1])/log(10)-phiScale[nLoc-1]
 */
        __pyx_t_91 = (__pyx_v_m + 1);
        __pyx_t_92 = __pyx_v_k;
        if (__pyx_t_91 < 0) __pyx_t_91 += __pyx_pybuffernd_mPhi.diminfo[0].shape;
        if (__pyx_t_92 < 0) __pyx_t_92 += __pyx_pybuffernd_mPhi.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mPhi.rcbuffer->pybuffer.buf, __pyx_t_91, __pyx_pybuffernd_mPhi.diminfo[0].strides, __pyx_t_92, __pyx_pybuffernd_mPhi.diminfo[1].strides) *= __pyx_v_dummy;
      }

      /* "fastphase/fastphase.pyx":1022
 *             tScale=0
 *         phiScale[m+1]=phiScale[m]+tScale
 *         if tScale > 0:             # <<<<<<<<<<<<<<
 *             dummy=myPow10(tScale)
 *             tSum[m+1]*=dummy
 */
    }
  }

  /* "fastphase/fastphase.pyx":1028
 *                 mPhi[m+1,k]*=dummy
 *     ## end calc Phi
 *     logLikelihood=log(tSum[nLoc-1])/log(10)-phiScale[nLoc-1]             # <<<<<<<<<<<<<<
 *     ##
 *     ## compute individual contributions top,bottom,jmk
 */
  __pyx_t_93 = (__pyx_v_nLoc - 1);
  if (__pyx_t_93 < 0) __pyx_t_93 += __pyx_pybuffernd_tSum.diminfo[0].shape;
  __pyx_t_43 = log((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_tSum.diminfo[0].strides)));
  __pyx_t_44 = log(10.0);
  if (unlikely(__pyx_t_44 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1028, __pyx_L1_error)
  }
  __pyx_t_94 = (__pyx_v_nLoc - 1);
  if (__pyx_t_94 < 0) __pyx_t_94 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
  __pyx_v_logLikelihood = ((__pyx_t_43 / __pyx_t_44) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_94, __pyx_pybuffernd_phiScale.diminfo[0].strides)));

  /* "fastphase/fastphase.pyx":1033
 *     ##
 *     # compute probZ see appx A at the end
 *     cdef np.ndarray[np.float64_t,ndim=2] probZ=mPhi*mBeta             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 */
  __pyx_t_3 = PyNumber_Multiply(((PyObject *)__pyx_v_mPhi), ((PyObject *)__pyx_v_mBeta)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1033, __pyx_L1_error)
  __pyx_t_95 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer, (PyObject*)__pyx_t_95, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_probZ = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1033, __pyx_L1_error)
    } else {__pyx_pybuffernd_probZ.diminfo[0].strides = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_probZ.diminfo[0].shape = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_probZ.diminfo[1].strides = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_probZ.diminfo[1].shape = __pyx_pybuffernd_probZ.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_95 = 0;
  __pyx_v_probZ = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":1034
 *     # compute probZ see appx A at the end
 *     cdef np.ndarray[np.float64_t,ndim=2] probZ=mPhi*mBeta
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1034, __pyx_L1_error)
  __pyx_t_96 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer, (PyObject*)__pyx_t_96, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_jmk = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1034, __pyx_L1_error)
    } else {__pyx_pybuffernd_jmk.diminfo[0].strides = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_jmk.diminfo[0].shape = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_jmk.diminfo[1].strides = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_jmk.diminfo[1].shape = __pyx_pybuffernd_jmk.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_96 = 0;
  __pyx_v_jmk = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":1035
 *     cdef np.ndarray[np.float64_t,ndim=2] probZ=mPhi*mBeta
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1035, __pyx_L1_error)
  __pyx_t_97 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_top.rcbuffer->pybuffer, (PyObject*)__pyx_t_97, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_top = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_top.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1035, __pyx_L1_error)
    } else {__pyx_pybuffernd_top.diminfo[0].strides = __pyx_pybuffernd_top.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_top.diminfo[0].shape = __pyx_pybuffernd_top.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_top.diminfo[1].strides = __pyx_pybuffernd_top.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_top.diminfo[1].shape = __pyx_pybuffernd_top.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_97 = 0;
  __pyx_v_top = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastphase/fastphase.pyx":1036
 *     cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     for m in range(nLoc):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nLoc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_nK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1036, __pyx_L1_error)
  __pyx_t_98 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bot.rcbuffer->pybuffer, (PyObject*)__pyx_t_98, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_bot = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1036, __pyx_L1_error)
    } else {__pyx_pybuffernd_bot.diminfo[0].strides = __pyx_pybuffernd_bot.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bot.diminfo[0].shape = __pyx_pybuffernd_bot.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_bot.diminfo[1].strides = __pyx_pybuffernd_bot.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_bot.diminfo[1].shape = __pyx_pybuffernd_bot.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_98 = 0;
  __pyx_v_bot = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":1038
 *     cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
 * 
 *     for m in range(nLoc):             # <<<<<<<<<<<<<<
 *         normC=0
 *         for k in range(nK):
 */
  __pyx_t_2 = __pyx_v_nLoc;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_m = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1039
 * 
 *     for m in range(nLoc):
 *         normC=0             # <<<<<<<<<<<<<<
 *         for k in range(nK):
 *             normC+=probZ[m,k]
 */
    __pyx_v_normC = 0.0;

    /* "fastphase/fastphase.pyx":1040
 *     for m in range(nLoc):
 *         normC=0
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             normC+=probZ[m,k]
 *         for k in range(nK):
 */
    __pyx_t_20 = __pyx_v_nK;
    __pyx_t_99 = __pyx_t_20;
    for (__pyx_t_100 = 0; __pyx_t_100 < __pyx_t_99; __pyx_t_100+=1) {
      __pyx_v_k = __pyx_t_100;

      /* "fastphase/fastphase.pyx":1041
 *         normC=0
 *         for k in range(nK):
 *             normC+=probZ[m,k]             # <<<<<<<<<<<<<<
 *         for k in range(nK):
 *             probZ[m,k]/=normC
 */
      __pyx_t_101 = __pyx_v_m;
      __pyx_t_102 = __pyx_v_k;
      if (__pyx_t_101 < 0) __pyx_t_101 += __pyx_pybuffernd_probZ.diminfo[0].shape;
      if (__pyx_t_102 < 0) __pyx_t_102 += __pyx_pybuffernd_probZ.diminfo[1].shape;
      __pyx_v_normC = (__pyx_v_normC + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_101, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_102, __pyx_pybuffernd_probZ.diminfo[1].strides)));
    }

    /* "fastphase/fastphase.pyx":1042
 *         for k in range(nK):
 *             normC+=probZ[m,k]
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             probZ[m,k]/=normC
 *     if up2pz:
 */
    __pyx_t_20 = __pyx_v_nK;
    __pyx_t_99 = __pyx_t_20;
    for (__pyx_t_100 = 0; __pyx_t_100 < __pyx_t_99; __pyx_t_100+=1) {
      __pyx_v_k = __pyx_t_100;

      /* "fastphase/fastphase.pyx":1043
 *             normC+=probZ[m,k]
 *         for k in range(nK):
 *             probZ[m,k]/=normC             # <<<<<<<<<<<<<<
 *     if up2pz:
 *         return probZ
 */
      __pyx_t_103 = __pyx_v_m;
      __pyx_t_104 = __pyx_v_k;
      if (__pyx_t_103 < 0) __pyx_t_103 += __pyx_pybuffernd_probZ.diminfo[0].shape;
      if (__pyx_t_104 < 0) __pyx_t_104 += __pyx_pybuffernd_probZ.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_103, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_104, __pyx_pybuffernd_probZ.diminfo[1].strides) /= __pyx_v_normC;
    }
  }

  /* "fastphase/fastphase.pyx":1044
 *         for k in range(nK):
 *             probZ[m,k]/=normC
 *     if up2pz:             # <<<<<<<<<<<<<<
 *         return probZ
 *     ## compute expected jum prob for each interval (appx C)
 */
  __pyx_t_45 = (__pyx_v_up2pz != 0);
  if (__pyx_t_45) {

    /* "fastphase/fastphase.pyx":1045
 *             probZ[m,k]/=normC
 *     if up2pz:
 *         return probZ             # <<<<<<<<<<<<<<
 *     ## compute expected jum prob for each interval (appx C)
 *     # locus 0
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_probZ));
    __pyx_r = ((PyObject *)__pyx_v_probZ);
    goto __pyx_L0;

    /* "fastphase/fastphase.pyx":1044
 *         for k in range(nK):
 *             probZ[m,k]/=normC
 *     if up2pz:             # <<<<<<<<<<<<<<
 *         return probZ
 *     ## compute expected jum prob for each interval (appx C)
 */
  }

  /* "fastphase/fastphase.pyx":1048
 *     ## compute expected jum prob for each interval (appx C)
 *     # locus 0
 *     for k in range(nK):             # <<<<<<<<<<<<<<
 *         ##jmk[0,k]=alpha[0,k]
 *         jmk[0, k] = probZ[ 0, k]
 */
  __pyx_t_2 = __pyx_v_nK;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_k = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1050
 *     for k in range(nK):
 *         ##jmk[0,k]=alpha[0,k]
 *         jmk[0, k] = probZ[ 0, k]             # <<<<<<<<<<<<<<
 *     for m in range(1,nLoc):
 *         dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 */
    __pyx_t_105 = 0;
    __pyx_t_106 = __pyx_v_k;
    if (__pyx_t_105 < 0) __pyx_t_105 += __pyx_pybuffernd_probZ.diminfo[0].shape;
    if (__pyx_t_106 < 0) __pyx_t_106 += __pyx_pybuffernd_probZ.diminfo[1].shape;
    __pyx_t_107 = 0;
    __pyx_t_108 = __pyx_v_k;
    if (__pyx_t_107 < 0) __pyx_t_107 += __pyx_pybuffernd_jmk.diminfo[0].shape;
    if (__pyx_t_108 < 0) __pyx_t_108 += __pyx_pybuffernd_jmk.diminfo[1].shape;
    *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_107, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_108, __pyx_pybuffernd_jmk.diminfo[1].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_105, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_106, __pyx_pybuffernd_probZ.diminfo[1].strides));
  }

  /* "fastphase/fastphase.pyx":1051
 *         ##jmk[0,k]=alpha[0,k]
 *         jmk[0, k] = probZ[ 0, k]
 *     for m in range(1,nLoc):             # <<<<<<<<<<<<<<
 *         dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):
 */
  __pyx_t_2 = __pyx_v_nLoc;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 1; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_m = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1052
 *         jmk[0, k] = probZ[ 0, k]
 *     for m in range(1,nLoc):
 *         dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])             # <<<<<<<<<<<<<<
 *         for k in range(nK):
 *             jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]
 */
    __pyx_t_109 = (__pyx_v_m - 1);
    if (__pyx_t_109 < 0) __pyx_t_109 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    __pyx_t_110 = __pyx_v_m;
    if (__pyx_t_110 < 0) __pyx_t_110 += __pyx_pybuffernd_betaScale.diminfo[0].shape;
    __pyx_t_111 = (__pyx_v_nLoc - 1);
    if (__pyx_t_111 < 0) __pyx_t_111 += __pyx_pybuffernd_phiScale.diminfo[0].shape;
    __pyx_v_dummy = __pyx_f_6hapflk_9fastphase_myPow10((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_109, __pyx_pybuffernd_phiScale.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_betaScale.rcbuffer->pybuffer.buf, __pyx_t_110, __pyx_pybuffernd_betaScale.diminfo[0].strides))) - (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_phiScale.rcbuffer->pybuffer.buf, __pyx_t_111, __pyx_pybuffernd_phiScale.diminfo[0].strides))));

    /* "fastphase/fastphase.pyx":1053
 *     for m in range(1,nLoc):
 *         dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):             # <<<<<<<<<<<<<<
 *             jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             jmk[m,k] *= alpha[m,k]
 */
    __pyx_t_20 = __pyx_v_nK;
    __pyx_t_99 = __pyx_t_20;
    for (__pyx_t_100 = 0; __pyx_t_100 < __pyx_t_99; __pyx_t_100+=1) {
      __pyx_v_k = __pyx_t_100;

      /* "fastphase/fastphase.pyx":1054
 *         dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
 *         for k in range(nK):
 *             jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]             # <<<<<<<<<<<<<<
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tSum[nLoc-1]
 */
      __pyx_t_112 = (__pyx_v_m - 1);
      if (__pyx_t_112 < 0) __pyx_t_112 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      __pyx_t_113 = __pyx_v_m;
      __pyx_t_114 = 0;
      if (__pyx_t_113 < 0) __pyx_t_113 += __pyx_pybuffernd_rho.diminfo[0].shape;
      if (__pyx_t_114 < 0) __pyx_t_114 += __pyx_pybuffernd_rho.diminfo[1].shape;
      __pyx_t_115 = __pyx_v_m;
      __pyx_t_116 = __pyx_v_k;
      if (__pyx_t_115 < 0) __pyx_t_115 += __pyx_pybuffernd_theta.diminfo[0].shape;
      if (__pyx_t_116 < 0) __pyx_t_116 += __pyx_pybuffernd_theta.diminfo[1].shape;
      __pyx_t_117 = __pyx_v_m;
      if (__pyx_t_117 < 0) __pyx_t_117 += __pyx_pybuffernd_hap.diminfo[0].shape;
      __pyx_t_118 = __pyx_v_m;
      __pyx_t_119 = __pyx_v_k;
      if (__pyx_t_118 < 0) __pyx_t_118 += __pyx_pybuffernd_mBeta.diminfo[0].shape;
      if (__pyx_t_119 < 0) __pyx_t_119 += __pyx_pybuffernd_mBeta.diminfo[1].shape;
      __pyx_t_120 = __pyx_v_m;
      __pyx_t_121 = __pyx_v_k;
      if (__pyx_t_120 < 0) __pyx_t_120 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_121 < 0) __pyx_t_121 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_120, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_121, __pyx_pybuffernd_jmk.diminfo[1].strides) = ((((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_112, __pyx_pybuffernd_tSum.diminfo[0].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_rho.rcbuffer->pybuffer.buf, __pyx_t_113, __pyx_pybuffernd_rho.diminfo[0].strides, __pyx_t_114, __pyx_pybuffernd_rho.diminfo[1].strides))) * __pyx_f_6hapflk_9fastphase_happrG((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_theta.rcbuffer->pybuffer.buf, __pyx_t_115, __pyx_pybuffernd_theta.diminfo[0].strides, __pyx_t_116, __pyx_pybuffernd_theta.diminfo[1].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_117, __pyx_pybuffernd_hap.diminfo[0].strides)))) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mBeta.rcbuffer->pybuffer.buf, __pyx_t_118, __pyx_pybuffernd_mBeta.diminfo[0].strides, __pyx_t_119, __pyx_pybuffernd_mBeta.diminfo[1].strides)));

      /* "fastphase/fastphase.pyx":1055
 *         for k in range(nK):
 *             jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             jmk[m,k] *= alpha[m,k]             # <<<<<<<<<<<<<<
 *             jmk[m,k] /= tSum[nLoc-1]
 *             jmk[m,k] /= dummy
 */
      __pyx_t_122 = __pyx_v_m;
      __pyx_t_123 = __pyx_v_k;
      if (__pyx_t_122 < 0) __pyx_t_122 += __pyx_pybuffernd_alpha.diminfo[0].shape;
      if (__pyx_t_123 < 0) __pyx_t_123 += __pyx_pybuffernd_alpha.diminfo[1].shape;
      __pyx_t_124 = __pyx_v_m;
      __pyx_t_125 = __pyx_v_k;
      if (__pyx_t_124 < 0) __pyx_t_124 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_125 < 0) __pyx_t_125 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_124, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_125, __pyx_pybuffernd_jmk.diminfo[1].strides) *= (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_alpha.rcbuffer->pybuffer.buf, __pyx_t_122, __pyx_pybuffernd_alpha.diminfo[0].strides, __pyx_t_123, __pyx_pybuffernd_alpha.diminfo[1].strides));

      /* "fastphase/fastphase.pyx":1056
 *             jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tSum[nLoc-1]             # <<<<<<<<<<<<<<
 *             jmk[m,k] /= dummy
 *             # if not np.isfinite(jmk[m,k]):
 */
      __pyx_t_126 = (__pyx_v_nLoc - 1);
      if (__pyx_t_126 < 0) __pyx_t_126 += __pyx_pybuffernd_tSum.diminfo[0].shape;
      __pyx_t_127 = __pyx_v_m;
      __pyx_t_128 = __pyx_v_k;
      if (__pyx_t_127 < 0) __pyx_t_127 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_128 < 0) __pyx_t_128 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_127, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_128, __pyx_pybuffernd_jmk.diminfo[1].strides) /= (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tSum.rcbuffer->pybuffer.buf, __pyx_t_126, __pyx_pybuffernd_tSum.diminfo[0].strides));

      /* "fastphase/fastphase.pyx":1057
 *             jmk[m,k] *= alpha[m,k]
 *             jmk[m,k] /= tSum[nLoc-1]
 *             jmk[m,k] /= dummy             # <<<<<<<<<<<<<<
 *             # if not np.isfinite(jmk[m,k]):
 *             #     print tSum[m-1],rho[m,0],theta[m,k],hap[m],mBeta[m,k]
 */
      __pyx_t_129 = __pyx_v_m;
      __pyx_t_130 = __pyx_v_k;
      if (__pyx_t_129 < 0) __pyx_t_129 += __pyx_pybuffernd_jmk.diminfo[0].shape;
      if (__pyx_t_130 < 0) __pyx_t_130 += __pyx_pybuffernd_jmk.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_jmk.rcbuffer->pybuffer.buf, __pyx_t_129, __pyx_pybuffernd_jmk.diminfo[0].strides, __pyx_t_130, __pyx_pybuffernd_jmk.diminfo[1].strides) /= __pyx_v_dummy;
    }
  }

  /* "fastphase/fastphase.pyx":1065
 *             #     raise ValueError
 *     # calc thetablock and its inner product with probZ (appx C.)
 *     for m in range(nLoc):             # <<<<<<<<<<<<<<
 *         if hap[m] == 0:
 *             for k in range(nK):
 */
  __pyx_t_2 = __pyx_v_nLoc;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_m = __pyx_t_11;

    /* "fastphase/fastphase.pyx":1066
 *     # calc thetablock and its inner product with probZ (appx C.)
 *     for m in range(nLoc):
 *         if hap[m] == 0:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 bot[m,k]=probZ[m,k]
 */
    __pyx_t_131 = __pyx_v_m;
    if (__pyx_t_131 < 0) __pyx_t_131 += __pyx_pybuffernd_hap.diminfo[0].shape;
    __pyx_t_45 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_131, __pyx_pybuffernd_hap.diminfo[0].strides)) == 0) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":1067
 *     for m in range(nLoc):
 *         if hap[m] == 0:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 bot[m,k]=probZ[m,k]
 *         elif hap[m]==1:
 */
      __pyx_t_20 = __pyx_v_nK;
      __pyx_t_99 = __pyx_t_20;
      for (__pyx_t_100 = 0; __pyx_t_100 < __pyx_t_99; __pyx_t_100+=1) {
        __pyx_v_k = __pyx_t_100;

        /* "fastphase/fastphase.pyx":1068
 *         if hap[m] == 0:
 *             for k in range(nK):
 *                 bot[m,k]=probZ[m,k]             # <<<<<<<<<<<<<<
 *         elif hap[m]==1:
 *             for k in range(nK):
 */
        __pyx_t_132 = __pyx_v_m;
        __pyx_t_133 = __pyx_v_k;
        if (__pyx_t_132 < 0) __pyx_t_132 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_133 < 0) __pyx_t_133 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        __pyx_t_134 = __pyx_v_m;
        __pyx_t_135 = __pyx_v_k;
        if (__pyx_t_134 < 0) __pyx_t_134 += __pyx_pybuffernd_bot.diminfo[0].shape;
        if (__pyx_t_135 < 0) __pyx_t_135 += __pyx_pybuffernd_bot.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf, __pyx_t_134, __pyx_pybuffernd_bot.diminfo[0].strides, __pyx_t_135, __pyx_pybuffernd_bot.diminfo[1].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_132, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_133, __pyx_pybuffernd_probZ.diminfo[1].strides));
      }

      /* "fastphase/fastphase.pyx":1066
 *     # calc thetablock and its inner product with probZ (appx C.)
 *     for m in range(nLoc):
 *         if hap[m] == 0:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 bot[m,k]=probZ[m,k]
 */
      goto __pyx_L43;
    }

    /* "fastphase/fastphase.pyx":1069
 *             for k in range(nK):
 *                 bot[m,k]=probZ[m,k]
 *         elif hap[m]==1:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k]=probZ[m,k]
 */
    __pyx_t_136 = __pyx_v_m;
    if (__pyx_t_136 < 0) __pyx_t_136 += __pyx_pybuffernd_hap.diminfo[0].shape;
    __pyx_t_45 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int_t *, __pyx_pybuffernd_hap.rcbuffer->pybuffer.buf, __pyx_t_136, __pyx_pybuffernd_hap.diminfo[0].strides)) == 1) != 0);
    if (__pyx_t_45) {

      /* "fastphase/fastphase.pyx":1070
 *                 bot[m,k]=probZ[m,k]
 *         elif hap[m]==1:
 *             for k in range(nK):             # <<<<<<<<<<<<<<
 *                 top[m,k]=probZ[m,k]
 *                 bot[m,k]=probZ[m,k]
 */
      __pyx_t_20 = __pyx_v_nK;
      __pyx_t_99 = __pyx_t_20;
      for (__pyx_t_100 = 0; __pyx_t_100 < __pyx_t_99; __pyx_t_100+=1) {
        __pyx_v_k = __pyx_t_100;

        /* "fastphase/fastphase.pyx":1071
 *         elif hap[m]==1:
 *             for k in range(nK):
 *                 top[m,k]=probZ[m,k]             # <<<<<<<<<<<<<<
 *                 bot[m,k]=probZ[m,k]
 *     return logLikelihood,top,bot,jmk
 */
        __pyx_t_137 = __pyx_v_m;
        __pyx_t_138 = __pyx_v_k;
        if (__pyx_t_137 < 0) __pyx_t_137 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_138 < 0) __pyx_t_138 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        __pyx_t_139 = __pyx_v_m;
        __pyx_t_140 = __pyx_v_k;
        if (__pyx_t_139 < 0) __pyx_t_139 += __pyx_pybuffernd_top.diminfo[0].shape;
        if (__pyx_t_140 < 0) __pyx_t_140 += __pyx_pybuffernd_top.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_top.rcbuffer->pybuffer.buf, __pyx_t_139, __pyx_pybuffernd_top.diminfo[0].strides, __pyx_t_140, __pyx_pybuffernd_top.diminfo[1].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_137, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_138, __pyx_pybuffernd_probZ.diminfo[1].strides));

        /* "fastphase/fastphase.pyx":1072
 *             for k in range(nK):
 *                 top[m,k]=probZ[m,k]
 *                 bot[m,k]=probZ[m,k]             # <<<<<<<<<<<<<<
 *     return logLikelihood,top,bot,jmk
 * 
 */
        __pyx_t_141 = __pyx_v_m;
        __pyx_t_142 = __pyx_v_k;
        if (__pyx_t_141 < 0) __pyx_t_141 += __pyx_pybuffernd_probZ.diminfo[0].shape;
        if (__pyx_t_142 < 0) __pyx_t_142 += __pyx_pybuffernd_probZ.diminfo[1].shape;
        __pyx_t_143 = __pyx_v_m;
        __pyx_t_144 = __pyx_v_k;
        if (__pyx_t_143 < 0) __pyx_t_143 += __pyx_pybuffernd_bot.diminfo[0].shape;
        if (__pyx_t_144 < 0) __pyx_t_144 += __pyx_pybuffernd_bot.diminfo[1].shape;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_bot.rcbuffer->pybuffer.buf, __pyx_t_143, __pyx_pybuffernd_bot.diminfo[0].strides, __pyx_t_144, __pyx_pybuffernd_bot.diminfo[1].strides) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_probZ.rcbuffer->pybuffer.buf, __pyx_t_141, __pyx_pybuffernd_probZ.diminfo[0].strides, __pyx_t_142, __pyx_pybuffernd_probZ.diminfo[1].strides));
      }

      /* "fastphase/fastphase.pyx":1069
 *             for k in range(nK):
 *                 bot[m,k]=probZ[m,k]
 *         elif hap[m]==1:             # <<<<<<<<<<<<<<
 *             for k in range(nK):
 *                 top[m,k]=probZ[m,k]
 */
    }
    __pyx_L43:;
  }

  /* "fastphase/fastphase.pyx":1073
 *                 top[m,k]=probZ[m,k]
 *                 bot[m,k]=probZ[m,k]
 *     return logLikelihood,top,bot,jmk             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_logLikelihood); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_top));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_top));
  PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_top));
  __Pyx_INCREF(((PyObject *)__pyx_v_bot));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bot));
  PyTuple_SET_ITEM(__pyx_t_6, 2, ((PyObject *)__pyx_v_bot));
  __Pyx_INCREF(((PyObject *)__pyx_v_jmk));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_jmk));
  PyTuple_SET_ITEM(__pyx_t_6, 3, ((PyObject *)__pyx_v_jmk));
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "fastphase/fastphase.pyx":943
 *     return soluce
 * 
 * cpdef hapCalc(aa,tt,rr,hh,u2p):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
 *     cdef np.ndarray[np.float64_t,ndim=2] theta = tt
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bot.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hap.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tSum.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_top.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("hapflk.fastphase.hapCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_alpha.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_betaScale.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bot.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hap.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_jmk.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mBeta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mPhi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phiScale.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_probZ.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rho.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tSum.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_theta.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_top.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_alpha);
  __Pyx_XDECREF((PyObject *)__pyx_v_theta);
  __Pyx_XDECREF((PyObject *)__pyx_v_rho);
  __Pyx_XDECREF((PyObject *)__pyx_v_hap);
  __Pyx_XDECREF((PyObject *)__pyx_v_betaScale);
  __Pyx_XDECREF((PyObject *)__pyx_v_tSum);
  __Pyx_XDECREF((PyObject *)__pyx_v_mBeta);
  __Pyx_XDECREF((PyObject *)__pyx_v_mPhi);
  __Pyx_XDECREF((PyObject *)__pyx_v_phiScale);
  __Pyx_XDECREF((PyObject *)__pyx_v_probZ);
  __Pyx_XDECREF((PyObject *)__pyx_v_jmk);
  __Pyx_XDECREF((PyObject *)__pyx_v_top);
  __Pyx_XDECREF((PyObject *)__pyx_v_bot);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6hapflk_9fastphase_23hapCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6hapflk_9fastphase_23hapCalc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aa = 0;
  PyObject *__pyx_v_tt = 0;
  PyObject *__pyx_v_rr = 0;
  PyObject *__pyx_v_hh = 0;
  PyObject *__pyx_v_u2p = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hapCalc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aa,&__pyx_n_s_tt,&__pyx_n_s_rr,&__pyx_n_s_hh,&__pyx_n_s_u2p,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aa)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapCalc", 1, 5, 5, 1); __PYX_ERR(0, 943, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapCalc", 1, 5, 5, 2); __PYX_ERR(0, 943, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapCalc", 1, 5, 5, 3); __PYX_ERR(0, 943, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u2p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hapCalc", 1, 5, 5, 4); __PYX_ERR(0, 943, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "hapCalc") < 0)) __PYX_ERR(0, 943, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_aa = values[0];
    __pyx_v_tt = values[1];
    __pyx_v_rr = values[2];
    __pyx_v_hh = values[3];
    __pyx_v_u2p = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hapCalc", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 943, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("hapflk.fastphase.hapCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6hapflk_9fastphase_22hapCalc(__pyx_self, __pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_hh, __pyx_v_u2p);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6hapflk_9fastphase_22hapCalc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aa, PyObject *__pyx_v_tt, PyObject *__pyx_v_rr, PyObject *__pyx_v_hh, PyObject *__pyx_v_u2p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("hapCalc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6hapflk_9fastphase_hapCalc(__pyx_v_aa, __pyx_v_tt, __pyx_v_rr, __pyx_v_hh, __pyx_v_u2p, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hapflk.fastphase.hapCalc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 272, __pyx_L1_error)

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 276, __pyx_L1_error)

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 306, __pyx_L1_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 325, __pyx_L1_error)
      break;
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(1, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(1, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 856, __pyx_L1_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 860, __pyx_L1_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 880, __pyx_L1_error)

        /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(1, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1036, __pyx_L3_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1042, __pyx_L3_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1048, __pyx_L3_error)

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *__pyx_freelist_6hapflk_9fastphase___pyx_scope_struct__optimfit[8];
static int __pyx_freecount_6hapflk_9fastphase___pyx_scope_struct__optimfit = 0;

static PyObject *__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct__optimfit(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct__optimfit > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit)))) {
    o = (PyObject*)__pyx_freelist_6hapflk_9fastphase___pyx_scope_struct__optimfit[--__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct__optimfit];
    memset(o, 0, sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6hapflk_9fastphase___pyx_scope_struct__optimfit(PyObject *o) {
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *p = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_imp);
  Py_CLEAR(p->__pyx_v_nClus);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct__optimfit < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit)))) {
    __pyx_freelist_6hapflk_9fastphase___pyx_scope_struct__optimfit[__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct__optimfit++] = ((struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6hapflk_9fastphase___pyx_scope_struct__optimfit(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *p = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)o;
  if (p->__pyx_v_imp) {
    e = (*v)(p->__pyx_v_imp, a); if (e) return e;
  }
  if (p->__pyx_v_nClus) {
    e = (*v)(p->__pyx_v_nClus, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6hapflk_9fastphase___pyx_scope_struct__optimfit(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *p = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit *)o;
  tmp = ((PyObject*)p->__pyx_v_imp);
  p->__pyx_v_imp = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_nClus);
  p->__pyx_v_nClus = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit = {
  PyVarObject_HEAD_INIT(0, 0)
  "hapflk.fastphase.__pyx_scope_struct__optimfit", /*tp_name*/
  sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct__optimfit), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6hapflk_9fastphase___pyx_scope_struct__optimfit, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6hapflk_9fastphase___pyx_scope_struct__optimfit, /*tp_traverse*/
  __pyx_tp_clear_6hapflk_9fastphase___pyx_scope_struct__optimfit, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct__optimfit, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *__pyx_freelist_6hapflk_9fastphase___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_6hapflk_9fastphase___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6hapflk_9fastphase___pyx_scope_struct_1_genexpr[--__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6hapflk_9fastphase___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_geno);
  Py_CLEAR(p->__pyx_v_ijump);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_4);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_6hapflk_9fastphase___pyx_scope_struct_1_genexpr[__pyx_freecount_6hapflk_9fastphase___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6hapflk_9fastphase___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_geno) {
    e = (*v)(p->__pyx_v_geno, a); if (e) return e;
  }
  if (p->__pyx_v_ijump) {
    e = (*v)(p->__pyx_v_ijump, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_4) {
    e = (*v)(p->__pyx_t_4, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hapflk.fastphase.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6hapflk_9fastphase___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6hapflk_9fastphase___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6hapflk_9fastphase___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6hapflk_9fastphase___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"calc_cost_matrix_haplo_tot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_7calc_cost_matrix_haplo_tot, METH_VARARGS|METH_KEYWORDS, 0},
  {"calc_cost_matrix_geno_tot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_11calc_cost_matrix_geno_tot, METH_VARARGS|METH_KEYWORDS, 0},
  {"genViterbi", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_17genViterbi, METH_VARARGS|METH_KEYWORDS, 0},
  {"genCalc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_19genCalc, METH_VARARGS|METH_KEYWORDS, 0},
  {"hapViterbi", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_21hapViterbi, METH_VARARGS|METH_KEYWORDS, 0},
  {"hapCalc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6hapflk_9fastphase_23hapCalc, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_fastphase(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_fastphase},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "fastphase",
    __pyx_k_Module_implementing_the_LD_mode, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_A_class_for_fastphase_model_par, __pyx_k_A_class_for_fastphase_model_par, sizeof(__pyx_k_A_class_for_fastphase_model_par), 0, 0, 1, 0},
  {&__pyx_kp_s_A_class_to_manipulate_and_contr, __pyx_k_A_class_to_manipulate_and_contr, sizeof(__pyx_k_A_class_to_manipulate_and_contr), 0, 0, 1, 0},
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_kp_u_Calculating_Costs, __pyx_k_Calculating_Costs, sizeof(__pyx_k_Calculating_Costs), 0, 1, 0, 0},
  {&__pyx_kp_u_Computing_optimum_permutations, __pyx_k_Computing_optimum_permutations, sizeof(__pyx_k_Computing_optimum_permutations), 0, 1, 0, 0},
  {&__pyx_n_u_EM, __pyx_k_EM, sizeof(__pyx_k_EM), 0, 1, 0, 1},
  {&__pyx_kp_u_EM_with_switched_parameters, __pyx_k_EM_with_switched_parameters, sizeof(__pyx_k_EM_with_switched_parameters), 0, 1, 0, 0},
  {&__pyx_kp_u_Fitting_fastphase_model, __pyx_k_Fitting_fastphase_model, sizeof(__pyx_k_Fitting_fastphase_model), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Genotypes, __pyx_k_Genotypes, sizeof(__pyx_k_Genotypes), 0, 1, 0, 0},
  {&__pyx_kp_u_Haplotypes, __pyx_k_Haplotypes, sizeof(__pyx_k_Haplotypes), 0, 1, 0, 0},
  {&__pyx_n_s_ID, __pyx_k_ID, sizeof(__pyx_k_ID), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_Imputations, __pyx_k_Imputations, sizeof(__pyx_k_Imputations), 0, 0, 1, 1},
  {&__pyx_kp_u_Init_EM, __pyx_k_Init_EM, sizeof(__pyx_k_Init_EM), 0, 1, 0, 0},
  {&__pyx_kp_u_Initial_Viterbi, __pyx_k_Initial_Viterbi, sizeof(__pyx_k_Initial_Viterbi), 0, 1, 0, 0},
  {&__pyx_kp_u_Iter, __pyx_k_Iter, sizeof(__pyx_k_Iter), 0, 1, 0, 0},
  {&__pyx_kp_u_Loci, __pyx_k_Loci, sizeof(__pyx_k_Loci), 0, 1, 0, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_kp_u_OPTIMFIT_RUN, __pyx_k_OPTIMFIT_RUN, sizeof(__pyx_k_OPTIMFIT_RUN), 0, 1, 0, 0},
  {&__pyx_n_s_Pool, __pyx_k_Pool, sizeof(__pyx_k_Pool), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_u_Usage_with_fastphase_nloc_nproc, __pyx_k_Usage_with_fastphase_nloc_nproc, sizeof(__pyx_k_Usage_with_fastphase_nloc_nproc), 0, 1, 0, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u_Wrong_Genotype_Size, __pyx_k_Wrong_Genotype_Size, sizeof(__pyx_k_Wrong_Genotype_Size), 0, 1, 0, 0},
  {&__pyx_kp_u_Wrong_Haplotype_Size, __pyx_k_Wrong_Haplotype_Size, sizeof(__pyx_k_Wrong_Haplotype_Size), 0, 1, 0, 0},
  {&__pyx_n_u_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 1, 0, 1},
  {&__pyx_n_s_aa, __pyx_k_aa, sizeof(__pyx_k_aa), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_addGenotype, __pyx_k_addGenotype, sizeof(__pyx_k_addGenotype), 0, 0, 1, 1},
  {&__pyx_n_s_addHaplotype, __pyx_k_addHaplotype, sizeof(__pyx_k_addHaplotype), 0, 0, 1, 1},
  {&__pyx_n_s_addIndivFit, __pyx_k_addIndivFit, sizeof(__pyx_k_addIndivFit), 0, 0, 1, 1},
  {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_alpha_up, __pyx_k_alpha_up, sizeof(__pyx_k_alpha_up), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_bot, __pyx_k_bot, sizeof(__pyx_k_bot), 0, 0, 1, 1},
  {&__pyx_n_s_calc_cost_matrix_geno, __pyx_k_calc_cost_matrix_geno, sizeof(__pyx_k_calc_cost_matrix_geno), 0, 0, 1, 1},
  {&__pyx_n_s_calc_cost_matrix_haplo, __pyx_k_calc_cost_matrix_haplo, sizeof(__pyx_k_calc_cost_matrix_haplo), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_kp_u_clusters, __pyx_k_clusters, sizeof(__pyx_k_clusters), 0, 1, 0, 0},
  {&__pyx_n_s_cost_mat, __pyx_k_cost_mat, sizeof(__pyx_k_cost_mat), 0, 0, 1, 1},
  {&__pyx_n_s_cpu_count, __pyx_k_cpu_count, sizeof(__pyx_k_cpu_count), 0, 0, 1, 1},
  {&__pyx_n_s_curclus, __pyx_k_curclus, sizeof(__pyx_k_curclus), 0, 0, 1, 1},
  {&__pyx_n_s_curpar, __pyx_k_curpar, sizeof(__pyx_k_curpar), 0, 0, 1, 1},
  {&__pyx_n_s_dat, __pyx_k_dat, sizeof(__pyx_k_dat), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_fast, __pyx_k_fast, sizeof(__pyx_k_fast), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase, __pyx_k_fastphase, sizeof(__pyx_k_fastphase), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase___enter, __pyx_k_fastphase___enter, sizeof(__pyx_k_fastphase___enter), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase___exit, __pyx_k_fastphase___exit, sizeof(__pyx_k_fastphase___exit), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase___init, __pyx_k_fastphase___init, sizeof(__pyx_k_fastphase___init), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_addGenotype, __pyx_k_fastphase_addGenotype, sizeof(__pyx_k_fastphase_addGenotype), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_addHaplotype, __pyx_k_fastphase_addHaplotype, sizeof(__pyx_k_fastphase_addHaplotype), 0, 0, 1, 1},
  {&__pyx_kp_s_fastphase_fastphase_pyx, __pyx_k_fastphase_fastphase_pyx, sizeof(__pyx_k_fastphase_fastphase_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_fastphase_fit, __pyx_k_fastphase_fit, sizeof(__pyx_k_fastphase_fit), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_flush, __pyx_k_fastphase_flush, sizeof(__pyx_k_fastphase_flush), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_gen2hap, __pyx_k_fastphase_gen2hap, sizeof(__pyx_k_fastphase_gen2hap), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_impute, __pyx_k_fastphase_impute, sizeof(__pyx_k_fastphase_impute), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_optimfit, __pyx_k_fastphase_optimfit, sizeof(__pyx_k_fastphase_optimfit), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_optimfit_locals_genexp, __pyx_k_fastphase_optimfit_locals_genexp, sizeof(__pyx_k_fastphase_optimfit_locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_fastphase_viterbi, __pyx_k_fastphase_viterbi, sizeof(__pyx_k_fastphase_viterbi), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_fit, __pyx_k_fit, sizeof(__pyx_k_fit), 0, 0, 1, 1},
  {&__pyx_n_s_fitInData, __pyx_k_fitInData, sizeof(__pyx_k_fitInData), 0, 0, 1, 1},
  {&__pyx_n_s_fitInData___init, __pyx_k_fitInData___init, sizeof(__pyx_k_fitInData___init), 0, 0, 1, 1},
  {&__pyx_n_s_fitOutData, __pyx_k_fitOutData, sizeof(__pyx_k_fitOutData), 0, 0, 1, 1},
  {&__pyx_n_s_fitOutData___init, __pyx_k_fitOutData___init, sizeof(__pyx_k_fitOutData___init), 0, 0, 1, 1},
  {&__pyx_n_s_fitter, __pyx_k_fitter, sizeof(__pyx_k_fitter), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_flog, __pyx_k_flog, sizeof(__pyx_k_flog), 0, 0, 1, 1},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_gLogLike, __pyx_k_gLogLike, sizeof(__pyx_k_gLogLike), 0, 0, 1, 1},
  {&__pyx_n_s_gen, __pyx_k_gen, sizeof(__pyx_k_gen), 0, 0, 1, 1},
  {&__pyx_n_s_gen2hap, __pyx_k_gen2hap, sizeof(__pyx_k_gen2hap), 0, 0, 1, 1},
  {&__pyx_n_s_gen_dat, __pyx_k_gen_dat, sizeof(__pyx_k_gen_dat), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_geno, __pyx_k_geno, sizeof(__pyx_k_geno), 0, 0, 1, 1},
  {&__pyx_n_u_geno, __pyx_k_geno, sizeof(__pyx_k_geno), 0, 1, 0, 1},
  {&__pyx_n_s_genotypes, __pyx_k_genotypes, sizeof(__pyx_k_genotypes), 0, 0, 1, 1},
  {&__pyx_n_s_gg, __pyx_k_gg, sizeof(__pyx_k_gg), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_hLogLike, __pyx_k_hLogLike, sizeof(__pyx_k_hLogLike), 0, 0, 1, 1},
  {&__pyx_n_s_hap, __pyx_k_hap, sizeof(__pyx_k_hap), 0, 0, 1, 1},
  {&__pyx_n_s_hapflk_fastphase, __pyx_k_hapflk_fastphase, sizeof(__pyx_k_hapflk_fastphase), 0, 0, 1, 1},
  {&__pyx_n_s_haplo, __pyx_k_haplo, sizeof(__pyx_k_haplo), 0, 0, 1, 1},
  {&__pyx_n_u_haplo, __pyx_k_haplo, sizeof(__pyx_k_haplo), 0, 1, 0, 1},
  {&__pyx_n_s_haplotypes, __pyx_k_haplotypes, sizeof(__pyx_k_haplotypes), 0, 0, 1, 1},
  {&__pyx_n_s_hh, __pyx_k_hh, sizeof(__pyx_k_hh), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_iEM, __pyx_k_iEM, sizeof(__pyx_k_iEM), 0, 0, 1, 1},
  {&__pyx_n_s_ijump, __pyx_k_ijump, sizeof(__pyx_k_ijump), 0, 0, 1, 1},
  {&__pyx_n_s_imp, __pyx_k_imp, sizeof(__pyx_k_imp), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_impute, __pyx_k_impute, sizeof(__pyx_k_impute), 0, 0, 1, 1},
  {&__pyx_n_s_imputeInData, __pyx_k_imputeInData, sizeof(__pyx_k_imputeInData), 0, 0, 1, 1},
  {&__pyx_n_s_imputeInData___init, __pyx_k_imputeInData___init, sizeof(__pyx_k_imputeInData___init), 0, 0, 1, 1},
  {&__pyx_n_s_imputeOutData, __pyx_k_imputeOutData, sizeof(__pyx_k_imputeOutData), 0, 0, 1, 1},
  {&__pyx_n_s_imputeOutData___init, __pyx_k_imputeOutData___init, sizeof(__pyx_k_imputeOutData___init), 0, 0, 1, 1},
  {&__pyx_n_s_imputer, __pyx_k_imputer, sizeof(__pyx_k_imputer), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_initUpdate, __pyx_k_initUpdate, sizeof(__pyx_k_initUpdate), 0, 0, 1, 1},
  {&__pyx_n_s_init_par, __pyx_k_init_par, sizeof(__pyx_k_init_par), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_kp_u_is_not, __pyx_k_is_not, sizeof(__pyx_k_is_not), 0, 1, 0, 0},
  {&__pyx_n_s_it, __pyx_k_it, sizeof(__pyx_k_it), 0, 0, 1, 1},
  {&__pyx_n_s_item, __pyx_k_item, sizeof(__pyx_k_item), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_u_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 1, 0, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_jm, __pyx_k_jm, sizeof(__pyx_k_jm), 0, 0, 1, 1},
  {&__pyx_n_s_jmk, __pyx_k_jmk, sizeof(__pyx_k_jmk), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_k1, __pyx_k_k1, sizeof(__pyx_k_k1), 0, 0, 1, 1},
  {&__pyx_n_s_k2, __pyx_k_k2, sizeof(__pyx_k_k2), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kp, __pyx_k_kp, sizeof(__pyx_k_kp), 0, 0, 1, 1},
  {&__pyx_n_s_kp1, __pyx_k_kp1, sizeof(__pyx_k_kp1), 0, 0, 1, 1},
  {&__pyx_n_s_kp2, __pyx_k_kp2, sizeof(__pyx_k_kp2), 0, 0, 1, 1},
  {&__pyx_n_s_lambda, __pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_liktraj, __pyx_k_liktraj, sizeof(__pyx_k_liktraj), 0, 0, 1, 1},
  {&__pyx_n_s_linear_sum_assignment, __pyx_k_linear_sum_assignment, sizeof(__pyx_k_linear_sum_assignment), 0, 0, 1, 1},
  {&__pyx_n_s_logLike, __pyx_k_logLike, sizeof(__pyx_k_logLike), 0, 0, 1, 1},
  {&__pyx_n_s_log_like, __pyx_k_log_like, sizeof(__pyx_k_log_like), 0, 0, 1, 1},
  {&__pyx_n_u_loglik, __pyx_k_loglik, sizeof(__pyx_k_loglik), 0, 1, 0, 1},
  {&__pyx_n_s_loglike, __pyx_k_loglike, sizeof(__pyx_k_loglike), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 0, 1, 1},
  {&__pyx_n_s_modParams, __pyx_k_modParams, sizeof(__pyx_k_modParams), 0, 0, 1, 1},
  {&__pyx_n_s_modParams___init, __pyx_k_modParams___init, sizeof(__pyx_k_modParams___init), 0, 0, 1, 1},
  {&__pyx_n_s_modParams_addIndivFit, __pyx_k_modParams_addIndivFit, sizeof(__pyx_k_modParams_addIndivFit), 0, 0, 1, 1},
  {&__pyx_n_s_modParams_initUpdate, __pyx_k_modParams_initUpdate, sizeof(__pyx_k_modParams_initUpdate), 0, 0, 1, 1},
  {&__pyx_n_s_modParams_update, __pyx_k_modParams_update, sizeof(__pyx_k_modParams_update), 0, 0, 1, 1},
  {&__pyx_n_s_modParams_write, __pyx_k_modParams_write, sizeof(__pyx_k_modParams_write), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessing, __pyx_k_multiprocessing, sizeof(__pyx_k_multiprocessing), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nClus, __pyx_k_nClus, sizeof(__pyx_k_nClus), 0, 0, 1, 1},
  {&__pyx_n_s_nEM, __pyx_k_nEM, sizeof(__pyx_k_nEM), 0, 0, 1, 1},
  {&__pyx_n_s_nK, __pyx_k_nK, sizeof(__pyx_k_nK), 0, 0, 1, 1},
  {&__pyx_n_s_nLoc, __pyx_k_nLoc, sizeof(__pyx_k_nLoc), 0, 0, 1, 1},
  {&__pyx_n_s_nLoci, __pyx_k_nLoci, sizeof(__pyx_k_nLoci), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_nbest, __pyx_k_nbest, sizeof(__pyx_k_nbest), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_newpar, __pyx_k_newpar, sizeof(__pyx_k_newpar), 0, 0, 1, 1},
  {&__pyx_n_s_nhap, __pyx_k_nhap, sizeof(__pyx_k_nhap), 0, 0, 1, 1},
  {&__pyx_n_s_niter, __pyx_k_niter, sizeof(__pyx_k_niter), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nproc, __pyx_k_nproc, sizeof(__pyx_k_nproc), 0, 0, 1, 1},
  {&__pyx_n_s_nsamp, __pyx_k_nsamp, sizeof(__pyx_k_nsamp), 0, 0, 1, 1},
  {&__pyx_n_s_nstep, __pyx_k_nstep, sizeof(__pyx_k_nstep), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_old_log_like, __pyx_k_old_log_like, sizeof(__pyx_k_old_log_like), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_optimfit, __pyx_k_optimfit, sizeof(__pyx_k_optimfit), 0, 0, 1, 1},
  {&__pyx_n_s_pZ, __pyx_k_pZ, sizeof(__pyx_k_pZ), 0, 0, 1, 1},
  {&__pyx_n_s_par, __pyx_k_par, sizeof(__pyx_k_par), 0, 0, 1, 1},
  {&__pyx_n_s_parList, __pyx_k_parList, sizeof(__pyx_k_parList), 0, 0, 1, 1},
  {&__pyx_n_s_par_s, __pyx_k_par_s, sizeof(__pyx_k_par_s), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_permut, __pyx_k_permut, sizeof(__pyx_k_permut), 0, 0, 1, 1},
  {&__pyx_n_s_pgeno, __pyx_k_pgeno, sizeof(__pyx_k_pgeno), 0, 0, 1, 1},
  {&__pyx_n_s_pool, __pyx_k_pool, sizeof(__pyx_k_pool), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_prfx, __pyx_k_prfx, sizeof(__pyx_k_prfx), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_probZ, __pyx_k_probZ, sizeof(__pyx_k_probZ), 0, 0, 1, 1},
  {&__pyx_n_s_pth, __pyx_k_pth, sizeof(__pyx_k_pth), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 0, 1, 1},
  {&__pyx_n_s_rho, __pyx_k_rho, sizeof(__pyx_k_rho), 0, 0, 1, 1},
  {&__pyx_n_u_rho, __pyx_k_rho, sizeof(__pyx_k_rho), 0, 1, 0, 1},
  {&__pyx_n_s_rhomin, __pyx_k_rhomin, sizeof(__pyx_k_rhomin), 0, 0, 1, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_s_rr, __pyx_k_rr, sizeof(__pyx_k_rr), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_optimize, __pyx_k_scipy_optimize, sizeof(__pyx_k_scipy_optimize), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_u_snp, __pyx_k_snp, sizeof(__pyx_k_snp), 0, 1, 0, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_switch_pars, __pyx_k_switch_pars, sizeof(__pyx_k_switch_pars), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_u_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 1, 0, 1},
  {&__pyx_n_s_tasks, __pyx_k_tasks, sizeof(__pyx_k_tasks), 0, 0, 1, 1},
  {&__pyx_n_s_terminate, __pyx_k_terminate, sizeof(__pyx_k_terminate), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theta, __pyx_k_theta, sizeof(__pyx_k_theta), 0, 0, 1, 1},
  {&__pyx_n_s_theta_up, __pyx_k_theta_up, sizeof(__pyx_k_theta_up), 0, 0, 1, 1},
  {&__pyx_kp_u_threads, __pyx_k_threads, sizeof(__pyx_k_threads), 0, 1, 0, 0},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tohap, __pyx_k_tohap, sizeof(__pyx_k_tohap), 0, 0, 1, 1},
  {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
  {&__pyx_n_s_tt, __pyx_k_tt, sizeof(__pyx_k_tt), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_u2p, __pyx_k_u2p, sizeof(__pyx_k_u2p), 0, 0, 1, 1},
  {&__pyx_n_s_u2z, __pyx_k_u2z, sizeof(__pyx_k_u2z), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_up2pz, __pyx_k_up2pz, sizeof(__pyx_k_up2pz), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_vectorize, __pyx_k_vectorize, sizeof(__pyx_k_vectorize), 0, 0, 1, 1},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_vit, __pyx_k_vit, sizeof(__pyx_k_vit), 0, 0, 1, 1},
  {&__pyx_n_s_viterber, __pyx_k_viterber, sizeof(__pyx_k_viterber), 0, 0, 1, 1},
  {&__pyx_n_s_viterbi, __pyx_k_viterbi, sizeof(__pyx_k_viterbi), 0, 0, 1, 1},
  {&__pyx_n_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 392, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(1, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "fastphase/fastphase.pyx":110
 *                     elif self.alpha[i,j]<0.001:
 *                         self.alpha[i,j]=0.001
 *                 self.alpha[i,:] /= np.sum(self.alpha[i,:])             # <<<<<<<<<<<<<<
 *         ## theta
 *         if self.theta_up:
 */
  __pyx_slice_ = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);

  /* "fastphase/fastphase.pyx":121
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)             # <<<<<<<<<<<<<<
 *         for i in range(self.nLoc):
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_n_u_snp); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_n_u_rho); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "fastphase/fastphase.pyx":202
 *             assert self.pool is not None
 *         except AssertionError:
 *             print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_Usage_with_fastphase_nloc_nproc); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "fastphase/fastphase.pyx":212
 *             par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
 *         if verbose:
 *             print( 'Fitting fastphase model',file=self.flog)             # <<<<<<<<<<<<<<
 *             print( '# clusters ',nClus, file=self.flog)
 *             print( '# threads ', self.nproc, file=self.flog)
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_Fitting_fastphase_model); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "fastphase/fastphase.pyx":246
 *         liktraj = []
 * 
 *         print("=== OPTIMFIT RUN ===",file=self.flog)             # <<<<<<<<<<<<<<
 *         ## Fit nEM EM to find an initial start
 *         if verbose:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_OPTIMFIT_RUN); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "fastphase/fastphase.pyx":249
 *         ## Fit nEM EM to find an initial start
 *         if verbose:
 *             print("*** Init EM", 1,file=self.flog)             # <<<<<<<<<<<<<<
 *         nbest = 0
 *         curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_kp_u_Init_EM, __pyx_int_1); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "fastphase/fastphase.pyx":264
 *         init_par = curpar
 *         ## Improve it
 *         print('Initial Viterbi',file=self.flog)             # <<<<<<<<<<<<<<
 *         self.flog.flush()
 *         imp = self.viterbi([curpar])
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Initial_Viterbi); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "fastphase/fastphase.pyx":274
 *             ## calculate costs
 *             if verbose:
 *                 print("Calculating Costs",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             ##cost_mat = np.zeros( (self.nLoci-1, nClus, nClus), dtype=np.float)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_Calculating_Costs); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "fastphase/fastphase.pyx":298
 *             ## combine
 *             if verbose:
 *                 print("Computing optimum permutations",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             ##cost_mat = cost_mat.reshape( ( self.nLoci-1, nClus, nClus))
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Computing_optimum_permutations); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "fastphase/fastphase.pyx":303
 *             ##args = [ cost_mat[i] for i in range( self.nLoci-1)]
 *             res = np.array( self.pool.map( linear_sum_assignment, cost_mat,  100))
 *             permut = res[:,1,:]             # <<<<<<<<<<<<<<
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:
 */
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_slice_, __pyx_int_1, __pyx_slice_); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "fastphase/fastphase.pyx":306
 *             newpar = switch_pars( curpar, permut)
 *             if verbose:
 *                 print("EM with switched parameters",file=self.flog)             # <<<<<<<<<<<<<<
 *                 self.flog.flush()
 *             if (it == niter-1): ## last iteration, longer and no imputation
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_EM_with_switched_parameters); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "fastphase/fastphase.pyx":319
 *         # par_s = self.fit(nClus = nClus, nstep = max(30,nstep), verbose = True, params = par_s, alpha_up = True, theta_up = False, fast = fast)
 *         # liktraj.append((nbest, it+2, par_s.loglike))
 *         print('EM','iter','loglik',file=self.flog)             # <<<<<<<<<<<<<<
 *         for dat in liktraj:
 *             print(*dat,file=self.flog)
 */
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_n_u_EM, __pyx_n_u_iter, __pyx_n_u_loglik); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "fastphase/fastphase.pyx":25
 * 
 * class fitInData():
 *     def __init__(self,type,alpha,theta,rho,data,u2z):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.alpha=alpha
 */
  __pyx_tuple__21 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_type, __pyx_n_s_alpha, __pyx_n_s_theta, __pyx_n_s_rho, __pyx_n_s_data, __pyx_n_s_u2z); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(7, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 25, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 25, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":34
 * 
 * class fitOutData():
 *     def __init__(self,logLike,top,bot,jmk,val):             # <<<<<<<<<<<<<<
 *         self.logLike=logLike
 *         self.top=top
 */
  __pyx_tuple__23 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_logLike, __pyx_n_s_top, __pyx_n_s_bot, __pyx_n_s_jmk, __pyx_n_s_val); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 34, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 34, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":42
 * 
 * class imputeInData():
 *     def __init__(self,type,parList,nLoc,name,data):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.parList=parList
 */
  __pyx_tuple__25 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_type, __pyx_n_s_parList, __pyx_n_s_nLoc, __pyx_n_s_name, __pyx_n_s_data); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":50
 * 
 * class imputeOutData():
 *     def __init__(self,pgeno,name,pZ,path=None):             # <<<<<<<<<<<<<<
 *         self.name=name
 *         self.pgeno=pgeno
 */
  __pyx_tuple__27 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_pgeno, __pyx_n_s_name, __pyx_n_s_pZ, __pyx_n_s_path); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 50, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_tuple__29 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "fastphase/fastphase.pyx":67
 *     -- rho (N x 1): jump probabilities in each interval
 *     '''
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):             # <<<<<<<<<<<<<<
 *         self.nLoc=nLoc
 *         self.nClus=nClus
 */
  __pyx_tuple__30 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_nLoc, __pyx_n_s_nClus, __pyx_n_s_rhomin, __pyx_n_s_alpha_up, __pyx_n_s_theta_up); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_tuple__32 = PyTuple_Pack(3, ((PyObject*)__pyx_float_1eneg_6), ((PyObject *)Py_True), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "fastphase/fastphase.pyx":78
 *         self.theta_up = theta_up
 *         self.loglike = 0
 *     def initUpdate(self):             # <<<<<<<<<<<<<<
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_initUpdate, 78, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 78, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":84
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):             # <<<<<<<<<<<<<<
 *         self.top += t
 *         self.bot += b
 */
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_t, __pyx_n_s_b, __pyx_n_s_j, __pyx_n_s_nhap); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_addIndivFit, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":90
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap
 *     def update(self):             # <<<<<<<<<<<<<<
 *         ''' Update parameters using top,bot,jmk jm probabilities'''
 *         ## rho
 */
  __pyx_tuple__37 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_update, 90, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 90, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":120
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):             # <<<<<<<<<<<<<<
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):
 */
  __pyx_tuple__39 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_stream, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_k, __pyx_n_s_k); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_write, 120, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 120, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":136
 *          ... do stuff ...
 *     '''
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):             # <<<<<<<<<<<<<<
 *         assert nLoci>0
 *         self.nLoci=nLoci
 */
  __pyx_tuple__41 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_nLoci, __pyx_n_s_nproc, __pyx_n_s_prfx); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_init, 136, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 136, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":145
 *         self.prfx = prfx
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_enter, 145, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 145, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":153
 *         return self
 * 
 *     def __exit__(self,*args):             # <<<<<<<<<<<<<<
 *         self.pool.terminate()
 *         if self.prfx is not None:
 */
  __pyx_tuple__45 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_args); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_exit, 153, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 153, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":158
 *             self.flog.close()
 * 
 *     def flush(self):             # <<<<<<<<<<<<<<
 *         '''
 *         remove data
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_flush, 158, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 158, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":165
 *         self.genotypes={}
 * 
 *     def addHaplotype(self,ID,hap,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add an haplotype to the model observations.
 */
  __pyx_tuple__49 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_ID, __pyx_n_s_hap, __pyx_n_s_missing); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_addHaplotype, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_tuple__51 = PyTuple_Pack(1, ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "fastphase/fastphase.pyx":177
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise
 *     def addGenotype(self,ID,gen,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add a genotype to the model observations.
 */
  __pyx_tuple__52 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_ID, __pyx_n_s_gen, __pyx_n_s_missing); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_addGenotype, 177, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_tuple__54 = PyTuple_Pack(1, ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "fastphase/fastphase.pyx":191
 * 
 *     @staticmethod
 *     def gen2hap(gen):             # <<<<<<<<<<<<<<
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_n_s_gen); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_gen2hap, 191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 191, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":194
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 *     def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Fit the model on observations with nCLus clusters using nstep EM iterations
 */
  __pyx_tuple__57 = PyTuple_Pack(19, __pyx_n_s_self, __pyx_n_s_nClus, __pyx_n_s_nstep, __pyx_n_s_params, __pyx_n_s_verbose, __pyx_n_s_rhomin, __pyx_n_s_alpha_up, __pyx_n_s_theta_up, __pyx_n_s_fast, __pyx_n_s_par, __pyx_n_s_old_log_like, __pyx_n_s_iEM, __pyx_n_s_log_like, __pyx_n_s_tasks, __pyx_n_s_results, __pyx_n_s_item, __pyx_n_s_hap, __pyx_n_s_gen, __pyx_n_s_gen); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(9, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_fit, 194, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_tuple__59 = PyTuple_Pack(8, ((PyObject *)__pyx_int_20), ((PyObject *)__pyx_int_20), ((PyObject *)Py_None), ((PyObject *)Py_False), ((PyObject*)__pyx_float_1eneg_6), ((PyObject *)Py_True), ((PyObject *)Py_True), ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */
  __pyx_tuple__60 = PyTuple_Pack(34, __pyx_n_s_self, __pyx_n_s_nClus, __pyx_n_s_nstep, __pyx_n_s_params, __pyx_n_s_verbose, __pyx_n_s_rhomin, __pyx_n_s_alpha_up, __pyx_n_s_fast, __pyx_n_s_nEM, __pyx_n_s_niter, __pyx_n_s_liktraj, __pyx_n_s_nbest, __pyx_n_s_curpar, __pyx_n_s_n, __pyx_n_s_par, __pyx_n_s_init_par, __pyx_n_s_imp, __pyx_n_s_it, __pyx_n_s_cost_mat, __pyx_n_s_gen_dat, __pyx_n_s_geno, __pyx_n_s_vit, __pyx_n_s_args, __pyx_n_s_res, __pyx_n_s_haplo, __pyx_n_s_permut, __pyx_n_s_newpar, __pyx_n_s_par_s, __pyx_n_s_dat, __pyx_n_s_i, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_ijump, __pyx_n_s_ijump); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(10, 0, 34, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_optimfit, 243, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_tuple__62 = PyTuple_Pack(9, ((PyObject *)__pyx_int_20), ((PyObject *)__pyx_int_10), ((PyObject *)Py_None), ((PyObject *)Py_False), ((PyObject*)__pyx_float_1eneg_6), ((PyObject *)Py_False), ((PyObject *)Py_False), ((PyObject *)__pyx_int_10), ((PyObject *)__pyx_int_5)); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "fastphase/fastphase.pyx":324
 *         return par_s
 * 
 *     def viterbi(self, parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */
  __pyx_tuple__63 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_parList, __pyx_n_s_tasks, __pyx_n_s_results, __pyx_n_s_Imputations, __pyx_n_s_item, __pyx_n_s_name, __pyx_n_s_hap, __pyx_n_s_name, __pyx_n_s_gen); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_viterbi, 324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":376
 *         #     tasks=[]
 * 
 *     def impute(self,parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */
  __pyx_tuple__65 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_parList, __pyx_n_s_tasks, __pyx_n_s_results, __pyx_n_s_Imputations, __pyx_n_s_item, __pyx_n_s_name, __pyx_n_s_hap, __pyx_n_s_name, __pyx_n_s_gen); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_impute, 376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 376, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":387
 * 
 * 
 * def fitter( item):             # <<<<<<<<<<<<<<
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 */
  __pyx_tuple__67 = PyTuple_Pack(7, __pyx_n_s_item, __pyx_n_s_hLogLike, __pyx_n_s_top, __pyx_n_s_bot, __pyx_n_s_jmk, __pyx_n_s_res, __pyx_n_s_gLogLike); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_fitter, 387, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 387, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":401
 *     return res
 * 
 * def viterber( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */
  __pyx_tuple__69 = PyTuple_Pack(10, __pyx_n_s_item, __pyx_n_s_i, __pyx_n_s_k, __pyx_n_s_nLoc, __pyx_n_s_x, __pyx_n_s_nsamp, __pyx_n_s_path, __pyx_n_s_par, __pyx_n_s_pth, __pyx_n_s_res); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_viterber, 401, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 401, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":421
 *     return res
 * 
 * def imputer( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */
  __pyx_tuple__71 = PyTuple_Pack(15, __pyx_n_s_item, __pyx_n_s_i, __pyx_n_s_k, __pyx_n_s_nLoc, __pyx_n_s_x, __pyx_n_s_nsamp, __pyx_n_s_pgeno, __pyx_n_s_probZ, __pyx_n_s_path, __pyx_n_s_par, __pyx_n_s_pZ, __pyx_n_s_pth, __pyx_n_s_res, __pyx_n_s_k1, __pyx_n_s_k2); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(1, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_imputer, 421, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 421, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":471
 *     return res
 * 
 * def calc_cost_matrix_haplo( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k = vit[0]
 */
  __pyx_tuple__73 = PyTuple_Pack(6, __pyx_n_s_args, __pyx_n_s_vit, __pyx_n_s_nK, __pyx_n_s_k, __pyx_n_s_kp, __pyx_n_s_cost_mat); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_calc_cost_matrix_haplo, 471, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 471, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":509
 *     return res
 * 
 * def calc_cost_matrix_geno( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k1, k2 = vit[0]
 */
  __pyx_tuple__75 = PyTuple_Pack(8, __pyx_n_s_args, __pyx_n_s_vit, __pyx_n_s_nK, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_kp1, __pyx_n_s_kp2, __pyx_n_s_cost_mat); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_calc_cost_matrix_geno, 509, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 509, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":538
 * 
 * 
 * def switch_pars(par, permut):             # <<<<<<<<<<<<<<
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):
 */
  __pyx_tuple__77 = PyTuple_Pack(6, __pyx_n_s_par, __pyx_n_s_permut, __pyx_n_s_newpar, __pyx_n_s_i, __pyx_n_s_curclus, __pyx_n_s_j); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastphase_fastphase_pyx, __pyx_n_s_switch_pars, 538, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_01 = PyFloat_FromDouble(0.01); if (unlikely(!__pyx_float_0_01)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_98 = PyFloat_FromDouble(0.98); if (unlikely(!__pyx_float_0_98)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_6 = PyFloat_FromDouble(1e-6); if (unlikely(!__pyx_float_1eneg_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_001 = PyFloat_FromDouble(0.001); if (unlikely(!__pyx_float_0_001)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_999 = PyFloat_FromDouble(0.999); if (unlikely(!__pyx_float_0_999)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit.tp_dictoffset && __pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6hapflk_9fastphase___pyx_scope_struct__optimfit = &__pyx_type_6hapflk_9fastphase___pyx_scope_struct__optimfit;
  if (PyType_Ready(&__pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6hapflk_9fastphase___pyx_scope_struct_1_genexpr = &__pyx_type_6hapflk_9fastphase___pyx_scope_struct_1_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfastphase(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfastphase(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_fastphase(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_fastphase(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_fastphase(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'fastphase' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_fastphase(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("fastphase", __pyx_methods, __pyx_k_Module_implementing_the_LD_mode, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_hapflk__fastphase) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "hapflk.fastphase")) {
      if (unlikely(PyDict_SetItemString(modules, "hapflk.fastphase", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "fastphase/fastphase.pyx":13
 * cimport cython
 * ## NUMPY import in cython
 * import numpy as np             # <<<<<<<<<<<<<<
 * from scipy.optimize import linear_sum_assignment
 * from libc.math cimport log
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":14
 * ## NUMPY import in cython
 * import numpy as np
 * from scipy.optimize import linear_sum_assignment             # <<<<<<<<<<<<<<
 * from libc.math cimport log
 * cimport numpy as np
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_linear_sum_assignment);
  __Pyx_GIVEREF(__pyx_n_s_linear_sum_assignment);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_linear_sum_assignment);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_optimize, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_linear_sum_assignment); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_linear_sum_assignment, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":17
 * from libc.math cimport log
 * cimport numpy as np
 * np.import_array()             # <<<<<<<<<<<<<<
 * from multiprocessing import Pool, cpu_count
 * import sys
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 17, __pyx_L1_error)

  /* "fastphase/fastphase.pyx":18
 * cimport numpy as np
 * np.import_array()
 * from multiprocessing import Pool, cpu_count             # <<<<<<<<<<<<<<
 * import sys
 * from functools import reduce
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Pool);
  __Pyx_GIVEREF(__pyx_n_s_Pool);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Pool);
  __Pyx_INCREF(__pyx_n_s_cpu_count);
  __Pyx_GIVEREF(__pyx_n_s_cpu_count);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_cpu_count);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_multiprocessing, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Pool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Pool, __pyx_t_2) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_cpu_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cpu_count, __pyx_t_2) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":19
 * np.import_array()
 * from multiprocessing import Pool, cpu_count
 * import sys             # <<<<<<<<<<<<<<
 * from functools import reduce
 * from operator import add
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":20
 * from multiprocessing import Pool, cpu_count
 * import sys
 * from functools import reduce             # <<<<<<<<<<<<<<
 * from operator import add
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_reduce);
  __Pyx_GIVEREF(__pyx_n_s_reduce);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_reduce);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_functools, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_reduce); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":21
 * import sys
 * from functools import reduce
 * from operator import add             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_add);
  __Pyx_GIVEREF(__pyx_n_s_add);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_add);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_operator, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_add); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_add, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":24
 * 
 * 
 * class fitInData():             # <<<<<<<<<<<<<<
 *     def __init__(self,type,alpha,theta,rho,data,u2z):
 *         self.type=type
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_fitInData, __pyx_n_s_fitInData, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":25
 * 
 * class fitInData():
 *     def __init__(self,type,alpha,theta,rho,data,u2z):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.alpha=alpha
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fitInData_1__init__, 0, __pyx_n_s_fitInData___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":24
 * 
 * 
 * class fitInData():             # <<<<<<<<<<<<<<
 *     def __init__(self,type,alpha,theta,rho,data,u2z):
 *         self.type=type
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_fitInData, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fitInData, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":33
 *         self.up2pz=u2z
 * 
 * class fitOutData():             # <<<<<<<<<<<<<<
 *     def __init__(self,logLike,top,bot,jmk,val):
 *         self.logLike=logLike
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_fitOutData, __pyx_n_s_fitOutData, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":34
 * 
 * class fitOutData():
 *     def __init__(self,logLike,top,bot,jmk,val):             # <<<<<<<<<<<<<<
 *         self.logLike=logLike
 *         self.top=top
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_10fitOutData_1__init__, 0, __pyx_n_s_fitOutData___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":33
 *         self.up2pz=u2z
 * 
 * class fitOutData():             # <<<<<<<<<<<<<<
 *     def __init__(self,logLike,top,bot,jmk,val):
 *         self.logLike=logLike
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_fitOutData, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fitOutData, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":41
 *         self.val=val
 * 
 * class imputeInData():             # <<<<<<<<<<<<<<
 *     def __init__(self,type,parList,nLoc,name,data):
 *         self.type=type
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_imputeInData, __pyx_n_s_imputeInData, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":42
 * 
 * class imputeInData():
 *     def __init__(self,type,parList,nLoc,name,data):             # <<<<<<<<<<<<<<
 *         self.type=type
 *         self.parList=parList
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_12imputeInData_1__init__, 0, __pyx_n_s_imputeInData___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":41
 *         self.val=val
 * 
 * class imputeInData():             # <<<<<<<<<<<<<<
 *     def __init__(self,type,parList,nLoc,name,data):
 *         self.type=type
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_imputeInData, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_imputeInData, __pyx_t_2) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":49
 *         self.data=data
 * 
 * class imputeOutData():             # <<<<<<<<<<<<<<
 *     def __init__(self,pgeno,name,pZ,path=None):
 *         self.name=name
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_imputeOutData, __pyx_n_s_imputeOutData, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":50
 * 
 * class imputeOutData():
 *     def __init__(self,pgeno,name,pZ,path=None):             # <<<<<<<<<<<<<<
 *         self.name=name
 *         self.pgeno=pgeno
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_13imputeOutData_1__init__, 0, __pyx_n_s_imputeOutData___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__29);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":49
 *         self.data=data
 * 
 * class imputeOutData():             # <<<<<<<<<<<<<<
 *     def __init__(self,pgeno,name,pZ,path=None):
 *         self.name=name
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_imputeOutData, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_imputeOutData, __pyx_t_2) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":56
 *         self.path=path
 * 
 * class modParams():             # <<<<<<<<<<<<<<
 *     '''
 *     A class for fastphase model parameters.
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_modParams, __pyx_n_s_modParams, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, __pyx_kp_s_A_class_for_fastphase_model_par); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":67
 *     -- rho (N x 1): jump probabilities in each interval
 *     '''
 *     def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):             # <<<<<<<<<<<<<<
 *         self.nLoc=nLoc
 *         self.nClus=nClus
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9modParams_1__init__, 0, __pyx_n_s_modParams___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__32);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":78
 *         self.theta_up = theta_up
 *         self.loglike = 0
 *     def initUpdate(self):             # <<<<<<<<<<<<<<
 *         self.top=np.zeros((self.nLoc,self.nClus))
 *         self.bot=np.zeros((self.nLoc,self.nClus))
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9modParams_3initUpdate, 0, __pyx_n_s_modParams_initUpdate, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_initUpdate, __pyx_t_2) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":84
 *         self.jm=np.zeros((self.nLoc,1))
 *         self.nhap=0.0
 *     def addIndivFit(self,t,b,j,nhap):             # <<<<<<<<<<<<<<
 *         self.top += t
 *         self.bot += b
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9modParams_5addIndivFit, 0, __pyx_n_s_modParams_addIndivFit, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_addIndivFit, __pyx_t_2) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":90
 *         self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
 *         self.nhap+=nhap
 *     def update(self):             # <<<<<<<<<<<<<<
 *         ''' Update parameters using top,bot,jmk jm probabilities'''
 *         ## rho
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9modParams_7update, 0, __pyx_n_s_modParams_update, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_update, __pyx_t_2) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":120
 *                     elif self.theta[i,j]<0.001:
 *                         self.theta[i,j]=0.001
 *     def write(self,stream=sys.stdout):             # <<<<<<<<<<<<<<
 *         print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
 *         for i in range(self.nLoc):
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9modParams_9write, 0, __pyx_n_s_modParams_write, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults), 1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_stdout); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_stream = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6hapflk_9fastphase_25__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_write, __pyx_t_2) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastphase/fastphase.pyx":56
 *         self.path=path
 * 
 * class modParams():             # <<<<<<<<<<<<<<
 *     '''
 *     A class for fastphase model parameters.
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_modParams, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_modParams, __pyx_t_2) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":125
 *             print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)
 * 
 * _tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))             # <<<<<<<<<<<<<<
 * 
 * class fastphase():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_vectorize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_24lambda, 0, __pyx_n_s_lambda, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tohap, __pyx_t_5) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":127
 * _tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))
 * 
 * class fastphase():             # <<<<<<<<<<<<<<
 *     '''
 *     A class to manipulate and control a fastphase model (Scheet and Stephens, 2006)
 */
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_fastphase, __pyx_n_s_fastphase, (PyObject *) NULL, __pyx_n_s_hapflk_fastphase, __pyx_kp_s_A_class_to_manipulate_and_contr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "fastphase/fastphase.pyx":136
 *          ... do stuff ...
 *     '''
 *     def __init__(self, nLoci, nproc = cpu_count(),prfx=None):             # <<<<<<<<<<<<<<
 *         assert nLoci>0
 *         self.nLoci=nLoci
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_1__init__, 0, __pyx_n_s_fastphase___init, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_1, sizeof(__pyx_defaults1), 1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cpu_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_1)->__pyx_arg_nproc = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_1, __pyx_pf_6hapflk_9fastphase_27__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_init, __pyx_t_1) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":145
 *         self.prfx = prfx
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self.pool = Pool ( self.nproc)
 *         if self.prfx is None:
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_3__enter__, 0, __pyx_n_s_fastphase___enter, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_enter, __pyx_t_1) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":153
 *         return self
 * 
 *     def __exit__(self,*args):             # <<<<<<<<<<<<<<
 *         self.pool.terminate()
 *         if self.prfx is not None:
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_5__exit__, 0, __pyx_n_s_fastphase___exit, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_exit, __pyx_t_1) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":158
 *             self.flog.close()
 * 
 *     def flush(self):             # <<<<<<<<<<<<<<
 *         '''
 *         remove data
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_7flush, 0, __pyx_n_s_fastphase_flush, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_flush, __pyx_t_1) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":165
 *         self.genotypes={}
 * 
 *     def addHaplotype(self,ID,hap,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add an haplotype to the model observations.
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_9addHaplotype, 0, __pyx_n_s_fastphase_addHaplotype, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__51);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_addHaplotype, __pyx_t_1) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":177
 *             print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
 *             raise
 *     def addGenotype(self,ID,gen,missing=-1):             # <<<<<<<<<<<<<<
 *         '''
 *         Add a genotype to the model observations.
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_11addGenotype, 0, __pyx_n_s_fastphase_addGenotype, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__54);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_addGenotype, __pyx_t_1) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastphase/fastphase.pyx":191
 * 
 *     @staticmethod
 *     def gen2hap(gen):             # <<<<<<<<<<<<<<
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_13gen2hap, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_fastphase_gen2hap, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fastphase/fastphase.pyx":190
 *             raise
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def gen2hap(gen):
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_gen2hap, __pyx_t_4) < 0) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":194
 *         return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
 * 
 *     def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Fit the model on observations with nCLus clusters using nstep EM iterations
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_15fit, 0, __pyx_n_s_fastphase_fit, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__59);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_fit, __pyx_t_4) < 0) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":243
 *         return par
 * 
 *     def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):             # <<<<<<<<<<<<<<
 *         liktraj = []
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_17optimfit, 0, __pyx_n_s_fastphase_optimfit, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__62);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_optimfit, __pyx_t_4) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":324
 *         return par_s
 * 
 *     def viterbi(self, parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_19viterbi, 0, __pyx_n_s_fastphase_viterbi, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_viterbi, __pyx_t_4) < 0) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":376
 *         #     tasks=[]
 * 
 *     def impute(self,parList):             # <<<<<<<<<<<<<<
 *         tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
 *         tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9fastphase_21impute, 0, __pyx_n_s_fastphase_impute, NULL, __pyx_n_s_hapflk_fastphase, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_impute, __pyx_t_4) < 0) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastphase/fastphase.pyx":127
 * _tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))
 * 
 * class fastphase():             # <<<<<<<<<<<<<<
 *     '''
 *     A class to manipulate and control a fastphase model (Scheet and Stephens, 2006)
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_fastphase, __pyx_empty_tuple, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fastphase, __pyx_t_4) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":387
 * 
 * 
 * def fitter( item):             # <<<<<<<<<<<<<<
 *     #type,alpha,theta,rho,data = item
 *     if item.type == 'haplo' :
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_1fitter, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fitter, __pyx_t_5) < 0) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":401
 *     return res
 * 
 * def viterber( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_3viterber, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_viterber, __pyx_t_5) < 0) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":421
 *     return res
 * 
 * def imputer( item):             # <<<<<<<<<<<<<<
 *     #cIter=iter
 *     cdef int i,k
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_5imputer, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_imputer, __pyx_t_5) < 0) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":471
 *     return res
 * 
 * def calc_cost_matrix_haplo( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k = vit[0]
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_9calc_cost_matrix_haplo, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_cost_matrix_haplo, __pyx_t_5) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":509
 *     return res
 * 
 * def calc_cost_matrix_geno( args):             # <<<<<<<<<<<<<<
 *     vit, nK = args
 *     k1, k2 = vit[0]
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_13calc_cost_matrix_geno, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_cost_matrix_geno, __pyx_t_5) < 0) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":538
 * 
 * 
 * def switch_pars(par, permut):             # <<<<<<<<<<<<<<
 *     newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
 *     for i in range(par.nClus):
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_6hapflk_9fastphase_15switch_pars, NULL, __pyx_n_s_hapflk_fastphase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_switch_pars, __pyx_t_5) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastphase/fastphase.pyx":1
 * #!python             # <<<<<<<<<<<<<<
 * #cython: language_level=3, boundscheck=False
 * '''
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "../../../../../usr/lib/python3/dist-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init hapflk.fastphase", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init hapflk.fastphase");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_sst_abs(size) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            if (unlikely(zerodivision_check && b == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");
                return NULL;
            }
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
    }
        
        PyFPE_START_PROTECT("add", return NULL)
        result = a + b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(operand) if (unlikely(zerodivision_check && ((operand) == 0))) {\
    PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");\
    return NULL;\
}
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(0) break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        }
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        __Pyx_PyFloat_TrueDivideCObj_ZeroDivisionError(b)
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* GetItemInt */
    static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
    #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractCObj(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long x;
        long b = PyInt_AS_LONG(op2);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            b = likely(size) ? digits[0] : 0;
            if (size == -1) b = -b;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
        double b = PyFloat_AS_DOUBLE(op2);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetTopmostException */
    #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* IterFinish */
    static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
    static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
    static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
    static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
    static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
    static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
    static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* None */
    static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
    static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* None */
    static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* BytesEquals */
    static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
    static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* SetItemInt */
    static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IsLittleEndian */
    static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
    static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
      static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* ExtTypeTest */
      static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* ArgTypeTest */
      static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyFloatBinop */
      #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* WriteUnraisableException */
        static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* None */
        static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* DictGetItem */
        #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObject_GenericGetAttrNoDict */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
        #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
        static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
        static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FetchCommonType */
        static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
        #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CalculateMetaclass */
        static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
        static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


        /* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* ToPyCTupleUtility */
        static PyObject* __pyx_convert__to_py___pyx_ctuple_int__and_int(__pyx_ctuple_int__and_int value) {
    PyObject* item = NULL;
    PyObject* result = PyTuple_New(2);
    if (!result) goto bad;
        item = __Pyx_PyInt_From_int(value.f0);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 0, item);
        item = __Pyx_PyInt_From_int(value.f1);
        if (!item) goto bad;
        PyTuple_SET_ITEM(result, 1, item);
    return result;
bad:
    Py_XDECREF(item);
    Py_XDECREF(result);
    return NULL;
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_long(npy_long value) {
    const npy_long neg_one = (npy_long) ((npy_long) 0 - (npy_long) 1), const_zero = (npy_long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
        #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
        #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
        #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
        #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
        static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
        #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
        static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
        #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
        static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
